/*  $Header: //depot/Software/include/identification/bootrec.h#7 $
 *
 *  Copyright (C) 2002 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* Some preprocessor magic to use this file with MVC, GCC and ARM Compiler */
#ifdef __GNUC__
#define GCC_PACKED __attribute__((packed))
#define ARM_PACKED
#define INLINE inline
#else
#ifdef WIN32
#define ARM_PACKED
#define GCC_PACKED
#define INLINE
#else
#define ARM_PACKED __packed
#define GCC_PACKED
#define INLINE __inline
#endif
#endif

/*
 * This file contains the definitions for Boot Records that
 * are used to identify (software/hardware) components.
 *
 * NOTE: All bootrecords start at a 32 bit aligned address
 * so the structs should NOT be packed
 */

#ifndef _BOOTREC_H
#define _BOOTREC_H

#define BR_MAX_BRA_SIZE		512	/* size in bytes */

#define BR_CODE_MIN			0x80000000
#define BR_CODE_COMPONENT_ID		0x80000001
#define BR_CODE_COMPONENT_VERSION	0x80000002
#define BR_CODE_DEPENDENT_IF		0x80000003
#define BR_CODE_EXPOSED_IF		0x80000004
#define BR_CODE_DESCR			0x80000101
#define BR_CODE_MAX			0x8FFFFFFF
#define BR_CODE_END_OF_BRA		0xFF0000FF
#define LEGACY_BR_CODE_END_OF_BRA	0xFFFFFFFF

#ifndef VER_STRING_SIZE
#define VER_STRING_SIZE 20		/* size in 24 bit words */
#endif

#ifndef __ASSEMBLER__	// skip te rest for assembler files

#ifndef BOOT_BOOTREC
/*
 * Each boot record is identified by a unique code, followed
 * by the length (in 32 bit words) of the data in the record.
 */
struct s_bootrec_hdr {
	unsigned long code;
	unsigned long length;
};

/* Generic bootrecord structure that can be mapped to each bootrecord */
struct s_bootrecord {
	struct s_bootrec_hdr hdr;
	char start_of_data;
};


/* The Component ID is a unique identification of the component
 * Component IDs are defined in "interface.h"
 */
struct s_bootrec_comp_id {
	struct s_bootrec_hdr hdr;
	long id;
};

/* The Version String is a zero terminated ASCII string */
struct s_bootrec_comp_ver {
	struct s_bootrec_hdr hdr;
	char ver_string[VER_STRING_SIZE * sizeof(long)];
};

/*
 * Interfaces are defined by the following scheme
 *
 *  -----------------------------------------------------
 * | 16 bits |  16 bits |  16 bits |  16 bits |  16 bits |
 * |  role   |   if ID  |  variant | btm comp | top comp |
 *  -----------------------------------------------------
 *
 * Role and Interface IDs are defined in "interface.h"
 */

/*
 * The dependent interface defines which interfaces a component
 * needs for its proper operation. These can be mached against
 * the exposed interfaces of other components that are available.
 */
ARM_PACKED struct s_bootrec_dep_if {
	unsigned short role;
	unsigned short if_id;
	unsigned short variant;
} GCC_PACKED;

/*
 * The exposed interface defines which interfaces a component
 * offers to other components. Bottom and Top compatibility define
 * with which variants this interface is compatible with.
 */
ARM_PACKED struct s_bootrec_exp_if {
	unsigned short role;
	unsigned short if_id;
	unsigned short variant;
	unsigned short btm_compat;
	unsigned short top_compat;
} GCC_PACKED;

/* The Bootrecord End defines the end of the bootrecords.
 * No bootrecords will follow after this one
 */
struct s_bootrec_end {
	struct s_bootrec_hdr hdr;
	short CRC;
	short pad;
	long MD5[4];
};

/* Some inline helper functions */

/*
 * Returns 0 if the passed bootrecord is valid, -1 otherwise.
 */
static INLINE int valid_bootrecord(const struct s_bootrecord *bootrec)
{
	/* length too big */
	if(bootrec->hdr.length > (BR_MAX_BRA_SIZE / sizeof(long)))
		return -1;

	if(bootrec->hdr.code == BR_CODE_END_OF_BRA)
		return 0;

	/* no valid code */
	if((bootrec->hdr.code < BR_CODE_MIN) ||
	   (bootrec->hdr.code > BR_CODE_MAX))
		return -1;

	return 0;
}

/*
 * Returns the number of dependent interfaces in a bootrecord
 * result < 0 if no valid dependent interface bootrecord is passed
 */
static INLINE int nof_dep_interfaces(const struct s_bootrecord *bootrec)
{
	if((bootrec->hdr.code != BR_CODE_DEPENDENT_IF) ||
	   (valid_bootrecord(bootrec)))
		return -1;

	return (bootrec->hdr.length * sizeof(long)) / sizeof(struct s_bootrec_dep_if);
}

/*
 * Returns the number of exposed interfaces in a bootrecord.
 * result < 0 if no valid exposed interface bootrecord is passed
 */
static INLINE int nof_exp_interfaces(const struct s_bootrecord *bootrec)
{
	if((bootrec->hdr.code != BR_CODE_EXPOSED_IF) ||
	   (valid_bootrecord(bootrec)))
		return -1;

	return (bootrec->hdr.length * sizeof(long)) / sizeof(struct s_bootrec_exp_if);
}

/*
 * Returns the size in bytes of a bootrecord, including the header.
 */
static INLINE int bootrecord_size(const struct s_bootrecord *bootrec)
{
	return (sizeof(struct s_bootrec_hdr) + (bootrec->hdr.length * sizeof(long)));
}

/*
 * Returns a pointer to the next bootrecord
 */
static INLINE struct s_bootrecord *next_bootrecord(const struct s_bootrecord *bootrec)
{
	struct s_bootrecord *next;

	next = (struct s_bootrecord *)((int)bootrec + bootrecord_size(bootrec));
	return next;
}

#else  // BOOT_BOOTREC

/*
 * Each boot record is identified by a unique code, followed
 * by the length (in 32 bit words) of the data in the record.
 */
struct s_bootrec_hdr {
	unsigned long code;
	unsigned long length;
};

/* Generic bootrecord structure that can be mapped to each bootrecord */
struct s_bootrecord {
	struct s_bootrec_hdr hdr;
	char start_of_data;
};


/* The Component ID is a unique identification of the component
 * Component IDs are defined in "interface.h"
 */
struct s_bootrec_comp_id {
	struct s_bootrec_hdr hdr;
	unsigned long id;
};

/* The Version String is a zero terminated ASCII string */
struct s_bootrec_comp_ver {
	struct s_bootrec_hdr hdr;
	char ver_string[VER_STRING_SIZE * sizeof(long)];
};

/*
 * Interfaces are defined by the following scheme
 *
 *  -----------------------------------------------------
 * | 16 bits |  16 bits |  16 bits |  16 bits |  16 bits |
 * |  role   |   if ID  |  variant | btm comp | top comp |
 *  -----------------------------------------------------
 *
 * Role and Interface IDs are defined in "interface.h"
 */

/*
 * The dependent interface defines which interfaces a component
 * needs for its proper operation. These can be mached against
 * the exposed interfaces of other components that are available.
 */
ARM_PACKED struct s_bootrec_dep_if {
	unsigned short role;
	unsigned short if_id;
	unsigned short variant;
} GCC_PACKED;

/*
 * The exposed interface defines which interfaces a component
 * offers to other components. Bottom and Top compatibility define
 * with which variants this interface is compatible with.
 */
ARM_PACKED struct s_bootrec_exp_if {
	unsigned short role;
	unsigned short if_id;
	unsigned short variant;
	unsigned short btm_compat;
	unsigned short top_compat;
} GCC_PACKED;

/* The Bootrecord End defines the end of the bootrecords.
 * No bootrecords will follow after this one
 */
struct s_bootrec_end {
	struct s_bootrec_hdr hdr;
	unsigned short CRC;
	unsigned short pad;
	unsigned long MD5[4];
};

/* Some inline helper functions */

/*
 * Returns 0 if the passed bootrecord is valid, -1 otherwise.
 */
static INLINE int valid_bootrecord(const struct s_bootrecord *bootrec)
{
	/* length too big */
	if(bootrec->hdr.length > (BR_MAX_BRA_SIZE / sizeof(long))) {
                printf("valid_bootrecord() -> ERROR: too long (length=%ld\n", bootrec->hdr.length); //REMOVE
                return -1;
        }

	if(bootrec->hdr.code == BR_CODE_END_OF_BRA)
		return 0;

	/* no valid code */
	if((bootrec->hdr.code < BR_CODE_MIN) ||
	   (bootrec->hdr.code > BR_CODE_MAX)) {
                printf("valid_bootrecord() -> ERROR: Illegal code\n"); //REMOVE
		return -1;
         }

	return 0;
}

/*
 * Returns the number of dependent interfaces in a bootrecord
 * result < 0 if no valid dependent interface bootrecord is passed
 */
static INLINE int nof_dep_interfaces(const struct s_bootrecord *bootrec)
{
	if((bootrec->hdr.code != BR_CODE_DEPENDENT_IF) ||
	   (valid_bootrecord(bootrec)))
		return -1;

	return (bootrec->hdr.length * sizeof(long)) / sizeof(struct s_bootrec_dep_if);
}

/*
 * Returns the number of exposed interfaces in a bootrecord.
 * result < 0 if no valid exposed interface bootrecord is passed
 */
static INLINE int nof_exp_interfaces(const struct s_bootrecord *bootrec)
{
	if((bootrec->hdr.code != BR_CODE_EXPOSED_IF) ||
	   (valid_bootrecord(bootrec)))
		return -1;

	return (bootrec->hdr.length * sizeof(long)) / sizeof(struct s_bootrec_exp_if);
}

/*
 * Returns the size in bytes of a bootrecord, including the header.
 */
static INLINE int bootrecord_size(const struct s_bootrecord *bootrec)
{
	return (sizeof(struct s_bootrec_hdr) + (bootrec->hdr.length * sizeof(long)));
}

/*
 * Returns a pointer to the next bootrecord
 */
static INLINE struct s_bootrecord *next_bootrecord(const struct s_bootrecord *bootrec)
{
	struct s_bootrecord *next;
	
	next = (struct s_bootrecord *)((int)bootrec + bootrecord_size(bootrec));	
	return next;
}

#endif // BOOT_BOOTREC

#endif // __ASSEMBLER__

#endif /* _BOOTREC_H */

