/*
 *  Boa, an http server
 *  Copyright (C) 1995 Paul Phillips <psp@well.com>
 *  Some changes Copyright (C) 1996 Larry Doolittle <ldoolitt@jlab.org>
 *  Some changes Copyright (C) 1996,97 Jon Nelson <nels0988@tc.umn.edu>
 *  Some changes Copyright (C) 1998 Martin Hinner <martin@tdp.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 1, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define NO_MMAP_BUFFER_SIZE (1024)
#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

#include "boa.h"


#ifdef SSI_MOD
int open_ssi_connection(request* req)
{

	struct sockaddr_un servaddr;
	int ssi_fd = -1;

	if ( (ssi_fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 ) {
		perror("Socket:");
	}

	bzero(&servaddr, sizeof(servaddr));
	servaddr.sun_family = AF_UNIX;
	strcpy(servaddr.sun_path, SSI_UNIX_DOMAIN_SOCKET);

	if (connect(ssi_fd, &servaddr, sizeof(servaddr.sun_family)
			+ strlen(servaddr.sun_path)) < 0) {
		perror("Connect:");
	}

	/* write pathname to SSI module */
	if (write(ssi_fd, req->pathname, strlen(req->pathname))
		!= strlen(req->pathname)) {
		perror("Write:");
	}

	/* write \n to SSI module */
	if (write(ssi_fd, "\n", 1) != 1) {
		perror("Write \\n");
	}

	return (ssi_fd);
}
#endif




/* Flag to prevent concurrent SSI requests */
static int ssi_active = 0;

/* Flag for upgrade mode */
static int upgrade_active = 0;

void get_goto_upgrade_mode(void) {
	upgrade_active = 1;
}


void get_reset_ssi_flag(void)
{
	if (ssi_active != 1)
	{
		fprintf(stderr, "reset_ssi_flag error\n");
		exit(1);
	}
	ssi_active = 0;
}

/*
 * Name: init_get
 * Description: Initializes a non-script GET or HEAD request.
 *
 * Return values:
 *   0: finished or error, request will be freed
 *   1: successfully initialized, added to ready queue
 */

int init_get(request * req)
{
	int data_fd;
	struct stat statbuf;
#ifdef SSI_MOD
	int eof = 0;
        int pathname_length = 0;
#endif

        req->cgi_env[req->cgi_env_index] = NULL;     /* terminate cgi env */

        if (upgrade_active) {

            /* check if the filename is in DocumentRoot 
	     * (during upgrade, the imginstall send a SIGUSR2 to
	     * boa. Boa then adds two aliases, one for /oem and
	     * one for /wizards/upgrade. These aliases are processes
	     * before this function. If req->pathname starts with
	     * the document root, this is a request outside these
	     * two aliases.
	     */
            if (strncmp(document_root, req->pathname, strlen(document_root)) == 0) {
                /* request is not in /tmp, change it */

                /* free original pathname */
                free(req->pathname);

                /* create a new one. (strdup because it's will be 'free()-ed' when
                 * the request is freed
                 */
                req->pathname = strdup(UPGRADE_INDEX);
            }
        }
	
        data_fd = open(req->pathname, O_RDONLY);

#ifdef SSI_MOD

        /* check if extension is .shtml */
        pathname_length = strlen(req->pathname);

        if (pathname_length > 6) {
            /* if the last 6 chars match .shtml, this is a SSI page */
            req->is_ssi = (strcmp(".shtml", req->pathname + (pathname_length - 6) ) == 0);
        }
#endif


	if (data_fd == -1) {		/* cannot open */
#ifdef GUNZIP
		sprintf(buf, "%s.gz", req->pathname);
		data_fd = open(buf, O_RDONLY);
		if (data_fd == -1) {
#endif
			int errno_save = errno;
			log_error_doc(req);
			errno = errno_save;
#if 0
			perror("document open");
#endif
			errno = errno_save;

			if (errno == ENOENT)
				send_r_not_found(req);
			else if (errno == EACCES)
				send_r_forbidden(req);
			else
				send_r_bad_request(req);
			return 0;
#ifdef GUNZIP
		}
		close(data_fd);

		req->response_status = R_REQUEST_OK;
		if (!req->simple) {
			req_write(req, "HTTP/1.0 200 OK-GUNZIP\r\n");
			print_http_headers(req);
			print_content_type(req);
			print_last_modified(req);
			req_write(req, "\r\n");
			req_flush(req);
		}
		if (req->method == M_HEAD)
			return 0;
		if (req->pathname)
			free(req->pathname);
		req->pathname = strdup(buf);
		return init_cgi(req);	/* 1 - OK, 2 - die */
#endif
	}

	fstat(data_fd, &statbuf);

	if (S_ISDIR(statbuf.st_mode)) {		/* directory */
		close(data_fd);			/* close dir */

		if (req->pathname[strlen(req->pathname) - 1] != '/') {
			char buffer[3 * MAX_PATH_LENGTH + 128];

			if (server_port != 80)
				sprintf(buffer, "http://%s:%d%s/", req->host?req->host:get_server_name(), server_port,
						req->request_uri);
			else
				sprintf(buffer, "http://%s%s/", req->host?req->host:get_server_name(), req->request_uri);
			send_redirect_perm(req, buffer);
			return 0;
		}
		data_fd = get_dir(req, &statbuf);	/* updates statbuf */

		if (data_fd == -1)		/* couldn't do it */
			return 0;			/* errors reported by get_dir */
		else if (data_fd == 0)
			return 1;
	}


	if (req->if_modified_since &&
		!modified_since(&(statbuf.st_mtime), req->if_modified_since)
#ifdef SSI_MOD
		&& (!req->is_ssi)
#endif
		)

	{
		send_r_not_modified(req);
		close(data_fd);
		return 0;
	}
	
#ifdef SSI_MOD
	if (req->is_ssi)
	{
		if (!ssi_active)
		{

			req->ssi_fd = open_ssi_connection(req);

			/* Flag that we have an ongoing SSI request */
			ssi_active = 1;
		}
		else
		{
			/* Already one SSI request active, move to blocked queue */

			/* Mark the req->ssi_id as 'not opened yet' */
			req->ssi_fd = -1;

		}

		/* filesize is unknown */
		req->filesize = -1;

	} else
#endif
	{
		req->filesize = statbuf.st_size;
	}


	req->last_modified = statbuf.st_mtime;


	if (req->method == M_HEAD) {
		send_r_request_ok(req);
		close(data_fd);
#ifdef SSI_MOD
		if (req->is_ssi > 0) {
			close(req->ssi_fd);
		}
#endif
		return 0;
	}

#ifdef HAVE_MMAP
	/* MAP_OPTIONS: see compat.h */
	req->data_mem = mmap(0, statbuf.st_size,
#ifdef USE_NLS
			PROT_READ|PROT_WRITE
#else
			PROT_READ
#endif
			, MAP_OPTIONS,data_fd, 0);
	close(data_fd);				/* close data file */

	if ((long) req->data_mem == -1) {
		boa_perror(req, "mmap");
		return 0;
	}

#else /* HAVE_MMAP */
	/* store data_fd and DON'T close it */
	req->file_fd = data_fd;
#endif /* HAVE_MMAP */

	send_r_request_ok(req);		/* All's well */
	{
		static int bob;

		bob = BUFFER_SIZE - req->buffer_end;
		if (bob > 0) {

#ifndef SSI_MOD /* don't know filesize if SSI module is used */
			if (!(req->is_ssi))
#endif
			{
				if (bob > req->filesize)
					bob = req->filesize;
			}

#ifdef SSI_MOD
			if (req->is_ssi) {

				if (req->ssi_fd != -1) {
					/* We don't know the filesize, just read 'bob' bytes or
					 * the maximum we get and adjust 'bob' .*/
					int new_bob = 0;
					new_bob = read(req->ssi_fd, req->buffer + req->buffer_end, bob); 
					bob = new_bob;
				}
				else
				{
					/* The connection to the SSI module is not yet opened, so we
					 * cannot fill our buffer
					 */
					bob = 0;
				}

			} else
#endif /* SSI_MOD */
			{
#ifdef HAVE_MMAP

				memcpy(req->buffer + req->buffer_end,
						req->data_mem + req->filepos,
						bob);
#else
				if ( read(req->file_fd, req->buffer + req->buffer_end, bob) == -1 ){
							perror("read");
				}
#endif
			}

			req->buffer_end += bob;
			req->filepos += bob;
		}
	}

#ifdef SSI_MOD
	if (req->is_ssi) {
		/* if the socket was closed, we are done */
		if (eof) {
			req->status = CLOSE;
			return 0; /* done */
		}
	} else
#endif
	{
		if (req->filepos == req->filesize) {
			req->status = CLOSE;
			return 0; /* done! */
		}
	}

	/* We lose statbuf here, so make sure response has been sent */
	return 1;
}

/*
 * Name: process_get
 * Description: Writes a chunk of data to the socket.
 *
 * Return values:
 *  -1: request blocked, move to blocked queue
 *   0: EOF or error, close it down
 *   1: successful write, recycle in ready queue
 */

int process_get(request * req)
{
	int bytes_written, bytes_to_write;

#ifndef HAVE_MMAP
        char buffer[NO_MMAP_BUFFER_SIZE];
        int bytes_to_read = 0;
        int bytes_to_read_in_buffer = 0;
        int bytes_written_from_buffer = 0;
#endif

#ifdef SSI_MOD
        int bytes_read = 0;

	/* When using the SSI modules, we don't know the filesize and
	 * therefore don't know the n/o bytes to write
	 */
	if (req->is_ssi) {
		bytes_to_write = -1;
	} else
#endif
	{
		bytes_to_write = req->filesize - req->filepos;
	}


#ifdef USE_NLS
	if (req->method != M_HEAD)
	{
		if (bytes_to_write>BYTES_TO_WRITE)
			bytes_to_write = BYTES_TO_WRITE;

		if (req->bytesconverted<req->filepos)
			req->bytesconverted = req->filepos;

		if (req->cp_table)
		{
			nls_convert(req->data_mem + req->bytesconverted,req->cp_table,
					req->filepos + bytes_to_write - req->bytesconverted );
			req->bytesconverted = req->filepos + bytes_to_write;
		}
	}
#endif

#ifdef SERVER_SSL
	if(req->ssl == NULL){
#endif /*SERVER_SSL*/

#ifdef SSI_MOD
	if (!(req->is_ssi))
#endif
	{
#ifdef HAVE_MMAP

		bytes_written = write(req->fd, req->data_mem + req->filepos,
							  bytes_to_write);
#else

                /* positioning input file */
                if (lseek(req->file_fd, req->filepos, SEEK_SET ) == -1 )
                                        perror("lseek");

                bytes_to_read = bytes_to_write;
                bytes_written = 0;

                bytes_to_read_in_buffer = 0;
                bytes_written_from_buffer = 0;

                while ((bytes_to_read > 0)
                        && (bytes_written_from_buffer == bytes_to_read_in_buffer)) {

                    bytes_to_read_in_buffer = min(NO_MMAP_BUFFER_SIZE, bytes_to_read);

                    if ( read( req->file_fd, buffer, bytes_to_read_in_buffer) == -1 )
		            perror("read");

                    bytes_written_from_buffer =  write(req->fd, buffer, bytes_to_read_in_buffer);

                    bytes_written = bytes_written + bytes_written_from_buffer;

                    bytes_to_read = bytes_to_read - bytes_to_read_in_buffer;
                }


#endif
	}
#ifdef SSI_MOD
	else
	{
		/* Check if this request has an open SSI connection */
		if (req->ssi_fd == -1) {

			/* Nope. Can we open it ? */
			if (!ssi_active) {
				/* Yes, we can */

				/* Open the connection */
				req->ssi_fd = open_ssi_connection(req);

				/* SSI connection active */
				ssi_active = 1;
			}
			else
			{
				/* Move to blocked queue */
				return -1;
			}
		}

		if (req->ssi_fd == -1) {
			/* No connection possible. We can't read, so cannot write */
			bytes_written = 0;
		}
		else
		{

			int written;
			bytes_written = 0;

			/* Check if we still got bytes left from our previous try */
			if (req->no_read_bytes > 0) {

				written = (write(req->fd, &(req->read_bytes[req->read_bytes_pos]), req->no_read_bytes)) ;

				if (written != -1) {

					/* wrote something */
					bytes_written = bytes_written + written;
					req->no_read_bytes = req->no_read_bytes - written;
					req->read_bytes_pos = req->read_bytes_pos + written;

					if (req->no_read_bytes < 0 )
					exit(1);
				}

			}


			bytes_read = SSI_BUFFER_SIZE; /* to help our while loop :-) */

			/* The while loop is only entered if req->read_bytes_pos = 0, which means that */
			/* all the previous read bytes are written                                     */

			while ((bytes_read == SSI_BUFFER_SIZE) && (req->no_read_bytes == 0)){

				bytes_read = read(req->ssi_fd, &(req->read_bytes[0]), SSI_BUFFER_SIZE);

				if (bytes_read > 0) {
					written = write(req->fd, &(req->read_bytes[0]), bytes_read);

					if ( written == bytes_read ) {
						/* We wrote everything we have read before */
						bytes_written =+ written;
					}
					else
					{
						/* We didn't write everything */

						/* written -1 means: wrote nothing */
						if ( written == -1 )
						{
							written = 0;
						}

						bytes_written =+ written;

						/* save our read data */
						req->read_bytes_pos = written;
						req->no_read_bytes = bytes_read - written;

					}
				}

			}

			if (bytes_written == 0 && bytes_read != 0) {
				/* We did read something but didn't write anything -> blocked write */
				bytes_written = -1;
			}
		}
	}
#endif



#ifdef SERVER_SSL
	}else{
		bytes_written = SSL_write(req->ssl, req->data_mem + req->filepos, bytes_to_write);
#if 0
		printf("SSL_write\n");
#endif /*0*/
	}
#endif /*SERVER_SSL*/
	if (bytes_written == -1)
		if (errno == EWOULDBLOCK || errno == EAGAIN)
			return -1;			/* request blocked at the pipe level, but keep going */
		else {
			if (errno != EPIPE) {
				log_error_doc(req);	/* Can generate lots of log entries, */
#if 0
				perror("write");	/* OK to disable if your logs get too big */
#endif
			}
			return 0;
		}
	req->filepos += bytes_written;

#ifdef SSI_MOD
	if (req->is_ssi) {

		/* If we didn't read anything, we are finished */
		if (bytes_read == 0) {
			return 0;
		}
		else
		{
			return 1; /* more to do */
		}
	} else
#endif
	{
		if (req->filepos == req->filesize)	/* EOF */
			return 0;
		else
			return 1;				/* more to do */
	}
}

/*
 * Name: get_dir
 * Description: Called from process_get if the request is a directory.
 * statbuf must describe directory on input, since we may need its
 *   device, inode, and mtime.
 * statbuf is updated, since we may need to check mtimes of a cache.
 * returns:
 *  -1 error
 *  0  cgi (either gunzip or auto-generated)
 *  >0  file descriptor of file
 */

int get_dir(request * req, struct stat *statbuf)
{

	char pathname_with_index[MAX_PATH_LENGTH];
	int data_fd;

	sprintf(pathname_with_index, "%s%s", req->pathname, directory_index);

	data_fd = open(pathname_with_index, O_RDONLY);

	if (data_fd != -1) {		/* user's index file */
		strcat(req->request_uri, directory_index); /* for mimetype */
		fstat(data_fd, statbuf);
		return data_fd;
	}
	if (errno == EACCES) {
		send_r_forbidden(req);
		return -1;
	}
#ifdef GUNZIP
	strcat(pathname_with_index, ".gz");
	data_fd = open(pathname_with_index, O_RDONLY);
	if (data_fd != -1) {		/* user's index file */
		close(data_fd);

		req->response_status = R_REQUEST_OK;
		if (req->method == M_HEAD) {
			req_write(req, "HTTP/1.0 200 OK-GUNZIP\r\n");
			print_http_headers(req);
			print_last_modified(req);
			req_write(req, "Content-Type: ");
			req_write(req, get_mime_type(directory_index));
			req_write(req, "\r\n\r\n");
			req_flush(req);
			return 0;
		}
		if (req->pathname)
			free(req->pathname);
		req->pathname = strdup(pathname_with_index);
		if (init_cgi(req) == 0)
			return -1;
		return 0;				/* in this case, 0 means success */
	} else
#endif

	if (dirmaker != NULL) {
		req->response_status = R_REQUEST_OK;

		/* the indexer should take care of all headers */
		if (!req->simple || req->method == M_HEAD) {
			req_write(req, "HTTP/1.0 200 OK\r\n");
			print_http_headers(req);
			print_last_modified(req);
			req_write(req, "Content-Type: ");
			req_write(req, get_mime_type(directory_index));
			req_write(req, "\r\n\r\n");
			req_flush(req);
		}
		if (req->method == M_HEAD)
			return 0;

		return (init_cgi(req) == 0 ? -1 : 0);
		/* in this case, 0 means success */
	}
	send_r_forbidden(req);
	return -1;					/* nothing worked */
}
