/*

aplog

This process listens on a unix socket for incoming request from 
clients. 
 
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <linux/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>

#include "apinfo.h" 

#include <aplog.h>
#include <aplog_defs.h>

//#define SOCKFILE "/var/run/aplog.sock"
//#define LOGFILE "/var/log/aplog.log"

#define CONFIG "/var/etc/aplog.conf"

#define LOGDB "/etc/aplogdb"
#define MAX(x,y) (x < y ? y : x)
//#define MAX_MSG_LEN 128
#define MAXCLIENTS 8
#define DEFLOGSIZE 100
#define MAXLOGSIZE 250
#define MAX_PRIORITY 8
#define NOF_PRIORITY	(MAX_PRIORITY)
#define MAX_FACILITY 20
#define NOF_FACILITY	(MAX_FACILITY + 1)

#define whine(a,b) if(a) fprintf(stderr, "%s : %s", b, strerror(errno));

#define SNMP_TRAP_HOST   0x01
#define SYSLOG_HOST      0x02

#define REPORTING_LOCAL       0x80
#define REPORTING_SNMP_TRAP   0x40
#define REPORTING_SYSLOG      0x20

struct loghost_t {
	struct in_addr    addr; 
	struct loghost_t *next;
};

struct config_t {
	int count_doubles;
	int logsize;
	int verbose;

	u_char faci_reporting[NOF_FACILITY];
	u_char prio_where[NOF_PRIORITY];
	struct loghost_t *syslog_hosts;	
	struct loghost_t *snmp_trap_hosts;
};

// event database

struct event_t {
	unsigned long eventid;
	int fptr;
	struct event_t *next;
};
	


//////////////////// Globals //////////////


static int clientfd[MAXCLIENTS];
static struct config_t config;

static int read_config_flag = 0;
static int dump_config_flag = 0;
static int flush_log_flag = 0;
static int remote_log = 0; /* remote log is disabled until SIGUSR1 is received */

static int config_read = 0; /* flag to indicate if the configfile has been read */

// event database

static struct event_t *eventlist = NULL;
FILE *eventdb_f;

// log database

FILE *log_f;
struct aplog_entry_t  entry;
struct aplog_header_t header;


//////////////////// Proto's //////////////

// server

int aplogd_main(int argc, char **argv);
static int doserver(void);
static int parse_msg(int cfd, char *buf);
static int parse_klog_msg(char *buf);

// log hanling 
static int log_init(int maxentries);
static int log_add(unsigned long evid, char **part, int parts);

// event db functions
static int eventdb_read(char *fname);
int eventdb_get_msg(unsigned long eventid, char *to, int tolen);

// syslog & snmp things
static void send_remote_log (struct aplog_entry_t* log_entry);
static int syslog_send      (struct in_addr* dest, struct aplog_entry_t* log_entry);
static int snmp_trap_send   (struct in_addr* dest, struct aplog_entry_t* log_entry);

// SIGNAL functions
static void read_config(void);
static void flush_log(void);
static void dump_config(void);

// IFACE snmp trap functions
static void process_rtnetlink(int s);
static void ifname_cache_init(void);
static void ifname_cache_get(int ifindex, char* ifname);
static void ifname_cache_store(int ifindex, char* ifname);
static int ifname2snmpindex(char* ifname);

// helper funcs
static int open_serversocket(char *fname);
static int open_netlinksocket();
static int split(char *str, char sep, char **part, int maxparts);
static int split_whitespace(char *str, char **part, int maxparts);
static unsigned long hex2long(char *orgstr);


///////////////////////////////////////////////////////////////////
///////////////////// some C code /////////////////////////////////
///////////////////////////////////////////////////////////////////
static void sighup(int a)
{
	read_config_flag = 1;
}

static void sigusr1(int a)
{
	if(!remote_log) {
		remote_log = 1;
		if (config_read) {
			flush_log_flag = 1;
		}
	}
}

static void sigusr2(int a)
{
	dump_config_flag = 1;
}

int aplogd_main(int argc, char **argv) 
{
	int i;
	
	config.count_doubles=0;
	config.logsize = DEFLOGSIZE;
	config.verbose = 1;
	
	for(i = 0; i < NOF_FACILITY; i++) {
		config.faci_reporting[i] = 0xFF;
	}

	for(i = 0; i < NOF_PRIORITY; i++) {
		config.prio_where[i] = 0x00;
	}

	config.syslog_hosts = NULL;
	config.snmp_trap_hosts = NULL;

	eventdb_read(LOGDB);

	// Sigpipe is ignored in case the remote socket closes when 
	// we're writing
	signal(SIGPIPE, SIG_IGN);
	
	/* ignore SIGHUP during startup */
	signal(SIGHUP, SIG_IGN);
	
	/* initialize ifname_cache */
	ifname_cache_init();
	
	doserver();

	return(0);
}


/*
 * Start server. Open listening socket and accept clients.
 * Holds an array of client fd's, unused fd's are set to -1
 * Read a msg (string) from the client and send to parsgmsg.
 */
 
int doserver(void)
{
	int fd, len, i, nls;
	int r, cfd;
	char buf[1024];
	struct sockaddr_un unix_addr;
	fd_set fds;
	int maxfd = 0;
	int kmsgfd = -1;

	log_init(config.logsize);
	
	nls = open_netlinksocket();

	kmsgfd = open("/proc/kmsg", O_RDONLY);

	if(kmsgfd == -1) {
		fprintf(stderr, "Error opening /proc/kmsg : %s", strerror(errno));
	}

	fd = open_serversocket(SOCKFILE);
	listen(fd, 5);
	
	for(i=0; i<MAXCLIENTS;i++) clientfd[i]=-1;

	read_config();
	
	/* SIGHUP will trigger reading of aplog.conf */
	signal(SIGHUP, sighup);
	
	/* SIGUSR1 will flush the local log to the log hosts */
	signal(SIGUSR1, sigusr1);
	
	/* SIGUSR2 will dump the aplog config to stdout */
	signal(SIGUSR2, sigusr2);
	
	while(0x2A) {

		FD_ZERO(&fds);
		maxfd = 0;
		
		if(kmsgfd != -1) {
			FD_SET(kmsgfd, &fds);
			maxfd = kmsgfd;
		}

		FD_SET(nls, &fds);
		maxfd = MAX(maxfd, nls);
			
		FD_SET(fd, &fds);
		maxfd = MAX(maxfd, fd);
		
		for(i=0; i<MAXCLIENTS; i++) {
			if(clientfd[i]>0) {
				FD_SET(clientfd[i], &fds);
				maxfd = MAX(clientfd[i], maxfd);
			}
		}

		r=select(maxfd+1, &fds, NULL, NULL, NULL);
		
		if(r <= 0) {
			
			/* Select interrupted */
			
			if(read_config_flag) {
				read_config();
			}

			if(dump_config_flag) {
				dump_config();
			}

			if(flush_log_flag) {
				flush_log();
			}
			
		}
		else {
			
			if(FD_ISSET(nls, &fds)) {
				process_rtnetlink(nls);
			}

			if((kmsgfd != -1) && (FD_ISSET(kmsgfd, &fds))) {
				buf[0] = '0';
				r=read(kmsgfd, buf, sizeof(buf)-1);
				buf[r]=0;
				parse_klog_msg(buf);
			}

			if(FD_ISSET(fd, &fds)) {
				len = sizeof(unix_addr);
				cfd = accept(fd, (struct sockaddr *)&unix_addr, &len);

				if(cfd<0) {
					printf("Error accepting : %s\n", strerror(errno));
				} else {
					for(i=0; i<MAXCLIENTS; i++) {
						if(clientfd[i]==-1) {
							clientfd[i] = cfd;
							break;
						}
					}

					if(i == MAXCLIENTS) {
						printf("Too many clients.\n");
						close(cfd);
					}
				}
			}

			for(i=0; i<MAXCLIENTS; i++) {
				cfd = clientfd[i];

				if((cfd != -1) && (FD_ISSET(cfd, &fds))) {
					r=read(cfd, buf, sizeof(buf));
					if (r<sizeof(buf)) buf[r] = 0;

					parse_msg(cfd, buf);

					close(cfd);
					clientfd[i] = -1;
				}
			}
		}
	}

	// should never get here ...

	return(0);
}

/*
 * Parse incoming message from a client. This is a semicolon separated
 * string describing the request
 */

int parse_msg(int cfd, char *buf)
{
	char *part[8];
	int parts;

	parts = split(buf, ';', part, 8);
	
	if(parts>=1) {
		if(strcmp(part[0], "LOG")==0) {
			log_add(atoi(part[1]), part+2, parts-2);
		}
	
	}

	return(0);
}



/*
 * Parse message received from /proc/kmsg and match for 
 * strings from the klogmatch[] array. If those occur, log
 * them.
 */
 
int parse_klog_msg(char *buf)
{
	char *p1, *p2, *p3;
	char *part[2];
	int i, prio;

	i=0;

	p1=buf;

	while((p1 = strstr(p1, "<"))) {
	
		if(p1) {
			p2 = strchr(p1, '\n');
			if(p2) { 
				*p2=0;
			} else {
				break;
			}

            /* Special case for Firewall logging */
			if(strncmp(p1, "<FW>", 4) == 0) {
				p3 = p1 + 5;
				prio = (*p3) - '0';
				part[0] = p3 + 2; /* After '>' */
				switch(prio) {
					case 1:
						log_add(EV_NET_FIREWALL_PR1, part, 1);
						break;
					case 2:
						log_add(EV_NET_FIREWALL_PR2, part, 1);
						break;
					case 3:
						log_add(EV_NET_FIREWALL_PR3, part, 1);
						break;
					case 4:
						log_add(EV_NET_FIREWALL_PR4, part, 1);
						break;
					case 5:
						log_add(EV_NET_FIREWALL_PR5, part, 1);
						break;
					case 6:
						log_add(EV_NET_FIREWALL_PR6, part, 1);
						break;
					case 7:
						log_add(EV_NET_FIREWALL_PR7, part, 1);
						break;
					case 8:
						log_add(EV_NET_FIREWALL_PR8, part, 1);
						break;
					default:
						log_add(EV_SYS_KLOG_MSG_ERR, part, 1);
						break;
				}				
			} else if (p1[1] >= '0' && p1[1] <= '3') {
		                /* We log KERN_EMERG ("<0>") through to KERN_ERR ("<3>") */
				part[0] = p1;
                switch(p1[1] - '0') {
                    case 0:
                        log_add(EV_SYS_KLOG_MSG_EMERG, part, 1);
                        break;
                    case 1:
                        log_add(EV_SYS_KLOG_MSG_ALERT, part, 1);
                        break;
                    case 2:
                        log_add(EV_SYS_KLOG_MSG_CRIT, part, 1);
                        break;
                    case 3:
                        log_add(EV_SYS_KLOG_MSG_ERR, part, 1);
                        break;
                    default:
                        log_add(EV_SYS_KLOG_MSG_ERR, part, 1);
                        break;
                }
			}
			
			p1=p2+1;
		}
	}
	
	return(0);	

}

///////////////////////////////////////////////////////////////////
//////////// Server part code. Log handling etc ///////////////////
///////////////////////////////////////////////////////////////////


void adapt_log_size(void) {
	int old_size, new_size;
	int i;
	
	
	new_size = config.logsize;
	old_size = header.num_slots;
	
	if(old_size == new_size) {
		/* nothing to do */
		return;
	}
	
	if((new_size > header.current_number) || (old_size > header.current_number)) {
		
		header.num_slots = new_size;

		/* write new header */
		fseek(log_f, 0, 0);
		fwrite(&header, sizeof(struct aplog_header_t), 1, log_f);
		
		if(new_size > old_size) {
			

			/* Bigger size. We should write extra entries at 
			 * the end of the current file */
			
			bzero(&entry, sizeof(struct aplog_entry_t));
			
			fseek(log_f, (old_size-1) * (sizeof(struct aplog_entry_t)), SEEK_CUR);			
			for(i = old_size; i < new_size ; i++) {
				fwrite(&entry, sizeof(struct aplog_entry_t), 1, log_f);
			}
			
		}
		else {
			/* file is padded with extra entries. Just leave it.... */
		}
				
	}
	else {
		/* Too much hassle to adapt size, just erase it*/
		fclose(log_f);
		log_init(new_size);

	}
	
}

/* 
 * Initialize log file : reserve space for <maxentries> entries.
 */
 
int log_init(int maxentries)
{
	int i;
	int r;

	log_f = fopen(APLOGFILE, "w+");

	if(log_f == NULL) {
		fprintf(stderr, "Can't write to %s : %s\n", APLOGFILE, strerror(errno));
		exit(1);
	}

	header.current_number = 0;
	header.head_slot = 0;
	header.tail_slot = 0;
	header.num_slots = maxentries;

	bzero(&entry, sizeof(struct aplog_entry_t));

	fwrite(&header, sizeof(struct aplog_header_t), 1, log_f);

	for(i=0; i<maxentries; i++) {
		r=fwrite(&entry, sizeof(struct aplog_entry_t), 1, log_f);
	}

	return(0);
}


/* 
 * Add entry to log
 */

int log_add(unsigned long evid, char **part, int parts)
{
	int i;
	char buf[MAX_MSG_LEN];
	int fptr;
	int sum;
	static int prevsum=0;
	static unsigned int count=1;
	static time_t prevfirsttime=0;
	static unsigned long record_num=0;
	char *p;
	int prio;
	int faci;
	u_char prio_mask;

	//
	// Lookup the event messsage, or put in a generic for unknown events
	//

	i=eventdb_get_msg(evid, buf, sizeof(buf));
	if(i == -1) {
		snprintf(buf, sizeof(buf), "Unknown event %lx : (%%s) (%%s) (%%s) (%%s)\n", evid);
	}

	snprintf(entry.msg, sizeof(entry.msg), buf, part[0], part[1], part[2], part[3]);
	
	/* derive priority and facility from eventid */
	prio = (evid % 0xFF000000) >> 24;
	faci = (evid & 0x00FF0000) >> 16;
	
	if(prio > MAX_PRIORITY || faci > MAX_FACILITY) {
		/* Invalid event */
		return (0);
	}

	/* Check if we have to log messages with this event id.
	 * Note that 'faci_reporting' is stored in SNMP bitmask format!
	 */
	prio_mask = 1 << (NOF_PRIORITY - prio);
	
	if(!(prio_mask & config.faci_reporting[faci])) {
		/* For this facility, message with this priority are not logged */
		return (0);
	}
	
	// Calculate a simple checksum over the last event. If this sum
	// is equal to the last message added, don't add a new record to
	// the file but update the count and timestamp of last one

	p=entry.msg;
	sum=0;
	while(*p) sum += *(p++);

	if(sum == prevsum) {

		// This is a duplicate, update the last record

		count++;
	
//		printf("ft=%d\n", prevfirsttime);

		entry.firsttime = prevfirsttime;
		entry.lasttime = time(NULL);
		entry.count = count;
		entry.eventid = evid;
	
	} else {

		// This is a new one. Add a new record

		count=1;
		record_num ++;

		entry.firsttime = time(NULL);
		entry.lasttime = time(NULL);
		entry.count = count;
		entry.eventid = evid;

		// Advance the record number and write the header to file

		header.current_number = record_num;
		header.head_slot  = (header.head_slot+1) % header.num_slots;
		if(header.tail_slot == header.head_slot) {
			header.tail_slot = (header.tail_slot+1) % header.num_slots;
		}
	
		fseek(log_f, 0, 0);
		fwrite(&header, sizeof(header), 1, log_f);
	
	}

	prevfirsttime = entry.firsttime;
	prevsum = sum;

	entry.lasttime = time(NULL);
	entry.count = count;
	entry.eventid = evid;
	entry.number = record_num;


	// Calculate where to store this log entry, and write it to the file

	fptr  = sizeof(header);
	fptr += header.head_slot * sizeof(entry);

	fseek(log_f, fptr, 0);
	fwrite(&entry, sizeof(entry), 1, log_f);
/*
	printf("header : num=%d head=%d tail=%d num=%d\n",
			header.current_number,
			header.head_slot,
			header.tail_slot,
			header.num_slots);


	printf("%03d %03d %s\n", header.current_number, header.head_slot, entry.msg);
*/

	// leave this seek in, causes the last record to sync !!
	fseek(log_f, 0, 0); 

	fsync(fileno(log_f));

	if(remote_log) {
		send_remote_log(&entry);
	}
		
	return(0);
}


///////////////////////////////////////////////////////////////////
//////// Read configfile                ///////////////////////////
///////////////////////////////////////////////////////////////////
static void read_config(void) {
	FILE* fp;
	char buf[80];
	int parts;
	char *part[5];
	int prio, faci;
	int i;
	struct in_addr ipaddr;
	u_char type;
	unsigned int h;
	struct loghost_t *loghost_entry;
	struct loghost_t *loghost_next_entry;
	
	read_config_flag = 0;
		
	fp = fopen(CONFIG, "r");
	
	if (!fp) {
		/* configfile not found. No problem. Configfile
		 * will be generated by datastore and is absent 
		 * during startup */
		return;
	}
	
	/* Configfile found, cleanup old config */

	loghost_entry = config.syslog_hosts;
	config.syslog_hosts = NULL;
	while(loghost_entry) {
		loghost_next_entry = loghost_entry->next;
		free(loghost_entry);
		loghost_entry = loghost_next_entry;
	}
	loghost_entry = config.snmp_trap_hosts;
	config.snmp_trap_hosts = NULL;
	while(loghost_entry) {
		loghost_next_entry = loghost_entry->next;
		free(loghost_entry);
		loghost_entry = loghost_next_entry;
	}

	for(i = 0; i < NOF_FACILITY; i++) {
		config.faci_reporting[i] = 0x00;
	}

	for(i = 0; i < NOF_PRIORITY; i++) {
		config.prio_where[i] = 0x00;
	}
	
	/* Read configfile */
	while(fgets(buf, sizeof(buf), fp) > 0) {

		/* chop */
		buf[strlen(buf)-1] = '\0';
		
		if(buf[0] == '#') {
			/* comment line */
			continue;
		}
		
		parts = split(buf, ' ', part, 5);
		
		if(strcmp(part[0], "EventControl") == 0) {
			prio = atoi(part[1]);
			sscanf(part[2], "%02X", &h);
			config.prio_where[prio-1] = h;
		} 
		else if(strcmp(part[0], "Facility") == 0) {
			faci = atoi(part[1]);
			sscanf(part[2], "%02X", &h);
			config.faci_reporting[faci] = h;
		} 
		else if (strcmp(part[0], "TrapDestination") == 0) {
			inet_aton(part[1], &ipaddr);
			type = atoi(part[2]);
			
			if ( type & SYSLOG_HOST ) {
				loghost_entry = (struct loghost_t *) malloc(sizeof(struct loghost_t));
				inet_aton(part[1], &(loghost_entry->addr));
				loghost_entry->next = config.syslog_hosts;
				config.syslog_hosts = loghost_entry;
			}
			
			if ( type & SNMP_TRAP_HOST ) {
				loghost_entry = (struct loghost_t *) malloc(sizeof(struct loghost_t));
				inet_aton(part[1], &(loghost_entry->addr));
				loghost_entry->next = config.snmp_trap_hosts;
				config.snmp_trap_hosts = loghost_entry;
			}
		}
		else if (strcmp(part[0], "MaxTableLength") == 0) {
			if(part[1]) {
				config.logsize = atoi(part[1]);
				if ((config.logsize < 0) || (config.logsize > MAXLOGSIZE)) {
					fprintf(stderr, "Parse error in aplog.conf. MaxTableLength out of range\n");
					config.logsize = DEFLOGSIZE;
				}
			}
			else {
				fprintf(stderr, "Parse error in aplog.conf. Missing value after MaxTableLength\n");
			}
		}
		else {
			fprintf(stderr, "Parse error in aplog.conf (%s)\n", part[0]);
		}
	}

	fclose(fp);
	
	if(!config_read) {
		/* This was the first time we read our config */
		config_read = 1;
		
		/* check if remote log already is enabled */
		if (remote_log) {
			flush_log_flag = 1;
		}
	}
	
	adapt_log_size();
}




///////////////////////////////////////////////////////////////////
//////// Build and send a syslog packet ///////////////////////////
///////////////////////////////////////////////////////////////////

static int syslog_send(struct in_addr *dest, struct aplog_entry_t* log_entry)
{
	struct servent *sp;
	int sock = -1;
	struct sockaddr_in server;
	char buf[132];
	
	int priority = (log_entry->eventid & 0xFF000000) >> 24;
	int facility = (log_entry->eventid & 0x00FF0000) >> 16;
	
	sp=getservbyname("syslog", "udp");
	bzero((char *) &server, sizeof(server));		
	server.sin_family = AF_INET;
	server.sin_port = sp->s_port;
	server.sin_addr.s_addr = dest->s_addr;	
	bzero (&server.sin_zero, sizeof (server.sin_zero));
	
	if((sock=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
		fprintf (stderr, "Can't open socket : %s\n", strerror (errno));
		return (-1);
	}
	
	snprintf(buf, sizeof(buf), "<%d>%s", (facility << 3)+(priority-1), log_entry->msg);	

	sendto(sock, buf, strlen(buf), 0, 
		(struct sockaddr *)&server, sizeof(server));


	close(sock);
	return(0);
}



///////////////////////////////////////////////////////////////////
/////////////// Code for generating SNMP trap packets /////////////
///////////////////////////////////////////////////////////////////

/* 
 * SNMP Trap packet generation - I.D.
 *
 * Generates and sends a SNMP trap packet. this code
 * is quite quick and dirty and should be replaced by
 * ucd-snmp library calls as soon as ld.so support is fixed on
 * arm-nommu. For now, this will do.
 *
 */

#define INTEGER 	0x02
#define OCTETSTRING 	0x04
#define SEQUENCE 	0x30
#define OID 		0x06
#define NETWORKADDRESS 	0x40
#define UNSIGNED32      0x42
#define TIMETICKS 	0x43
#define TRAPDATA	0xA7
#define TRAPDATAV1      0xA4	


struct ber_rec_t {
	int type;
	unsigned char *ptr;
	int len;
};

struct ber_rec_t ber_rec[30];

unsigned char berbuf[2024];
unsigned char *berptr;
int berlvl;

static int ber_push(int type)
{
	*berptr = type;

	berlvl++;

	ber_rec[berlvl].type = type;
	ber_rec[berlvl].ptr = berptr;
	ber_rec[berlvl].len = -1; 	// not yet known
	
	berptr += 6; /* long form length */

	return(0);
}

static int ber_add(char *data)
{	
	long tmp;
	char *p;
	int a;
	char buf[64];
	
	switch(ber_rec[berlvl].type) {
		case INTEGER : 
		case UNSIGNED32 :
		case TIMETICKS :
			tmp = atol(data);
			if(tmp>=(2^24)) *berptr++ = (tmp>>24) % 256;
			if(tmp>=(2^16)) *berptr++ = (tmp>>16) % 256;
			if(tmp>=(2^8))  *berptr++ = (tmp>> 8) % 256;
					*berptr++ = tmp % 256;
			
			break;
	
		case OCTETSTRING :
			memcpy(berptr, data, strlen(data));
			berptr += strlen(data);
			break;
	
		case NETWORKADDRESS :
			inet_aton(data, (struct in_addr *)&tmp);
					*berptr++ = tmp % 256;
			if(tmp>=(2^8))  *berptr++ = (tmp>> 8) % 256;
			if(tmp>=(2^16)) *berptr++ = (tmp>>16) % 256;
			if(tmp>=(2^24)) *berptr++ = (tmp>>24) % 256;

			break;
	
		case OID :
			// OID encoding is a bit funny...

			strncpy(buf, data, sizeof(buf));

			a=atoi(strtok(buf, "."))*40;
			a+=atoi(strtok(NULL, "."));

			while(42) {
				if(a>127) {
					*berptr++ = (a>>7)|128;
					*berptr++ = (a%128);
				} else {
					*berptr++ = (a%128);
				}
		
				if( (p=(strtok(NULL, "."))) ==NULL) break;
				a=atoi(p);
			}
			
			break;
	
	/*	case TIMETICKS :
			*(unsigned long *)berptr = htonl(atoi(data));
			berptr += 4;
			break; */
	
		default :
			fprintf(stderr, "unknown type %d\n", ber_rec[berlvl].type);
			break;
	}

	return(0);
}

static int ber_pop(void)
{
	unsigned int len;

	len = (berptr - ber_rec[berlvl].ptr) - 6;
	
	*(ber_rec[berlvl].ptr+1) = 0x84; /* long format, 4 octets */
	*(ber_rec[berlvl].ptr+2) = ( len & 0xFF000000) >> 24;
	*(ber_rec[berlvl].ptr+3) = ( len & 0x00FF0000) >> 16;
	*(ber_rec[berlvl].ptr+4) = ( len & 0x0000FF00) >> 8;
	*(ber_rec[berlvl].ptr+5) = ( len & 0x000000FF) ;
	
	berlvl --;

	return(0);
}

		
static int snmp_genpkt(struct aplog_entry_t* log_entry)
{
	static unsigned int reqid;
	
	char intbuf[11]; /* can hold max integer (4294967296) */
	char oidbuf[64];
	int index = log_entry->number;
	int level = (log_entry->eventid & 0xFF000000) >> 24;
	int facility = (log_entry->eventid & 0x00FF0000) >> 16;
	int eventid = (log_entry->eventid & 0x0000FFFF);
		
	berptr = berbuf;
	berlvl = 0;

	ber_push(SEQUENCE);
	 ber_push(INTEGER); 
	  ber_add("1");				//version 1 == SNMPV2c
  	 ber_pop();
	 ber_push(OCTETSTRING); 
	  ber_add("public");
	 ber_pop();
	 ber_push(TRAPDATA);
	  ber_push(INTEGER);  // reqid
	   snprintf(intbuf, sizeof(intbuf), "%ud", reqid++);
	   ber_add(intbuf);
	  ber_pop();
	  ber_push(INTEGER);  // error-status
	   ber_add("0");
	  ber_pop();
	  ber_push(INTEGER);  // error-index
	   ber_add("0");
	  ber_pop(); 
	  ber_push(SEQUENCE);	// Var bindings
  	   
	   ber_push(SEQUENCE);	// SysUptime
	    ber_push(OID);
	     ber_add("1.3.6.1.2.1.1.3.0");
	    ber_pop();
	    ber_push(TIMETICKS);
	     snprintf(intbuf, sizeof(intbuf), "%ud", (unsigned int) log_entry->lasttime);
	     ber_add(intbuf);
	    ber_pop();
	   ber_pop(); /* SEQUENCE */ 
	  
	   ber_push(SEQUENCE); // snmpTrapOID
	    ber_push(OID);
	     ber_add("1.3.6.1.6.3.1.1.4.1.0");
	    ber_pop();
	    ber_push(OID);
	     ber_add("1.3.6.1.4.1.937.2.3.4.0.1");
	    ber_pop();
	   ber_pop(); // END snmpTrapOID
	   
	   ber_push(SEQUENCE); // eventReportLevel
	    ber_push(OID);
	     snprintf(oidbuf, sizeof(oidbuf), "1.3.6.1.4.1.937.2.3.3.4.1.5.%i", index);
	     ber_add(oidbuf);  
	    ber_pop();
	    ber_push(INTEGER);
	     snprintf(intbuf, sizeof(intbuf), "%ud", level);
	     ber_add(intbuf);
	    ber_pop();
	   ber_pop(); // END eventReportLevel
	   
   	   ber_push(SEQUENCE); // eventReportFacility
	    ber_push(OID);
	     snprintf(oidbuf, sizeof(oidbuf), "1.3.6.1.4.1.937.2.3.3.4.1.6.%i", index);
	     ber_add(oidbuf);
	    ber_pop();
	    ber_push(UNSIGNED32);
	     snprintf(intbuf, sizeof(intbuf), "%ud", facility);
	     ber_add(intbuf);
	    ber_pop();
	   ber_pop(); // END eventReportLevel

	   ber_push(SEQUENCE); // eventReportID
	    ber_push(OID);
	     snprintf(oidbuf, sizeof(oidbuf), "1.3.6.1.4.1.937.2.3.3.4.1.7.%i", index);
	     ber_add(oidbuf);
	    ber_pop();
	    ber_push(UNSIGNED32);
	     snprintf(intbuf, sizeof(intbuf), "%ud", eventid);
	     ber_add(intbuf);
	    ber_pop();
	   ber_pop(); // END eventReportID
	 
	   ber_push(SEQUENCE); // eventReportText
	    ber_push(OID);
	     snprintf(oidbuf, sizeof(oidbuf), "1.3.6.1.4.1.937.2.3.3.4.1.8.%i", index);
	     ber_add(oidbuf);
	    ber_pop();
	    ber_push(OCTETSTRING);
	     ber_add(log_entry->msg);
	    ber_pop();
	   ber_pop(); // END eventReportID
	   
	  ber_pop(); // Var bindings
	 ber_pop();
	ber_pop();

	return(0);
}

static int snmp_genv1pkt(struct aplog_entry_t* log_entry, int sock)
{
	char intbuf[11]; /* can hold max integer (4294967296) */
	char buf[64];
	static struct ifreq ifr;
	static struct sockaddr_in *p;

	// Fetch the IP address of the interface

	strncpy(ifr.ifr_name, "br0:0", sizeof(ifr.ifr_name));
	if ( ioctl(sock, SIOCGIFADDR, &ifr) < 0) {
		fprintf(stderr, "Error in snmpd_send_trap could not obtain IP address of iface br0:0\n");
		return 0;
	} else {
		p = (struct sockaddr_in *)&(ifr.ifr_addr);
	}
	
		
	berptr = berbuf;
	berlvl = 0;

	ber_push(SEQUENCE);
	  ber_push(INTEGER);
	    ber_add("0");				//version
	  ber_pop();
	  ber_push(OCTETSTRING);
	    ber_add("public");
	  ber_pop();
	  ber_push(TRAPDATAV1);
	    ber_push(OID);
	      ber_add(apinfo_get_SysObjectID() + 1); /* skip first dot in OID */
	    ber_pop();
	
	    ber_push(NETWORKADDRESS);
	      snprintf(buf, 16, "%d.%d.%d.%d",
	        (p->sin_addr.s_addr >>  0) & 0xff,
	        (p->sin_addr.s_addr >>  8) & 0xff,
	        (p->sin_addr.s_addr >> 16) & 0xff,
	        (p->sin_addr.s_addr >> 24) & 0xff);
	      ber_add(buf);
	    ber_pop();
	
	    ber_push(INTEGER);
	      switch(log_entry->eventid) {
		  case EV_SYS_BOOT_START:
			ber_add("0");
			break;
		  case EV_NET_IFACE_UP:
			ber_add("3");
			break;
		  case EV_NET_IFACE_DOWN:
			ber_add("2");
			break;
		  default:
			ber_add("7");
			break;
	      }	
	    ber_pop();
	
	    ber_push(INTEGER);
	      ber_add("0");			//specifictrap
	    ber_pop();

	    ber_push(TIMETICKS);
              snprintf(intbuf, sizeof(intbuf), "%ud", (unsigned int) log_entry->lasttime);
	      ber_add(intbuf);
	    ber_pop();
            
	    ber_push(SEQUENCE); /* Varbindings */
	    if((log_entry->eventid == EV_NET_IFACE_UP) 
			|| (log_entry->eventid == EV_NET_IFACE_DOWN)) {
		int ifindex;
		char* eventtext;
		char* part[3];

		eventtext = strdup(log_entry->msg);
		
		/* eventtext = "Interface <X> up/down". Second word is 
		   interface name */		
		split(eventtext, ' ', part, 3); 
		ifindex = ifname2snmpindex(part[1]);
		
		free(eventtext);
		
		snprintf(buf, sizeof(buf), ".1.3.6.1.2.1.2.2.1.1.%u", ifindex);
		snprintf(intbuf, sizeof(intbuf), "%u", ifindex);
		
		  ber_push(SEQUENCE);
		     ber_push(OID);
		       ber_add(buf);
		     ber_pop();
	
		     ber_push(INTEGER);
		       ber_add(intbuf);
		     ber_pop();	
	          ber_pop(); /* name/value */

	    }
  	    ber_pop(); /* SEQUENCE varbinding*/
	  ber_pop(); /* TRAPDATA */
	ber_pop(); /* SEQUENCE */

	return(0);

}




static int snmp_v1_trap_send(struct in_addr* dest, struct aplog_entry_t *log_entry) {
	struct servent *sp;
	int sock = -1;
	struct sockaddr_in server;

	sp=getservbyname("snmptrap", "udp");
	bzero((char *) &server, sizeof(server));		
	server.sin_family = AF_INET;
	server.sin_port = sp->s_port;
	server.sin_addr.s_addr = dest->s_addr;	
	bzero (&server.sin_zero, sizeof (server.sin_zero));
	
	if((sock=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
		fprintf (stderr, "Can't open socket : %s\n", strerror (errno));
		exit(0);
		return (-1);
	}
	
	snmp_genv1pkt(log_entry, sock);
	
	sendto(sock, berbuf, berptr-berbuf, 0, 
		(struct sockaddr *)&server, sizeof(server));
	
	close(sock);

	return(0);
}

static int snmp_trap_send(struct in_addr* dest, struct aplog_entry_t *log_entry)
{
	struct servent *sp;
	int sock = -1;
	struct sockaddr_in server;
	
	sp=getservbyname("snmptrap", "udp");
	bzero((char *) &server, sizeof(server));		
	server.sin_family = AF_INET;
	server.sin_port = sp->s_port;
	server.sin_addr.s_addr = dest->s_addr;	
	bzero (&server.sin_zero, sizeof (server.sin_zero));
	
	if((sock=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
		fprintf (stderr, "Can't open socket : %s\n", strerror (errno));
		exit(0);
		return (-1);
	}
	
	snmp_genpkt(log_entry);
	
	sendto(sock, berbuf, berptr-berbuf, 0, 
		(struct sockaddr *)&server, sizeof(server));
	
	close(sock);

	return(0);
}


/////////////////////////////////////////////////////////////////////
// Functions for reading the event database and handling the lists //
/////////////////////////////////////////////////////////////////////


/* 
 * This reads the event database into a linked list with filepointers
 * to the event messages.
 */

int eventdb_read(char *fname)
{
	char buf[256];
	char *p;
	char *part[5];
	int parts;
	char *r;
	int fptr;

	struct event_t *newe, *tmpe;

	eventdb_f=fopen(fname, "ro");

	if(eventdb_f==NULL) {
		printf("Error opening event database %s\n", fname);
		return(1);
	}

	while(42) {
		fptr = ftell(eventdb_f);
		r = fgets(buf, sizeof(buf), eventdb_f);
	
		if(r==NULL) break;

		p=strchr(buf, '#');
		if(p) *p=0;
		p=strchr(buf, '\n');
		if(p) *p=0;

		parts = split_whitespace(buf, part, 4);
		
		if(parts == 4) {
	
			if(strcmp(part[0], "ev")==0) {
				newe = (struct event_t *) malloc(sizeof(struct event_t));
				if(eventlist == NULL) {
					eventlist = newe;
					newe->next = NULL;
				} else {
					tmpe = eventlist;
					eventlist = newe;
					newe->next = tmpe;
				}
				newe->eventid = hex2long(part[1]);
				newe->fptr = fptr + part[3] - buf;
			}
		}
	}

	return(0);
}


/*
 * Get the event message text for the given event
 */

int eventdb_get_msg(unsigned long eventid, char *to, int tolen)
{
	struct event_t *ev = eventlist;
	char *p;

	while(ev) {
		if(ev->eventid == eventid) {
			fseek(eventdb_f, ev->fptr, 0);
			fgets(to, tolen, eventdb_f);
			p=strchr(to, '\n');
			if(p) *p=0;
			return(0);
		}	
		ev=ev->next;
	}

	return(-1);
}


///////////////////////////////////////////////////////////////////
///////////// Generic helper functions ////////////////////////////
///////////////////////////////////////////////////////////////////

/* 
 * Open the rt netlink socket 
 */
static int open_netlinksocket() 
{
	struct sockaddr_nl  local;
	int rtnl_fd;

    	if ((rtnl_fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE)) < 0) {
        	perror("Cannot open route netlink socket.\n");
        	return -1;
    	}

    	local.nl_family = AF_NETLINK;
    	local.nl_groups = RTMGRP_LINK;

    	if(bind(rtnl_fd, (struct sockaddr *)&local, sizeof(local)) < 0) {
        	perror("Cannot bind to route netlink socket");
        	goto err_out;
    	}

    	return rtnl_fd;;

err_out:
	close(rtnl_fd);
	return -1;
}

/*
 * Receive a route netlink packet and copy it to the buffer pointed to by nlmsg
 * On success, the length of the received message is returned, otherwise -1.
 */
static int rtnl_recv(int nls, struct nlmsghdr *nlmsg, int max_nlmsglen)
{
    int status;
    int nlmsglen;
    int payload_len;

    struct nlmsghdr *h;
    struct sockaddr_nl nladdr;
    struct iovec iov;

    struct msghdr msg = {
            (void*)&nladdr, sizeof(nladdr),
            &iov,	1,
            NULL,	0,
            0
    };

    memset(&nladdr, 0, sizeof(nladdr));
    nladdr.nl_family = AF_NETLINK;
    nladdr.nl_pid = 0;
    nladdr.nl_groups = 0;

    iov.iov_base = nlmsg;
    iov.iov_len = max_nlmsglen;

    memset(nlmsg, 0, max_nlmsglen);

retry:
    status = recvmsg(nls, &msg, 0);

    if (status < 0) {
            if (errno == EINTR)
                    goto retry;

            perror("Failed to receive route netlink msg");
            return -1;
    }
    
    if (status == 0) {
            printf("EOF on route netlink\n");
            return -1;
    }

    h = (struct nlmsghdr *)nlmsg;

    nlmsglen = h->nlmsg_len;
    payload_len = nlmsglen - sizeof(struct nlmsghdr);

#ifdef DEBUG
    printf("Netlink packet received of length %d, payload_len %d status = %d, seq = %d\n",
            nlmsglen, payload_len, status, h->nlmsg_seq);
#endif

    if (h->nlmsg_type == NLMSG_ERROR) {
            struct nlmsgerr *nlerr = (struct nlmsgerr *)NLMSG_DATA(h);
            if (payload_len < sizeof(struct nlmsgerr)) {
                    fprintf(stderr, "Netlink error truncated\n");
            } else {
                    errno = -nlerr->error;
                    if (errno == 0) {
                            return nlmsglen;
                    }
                    perror("netlink error");
            }
            return -1;
    }

    return nlmsglen;
}

static char *
idxtoname (unsigned int ifindex, char *ifname)
{
  struct ifreq ifr;
  int fd;

  fd = socket (AF_INET, SOCK_DGRAM, 0);

  if (fd < 0)
	return NULL;

   ifr.ifr_ifindex = ifindex;
   ioctl (fd, SIOCGIFNAME, &ifr);
   close (fd);
   
   return strncpy (ifname, ifr.ifr_name, IFNAMSIZ);
   
}

struct ifname_cache_t {
	int ifindex;
	char* ifname;
};

static struct ifname_cache_t *ifname_cache = NULL;

#define IFNAME_CACHE_SIZE 24 /* NO SNMP interfaces + 1 */

static void ifname_cache_init(void) {
	ifname_cache = calloc(IFNAME_CACHE_SIZE, sizeof(struct ifname_cache_t));
	if(!ifname_cache) {
		fprintf(stderr, "Cannot allocate ifname_cache\n");
		return;
	}
	
	ifname_cache[0].ifname  = strdup("unknown");
	ifname_cache[1].ifname  = strdup("lo");
	ifname_cache[2].ifname  = strdup("eth0");
	ifname_cache[3].ifname  = strdup("eth1");
	ifname_cache[4].ifname  = strdup("eth2");
	ifname_cache[5].ifname  = strdup("eth3");
	ifname_cache[6].ifname  = strdup("br0");
	ifname_cache[7].ifname  =  strdup("ppp0");
	ifname_cache[8].ifname  = strdup("eth1_0");
	ifname_cache[9].ifname  = strdup("eth1_1");
	ifname_cache[10].ifname = strdup("eth1_2");
	ifname_cache[11].ifname = strdup("eth1_3");
	ifname_cache[12].ifname = strdup("eth1_4");
	ifname_cache[13].ifname = strdup("eth1_5");
	ifname_cache[14].ifname = strdup("eth1_6");
	ifname_cache[15].ifname = strdup("eth1_7");
	ifname_cache[16].ifname = strdup("eth3_0");
	ifname_cache[17].ifname = strdup("eth3_1");
	ifname_cache[18].ifname = strdup("eth3_2");
	ifname_cache[19].ifname = strdup("eth3_3");
	ifname_cache[20].ifname = strdup("eth3_4");
	ifname_cache[21].ifname = strdup("eth3_5");
	ifname_cache[22].ifname = strdup("eth3_6");
	ifname_cache[23].ifname = strdup("eth3_7");
	
	return;	
	
}

static void ifname_cache_get(int ifindex, char* ifname) {
	int i;
	
	if(!ifname_cache) {
		fprintf(stderr, "Error, ifname_cache not initialised\n");
		goto ifname_cache_get_out;
	}
	
	for(i=0; i < IFNAME_CACHE_SIZE; i++) {
		if(ifindex == ifname_cache[i].ifindex) {
			strncpy(ifname, ifname_cache[i].ifname, IFNAMSIZ);
			return;
		}
	}
	
	fprintf(stderr, "Error, unknown interface get from ifname_cache\n");

ifname_cache_get_out:	
	ifname[0] = '\0';	
	return;
}

static void ifname_cache_store(int ifindex, char* ifname) {
	int i;
	
	if(!ifname_cache) {
		fprintf(stderr, "Error, ifname_cache not initialized\n");
		return;
	}
	
	for(i=0; i < IFNAME_CACHE_SIZE; i++) {
		if(strncmp(ifname_cache[i].ifname, ifname, IFNAMSIZ) == 0) {
			ifname_cache[i].ifindex = ifindex;
			return;
		}
	}
	
	fprintf(stderr, "Error, unknown interface store in ifname_cache\n");
	return;
}

static int ifname2snmpindex(char* ifname) {
	int i;
	
	if(!ifname_cache) {
		fprintf(stderr, "Error, ifname_cache not initialized\n");
		return -1;
	}
	
	for(i=0; i < IFNAME_CACHE_SIZE; i++) {
		if(strncmp(ifname_cache[i].ifname, ifname, IFNAMSIZ) == 0) {
			return i;
		}
	}
	
	return -1;
}

/* processes the netlink packet waiting in socket s and 
 * returns the interface which status was changed */
static void process_rtnetlink(int s) 
{
	int status, packetsize;
	struct nlmsghdr *packet;
	struct ifinfomsg *ifinfo;
	int ifindex;
	unsigned ifflags;
	unsigned ifchange;
	char ifname[IFNAMSIZ];
    
    	/* Packetsize = size of nlmsghdr + size of ifinfomsg */
    	packetsize = (NLMSG_SPACE(sizeof(struct ifinfomsg)));
    
    	packet = malloc(packetsize);
    	if(!packet) {
        	fprintf(stderr, "Cannot allocate NETLINK packet");
		return;
    	}
    
    	status = rtnl_recv(s, packet, packetsize);
    	if(status < 0) {
        	printf("netlink_recv failed");
		free(packet);
        	return;
    	}

    	ifinfo = NLMSG_DATA(packet);
    
	ifindex = ifinfo->ifi_index;
	ifflags = ifinfo->ifi_flags;
	ifchange = ifinfo->ifi_change;
	
	/* get ifname for ifindex */
	idxtoname(ifindex, ifname);
	
	/* Interface that are removed (i.e. WDS links, PPP) do not have
	 * interface names anymore in the kernel. To solve this problem
	 * the ifname_cache is implemented. */
	if(ifname[0] == '\0') {
		/* interface does not exist anymore. Find it in our
		 * internal administration */
		ifname_cache_get(ifindex, ifname);
	}
	else {
		/* store ifindex <-> ifname combination */
		ifname_cache_store(ifindex, ifname);
	}
	
	/* we're receiving more than 1 message if an interface status 
	 * changes. Process only the ones with the IFF_RUNNING flag set
	 * When a interface is created (i.e. a WDS link), a netlink
	 * message with ifi_flags = 0xFFFFFFFF is received. This 
	 * message is ignored to prevent a 'interface down' message
	 * just before the 'interface up' message
	 */
#ifdef DEBUG	
	printf("iff flags %X, %X, interface=(%s)\n", ifflags, ifchange, ifname);
#endif	
	if((ifchange & IFF_RUNNING) && (ifchange != 0xFFFFFFFF)) {
		char*  part[1];
		part[0] = ifname;
		if(ifflags & IFF_UP) {
			log_add(EV_NET_IFACE_UP, part, 1);
		}
		else {
			log_add(EV_NET_IFACE_DOWN, part, 1);
		}
	}
			
	free(packet);
	
	return;
	
}



/*
 * Open a unix domain socket
 */

static int open_serversocket(char *fname)
{
	int fd;
	struct sockaddr_un unix_addr;
	int r;

	fd = socket(AF_UNIX, SOCK_STREAM, 0);

	if(fd<0) {
		fprintf(stderr, "Error opening socket : %s\n", strerror(errno));
		exit(1);
	}

	unlink(fname);

	// Fill in unix_addr struct

	memset(&unix_addr, 0, sizeof(unix_addr));
	unix_addr.sun_family=AF_UNIX;
	strncpy(unix_addr.sun_path, fname, sizeof(unix_addr.sun_path));

	r = bind(fd, (struct sockaddr *)&unix_addr, sizeof(unix_addr));

	if(r<0) {
		fprintf(stderr, "Error connection to log daemon : %s\n", strerror(errno));
		exit(1);
	}

	return(fd);
}



/*
 * Split a string into seperate 0-terminated parts
 */ 

int split(char *str, char sep, char **part, int maxparts)
{
	int parts = 0;
	char *p1 = str;
	char *p2;

	part[parts++] = p1;

	while ((p2 = strchr (p1, sep)) && (parts < maxparts)) {
		*p2 = 0;
		p1 = p2 + 1;
		part[parts++] = p1;
	}

	return parts;
}



/*
 * Split a string, seperated by one ore more whitespace, in 0-term parts
 */
 
int split_whitespace(char *str, char **part, int maxparts)
{
	char *p = str;
	int parts=0;

	while(*p) {
		while( (*p == ' ') || (*p == '\t')) p++;
		part[parts]=p;
		parts++;
		if(parts>=maxparts) return(parts);
		while( (*p != ' ') && (*p != '\t') && (*p != 0)) p++;
		*p=0;
		p++;
	}

	return(parts);
}



/*
 * hex2long.... ??
 */

unsigned long hex2long(char *orgstr)
{
	unsigned long t=0;
	char *p;
	char str[9];

	p=strncpy(str, orgstr, sizeof(str));

	while(*p) {
		*p=*p|32;
		if(*p>='a') *p-=39;
		*p -= 48;
		if((*p)>15) return(0);
		t=(t<<4)|(*p++);
	}
	
	return(t);
}


static void send_remote_log(struct aplog_entry_t *log_entry) {
	struct loghost_t *loghost_entry;	
	int prio = (log_entry->eventid & 0xFF000000) >> 24;
	
	/* check if we have to syslog this event */
	if( config.prio_where[prio-1] & REPORTING_SYSLOG) {
		/* send entry to all configured syslog hosts */
		loghost_entry = config.syslog_hosts;
		while(loghost_entry) {
			syslog_send(&(loghost_entry->addr), log_entry);
			loghost_entry = loghost_entry->next;
		}
	}
	
	/* check if we have to snmp_trap this event */
	if( config.prio_where[prio-1] & REPORTING_SNMP_TRAP) {
		/* send entry to all configured snmp trap hosts */
		loghost_entry = config.snmp_trap_hosts;
		while(loghost_entry) {
			snmp_trap_send(&(loghost_entry->addr), log_entry);
			loghost_entry = loghost_entry->next;
		}
	}
	
	/* Check if this is an event for which we should send snmpv1 traps */
	if((log_entry->eventid == EV_NET_IFACE_UP) 
		|| (log_entry->eventid == EV_NET_IFACE_DOWN)
		|| (log_entry->eventid == EV_SYS_BOOT_START)) {
		/* send this entry to all configured snmp trap hosts 
		   as snmpv1 trap */
		loghost_entry = config.snmp_trap_hosts;
		while(loghost_entry) {
			snmp_v1_trap_send(&(loghost_entry->addr), log_entry);
			loghost_entry = loghost_entry->next;
		}
	}
}
	
static void flush_log() {

	unsigned long curnum;
	int max;
	long index;
	static FILE *f;
	int entrynum;
	int fptr;
	struct aplog_entry_t  log_entry;
	struct aplog_header_t log_header;	

	flush_log_flag = 0;

	f=fopen(APLOGFILE, "r");

	if(f==NULL) {
		fprintf(stderr, "Error opening %s : %s", APLOGFILE, strerror(errno));
		exit(1);
	}

	fread(&log_header, sizeof(log_header), 1, f);
	
	curnum  = log_header.current_number;
	max     = log_header.num_slots;
	

	index=curnum-max+1;
	if(index<1) index=1;	
	
	for(; index<=curnum; index++) {
		entrynum = (index % log_header.num_slots);
		fptr = sizeof(log_header) + entrynum * sizeof(log_entry);
	
		fseek(f, fptr, 0);
		fread(&log_entry, sizeof(log_entry), 1, f);
		
		send_remote_log(&log_entry);
	}


	fclose(f);
	
}

static void dump_config() {
	int i;
	struct loghost_t* entry;

	dump_config_flag = 0;
			
	fprintf(stderr, "APLOG config dump\n");
	fprintf(stderr, "=======================\n");
	
	fprintf(stderr, "count_doubles = %i\n", config.count_doubles);
	fprintf(stderr, "logsize       = %i\n", config.logsize);
	fprintf(stderr, "verbose       = %i\n", config.verbose);

	fprintf(stderr, "=======================\n");
	
	for(i = 0; i < NOF_FACILITY; i++) {
		fprintf(stderr, "Facility %i : %02X\n", i, config.faci_reporting[i]);
	}
	
	fprintf(stderr, "=======================\n");

	for(i = 0; i < NOF_PRIORITY; i++) {
		fprintf(stderr, "Priority %i : %02X\n", i, config.prio_where[i]);
	}
	
	fprintf(stderr, "=======================\n");
	
	fprintf(stderr, "SYSLOG hosts:\n");
	entry = config.syslog_hosts;
	while(entry) {
		fprintf(stderr, "ip: %s\n", inet_ntoa(entry->addr));
		entry = entry->next;
	}

	fprintf(stderr, "=======================\n");
	
	fprintf(stderr, "SNMP_TRAP hosts:\n");
	entry = config.snmp_trap_hosts;
	while(entry) {
		fprintf(stderr, "ip: %s\n", inet_ntoa(entry->addr));
		entry = entry->next;
	}

	fprintf(stderr, "=====END DUMP===========\n");	
}

// eof..
