/*  $Header$
 *
 *  Copyright (C) 2002 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  CONFIGURATION FILE PARSER
 *
 *  The config file parser checks for each token in the default config file
 *  if it is present in the user config file. If all token are found in the
 *  user configuration file, this file is up to date and no further action is taken.
 *  If a token is not found in the user configuration file, the default config
 *  file is newer and the two are merged.
 *  Merging is done by concatenating the default and the user config file. The
 *  SNMP agent will then automatically perform the merge for us, because it
 *  parses all tokens in the config file and when the same token is present
 *  multiple times, it uses the last occurrence.
 */
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <syslog.h>
/* Inl lib functions */
#include <inl_debug.h>
#include <confparse.h>
#include "aplog.h"

/* Define this for debugging output */
#define D(x)

/*
 * Maximum total path length, including filename.
 */
#define MAX_PATH 512

/*
 * Maximum number of tokens in a config file and the max Token length
 */
#define MAX_LINE_LEN	128
#define MAX_TOKENS	(2048 * 4)

/*
 * Parsed token structures.
 * A line in a config file (token) has the following format:
 * "[TokenIdentifier][spaces or \t][TokenValue][\n]"
 * Comments start with a "#".
 */
typedef struct
{
	int nrTokens;                 /* The number of tokens curently in the structure */
	char *Tokens[MAX_TOKENS];
} TokenStruct;

static TokenStruct targettokens;

extern void set_malloc_mmap_size(size_t size);

static void init_token_struct(TokenStruct *tokens)
{
	int i;

	for(i = 0; i < MAX_TOKENS; i++)
		tokens->Tokens[i] = NULL;
	tokens->nrTokens = 0;
}

/*****************************************************************************
 * NAME:
 *	compare_tokens
 *
 * DESCRIPTION:
 *	compare_tokens searches the tokenlist 'tokens' for a matching token
 *	identifier of the token that is passed in 'buffer'.
 * PARAMETERS:
 *	TokenStruct *tokens	: tokenlist
 *	const char *buffer	: token to find
 * RETURN:
 *	 0 if match is found
 *	-1 if no match is found
 *****************************************************************************/
static int compare_tokens(TokenStruct *tokens, const char *buffer)
{
	int length, index = 0;

	/* get length of token identifier */
	length = strcspn(buffer, " \t");
	while (index < tokens->nrTokens) {
		if(strncmp(tokens->Tokens[index], buffer, length) == 0)
			return 0;
		index++;
	}
	return -1;
}

static inline int iscomment(const char *string)
{
	int comment = 0;
	char *c = (char *)string;

	while(!comment &&
	      ((*c == '#') || (*c == ' ')) && *c) {
		if(*c == '#')
			comment = 1;
		c++;
	}
	return comment;
}

/*****************************************************************************
 * NAME:
 *	read_config_file
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 *	char *filename		: name of file to read
 *	TokenStruct *tokens	: tokenlist for storing the tokens read
 *
 * RETURN:
 *	 0 on success
 *	-1 on failure, tokens->nrTokens holds the number of lines that
 *	   were read.
 *****************************************************************************/
static int read_config_file(char *filename, TokenStruct* tokens)
{
	int retval = 0, len = 0;
	FILE *fp;
	static char buffer[MAX_LINE_LEN];

	if((fp = fopen(filename, "r")) == NULL) {
		return -1;      /* failure */
	}

	tokens->nrTokens = 0;
	/* parse all lines from the configuration file */
	while(!feof( fp ) && !ferror(fp) && !retval) {
		if(tokens->nrTokens == MAX_TOKENS) {
			error("Too many tokens %d in config file '%s'",
				MAX_TOKENS, filename );
			retval = -1;
		} else {
			if(!fgets(buffer, MAX_LINE_LEN, fp)) {
				continue;
			}
			if(iscomment(buffer)) {
				continue;
			}
			len = strlen(buffer) + 1;
			printf("buffer len is %d\n", len);
			if (len > MAX_LINE_LEN)
			    len = MAX_LINE_LEN;

			if(!tokens->Tokens[tokens->nrTokens])
				tokens->Tokens[tokens->nrTokens] = (char *)malloc(len);
				if (tokens->Tokens[tokens->nrTokens] == NULL)
				{
				   error("Malloc of token failed\n");
				   fclose(fp);
				   return retval;
			    	}

			if(tokens->Tokens[tokens->nrTokens]) {
				strncpy(tokens->Tokens[tokens->nrTokens], buffer, len);
				tokens->nrTokens++;
			} else {
				error("Out of memory while reading config file '%s'", filename );
				retval = -1;	/* Out of Memory */
			}
		}
	}
	fclose(fp);
	return retval;
}

/*****************************************************************************
 * NAME:
 *	filecopy
 *
 * DESCRIPTION:
 *	copy sourcefile to destinationfile. The destination file will be
 *	overwritten if it exists.
 * PARAMETERS:
 *	const char *source	: source filename
 *	const char *dest	: destination filename
 * RETURN:
 *	 0 on success
 *	-1 on failure
 *****************************************************************************/
static int filecopy(const char *source, const char *dest)
{
    FILE *sourcefile, *destfile;
    int c;

    if(( sourcefile = fopen( source, "r" )) == NULL) {
        error("filecopy: Can't open source file '%s'", source );
        return -1;      /* failure */
    }
    if(( destfile = fopen( dest, "w" )) == NULL) {
        error("filecopy Can't open destination file '%s'", dest );
		fclose(sourcefile);
        return -1;      /* failure */
    }

	while((c = getc(sourcefile)) != EOF) {
		putc(c, destfile);
	}

	fclose(sourcefile);
	fclose(destfile);
	return 0;
}

/*****************************************************************************
 * NAME:
 *	write_config_file
 *
 * DESCRIPTION:
 *	Write tokens from the file "sourcefilename" concatenated
 * 	with the tokens in "tokens" to a file "targetfilename"
 * PARAMETERS:
 *	const char *targetfilename	: file that is written to
 *	const char *sourcefilename	: file to read tokens from
 *	TokenStruct *tokens		: tokenlist
 * RETURN:
 *	  0 on success
 *	 <0 on Error.
 *****************************************************************************/
static int write_config_file(const char *targetfilename, const char *sourcefilename,
				TokenStruct *tokens)
{
	FILE *sourcefile, *destfile;
	int index = 0, retval = 0, c;

	if(( sourcefile = fopen(sourcefilename, "r" )) == NULL) {
		error("Can't open source file '%s'", sourcefilename);
		return -1;      /* failure */
	}
	if(( destfile = fopen(targetfilename, "w")) == NULL) {
		error("Can't open destination config file for writing '%s'", targetfilename);
		return -1;      /* failure */
	}
	/* First copy the default config file to the user config file */
	while((c = getc(sourcefile)) != EOF) {
		putc(c, destfile);
	}
	/* Write out an extra newline, just in case */
	if(!retval)
		retval = fputs("\n", destfile);

	/* Next, concatenate the contents of the user config file */
	while ( (retval >= 0) && (index < tokens->nrTokens) ) {
		retval = fputs(tokens->Tokens[index], destfile);
		index++;
	}
	fclose(sourcefile);
	fclose(destfile);
	return retval;
}

static void mkfullpath(char *full_path, const char *path, const char *name)
{
	strcpy(full_path, path);
	strcat(full_path, "/");
	strcat(full_path, name);
}

/*
 */
/*****************************************************************************
 * NAME:
 *	merge_config_files
 *
 * DESCRIPTION:
 *	This is the core of the config file parser. It compares the default
 *	and user config file and merges these files if necessary.
 *
 * PARAMETERS:
 *	char *defaultfilename	: default configuration file name
 *	char *userfilename	: user configuration file name
 * RETURN:
 *	  0 on success
 *	<>0 on Error.
 *****************************************************************************/
static int merge_config_files(char *defaultfilename, char *userfilename)
{
	int length;
	FILE *fp;
	char buffer[MAX_LINE_LEN];
	int written = 0, retval = 0;

	if((retval = read_config_file(userfilename, &targettokens)) < 0) {
		error("Reading user config file '%s' failed, error %d",
			userfilename, retval);
		return -1;
	}

	if((fp = fopen(defaultfilename, "r")) == NULL) {
		error("Can't open default config file '%s' for merging", defaultfilename);
		return -1;      /* failure */
	}

	/* parse all lines from the default configuration file */
	while(!feof( fp ) && !ferror(fp) && !written) {
		length = MAX_LINE_LEN;
		read_text_line(buffer, &length, fp);
		/* get token identifier */
		if (strtok(buffer, "#") == NULL)
			continue;

		if(compare_tokens(&targettokens, buffer) != 0) {
			/* Token not found, write out concatenated default
			 * and user config file
			 */
			aplog(EV_MERGE_NEW_DEFAULTS, defaultfilename, NULL);
			if((retval = write_config_file(userfilename, defaultfilename, &targettokens)) < 0)
				error("Could not write merged configuration file %s, error %d",
					userfilename, retval);
			written = 1;
		}
	}

	while (targettokens.nrTokens > 0)
	{
	    free(targettokens.Tokens[targettokens.nrTokens]);
	    targettokens.nrTokens--;
        }
	init_token_struct(&targettokens);

	fclose(fp);
	return retval;
}

/*
 * Main function navigates through all the default config files.
 * in order to find posible problems in the user config files.
 * Returns: Zero is succesfull, <0 on error.
 */
int main(int argc, char **argv)
{
	struct dirent 	*dp, *usr_dp;
	DIR	*dfd, *usr_dfd;
	int user_conf_dir_found = 1, user_config_file_found = 0;
	char *defconfig_path, *userconfig_path;
	char defaultfilename[MAX_PATH], userfilename[MAX_PATH];

	openlog ("confparser", LOG_PID, LOG_LOCAL0);

	if(argc != 3) {
		error("Usage: %s [default configdir] [user configdir]\n", argv[0]);
		return -1;
	}

	set_malloc_mmap_size(64*1024);
	defconfig_path = argv[1];
	userconfig_path = argv[2];
	init_token_struct(&targettokens);

	note("Start parsing config files, to detect possible errors");

	/* Open defconfig_path dir */
	if ((dfd = opendir(defconfig_path)) == NULL) {
		error("Error can't open default directory %s", defconfig_path);
		return -1;
	}

	/* Check if user config dir exists */
	if ((usr_dfd = opendir(userconfig_path)) == NULL) {
		user_conf_dir_found = 0;
		mkdir(userconfig_path, 0666);
		aplog(EV_NO_USER_CONFIG, NULL);
	}
	else
		closedir(usr_dfd);

	while ((dp = readdir(dfd)) != NULL)
	{
		D(printf("Read files in dir --%s--\n",dp->d_name));
		if (strcmp(dp->d_name, ".") == 0 ||
			strcmp(dp->d_name, "..") == 0)
				continue;

		if (strlen(defconfig_path)+strlen(dp->d_name)+2 > sizeof(defaultfilename)) {
			error ("name %s/%s too long, file skipped", defconfig_path, dp->d_name);
			continue;
		}

		D(fprintf (stderr, "confparse: Found file %s/%s\n", defconfig_path, dp->d_name));

		mkfullpath(defaultfilename, defconfig_path, dp->d_name);

		if (user_conf_dir_found == 0) {
			/*
			 * If the userconfig_path dir isn't found just copy all conf files.
			 */
			note("User configuration file %s not found, using default file", dp->d_name);
			mkfullpath(userfilename, userconfig_path, dp->d_name);
			filecopy (defaultfilename, userfilename);
			continue;
		}
		/* Open userconfig_path dir */
		if ((usr_dfd = opendir(userconfig_path)) == NULL) {
			error("Can't open %s", userconfig_path);
			closedir(dfd);
			return -1;
		}
		while (((usr_dp = readdir(usr_dfd)) != NULL) &&
			!user_config_file_found) {
			if (strcmp(usr_dp->d_name, dp->d_name) == 0)
			{
				user_config_file_found = 1;
		        }
		        printf("user ocnfig file not yet found\n");
		}
		mkfullpath(userfilename, userconfig_path, dp->d_name);

		if (user_config_file_found) {
			D(fprintf (stderr, "Merging config files %s %s\n",
				defaultfilename, userfilename));
			if (merge_config_files(defaultfilename, userfilename) < 0) {
				note("Merging configuration files failed, using default file", usr_dp->d_name);
				aplog(EV_MERGE_FAILED, usr_dp->d_name, NULL);
				filecopy(defaultfilename, userfilename);
			}
		} else {
			note("User configuration file %s not found, using default file", dp->d_name);
			filecopy(defaultfilename, userfilename);
			aplog(EV_NO_USER_CONFIGFILE, dp->d_name, NULL);
		}
		user_config_file_found = 0;
		closedir(usr_dfd);
	} /* End while ((dp = readdir(dfd)) != NULL) */

	closedir(dfd);
	note("Finished parsing config files");
	return 0;
}

