#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/ether.h>
#include <getopt.h>
#include "../include/ebtables_u.h"
#include <linux/netfilter_bridge/ebt_macnat.h>

#define NAT_S_BRIF '1'
#define NAT_S_TARGET '2'
#define NAT_S_BOOTPMETHOD '3'
#define NAT_D_BRIF '1'
#define NAT_D_TARGET '2'

static struct option opts_macs[] =
{
	{ "snat-brif"     , required_argument, 0, NAT_S_BRIF },
	{ "snat-target"   , required_argument, 0, NAT_S_TARGET },
	{ "alg_bootp", required_argument, 0, NAT_S_BOOTPMETHOD},
	{ 0 }
};

static struct option opts_macd[] =
{
	{ "dnat-brif"     , required_argument, 0, NAT_D_BRIF},
	{ "dnat-target"   , required_argument, 0, NAT_D_TARGET },
	{ 0 }
};

static void print_help_s()
{
	printf(
	"macsnat options:\n"
	" --snat-brif ifname       : bridge interface's name, ex: br0\n"
	" --snat-target target   : ACCEPT, DROP, RETURN or CONTINUE\n"
	" --alg_bootp method: broadcast or relay\n");
	
}

static void print_help_d()
{
	printf(
	"macdnat options: \n"
	" --dnat-brif ifname       : bridge interface's name, ex: br0\n"
	" --dnat-target target   : ACCEPT, DROP, RETURN or CONTINUE\n");
}

static void init_s(struct ebt_entry_target *target)
{
	struct ebt_macnat_info *natinfo = (struct ebt_macnat_info *)target->data;
	
	strcpy(natinfo->brifname, "br0");
	natinfo->target = EBT_ACCEPT;
	natinfo->alg_bootp_method = EBT_ALG_BOOTP_RELAY;
	return;
}

static void init_d(struct ebt_entry_target *target)
{
	struct ebt_macnat_info *natinfo = (struct ebt_macnat_info *)target->data;

	strcpy(natinfo->brifname, "br0");
	natinfo->target = EBT_ACCEPT;
	return;
}

#define OPT_SNAT_BRIF    0x01
#define OPT_SNAT_TARGET  0x02
#define OPT_SNAT_BOOTP	 0x04
static int parse_s(int c, char **argv, int argc,
   const struct ebt_u_entry *entry, unsigned int *flags,
   struct ebt_entry_target **target)
{
	struct ebt_macnat_info *natinfo = (struct ebt_macnat_info *)(*target)->data;

	switch (c) {
	case NAT_S_BRIF:
		check_option(flags, OPT_SNAT_BRIF);
		strncpy(natinfo->brifname, optarg, sizeof(natinfo->brifname)-1);
/*		if (!(addr = ether_aton(optarg)))
			print_error("Problem with specified --to-source mac");
		memcpy(natinfo->mac, addr, ETH_ALEN); */
		break;
	case NAT_S_TARGET:
		check_option(flags, OPT_SNAT_TARGET);
		if (FILL_TARGET(optarg, natinfo->target))
			print_error("Illegal --snat-target target");
		break;
	case NAT_S_BOOTPMETHOD:
		check_option(flags, OPT_SNAT_BOOTP);
		if(strcmp(optarg, "broadcast")==0)
			natinfo->alg_bootp_method = EBT_ALG_BOOTP_BROADCAST;
		else if(strcmp(optarg, "relay")==0)
			natinfo->alg_bootp_method = EBT_ALG_BOOTP_RELAY;
		else
			natinfo->alg_bootp_method = EBT_ALG_BOOTP_RELAY;
		break;
			
	default:
		return 0;
	}
	return 1;
}

#define OPT_DNAT_BRIF	0x01
#define OPT_DNAT_TARGET 0x02

static int parse_d(int c, char **argv, int argc,
   const struct ebt_u_entry *entry, unsigned int *flags,
   struct ebt_entry_target **target)
{
	struct ebt_macnat_info *natinfo = (struct ebt_macnat_info *)(*target)->data;
	
	switch(c){
	case NAT_D_BRIF:
		check_option(flags, OPT_DNAT_BRIF);
		strncpy(natinfo->brifname, optarg, sizeof(natinfo->brifname)-1);
		break;
	case NAT_D_TARGET:
		check_option(flags, OPT_DNAT_TARGET);
		if (FILL_TARGET(optarg, natinfo->target))
			print_error("Illegal --dnat-target target");
		break;
	default:
		return 0;
	}
	return 1;
}

static void final_check_s(const struct ebt_u_entry *entry,
   const struct ebt_entry_target *target, const char *name,
   unsigned int hookmask, unsigned int time)
{
	struct ebt_macnat_info *natinfo = (struct ebt_macnat_info *)target->data;

	if (BASE_CHAIN && natinfo->target == EBT_RETURN)
		print_error("--snat-target RETURN not allowed on base chain");
	CLEAR_BASE_CHAIN_BIT;
	if ((hookmask & ~(1 << NF_BR_POST_ROUTING)) || strcmp(name, "nat"))
		print_error("Wrong chain for snat");
}

static void final_check_d(const struct ebt_u_entry *entry,
   const struct ebt_entry_target *target, const char *name,
   unsigned int hookmask, unsigned int time)
{
	struct ebt_macnat_info *natinfo = (struct ebt_macnat_info *)target->data;

	if (BASE_CHAIN && natinfo->target == EBT_RETURN)
		print_error("--dnat-target RETURN not allowed on base chain");
	CLEAR_BASE_CHAIN_BIT;
	if (((hookmask & ~((1 << NF_BR_PRE_ROUTING) | (1 << NF_BR_LOCAL_OUT)))
	   || strcmp(name, "nat")) &&
	   ((hookmask & ~(1 << NF_BR_BROUTING)) || strcmp(name, "broute")))
		print_error("Wrong chain for dnat");
}

static void print_s(const struct ebt_u_entry *entry,
   const struct ebt_entry_target *target)
{
	struct ebt_macnat_info *natinfo = (struct ebt_macnat_info *)target->data;

	printf("--snat-brif ");
	printf("%s", natinfo->brifname);
	printf(" --alg_bootp %s", 
		natinfo->alg_bootp_method==EBT_ALG_BOOTP_BROADCAST?
		"broadcast": "relay");
	printf(" --snat-target %s", TARGET_NAME(natinfo->target));
}

static void print_d(const struct ebt_u_entry *entry,
   const struct ebt_entry_target *target)
{
	struct ebt_macnat_info *natinfo = (struct ebt_macnat_info *)target->data;
	
	printf("--dnat-brif ");
	printf("%s", natinfo->brifname);
	printf(" --dnat-target %s", TARGET_NAME(natinfo->target));
}

static int compare(const struct ebt_entry_target *t1,
   const struct ebt_entry_target *t2)
{
	struct ebt_macnat_info *natinfo1 = (struct ebt_macnat_info *)t1->data;
	struct ebt_macnat_info *natinfo2 = (struct ebt_macnat_info *)t2->data;


	return !memcmp(natinfo1, natinfo2, sizeof(struct ebt_macnat_info)) ;
}

static struct ebt_u_target macsnat_target =
{
	EBT_MACSNAT_TARGET,
	sizeof(struct ebt_macnat_info),
	print_help_s,
	init_s,
	parse_s,
	final_check_s,
	print_s,
	compare,
	opts_macs,
};

static struct ebt_u_target macdnat_target =
{
	EBT_MACDNAT_TARGET,
	sizeof(struct ebt_macnat_info),
	print_help_d,
	init_d,
	parse_d,
	final_check_d,
	print_d,
	compare,
	opts_macd
};

//static void _init(void) __attribute__ ((constructor));
//static void _init(void)
void _init_macnat(void)
{
	//printf("Register MACNAT(source NAT and destination NAT) target\n");
	register_target(&macsnat_target);
	register_target(&macdnat_target);
}
