/*
 * flash.c - flash handling for installing an image
 *
 * Copyright (C) 2002 by Intersil Corp.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <linux/mtd/mtd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>

#include <identification/identification.h>
#include <identification/image_header.h>
#include <inl_debug.h>
#include <sockutil.h>

#include "imginstall.h"

#ifndef offsetof
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

  /* global declarations */
extern char *sector_buf;

static int flash_install_image(int fd, const char *device, int start_sect, int status_offset, int size);
static int mtd_write_loc(int fd, int offset,const void *buf,int count);


int flash_install(int fd, int type, int size)
{
   switch (type) {
      case image_apfirmware:
         return flash_install_image(fd, MTD_APFIRMWARE, 0,
                              STATUS_INSTALLING_APFIRMWARE, size);
      default:
         warn ("flash_install: invalid image type %d", type);
         return -EINVAL;
   }
}
/*-----------------------------------------------------------------------------*
 | NAME:
 |   flash_install_image
 |
 | DESCRIPTION:
 |   Install an image in the given device.
 |   The following steps are taken:
 |   1. clear OKAY from new image (invalidate it)
 |   2. program new image
 |   3. make new image valid (write OKAY)
 |
 | PARAMETERS:
 |   int fd                 - IN  - filehandle to the image
 |   const char * device    - IN  - name of the device on which the system
 |                                  must be installed
 |   int  start_sect        - IN  - destination sector for image
 |   int size               - IN  - size of image, in bytes
 |
 | RETURN:
 |    0 - success
 |   <0 - failure:
 |         -ENOSPC if image does not fit in flash
 |         -EINVAL upon other failures
 *-----------------------------------------------------------------------------*/
static int flash_install_image(int fd, const char *device, int start_sect, int status_offset, int size)
{
   erase_info_t eraseinfo;
   int          devfd            = -1;
   int          rc               =  0;
   int          sectorcount      =  0;
   int          offset;
   off_t        seekpos, flashsize;
   int          sectors_needed   =  0;
   int          percentage       =  0;
   int          totalread        =  0;
   u_long       magic1;

   if ((devfd = open( device, O_RDWR )) == -1) {
      warn( "Failed to open device %s : %m", device );
      return -EINVAL;
   }

   flashsize = lseek( devfd, 0, SEEK_END);
   if (flashsize < 0) {
      warn( "lseek to end failed : %m" );
      return -EINVAL;
   }

   if ( ((start_sect * SECTOR_SIZE) + size) > flashsize) {
      warn( "image (%d) too big for flash (%d)\n", size, flashsize);
      return -ENOSPC;
   }

   do {

      sectors_needed = (size / SECTOR_SIZE) + 1;

      seekpos = start_sect * SECTOR_SIZE;
      if (lseek( devfd, seekpos, SEEK_SET) != seekpos) {
         warn( "lseek failed : %m" );
         rc = -EINVAL;
         break;
      }

      while (totalread < size) {

         int bytesread;

         note("Programming sector %d", sectorcount);
         memset(sector_buf, '\xFF', SECTOR_SIZE);

         if (totalread < (size - SECTOR_SIZE)) {
            bytesread = readn(fd, sector_buf, SECTOR_SIZE);
         } else {
            bytesread = readn(fd, sector_buf, (size - totalread));
         }

         if (bytesread == 0) {
            continue;
         }

         if (bytesread < 0) {
            warn("Read failed during programming on fd %d %m", fd);
            rc = -EINVAL;
            break;
         }

         totalread += bytesread;
         percentage = (100 * totalread) / size;
         imginstall_write_status( status_offset + percentage);

         /* 1. clear the OKAY */

         if (sectorcount == 0 ) {
            memset(sector_buf, '\xFF', 4);
         }

         eraseinfo.start = (start_sect + sectorcount) * SECTOR_SIZE;
         eraseinfo.length = SECTOR_SIZE;

         if (ioctl( devfd, MEMERASE, &eraseinfo ) != 0) {
            warn( "Flash erase of device %s failed : %m", device );
            rc = -EINVAL;
            break;
         }

         /* 2. program image in flash */

         if (writen( devfd, sector_buf, bytesread ) < 0) {
            warn( "Failed to write to flash : %m");
            rc = -EINVAL;
            break;
         }
         sectorcount++;
      }

      if (rc) {
         break;
      }

      /* 3. validate new image */
      magic1 = MAGIC_OKAY;
      offset = start_sect * SECTOR_SIZE + offsetof(image_header_t, magic1);
      if (mtd_write_loc( devfd, offset, &magic1, sizeof(magic1)) < 0) {
         rc = -EINVAL;
         break;
      }

   } while( 0 );

   if( devfd != -1 ) {
      close( devfd );
   }

   return rc;
}

/*---------------------------------*
 | Write a buffer to an MTD device |
 *---------------------------------*/
static int mtd_write_loc( int fd, int offset, const void *buf, int count )
{
   int cnt;

   if (lseek( fd, offset, SEEK_SET ) == -1) {
      warn( "lseek failed: %m" );
      return -1;
   }
   if ((cnt = writen( fd, buf, count)) < 0) {
      warn( "write failed: %m" );
      return -1;
   }
   return cnt;
}
