/*
 * imginstall.c - image installer program
 *
 * Copyright (C) 2002 by Intersil Corp.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <dirent.h>
#include <signal.h>
#include <sys/reboot.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <identification/bootrec.h>
#include <identification/identification.h>
#include <identification/image_header.h>
#include <hardconfig.h>
#include <inl_debug.h>

#include "imginstall.h"
#include "files.h"
#include "flash.h"

struct image_info_t {
        int type;
        int size;
        int(*handler)(char *, int);
	char* fname;
};

char *sector_buf        = NULL;
char *full_filename     = NULL;                /* image file to install */
char *working_dir;                             /* working directory */

int num_apfirmware  = 0;
int num_uapfirmware = 0;
int num_webfileset  = 0;

#define MAXIMAGES	7

/* Image header store for compatibility checking */
int image_index = 0;
struct image_info_t imginfo[MAXIMAGES];
image_header_t *image_headers[MAXIMAGES];


static int validate_file(char *fname);
static int handle_uploaded_file(char* fname);
static int write_part(char *fname, int image);
static int flash_part(char *fname, int image);

static void usage( char * appname )
{
    fprintf( stderr, "Usage: %s [-k] <filename>\n", appname);
    exit(1);
}

/* Image installer must always reboot on exit, because either a new
 * image is installed or the upgrade failed and we don't know exactly
 * what mess we leave behind
 */
static void exit_reboot(void)
{
   imginstall_write_status(STATUS_REBOOT);
   note( "Rebooting the system in %d seconds...", REBOOT_TIME );

   sleep(REBOOT_TIME);
   reboot(RB_AUTOBOOT);
}

void imginstall_write_status(int status)
{
   FILE *status_fp;
   char stat_buf[10];
   status_fp = fopen("/tmp/upgrade_status", "w");

   if (status_fp != NULL) {
      sprintf(stat_buf, "%i", status);

      fputs(stat_buf ,status_fp);

      fclose(status_fp);
   }
}

static void write_error(int status)
{
   imginstall_write_status(status);

   sleep(5);

   /* delete status file */
   unlink("/tmp/upgrade_status");

   /* cleanup image file */
   unlink(full_filename);
}

void write_image_check_status(upg_image_check_t status)
{
   switch(status) {
      case upg_image_unsupported_oem_id:
         write_error(STATUS_ERROR_UNSUPPORTED_OEM);
         break;
      case upg_image_unsupported_product:
         write_error(STATUS_ERROR_UNSUPPORTED_PRODUCT);
         break;
      case upg_image_unsupported_hardware_revs:
         write_error(STATUS_ERROR_UNSUPPORTED_HARDWARE);
         break;
      case upg_image_invalid:
      default:
         write_error(STATUS_ERROR_NO_VALID_IMAGES_INSTALLED);
         break;
   }
}

void write_image_program_status(int ret, int imagetype)
{
   int status;

   switch(imagetype) {
      case image_apfirmware:
         status = STATUS_INSTALLING_APFIRMWARE;
         break;
      case image_uapfirmware:
         status = STATUS_INSTALLING_UAPFIRMWARE;
         break;
      case image_webfileset:
         status = STATUS_INSTALLING_WEBFILESET;
         break;
      default:
         write_error(STATUS_ERROR_NO_VALID_IMAGES_INSTALLED);
         return;
    }

   if(ret < 0) {
       status += STATUS_ERROR_OFFSET;
       write_error(status);
   } else {
       imginstall_write_status(status);
   }
}


/*****************************************************************************
 * NAME:
 *        kill_unneeded_apps
 *
 * DESCRIPTION:
 *        Kill all applications that are "unwanted" during an upgrade. This
 *        saves memory (we need some for storing images temporarily on the
 *        ramdisk) and prevents unwanted user interaction during an upgrade.
 *        The following apps are not killed:
 *        init - of course; boa/ssi/webupgrade - we want to keep web interface
 *        during upgrading; dhcpcd - we want to keep our DHCP ip address.
 *
 * PARAMETERS:
 *        none
 *
 * RETURN:
 *        -1 - error
 *        0 - success
 *****************************************************************************/

#define READ_BUF_SIZE   50

static int kill_unneeded_apps(const char* myname)
{
   DIR *dir;
   struct dirent *next;
   int i, foundkill;

   static char *apps_not_to_kill[] =  {
       "init", "sysd", "boa", "ssi", "webupgrade", "dhcpcd", "autoip", "respawnd", '\0'
   };
       
   static char *apps_to_stop[] = {
       "respawnd", '\0'
   };


   warn("Stopping applications which could interfere with the upgrade process.");
   dir = opendir("/proc");
   if (!dir) {
       error("Cannot open /proc. Unable to kill unwanted applications");
       return -1;
   }
   
   warn("Freeing memory for upgrade by killing applications");
   while ((next = readdir(dir)) != NULL) {
      FILE *status;
      char filename[READ_BUF_SIZE];
      char buffer[READ_BUF_SIZE];
      char name[READ_BUF_SIZE];
      int pid;

      /* If it isn't a number, we don't want it */
      if (!isdigit(*next->d_name))
         continue;

      sprintf(filename, "/proc/%s/status", next->d_name);
      if (! (status = fopen(filename, "r")) ) {
         continue;
      }
      if (fgets(buffer, READ_BUF_SIZE-1, status) == NULL) {
         fclose(status);
         continue;
      }
      fclose(status);

      /* Buffer should contain a string like "Name:   binary_name" */
      sscanf(buffer, "%*s %s", name);

      if( strcmp(name, myname) == 0 )
          continue;
      
      for( i = 0; apps_to_stop[i]; i++) {
         if (strcmp(name, apps_to_stop[i]) == 0 ) {
            pid = strtol(next->d_name, NULL, 0);
	    kill(pid, SIGSTOP);
	    break;
         }
      }
	   
   }
   
   dir = opendir("/proc");
   if (!dir) {
       error("Cannot open /proc. Unable to kill unwanted applications");
       return -1;
   }

   while ((next = readdir(dir)) != NULL) {
      FILE *status;
      char filename[READ_BUF_SIZE];
      char buffer[READ_BUF_SIZE];
      char name[READ_BUF_SIZE];
      int pid;

      /* If it isn't a number, we don't want it */
      if (!isdigit(*next->d_name))
         continue;

      sprintf(filename, "/proc/%s/status", next->d_name);
      if (! (status = fopen(filename, "r")) ) {
         continue;
      }
      if (fgets(buffer, READ_BUF_SIZE-1, status) == NULL) {
         fclose(status);
         continue;
      }
      fclose(status);
         /* Buffer should contain a string like "Name:   binary_name" */
         sscanf(buffer, "%*s %s", name);

         if( strcmp(name, myname) == 0 )
            continue;

      foundkill = 0;
      for( i = 0; apps_not_to_kill[i]; i++) {
         if (strcmp(name, apps_not_to_kill[i]) == 0 ) {
            foundkill = 1;
         }
      }
      
      pid = strtol(next->d_name, NULL, 0);
      if ((pid >= 8) && (!foundkill)) {
         kill(pid, SIGTERM);
         sleep(1);
         kill(pid, SIGKILL);
      }
      
   }
   return 0;
}


/*****************************************************************************
 * NAME:
 *        copy fileset
 *
 * DESCRIPTION:
 *	  Copy the for upgrade necessary files from the Web File Set. 
 *        Signal boa.
 *
 * PARAMETERS:
 *        none
 *
 * RETURN:
 *	  none
 *****************************************************************************/

static void copy_fileset(void) {
	FILE* fp;
	
	/* check if a fileset is present */
	fp = fopen("/usr/www/html/_standardfiles/upgrade/index.shtml", "r");
	if(!fp) {
		/* no fileset, no problem */
		return;
	}	
	fclose(fp);
	
	system("/bin/cp_fileset");
}


int main(int argc, char **argv)
{
   int                i;
   int                kill_apps  = 1;
   char              *appname;
   char              *filename;
   struct stat        stats;
   int                image;
   upg_image_check_t  status     = upg_image_invalid;

   set_malloc_mmap_size(16000);

   /* determine application name from argumentlist */
   appname = strrchr(argv[0], '/');
   if (!appname )
      appname = argv[0];
   else
      appname++;

   /* parse arguments */
   if (argc < 2 )
      usage( appname );

   for (i = 1; i < argc; i++ ) {
      if (!strcmp( argv[i], "-k" )) {
         kill_apps = 0;
      } else {
         full_filename = argv[i];
      }
   }
   if (!full_filename)
      usage( appname );

   imginstall_write_status(STATUS_INITIALIZING);
   note( "Image installer started" );
   
   /* determine path and filename */
   filename = strrchr( full_filename, '/');
   if (!filename) {
      working_dir = ".";
      filename = full_filename;
   } else {
      *filename = '\0';
      filename++;
      working_dir = full_filename;
   }

   if (kill_apps)
      kill_unneeded_apps(appname);
  
	// blink DIAG LED
	{ FILE *ledcontrol = fopen("/proc/ledcontrol", "w");
	  char cmd[] = "0 blink 50";
		if(ledcontrol!=NULL){
			fwrite(cmd, sizeof(cmd), 1, ledcontrol);
			fclose(ledcontrol);
		}
	} 
   /* copy fileset parts */
   copy_fileset();

   /* Allocate the sector buffer. We use mmap because we need
    * 64kB and malloc has a few bytes overhead so it would
    * allocate 128kB at the kernel.
    */
   sector_buf = (char *) mmap((void *) 0, SECTOR_SIZE,
                              PROT_READ | PROT_WRITE,
                              MAP_SHARED | MAP_ANONYMOUS,
                              0, 0
                             );
   if (sector_buf == (char *) MAP_FAILED) {
      error("mmap allocation of %d bytes failed : %m", SECTOR_SIZE);
      write_error(STATUS_ERROR_MALLOC);
      exit_reboot();
   }

   /* Check input file(s) */
   if (chdir(working_dir) != 0) {
      warn( "Failed to change working directory to %s: %m", working_dir );
      write_error(STATUS_ERROR_CHANGE_WORKINGDIR_FAILED);
      exit_reboot();
   }

   if (stat(filename, &stats) < 0) {
      warn("Error on file %s : %m", filename);
      write_error(STATUS_ERROR_NO_VALID_IMAGES_INSTALLED);
      /* reboot */
      exit_reboot();
   }

   if (handle_uploaded_file(filename) < 0) {
      warn("Validation of files failed");
      write_error(STATUS_ERROR_NO_VALID_IMAGES_INSTALLED);
      /* reboot */
      exit_reboot();
   }

   if (image_index > 0) {
      /* images array must be zero terminated */
      image_headers[image_index] = NULL;
      status = image_file_check(image_headers);
   } else {
      warn("No images found, exiting");
      write_error(STATUS_ERROR_NO_VALID_IMAGES_INSTALLED);
      exit_reboot();
   }

   if ((status != upg_image_ok) ||
       (num_apfirmware > 1) || (num_uapfirmware > 1) ||
       (num_webfileset > 1))
   {
      warn("No valid images, status %d, (%d,%d,%d)", status,
           num_apfirmware, num_uapfirmware, num_webfileset);
      write_image_check_status(status);
      /* reboot */
      exit_reboot();
   }

   note( "Image validation OK, now starting programming" );

   /* Input files OK, start programming images */
   for(image = 0; image < image_index; image++) {
      int ret;

      note("Now programming image %d, imgtype %d", image, imginfo[image].type);
      write_image_program_status(0, imginfo[image].type);

      /* Call the handler for this image that will flash/program it */
      ret = imginfo[image].handler(imginfo[image].fname, image);

      if(ret < 0) {
         note("Error during programming: %d", ret);
         if(ret == -ENOSPC) {
            write_error(STATUS_ERROR_IMAGE_TOO_BIG);
         } else {
	    write_image_program_status(ret, imginfo[image].type);
         }
         goto out;
      }
      note("Programming successful");
   }

out:
   /* reboot */
   exit_reboot();
   return(0);
}



/*
 * Install the image in flash.
 */
static int flash_part(char *fname, int image)
{
   int size;
   int ret;
   int fd;

   fd = open_image(fname, &size);
   if (fd < 0)
      return -EINVAL;

   ret = flash_install(fd, imginfo[image].type, size);

   close(fd);
   return(ret);
}

/*
 * Install this image as a file
 */
static int write_part(char *fname, int image)
{
   int size;
   int ret;
   int fd;

   fd = open_image(fname, &size);
   if (fd < 0)
      return -EINVAL;

   ret = file_install(fd, imginfo[image].type, size);

   close(fd);
   return(ret);
}


/* This functions determines if the uploaded file is a
 * single or combined image. It will call validate_file
 * for each found image.
 */
static int handle_uploaded_file(char *fname) {
	
	char *buf;
	long bytesread = 0;
   	ulong component_id;
   	int fd;
	char strbuf[160];
	int size;
	DIR *dir;
	struct dirent *next;
	int ret;
	
	/* get the component ID for this file */
	fd = open_image(fname, &size);
	if (fd < 0) {
		return -EINVAL;
	}
	
	buf = (char *)malloc(BR_MAX_BRA_SIZE);
	if(!buf) {
		close(fd);
		return -EINVAL;
	}
	
	bytesread = read(fd, buf, BR_MAX_BRA_SIZE);
	if (bytesread != BR_MAX_BRA_SIZE) {
		close(fd);
		free(buf);
		return -EINVAL;
	}

   	component_id = ((struct image_header_extended *)buf)->component_id;
	close(fd);
	free(buf);

	if(component_id != BR_COMP_ID_DBCO) {
		/* This is not a combined image, just call validate_file */
		return(validate_file(fname));
	}
	
	/* This file is an combined image. Mount it (cramfs) to access the
	 * images inside it */
	snprintf(strbuf, sizeof(strbuf), "mount -o loop -t cramfs %s /mnt", fname);
	system(strbuf);
	
	dir = opendir("/mnt");
	if(!dir) {
		error("Cannot open /mnt. Unable to find images");
		return -1;
	}
	
	while ((next = readdir(dir)) != NULL) {
		snprintf(strbuf, sizeof(strbuf), "/mnt/%s", next->d_name);
		ret = validate_file(strbuf);
		if (ret < 0) {
			return (ret);
		}		
	}
	
	return(ret);
		
}

/*
 * This function is called for each found image. The images' id
 * and size is checked and put in the imginfo[] array
 */
static int validate_file(char *fname)
{
   char *buf;
   int ret = 0;
   int size;
   long bytesread = 0;
   ulong component_id;
   int fd;

   fd = open_image(fname, &size);
   if (fd < 0)
      return -EINVAL;

   buf = (char *)malloc(BR_MAX_BRA_SIZE);
   if (!buf) {
      close(fd);
      return -EINVAL;
   }

   bytesread = read (fd, buf, BR_MAX_BRA_SIZE);

   if (bytesread != BR_MAX_BRA_SIZE)
      goto err_out;

   /* Store the image header for later compatibility checking */
   imginfo[image_index].type = image_nok;
   imginfo[image_index].fname = strdup(fname);
   image_headers[image_index] = (image_header_t *)buf;
	
   component_id = ((struct image_header_extended *)buf)->component_id;
   switch (component_id) {
      case BR_COMP_ID_DAP1:
         imginfo[image_index].type = image_apfirmware;
         imginfo[image_index].handler = flash_part;
         num_apfirmware++;
         break;
      case BR_COMP_ID_ITM1:	/* Indigo uAP */
      case BR_COMP_ID_LTM1:	/* Liberty uAP */
         imginfo[image_index].type = image_uapfirmware;
         imginfo[image_index].handler = write_part;
         num_uapfirmware++;
         break;
      case BR_COMP_ID_DFS1:
         imginfo[image_index].type = image_webfileset;
         imginfo[image_index].handler = write_part;
         num_webfileset++;
         break;
      default:
         goto err_out;
   }

   /* This is a valid image */
   imginfo[image_index].size = size;
   image_index++;
   goto out;

err_out:
   ret = -EINVAL;
   free(buf);
out:
   close(fd);
   return(ret);
}

