/* inetd.c: Start up network services
 *
 * Copyright (C) 1998  Kenneth Albanowski <kjahds@kjahds.com>
 * Copyright (C) 1999  D. Jeff Dionne     <jeff@rt-control.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <linux/fs.h>
#include "inetd.h"

#define MAX_SERVICES 4
#define MAX_CONNECT 8

//#define DEBUG
//#define USE_SIGPIPE
//#define USE_SIGSTOP
//#define USE_SIGTSTP
//#define USE_SIGCONT
//#define USE_SIGCHLD
#define USE_SIGHUP

/* external function prototypes */
extern char ** cfgread(FILE *fp);				/* in config.c */
extern char ** cfgfind(FILE *fp, char *var);


void close_on_exec(int f)
{
	if (fcntl(f, F_SETFD, 1))
		perror("close on exec: ");
}

void close_all_fds(int keep_top)
{
	int i;
	for(i=keep_top+1;i<getdtablesize();i++)
		close(i);
}

struct {
	int enabled;
        char *args[6];
        char arg[128];
	int port;
	int tcp;
	int reconfig;

	int master_socket;

	int limit;
	int current;
	int pid[16];

	int changed;
} service[MAX_SERVICES];


int reap_child(int pid) {
  int i;
  char buf[64];
  int reaped = 0;

  for(i=0;i<MAX_SERVICES;i++) {
    int j;
    for(j=0;j<16;j++)
      if (service[i].pid[j] == pid) {
        service[i].pid[j] = 0;
        service[i].current--;
      }
  }
  return reaped;
}

int generate_select_fds(fd_set * readmask, fd_set * writemask)
{
  int i;
  int max=0;
  FD_ZERO(readmask);
  FD_ZERO(writemask);
#define FD_MSET(x,y) FD_SET((x),(y)); if ((x)>max) max = (x);

  for(i=0;i<MAX_SERVICES;i++) {
#ifdef DEBUG
	  printf("generate_select_fds: service[%d].args[0]: %s\n", i, service[i].args[0]);
#endif
	  if (service[i].args[0] == NULL)
	  {
		  continue;
	  }
	  if (!strlen(service[i].args[0]) || (service[i].current >= service[i].limit))
      continue;

    FD_MSET(service[i].master_socket, readmask);

  }

  return max+1;
}

int handle_incoming_fds(fd_set * readmask, fd_set * writemask)
{
  int i;


  for(i=0;i<MAX_SERVICES;i++) {
    int fd;
    if (service[i].master_socket && FD_ISSET(service[i].master_socket, readmask)) {
      int j;
      for(j=0;j<16;j++)
        if (service[i].pid[j] == 0)
          break;

      if (service[i].tcp) {
        struct sockaddr_in remote;
        int remotelen = sizeof(remote);
        fd = accept(service[i].master_socket, (struct sockaddr*)&remote, &remotelen);
        if (fd < 0) {
          printf("accept failed\n");
          break;
        }
      } else {
        fd = service[i].master_socket;
      }

      if (!(service[i].pid[j] = vfork())) {
        if (fd != 0)
          dup2(fd, 0);
        if (fd != 1)
          dup2(fd, 1);
        if (fd != 2)
          dup2(fd, 2);
        if (fd > 2)
          close(fd);
        close_all_fds(2);
        service[i].current++;
        execlp(service[i].args[0],
	       service[i].args[0],
	       service[i].args[1],
	       service[i].args[2],
	       service[i].args[3],
	       service[i].args[4],
	       service[i].args[5],
	       NULL /* just to be safe */
	       );

        service[i].current--;
        close(service[i].master_socket);
        service[i].enabled = 0;
        service[i].master_socket = 0;
        _exit(0);
      }

      if (service[i].tcp) {
        close(fd);
      }
    }
  }

}

int start_services(void) {
  int s;
  int i;
  struct server_sockaddr;

  for(i=0;i<MAX_SERVICES;i++) {
    struct sockaddr_in server_sockaddr;
#ifdef DEBUG
	printf("start_services: starting service[%d].args[0] = %s\n", i, service[i].args[0]);
#endif
	if (service[i].args[0] == NULL)
	{
		continue;
	}

    if (service[i].master_socket || !strlen(service[i].args[0]))
      continue;

    if (service[i].tcp) {
      int true;

      if ((s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
        printf("Unable to create socket\n");
      }

      close_on_exec(s);

      server_sockaddr.sin_family = AF_INET;
      server_sockaddr.sin_port = htons(service[i].port);
      server_sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);

      true = 1;

      if((setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (void *)&true,
         sizeof(true))) == -1) {
        perror("setsockopt: ");
      }

      if(bind(s, (struct sockaddr *)&server_sockaddr,
        sizeof(server_sockaddr)) == -1)  {
        perror("Unable to bind server socket: ");
        close(s);
      }

      if(listen(s, 1) == -1) {
        printf("Unable to listen to socket\n");
      }

    } else {

      if ((s = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
        printf("Unable to create socket\n");
      }

      close_on_exec(s);

      server_sockaddr.sin_family = AF_INET;
      server_sockaddr.sin_port = htons(service[i].port);
      server_sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);

      if(bind(s, (struct sockaddr *)&server_sockaddr,
        sizeof(server_sockaddr)) == -1)  {
        printf("Unable to bind socket\n");
        close(s);
      }
    }
    service[i].master_socket = s;
    service[i].reconfig = 1;
    service[i].changed = 0;
  }
}

int stop_services(void) {
  int s;
  int i;
  struct server_sockaddr;

  for(i=0;i<MAX_SERVICES;i++) {
    if(service[i].master_socket)
      close(service[i].master_socket);
    service[i].master_socket = 0;
  }
}

void reap_children(void)
{
  int child;
  int status;
  while ((child = waitpid(-1, &status, WNOHANG)) > 0) {
    reap_child(child);
  }
}

#if defined(USE_SIGHUP)
volatile int got_hup;
void hup_handler(int signo)
{
	got_hup = 1;

}
#endif /* defined(USE_SIGHUP) */

#if defined(USE_SIGCONT)
volatile int got_cont;
void cont_handler()
{
	got_cont = 1;
}
#endif /* defined(USE SIGCONT) */

#if defined(USE_SIGSTOP) || defined(USE_SIGTSTP)
void stop_handler()
{
	got_cont = 0;
	/* To reduce memory usage & prevent callers from getting gummed up */
	stop_services();
	while(!got_cont) {
		pause();
		reap_children();
	}
	got_cont = 0;
}
#endif /* defined(USE_SIGSTOP) || defined(USE_SIGTSTP) */

#if defined(USE_SIGCHLD)
void child_handler(int signo)
{
	/* Don't reap, just interrupt the syscall */
}
#endif /* defined(USE_SIGCHLD) */

char * pstrdup(char * c) {
	char * _c = c;
	if (_c)
		_c = strdup(_c);
	return _c;
}

void kill_changed_things(void)
{
	int i;

	for(i=0;i<MAX_SERVICES;i++) {
		int j;

		if (!service[i].changed || !service[i].reconfig)
			continue;

		if (service[i].master_socket) {
			close(service[i].master_socket);
			service[i].master_socket = 0;
		}

		for (j=0;j<16;j++) {
			if (service[i].pid[j] != 0) {
				kill(service[i].pid[j], SIGTERM);
				kill(service[i].pid[j], SIGHUP);
				service[i].pid[j] = 0;
			}
		}
		service[i].changed = service[i].reconfig = 0;
	}
}

void run(void)
{
  fd_set rfds, wfds;
  struct timeval tv;
  int max;
  for(;;) {
    reap_children();
    start_services();

    max = generate_select_fds(&rfds, &wfds);
    tv.tv_sec = 1;
    tv.tv_usec = 0;

    if (select(max, &rfds, &wfds, 0, &tv) > 0) {
      handle_incoming_fds(&rfds, &wfds);
    }
    if (got_hup) {
    	got_hup = 0;
		read_config();
    	kill_changed_things();
    	continue;
    }
  }
}

/* Very simple file format, lines of the form
 * port# ignored tcp/udp ignored ignored <exec path> args...
 * this should be compatible with berkeley derived inetd
 */
int
read_config(char *inetd_conf)
{
  FILE *cfp;
  FILE *sfp;
  char **args;
  char *ap;

  int i;
  int j;

  if (!(cfp = fopen(inetd_conf,"r"))) {
    perror("Can't open specified argument");
    exit(1);
  }

  if (!(sfp = fopen(SERVICES,"r"))) {
    perror("Can't open " SERVICES);
    fclose(cfp);
    exit(1);
  }

  for (i=0; i<MAX_SERVICES;) {
    if (!(args = cfgread(cfp))) break;

    for (j=0; j<11 && args[j]; j++);
    if (j<6) {
      fprintf(stderr,"Bad line in config file %s...\n",
	      args[0] ? args[0] : "");
      continue;
    }

    /* copy in the args and exec path. */
    service[i].args[j - 5] = NULL;
    ap = service[i].arg;

    for (j-- ;j > 4; j--) {
      service[i].args[j - 5] = ap;
      strncpy(ap, args[j], 32);

      ap += strlen(args[j]) + 1;
      /* FIXME: bounds check */
    }

    service[i].tcp = !strcmp(args[2], "tcp") ? 1 : 0;

    if (!(args = cfgfind(sfp, args[0]))) {
      fprintf(stderr, "can't find service\n");
      continue;
    }

    service[i].port = atoi(args[1]);

    service[i].master_socket = 0;
    service[i].limit = MAX_CONNECT;
    service[i].changed = 1;
    service[i].enabled = 1;

#ifdef DEBUG
    printf("service %s port %d %s %s\n",
	   service[i].args[0],
	   service[i].port,
	   service[i].tcp ? "tcp" : "udp",
	   service[i].enabled ? "enabled" : "disabled");
#endif
    i++;
  }

  fclose(sfp);
  fclose(cfp);
  return i;
}

main(int argc, char *argv[], char *env[])
{
#if defined(USE_SIGPIPE)
  __signal(SIGPIPE, SIG_IGN, 0);
#endif /* defined(USE_SIGPIPE) */
#if defined(USE_SIGSTOP)
  __signal(SIGSTOP, stop_handler, 0);
#endif /* defined(USE_SIGSTOP) */
#if defined(USE_SIGTSTP)
  __signal(SIGTSTP, stop_handler, 0);
#endif /* defined(USE_SIGTSTP) */
#if defined(USE_SIGCONT)
  __signal(SIGCONT, cont_handler, 0);
#endif /* defined(USE_SIGCONT) */
#if defined(USE_SIGCHLD)
  __signal(SIGCHLD, child_handler, SA_INTERRUPT);
#endif /* defined(USE_SIGCHLD) */
#if defined(USE_SIGHUP)
  __signal(SIGHUP, hup_handler, 0);
#endif /* defined(USE_SIGHUP) */

  printf("PATH_MAX=%d\n",PATH_MAX);

  if (argc < 2)
  {
	  printf("INETD: no arguments found using default conf: %s\n", INETD_CONF);
	  read_config(INETD_CONF);
  }
  else
  {
	  printf("INETD: arguments found using conf: %s\n", argv[1]);
	  read_config(argv[1]);
  }

  run();

  /* not reached */
}
