/*  $Header$
 *  
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <ctype.h>

#include "identification/identification.h"
#include "identification/image_header.h"
#include "image.h"

struct meminfo {
	int total;
	int used;
};

struct slabinfo {
	int active_size;
	int total_size;
	int kmem_active_size;
	int kmem_total_size;
};

static char	filename[255];
static char	line[255];

static int read_proc_iomem(int *ramsize, int *mvcsize, int *kernelsize);
static int read_proc_meminfo(struct meminfo *system_mem, int *buffers, int *cached);
static int read_proc_slabinfo(struct slabinfo *slab);
static int read_proc_statm(int pid, int *memtotal, int *txtsize, int *datasize);
static int read_proc_mpu(int pid, int *reprogram_count);
static int get_execname(int pid, char *exec);
static int show_mtd(void);

int main(int argc, char *argv[])
{
	DIR		*procdir;
	struct dirent	*dp;
	char		exec[255];
	int		i, pid, found;
	int		buffers, cached, sumused = 0, sumfree = 0;
	int		ramsize, mvcsize, kernelsize;
	int		reprogram_count;
	int		ret = 0;
	struct meminfo	system_mem;

	if (argc != 1) {
		printf("Usage: %s\n", argv[0]);
		printf("   Prints out an overview of the memory usage:\n");
		printf("       1. RAM layout gives an overview of the MVC and Kernel size\n");
		printf("       2. Linux Kernel Memory Usage prints the memory used for buffers, page cache and slab cache\n");
		printf("       3. Applications prints the memory used by each running application\n");
		exit(0);
	}

	show_mtd();

	printf("\n1. RAM Layout\n");
	printf("========================================================\n");
	read_proc_iomem(&ramsize, &mvcsize, &kernelsize);
	printf("RAM available\t%5d kB\n", ramsize/1024);
	printf("MVC\t\t%5d kB\n", mvcsize/1024);
	printf("Kernel\t\t%5d kB\n", kernelsize/1024);
	printf("Available\t%5d kB\n\n", (ramsize - kernelsize - mvcsize)/1024);

	printf("\n2. Linux Kernel Memory Usage\n");
	printf("========================================================\n");
	printf("\t\t  TOTAL\t\t   USED\t\t   FREE\n");
	if(!(ret = read_proc_meminfo(&system_mem, &buffers, &cached))) {
		printf("System Memory");
		printf("\t%5d kB\t%5d kB\t%5d kB\n", system_mem.total/1024, system_mem.used/1024,
			(system_mem.total - system_mem.used)/1024);
		printf("\nBuffers\t\t\t\t%5d kB\n", buffers/1024);
		printf("Cached\t\t\t\t%5d kB\n", cached/1024);
		sumused = buffers + cached;
		sumfree = system_mem.total - system_mem.used;
	}
	
	if(!ret) {
		struct slabinfo slab;

		ret = read_proc_slabinfo(&slab);
		printf("Slabs (used/active)\t      %5d / %d kB\t%5d kB\n", slab.total_size/1024, 
			slab.active_size/1024, (slab.total_size - slab.active_size)/1024);
		printf(" [kmem (used/active)\t      %5d / %d kB\t%5d kB]\n", slab.kmem_total_size/1024,
			slab.kmem_active_size/1024, (slab.kmem_total_size - slab.kmem_active_size)/1024);
		sumused += slab.total_size;
		sumfree += slab.total_size - slab.active_size;
	}
	printf("\t\t---------------------------------------- +\n");
	printf("\t\t\t\t%5d kB\t%5d kB\n", sumused/1024, sumfree/1024);

	if ((procdir = opendir("/proc")) == NULL) {
		fprintf(stderr, "Error can't open directory /proc\n");
		return -1;
	}
	/* Note that all memory that is allocated by applications is accounted for
	 * in the slab administration. We don't want to count things double, so we
	 * reset the total summed memory
	 */
	sumused = 0;
	printf("\n3.Applications\n");
	printf("========================================================\n");
	printf("NAME\t   TOTAL      TEXT      DATA    MPU CNT\n");
	while (!ret && ((dp = readdir(procdir)) != NULL))
	{
		int memtotal, txtsize, datasize;
		if (strcmp(dp->d_name, ".") == 0 ||
			strcmp(dp->d_name, "..") == 0)
				continue;
		/* only process numeric entries */
		found = i = 0;
		while(!found && (i < strlen(dp->d_name))) {
			if(!isdigit(dp->d_name[i]))
				found = 1;
			i++;
		}
		if(found)
			continue;

		pid = atoi(dp->d_name);
		if(get_execname(pid, exec)) {
			fprintf(stderr, "Unable to determine process name for pid %d\n", pid);
			strcpy(exec, "Unknown");
		}
		read_proc_mpu(pid, &reprogram_count);
		if( !(ret = read_proc_statm(pid, &memtotal, &txtsize, &datasize)) ) {
			if(memtotal != 0) {
				printf("%.7s\t%5d kB  %5d kB  %5d kB  %5lu\n", exec,
					memtotal/1024, txtsize/1024, datasize/1024,
					reprogram_count);
				sumused += memtotal;
			}
		}
	}
	printf("\t---------------------------------------------- +\n");
	printf("\t%5d kB\n", sumused/1024);

	exit(ret);
}

static int read_proc_iomem(int *ramsize, int *mvcsize, int *kernelsize)
{
	FILE	*statf;
	int	start, end, kernelstart = 0 , ramstart = 0, i = 0;

	statf = fopen("/proc/iomem", "r");
	if (!statf) {
		perror("fopen");
		return -1;
	}
	while(fgets(line, sizeof(line), statf)) {
		if(sscanf(line, "%x-%x %*s", &start, &end) == 2) {
			switch(i) {
			case 0:
				/* RAM start/end */
				ramstart = start;
				*ramsize = end - start;
				break;
			case 1:
				/* Kernel text start/end */
				kernelstart = start;
				*mvcsize = kernelstart - ramstart;
				break;
			case 2:
				/* Kernel data start/end */
				*kernelsize = end - kernelstart;
				break;
			default:
				/* nothing */
				break;
			}
			i++;
		}
	}
	fclose(statf);
	return 0;
}

static int read_proc_meminfo(struct meminfo *system_mem, int *buffers, int *cached)
{
	FILE	*statf;
	int	total, used;

	statf = fopen("/proc/meminfo", "r");
	if (!statf) {
		perror("fopen");
		return -1;
	}

	memset(system_mem, 0, sizeof(struct meminfo));
	*buffers = 0;
	*cached = 0;
	while(fgets(line, sizeof(line), statf)) {
		if(strstr(line, "Mem:")) {
			if(sscanf(line, "%*s %d %d %*d %*d %d %d", &total, &used, buffers, cached) == 4) {
				system_mem->total = total;
				system_mem->used = used;
			}
		}
	}
	fclose(statf);
	return 0;
}

static int read_proc_slabinfo(struct slabinfo *slab)
{
	FILE	*statf;
	int	active, total, pages_per_slab;
	char	type[255];

	statf = fopen("/proc/slabinfo", "r");
	if (!statf) {
		perror("fopen");
		return -1;
	}

	memset(slab, 0, sizeof(struct slabinfo));
	while(fgets(line, sizeof(line), statf)) {
		if(sscanf(line, "%s %*d %*d %*d %d %d %d", type, &active, &total, &pages_per_slab) == 4) {
			if(strstr(type, "size-")) {
				slab->kmem_active_size += (active * pages_per_slab * 4096);
				slab->kmem_total_size += (total * pages_per_slab * 4096);
			}
			slab->active_size += (active * pages_per_slab * 4096);
			slab->total_size += (total * pages_per_slab * 4096);
		}
	}
	fclose(statf);
	return 0;
}

static int read_proc_mpu(int pid, int *reprogram_count)
{
	FILE	*statf;
	char	type[255];

	sprintf(filename, "/proc/%d/mpu", pid);
	statf = fopen(filename, "r");
	if (!statf) {
	    	/* This may fail if we have no MPU configured */
	    	*reprogram_count = 0;
		return -1;
	}

	while(fgets(line, sizeof(line), statf)) {
		if(sscanf(line, "%s %d", type, reprogram_count) == 2) {
			if(strstr(type, "ReprogCnt")) {
				break;
			}
		}
	}
	fclose(statf);
	return 0;
    
}

static int get_execname(int pid, char *exec)
{
	FILE	*statf;

	sprintf(filename, "/proc/%d/status", pid);
	statf = fopen(filename, "r");
	if (!statf) {
		perror("fopen");
		return -1;
	}

	if(fgets(line, sizeof(line), statf)) {
		sscanf(line, "%*s %s", exec);
	}
	fclose(statf);
	return 0;
}

static int read_proc_statm(int pid, int *memtotal, int *txtsize, int *datasize)
{
	FILE	*statf;
	
	sprintf(filename, "/proc/%d/statm", pid);
	statf = fopen(filename, "r");
	if (!statf) {
		perror("fopen");
		return -1;
	}

	if(fgets(line, sizeof(line), statf)) {
			/* size, resident, share, txtrs, librs, datars, dirty */
		sscanf(line, "%d %*d %*d %d %*d %d %*d", memtotal, txtsize, datasize);
	}
	fclose(statf);
	return 0;
}

#if 0  //was #ifndef TURBO_AP  /*---------------------- NO TURBO_AP ---------------------*/

static int show_mtd(void)
{
	FILE *f;
	image_header_t hdr;
	char ts[32];
	int i;
	int r;

	f=fopen("/dev/mtd1", "r");
	
	if(f==NULL) {
		fprintf(stderr, "Can't open dev");
		exit(1);
	}

	printf("Layout of /dev/mtd1 :\n\n");

	for(i=0; i<5; i++) {
		r=fseek(f, i * 0x10000, SEEK_SET);
		r=fread(&hdr, 1, sizeof(hdr),  f);
	
		strcpy(ts, "-");
		switch (hdr.magic2) {
			case BR_COMP_ID_MVC1:	strcpy(ts, "Blob"); 	break; 
			case BR_COMP_ID_LDV1:	strcpy(ts, "Mtfw"); 	break; 
			case BR_COMP_ID_RESC:	strcpy(ts, "Rescue"); 	break; 
			default:		strcpy(ts, "Unknown");	break;
		}

		if(hdr.status & IMAGE_PREFERRED) {
			strcat(ts, " (preferred)");
		}
	
		if(hdr.magic1 != MAGIC_OKAY) {
			strcpy(ts, "-");
		}
	 
		printf("  %d %s\n", i, ts); 
	}

	printf("\n");
  
	return(0);
}

#else TURBO_AP  /*------------------------ TURBO_AP ----------------------*/

static int show_mtd(void)
{
   FILE *fp;
   image_header_ext hdre;
   ulong component_id;
   ulong magic;
   ulong cnt;
   ulong offset;
   char ts[32];
   int i;
   int r;

   fp = fopen("/dev/mtd1", "r");
   
   if (fp == NULL) {
      fprintf(stderr, "Can't open dev");
      exit(1);
   }

   printf("Layout of /dev/mtd1 :\n\n");

   for (i=0; i<10; i++) {
      r = fseek( fp, i * 0x10000, SEEK_SET);
      r = fread( &hdre, (size_t)1, sizeof(hdre), fp);
   
      strcpy(ts, "-");

      if (hdre.magic1 == MAGIC_OKAY) {
         component_id = hdre.component_id;
         switch (component_id) {
            case BR_COMP_ID_DAP1:
               strcpy(ts, "Turbo AP Firmware");
               break; 
            case BR_COMP_ID_ITM1:
               strcpy(ts, "Micro AP Firmware");
               break; 
            case BR_COMP_ID_DFS1:
               strcpy(ts, "FileSet");
               break;
            case BR_COMP_ID_DMT1:
               strcpy(ts, "MTFW");
               break;
            case BR_COMP_ID_DBO1:
               strcpy(ts, "Boot (Bootloader + PDA)");
               break;
            default:
               strcpy(ts, "Unknown");
               break;
         }

         printf("  %d %s\n", i, ts); 

         if (component_id == BR_COMP_ID_DAP1) {
            for (cnt=0; cnt<7424; cnt++) {
               offset = (i * 0x10000) + (cnt * 512);
               r = fseek( fp, offset, SEEK_SET);
               r = fread( &magic, (size_t)1, sizeof(magic), fp);
               switch (magic) {
                  case 0x52545344:
                     printf("       DAP Starter at offset 0x%lx\n", offset);
                     break;
                  case BR_COMP_ID_MVC1:
                     printf("       MVC at offset 0x%lx\n", offset);
                     break;
                  case BR_COMP_ID_SYST:
                     printf("       Linux System at offset 0x%lx\n", offset);
                     cnt = 8000; /* stop the for-loop */
                     break;
               }
            }
         }
      } else {
         printf("  %d -\n", i); 
      }
   }

   printf("\n");
  
   return(0);
}
#endif
