/* pptp.c ... client shell to launch call managers, data handlers, and
 *            the pppd from the command line.
 *            C. Scott Ananian <cananian@alumni.princeton.edu>
 *
 * $Id: pptp.c,v 1.7 1997/12/16 11:38:48 cananian Exp $
 */


#include <getopt.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>
#include <sys/wait.h>
#include <aplog_defs.h>
#include "aplog.h"
#include "pptp_callmgr.h"
#include "pptp_gre.h"
#include "version.h"
#include "inststr.h"
#include "util.h"
#include "pty.h"

#include "inl_misc.h"

#ifndef PPPD_BINARY
#define PPPD_BINARY "pppd"
#endif

struct in_addr get_ip_address(char *name);
int open_callmgr(struct in_addr inetaddr, char* phonenr, int argc,char **argv,char **envp);
void launch_callmgr(struct in_addr inetaddr, char* phonenr,  int argc,char **argv,char **envp);
void get_call_id(int sock, pid_t gre, pid_t pppd, 
		 u_int16_t *call_id, u_int16_t *peer_call_id);
void launch_pppd(char *ttydev, int argc, char **argv);

void usage(char *progname) {
  fprintf(stderr,
	  "%s\n"
	 "Usage:\n"
	 " %s hostname [pppd options]\n", version, progname);
  exit(1);
}

int signaled = 0;

void do_nothing(int sig) { 
    /* do nothing signal handler. Better than SIG_IGN. */
    signaled = 1;
}

sigjmp_buf env;
void sighandler(int sig) {
  siglongjmp(env, 1);
}

int main(int argc, char **argv, char **envp) {
  struct in_addr inetaddr;
  int callmgr_sock;
  char ptydev[PTYMAX], ttydev[TTYMAX];
  int pty_fd;
  pid_t parent_pid, child_pid;
  u_int16_t call_id, peer_call_id;
  int pppdargc;
  char **pppdargv;
  char phonenrbuf[65];
  char* volatile phonenr = NULL;

  // ID MPU fix
  set_malloc_mmap_size(16000);
  
  if (argc < 2)
    usage(argv[0]);

  /* Step 1: Get IP address for the hostname in argv[1] */
  inetaddr = get_ip_address(argv[1]);

	argc--;
	argv++;
	while(1)
	{
  
	   static struct option long_options[] = {
		{"phone", 1, 0, 0},
		{0,0,0,0}
	   };
	  int option_index = 0;
	  int c;	   
	 
	   opterr=0;
	   c = getopt_long (argc,argv,"",long_options, &option_index);
	   if(c==-1) break;
	   switch(c)
	   {
	   	case 0:
		if(option_index == 0)
		{
		  strncpy(phonenrbuf,optarg,sizeof(phonenrbuf));
		  phonenrbuf[sizeof(phonenrbuf)-1]='\0';
		  phonenr=phonenrbuf;
		}
		break;
		case '?':
		default:
		 c = -1;
		break;
	   }
	   if ( c==-1) break;
	}
   pppdargc = argc - optind;
   pppdargv = argv + optind;

  /* Step 2: Open connection to call manager
   *         (Launch call manager if necessary.)
   */
  callmgr_sock = open_callmgr(inetaddr, phonenr, argc,argv,envp);

  /* Step 3: Find an open pty/tty pair. */
  pty_fd = getpseudotty(ttydev, ptydev);
  if (pty_fd < 0) { close(callmgr_sock); fatal("Could not find free pty."); }


  // ID The following code used to for and do a wait() in the pptp
  // child. When the callmgr is sucessfully spawned, the child is 
  // given a signal so it continues. I changed this in exec'ing ppp,
  // and sending it a SIGSTOP immeadiately from the parent. After
  // the callmgr started, it's sent a SIGCONT... The pppd process
  // runs a very short time before STOPPED, but this doesn't seem
  // to be a problem...
  
  
  /* Step 4: fork and wait. */
  parent_pid = getpid();

  // ID Switched parent and child because of vfork

  switch (child_pid = vfork()) {
  	case -1:
	    close (pty_fd); close (callmgr_sock);
	    perror("Error");
	    fatal("Could not fork pppd process");

	  case 0: 

	    launch_pppd(ttydev, pppdargc, pppdargv); /* launch pppd */
	    perror("Error");
	    fatal("Could not launch pppd");

	  default:
	  
	    break;
  }

  // ID Step 5a : SIGSTOP the newly spawned pppd process

  kill(child_pid, SIGSTOP);


  /* Step 5b: Exchange PIDs, get call ID */
  get_call_id(callmgr_sock, parent_pid, child_pid, 
	      &call_id, &peer_call_id);


  /* Step 5c: Send signal to wake up pppd task */

  kill(child_pid, SIGCONT);
  aplog(EV_SYS_SUBSYS_START, "pptp", NULL);
  {
    char buf[128];
    snprintf(buf, sizeof(buf), "pptp: GRE-to-PPP gateway on %s", ptydev);
    inststr(argc,argv,envp, buf);
  }

  // ID Catch these signals, so the pptp_gre_copy loop knows
  // we want to exit and clean up.[B

  signal(SIGINT, do_nothing);
  signal(SIGTERM, do_nothing);
  signal(SIGHUP, do_nothing);

  /* Step 6: Do GRE copy until close. */
  pptp_gre_copy(call_id, peer_call_id, pty_fd, inetaddr);


  /* on close, kill em all. */

  kill(child_pid, SIGKILL);
  close(pty_fd);
  close(callmgr_sock);
  aplog(EV_SYS_SUBSYS_STOP, "pptp", NULL);
  exit(0);
}

struct in_addr get_ip_address(char *name) {
  struct in_addr retval;
  struct hostent *host = gethostbyname(name);
  if (host==NULL) {
    if (h_errno == HOST_NOT_FOUND) {
      aplog(EV_NET_PPTP_ERR, "gethostbyname: HOST NOT FOUND", NULL);
      fatal("gethostbyname: HOST NOT FOUND");
    }
    else if (h_errno == NO_ADDRESS) {
      aplog(EV_NET_PPTP_ERR, "gethostbyname: NO IP ADDRESS", NULL);
      fatal("gethostbyname: NO IP ADDRESS");
    }
    else {
      aplog(EV_NET_PPTP_ERR, "gethostbyname: name server error", NULL);
      fatal("gethostbyname: name server error");
    }
  }
  
  if (host->h_addrtype != AF_INET) {
    aplog(EV_NET_PPTP_ERR, "Host has non-internet address", NULL);
    fatal("Host has non-internet address");
  }
  
  memcpy(&retval.s_addr, host->h_addr, sizeof(retval.s_addr));
  return retval;
}

int open_callmgr(struct in_addr inetaddr, char* phonenr, int argc, char **argv, char **envp) {
  /* Try to open unix domain socket to call manager. */
  struct sockaddr_un where;
  const int NUM_TRIES = 1; /* the respawnd daemon will do the retries */
  int i, fd;


      
  /* Open socket */
  if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
    fatal("Could not create unix domain socket: %s", strerror(errno));
  }

  /* Make address */
  where.sun_family = AF_UNIX;
  snprintf(where.sun_path, sizeof(where.sun_path), 
	   PPTP_SOCKET_PREFIX "%s", inet_ntoa(inetaddr));

  if (connect(fd, (struct sockaddr *) &where, sizeof(where)) < 0) {
      /* couldn't connect.  We'll have to launch this guy. */
      // ID Remove UNIX socket if exists
      unlink(where.sun_path);

      launch_callmgr(inetaddr, phonenr,  argc,argv,envp);
      
      /* Try NUM_TRIES times to connect to the call manager */
      for (i=0; i<NUM_TRIES; i++) {
 	sleep(1);
        if (connect(fd, (struct sockaddr *) &where, sizeof(where)) >= 0) {
		return fd;
	} 
      }    
  
  }
  else{ 
    return fd;
  }
  
  close(fd);
  fatal("Could not connect to call manager after %d seconds.", i);
  /* just to be sure */
  system("/bin/killall -9 pptp_callmgr");
  return -1;   /* make gcc happy */
}

void launch_callmgr(struct in_addr inetaddr, char* phonenr, int argc,char**argv,char**envp) {

  char* cmd[5];
  int pid;
  
  cmd[0] = "/bin/pptp_callmgr";
  cmd[1] = inet_ntoa(inetaddr);	      
  cmd[2] = phonenr;
  cmd[3] = NULL;
	     
  inl_exec(cmd);
}		

/* XXX need better error checking XXX */
void get_call_id(int sock, pid_t gre, pid_t pppd, 
		 u_int16_t *call_id, u_int16_t *peer_call_id) {
  u_int16_t m_call_id, m_peer_call_id;
  /* write pid's to socket */
  /* don't bother with network byte order, because pid's are meaningless
   * outside the local host.
   */
  write(sock, &gre, sizeof(gre));
  write(sock, &pppd, sizeof(pppd));
  read(sock,  &m_call_id, sizeof(m_call_id));
  read(sock,  &m_peer_call_id, sizeof(m_peer_call_id));
  /* XXX FIX ME ... DO ERROR CHECKING & TIME-OUTS XXX */
  *call_id = m_call_id;
  *peer_call_id = m_peer_call_id;
}

void launch_pppd(char *ttydev, int argc, char **argv) {
  char *new_argv[argc+4]; /* XXX if not using GCC, hard code a limit here. */
  int i;

  new_argv[0] = PPPD_BINARY;
  new_argv[1] = ttydev;
//  new_argv[2] = "38400";
  for (i=0; i<argc; i++)
    new_argv[i+2] = argv[i];
  new_argv[i+2] = NULL;
  
  printf("Exec pppd : %s %s %s %s %s %s %s %s\n", new_argv[0],new_argv[1],new_argv[2],new_argv[3],new_argv[4],new_argv[5],new_argv[6],new_argv[7]);
  execvp(new_argv[0], new_argv);
}

/*************** COMPILE call manager into same binary *********/
#define main       callmgr_main
#define sighandler callmgr_sighandler
#define do_nothing callmgr_do_nothing
#include "pptp_callmgr.c"
