/*  $Header$
 *
 *  respawnd - daemon for respawning crashed processes
 *  Copyright (C) 2002 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * respawnd monitors a list of processes periodically (each RESPAWN_SLEEP
 * seconds).
 * If one of the monitored processes is no longer running, it will be restarted.
 * respawnd is passed a directory that contains two subdirectories (fatal and 
 * non_fatal) that must contain zero or more files, where it takes the name of
 * all these files as the processes to monitor. A process can make respawnd
 * monitor it by creating an (empty) file with the processname in this
 * directory.
 * 
 * If a monitored process "foo" is not running anymore, respawnd will exec the
 * one of the following commands:
 *
 * - If the Execute bit of the file in the passed directory is set, respawnd
     will assume the file is a script and will execute it.
 * - If the Execute bit is off, respawnd will assume that it is an 'old style'
 *   file and will execute:  "/etc/init.d/foo restart" 

 * The subdirectory fatal can be used to monitor critical processes. If
 * the process can not be restarted after MAX_RESPAWN_COUNT times a reboot is forced.
 * The subdirectory non_fatal can be used to monitor non critical processes. 
 * Respawnd will always attempt to restart these processes without attempting
 * to reboot.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <ctype.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <aplog.h>
#include <sys/stat.h>

#define DEBUG(x)

extern void set_malloc_mmap_size(size_t);

#define DIR_BUFFER_LENGTH	40
#define MAX_PROCESSES		20
#define READ_BUF_SIZE   	50
#define MAX_RESPAWN_COUNT	5
#define RESPAWN_SLEEP		5	/* in seconds */

struct respawnlist_s {
	char *name;
};

struct respawnlist_s respawnlist[MAX_PROCESSES];
char foundlist[MAX_PROCESSES];

char *procname;
char *directory_name;
int do_exit = 0;

static void sigterm(int);
static void reboot_system(void);
static int read_procfs(int);
static int read_respawnlist(char *, int*);
static void execute(char *);

int main(int argc, char *argv[])
{
	int nof_entries = 0;
	int ret;
	int respawn_count = 0;
	char fatal_dir[DIR_BUFFER_LENGTH];
	char non_fatal_dir[DIR_BUFFER_LENGTH];
	
	procname = argv[0];
	if(argc != 2) {
		fprintf(stderr, "Usage: %s respawn_dir\n", argv[0]);
		return -1;
	}

	signal(SIGTERM, sigterm);
	set_malloc_mmap_size(8000);
	
	/* From the passed directory in argv[1] two subdirectories
	 * are determined <argv[1]>fatal/ and <argv[1]>non_fatal/
	 */
	ret = snprintf( (char *)&fatal_dir, sizeof(fatal_dir), 
		        "%sfatal/", argv[1] );
	/* Make sure the subdirectory name fits. */
	if ( ret > sizeof(fatal_dir) ) {
		fprintf(stderr, "Argument passed to respawnd is too big\n");
		return 0;
	}

	ret = snprintf( (char *)&non_fatal_dir, sizeof(non_fatal_dir), 
			"%snon_fatal/", argv[1]);
	if ( ret > sizeof(non_fatal_dir) ) {
		fprintf(stderr, "Argument passed to respawnd is too big\n");
		return 0;
	}

	while(!do_exit) {
		// First check if fatal processes are scheduled.
		directory_name = (char *)&fatal_dir;
		ret = read_respawnlist(directory_name, &nof_entries);
		if(ret < 0) {
			break;
		}

		if(nof_entries > 0) {
			ret = read_procfs(nof_entries);
			/* If we respawned a process from the list on
			 * the fatal directory, increase the respawn
			 * count, otherwise decrease */
			if(ret)
				respawn_count += 1;
			else
				respawn_count -= 1;
		}

		if(respawn_count > MAX_RESPAWN_COUNT)
			reboot_system();

		if(respawn_count < 0)
			respawn_count = 0;

		/* Do the same check for the non fatal processes
		 * but do not update the respawn_count.
		 */
		directory_name = (char *)&non_fatal_dir;
		ret = read_respawnlist(directory_name, &nof_entries);
		if(ret < 0) {
			break;
		}

		if(nof_entries > 0) {
			ret = read_procfs(nof_entries);
		}
		sleep(RESPAWN_SLEEP);
	}
	return 0;
}

static void sigterm(int signal)
{
	aplog(EV_SYS_RSPWN_TERM,"",NULL);
	DEBUG(printf("%s: Terminating\n", procname));
	do_exit = 1;
}

static void reboot_system(void)
{
	aplog(EV_SYS_RSPWN_REBOOT,"",NULL);
	DEBUG(printf("%s: Rebooting the system\n", procname));
	/* do a reboot */
	kill(1, SIGTERM);
}

static int read_respawnlist(char *respawn_dir, int *nof_entries)
{
	DIR *dir;
	int i;
	struct dirent *next;

	/* Free strings that were strdup-ed */
	for (i = 0; i < *nof_entries; i++)
		free(respawnlist[i].name);

	*nof_entries = 0;
	dir = opendir(respawn_dir);
	if (!dir) {
		fprintf(stderr, "%s: Cannot open dir %s, exiting.\n", procname, respawn_dir);
		return -1;
	}

	memset(respawnlist, 0, sizeof(respawnlist));
	while ((*nof_entries < MAX_PROCESSES) &&
		(next = readdir(dir)) != NULL) {
		char *p;

		if(!strcmp(next->d_name, ".") || !strcmp(next->d_name, ".."))
			continue;

		p = strdup(next->d_name);
		if(!p) {
			fprintf(stderr, "%s: Out of memory\n", procname);
			closedir(dir);
			return -1;
		}
		DEBUG(printf("Adding %s to respawn list\n", p));
		respawnlist[*nof_entries].name = p;
		(*nof_entries)++;
	}
	closedir(dir);
	return 0;
}

static int read_procfs(int nof_entries)
{
	DIR *dir;
	struct dirent *next;
	int i, ret = 0;

	dir = opendir("/proc");
	if (!dir) {
		fprintf(stderr, "Cannot open /proc.");
		return ret;
	}
	memset(foundlist, 0, sizeof(foundlist));
	while ((next = readdir(dir)) != NULL) {
		FILE *fp;
		char filename[READ_BUF_SIZE];
		char buffer[READ_BUF_SIZE];
		char name[READ_BUF_SIZE];

		/* If it isn't a number, we don't want it */
		if (!isdigit(*next->d_name))
			continue;

		sprintf(filename, "/proc/%s/status", next->d_name);
		if (! (fp = fopen(filename, "r")) ) {
			continue;
		}
		if (fgets(buffer, READ_BUF_SIZE - 1, fp) == NULL) {
			fclose(fp);
			continue;
		}
		fclose(fp);

		/* Buffer should contain a string like "Name:   binary_name" */
		sscanf(buffer, "%*s %s", name);

		for (i = 0; i < nof_entries; i++) {
			if(strcmp(respawnlist[i].name, name))
				continue;

			DEBUG(printf("Found %s, pid %ld\n", name, strtol(next->d_name, NULL, 0)));
			foundlist[i] = 1;
			break;
		}
	}
	closedir(dir);
	for (i = 0; i < nof_entries; i++) {
		if(!foundlist[i]) {
			ret = 1;
			aplog(EV_SYS_RSPWN_RESPAWNING, respawnlist[i].name, NULL);
			execute(respawnlist[i].name);
		}
	}
	return ret;
}

static void execute(char *filename)
{
	pid_t pid, pid2;
	char *cmd[4];
	char buffer[READ_BUF_SIZE];
	struct stat st;
	/* get the execute bit of the file 'filename' in 'directory_name' */
	snprintf(buffer, sizeof(buffer), "%s/%s", directory_name, filename);
	
	if(stat(buffer, &st) != 0) {
		fprintf(stderr, "respawnd: File %s does not exist\n", buffer);		
	}

	if(st.st_mode &  S_IXUSR) {
		/* execute bit is not set, execute file */
		cmd[0] = "/bin/sh";
		cmd[1] = buffer;
		cmd[2] = NULL;
	}
	else {
		/* old style */
		sprintf(buffer, "/etc/init.d/%s", filename);
		cmd[0] = "/bin/sh";
		cmd[1] = buffer;
		cmd[2] = "restart";
		cmd[3] = NULL;
	}
	
	DEBUG(printf("%s: respawning %s\n", procname, buffer));
	
	/* Do a double vfork to make the respawned process a child of init */
	if ( (pid = vfork()) == 0 )
	{
		if ( ( pid2 = vfork()) == 0)
		{
			execv(cmd[0], cmd);
			/* Something went wrong */
			exit(-1);
		}
		exit(0);
	} else {
		/* wait until child 1 ends */
		waitpid(pid, NULL, 0);
	}
}
