/*  $Header$
 *  
 *  sdhcp - simple dhcp server
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 
/* include files */
#include <errno.h>
#include <arpa/inet.h>
#include <sys/socket.h>

#include "debug.h"
#include "dhcp.h"
#include "sdhcpd.h"

/* function prototypes */
extern int write_lease_file( const char *filename );

static void dhcpdiscover( struct packet_t *packet );
static void dhcprequest( struct packet_t *packet );
static void dhcprelease( struct packet_t *packet );
static void dhcpdecline( struct packet_t *packet );
static void dhcpinform( struct packet_t *packet );
static void dhcpoffer( struct packet_t *packet,
		       struct lease_t *lease,
		       u_int32_t lease_time );
static void dhcpack( struct packet_t *packet,
		     struct lease_t *lease,
		     u_int32_t lease_time );
static void dhcpnak( struct packet_t *packet, char *errmsg );
static void dhcpreply(struct packet_t *packet,
		      struct lease_t *lease,
		      unsigned char msgtype,
		      u_int32_t lease_time );
static int add_dhcp_option( struct dhcp_packet *packet, unsigned char type,
			    unsigned char len, unsigned char *data);
static int add_mandatory_dhcp_options(struct dhcp_packet *packet,
				      unsigned char msg_type);

#define INFINITY	0xffffffff

/*****************************************************************************
 * NAME:
 *	dhcp
 *
 * DESCRIPTION:
 *	dispatch the packet to the appropriate handler
 *	depending on the message type
 *
 * PARAMETERS:
 *	*packet		- IN  - link to parsed packet
 *
 * RETURN:
 *	NONE
 *****************************************************************************/
void
dhcp( struct packet_t *packet )
{
    switch( packet->msg_type ) {
    case DHCPDISCOVER:
	dhcpdiscover( packet );
	break;

    case DHCPREQUEST:
	dhcprequest( packet );
	break;

    case DHCPRELEASE:
	dhcprelease( packet );
	break;

    case DHCPDECLINE:
	dhcpdecline( packet );
	break;

    case DHCPINFORM:
	dhcpinform( packet );
	break;

    default:
	note( "Unknown DHCP message: %02X from %s via %s", packet->msg_type,
	  print_hw_addr( packet->raw->chaddr, packet->raw->hlen ),
	  packet->raw->giaddr.s_addr ?
	  inet_ntoa( packet->raw->giaddr ) : server_ifinfo.name );
	break;
    }
}

/*****************************************************************************
 * NAME:
 *	dhcpdiscover
 *
 * DESCRIPTION:
 *	On reception of a discover message the server
 *
 * PARAMETERS:
 *	*packet		- IN  - link to parsed packet
 *
 * RETURN:
 *	NONE
 *****************************************************************************/
static void
dhcpdiscover( struct packet_t *packet )
{
    struct lease_t *lease = (struct lease_t *)0;
    struct hardware_t from_hwaddr;
    struct in_addr requested_addr;
    int size;

    do {
	note( "DISCOVER from %s via %s",
	      print_hw_addr( packet->raw->chaddr, packet->raw->hlen ),
	      packet->raw->giaddr.s_addr ?
	      inet_ntoa( packet->raw->giaddr ) : server_ifinfo.name );

	/* get hardware address we received this packet from */
	from_hwaddr.type = packet->raw->htype;
	from_hwaddr.len = packet->raw->hlen;
	memcpy( &from_hwaddr.u.chaddr, packet->raw->chaddr, from_hwaddr.len );

	/* check if we already offered an (old) response we should reuse */
	lease = find_offered_lease_by_hwaddr( &from_hwaddr );
	if( lease ) {
	    /* verify if address is still free */
	    if( ping_check( lease->ip_addr, PING_TIMEOUT ) != 1 ){
		/* it's no good anymore, remove it from administration */
		remove_lease( &pending_lease_list, lease );
		remove_lease( &own_lease_list, lease );
		break;
	    }

	    /* our old entry is still useable -- use it */
	    lease->state = cs_selecting_e;
	    dhcpoffer( packet, lease, INFINITY );
	    break;
	}

	/*
	 * Apparently it's the first time we receive a discover from this
	 * host (or an earlier discover has expired), so we try to allocate
	 * a new lease
	 */

	/* get the requested address and use that if in subnet */
	size = sizeof(requested_addr);
	if( get_dhcp_option( packet->raw, packet->length,
			     DHO_DHCP_REQUESTED_ADDRESS,
                             (unsigned char *)&requested_addr, &size) < 0) {
	    warn( "Failed to parse dhcp options -- ignoring message" );
	    break;
	}

	/* if no lease is found, try to allocate one */
	lease = add_pending_lease( packet, requested_addr );
	if( lease ) {
	    dhcpoffer( packet, lease, INFINITY );
	    break;
	}

	/* TODO: remove the old leases from the lease list to free addresses */

	/* If we get here we didn't succeed in creating an offer */
	warn( "All avaliable leases taken" );

    } while( 0 );

    print_leases();

}

/*****************************************************************************
 * NAME:
 *	dhcprequest
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 *	*packet		- IN  - link to parsed packet
 *
 * RETURN:
 *	NONE
 *****************************************************************************/
static void
dhcprequest( struct packet_t *packet )
{
    struct lease_t *lease = (struct lease_t *)0;
    struct hardware_t from_hwaddr;
    struct in_addr server_id, requested_addr;
    int size;

    do {
	note( "REQUEST from %s via %s",
	      print_hw_addr( packet->raw->chaddr, packet->raw->hlen ),
	      packet->raw->giaddr.s_addr ?
	      inet_ntoa( packet->raw->giaddr ) : server_ifinfo.name );

	/* get hardware address we received this packet from */
	from_hwaddr.type = packet->raw->htype;
	from_hwaddr.len = packet->raw->hlen;
	memcpy( &from_hwaddr.u.chaddr, packet->raw->chaddr, from_hwaddr.len );

	/* retrieve server identifier */
	size = sizeof(server_id);
	if( get_dhcp_option( packet->raw, packet->length,
			     DHO_DHCP_SERVER_IDENTIFIER,
                             (unsigned char *)&server_id, &size) < 0) {
	    warn( "Failed to parse dhcp options -- ignoring message" );
	    break;
	}

	/* retrieve the requested IP address */
	size = sizeof(requested_addr);
	if( get_dhcp_option( packet->raw, packet->length,
			     DHO_DHCP_REQUESTED_ADDRESS,
                             (unsigned char *)&requested_addr, &size) < 0) {
	    warn( "Failed to parse dhcp options -- ignoring message" );
	    break;
	}

	/* it should be a client we know because of the request */
	lease = find_offered_lease_by_hwaddr( &from_hwaddr );
	if( !lease ) {
	    /* probably a client rebooting that has moved or something */
	    break;
	}

	/*
	 * only now we can determine the client state for a bound client
	 * by examining the destination address of the request
	 * unicast for RENEWING, broadcast for REBINDING
	 */
	if( lease->state == cs_bound_e) {
	    if( packet->raw->yiaddr.s_addr == htonl( INADDR_BROADCAST)) {
		lease->state = cs_rebinding_e;
	    }else{
		lease->state = cs_renewing_e;
	    }
	} else if( lease->state == cs_init_reboot_e ) {
	    /* maybe we did reset and our leases are renewing / rebinding */
	    if( packet->raw->yiaddr.s_addr == htonl( INADDR_BROADCAST )) {
		/* really init-reboot or rebinding */
		if( packet->raw->ciaddr.s_addr != htonl( 0 )) {
		    lease->state = cs_rebinding_e;
		}
	    } else {
		lease->state = cs_renewing_e;
	    }
	}

	/* depending on the client state we must change our behaviour */
	switch( lease->state ) {
	case cs_selecting_e:
	    /* server id must be filled in */
	    if( !server_id.s_addr ) {
		dhcpnak( packet, "No server identifier in SELECTING" );
		lease->state = cs_init_e;
		break;
	    }
	    /* are we the selected server? */
	    if( server_id.s_addr != server_ifinfo.ip_addr.s_addr ) {
		remove_lease( &pending_lease_list, lease );
		break;
	    }
	    /* ciaddr must be zero */
	    if( packet->raw->ciaddr.s_addr ) {
		dhcpnak( packet, "ciaddr filled in in SELECTING" );
		lease->state = cs_init_e;
		break;
	    }
	    /* requested IP address must be same as from offer */
	    if( requested_addr.s_addr != lease->ip_addr.s_addr ) {
		dhcpnak( packet, "Requested IP address filled out with a " \
			 "different value than offered" );
		lease->state = cs_init_e;
		break;
	    }
	    /* all is fine -- ack it */
	    dhcpack( packet, lease, INFINITY );
	    lease->state = cs_bound_e;
	    /* move lease to own lease list (if needed) */
	    if(find_lease_by_hwaddr(pending_lease_list,&from_hwaddr)==lease){
		if(move_lease(lease,&own_lease_list,&pending_lease_list)<0)
		    warn("Internal error in administration");

		/*
		 * make own lease administration permanent
		 * we a do it for every new lease managed by us
		 * since we use fixed leases this should not happen often
		 * after the initial learning of the clients on the subnet
		 */
		(void)write_lease_file( lease_file );
	    }
	    break;

	case cs_init_reboot_e:
	    /* server id must not be filled in */
	    if( server_id.s_addr ) {
		dhcpnak( packet, "Server identifier in INIT-REBOOT" );
		lease->state = cs_init_e;
		break;
	    }
	    /* ciaddr must be zero */
	    if( packet->raw->ciaddr.s_addr ) {
		dhcpnak( packet, "ciaddr filled in in INIT-REBOOT" );
		lease->state = cs_init_e;
		break;
	    }
	    /* requested IP address must be filled in */
	    if( !requested_addr.s_addr ) {
		dhcpnak( packet, "No requested IP address in INIT-REBOOT" );
		lease->state = cs_init_e;
		break;
	    }
	    /* check if client is on the correct subnet */
	    if( (requested_addr.s_addr & netmask.s_addr) != subnet.s_addr ) {
		dhcpnak( packet, "Client is on the wrong subnet" );
		lease->state = cs_init_e;
		break;
	    }
	    /* check if requested address matches the one in our admin */
	    if( requested_addr.s_addr != lease->ip_addr.s_addr) {
		dhcpnak( packet, "Unexpected requested IP address" );
		lease->state = cs_init_e;
		break;
	    }
	    /* all is fine -- ack it */
	    dhcpack( packet, lease, INFINITY );
	    lease->state = cs_bound_e;
	    break;

	case cs_renewing_e:
	    /* server id must not be filled in */
	    if( server_id.s_addr ) {
		dhcpnak( packet, "Server identifier in RENEWING" );
		lease->state = cs_init_e;
		break;
	    }
	    /* requested IP address must not be filled in */
	    if( requested_addr.s_addr ) {
		dhcpnak(packet,"No requested IP address allowed in RENEWING");
		lease->state = cs_init_e;
		break;
	    }
	    /* ciaddr must be filled in with clients address */
	    if( packet->raw->ciaddr.s_addr != lease->ip_addr.s_addr ) {
		dhcpnak( packet, "Wrong ciaddr in RENEWING" );
		lease->state = cs_init_e;
		break;
	    }
	    /* all is fine -- ack it */
	    dhcpack( packet, lease, INFINITY );
	    lease->state = cs_bound_e;
	    break;

	case cs_rebinding_e:
	    /* server id must not be filled in */
	    if( server_id.s_addr ) {
		dhcpnak( packet, "Server identifier in REBINDING" );
		lease->state = cs_init_e;
		break;
	    }
	    /* requested IP address must not be filled in */
	    if( requested_addr.s_addr ) {
		dhcpnak(packet,"No requested IP address allowed in REBINDING");
		lease->state = cs_init_e;
		break;
	    }
	    /* ciaddr must be filled in with clients address */
	    if( packet->raw->ciaddr.s_addr != lease->ip_addr.s_addr ) {
		dhcpnak( packet, "Wrong ciaddr in REBINDING" );
		lease->state = cs_init_e;
		break;
	    }
	    /* IP address should be used (by this client) -- test it */
	    if( ping_check( lease->ip_addr, PING_TIMEOUT ) == 1){
		dhcpnak( packet, "IP address unexpectedly released" );
		lease->state = cs_init_e;
		break;
	    }

	    /* all is fine -- ack it */
	    dhcpack( packet, lease, INFINITY );
	    lease->state = cs_bound_e;
	    break;

	default:
	    warn( "Unexpected request, client state = %s",
		  print_client_state( lease->state ));
	    break;
	}
    } while( 0 );
    print_leases();
}

/*****************************************************************************
 * NAME:
 *	dhcprelease
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 *	*packet		- IN  - link to parsed packet
 *
 * RETURN:
 *	NONE
 *****************************************************************************/
static void
dhcprelease( struct packet_t *packet )
{
    struct lease_t *lease = (struct lease_t *)0;
    struct hardware_t from_hwaddr;

    note( "RELEASE from %s via %s",
	  print_hw_addr( packet->raw->chaddr, packet->raw->hlen ),
	  packet->raw->giaddr.s_addr ?
	  inet_ntoa( packet->raw->giaddr ) : server_ifinfo.name );

    /* get hardware address we received this packet from */
    from_hwaddr.type = packet->raw->htype;
    from_hwaddr.len = packet->raw->hlen;
    memcpy( &from_hwaddr.u.chaddr, packet->raw->chaddr, from_hwaddr.len );

    /* update administration the lease is inactive */
    lease = find_offered_lease_by_hwaddr( &from_hwaddr );
    if( lease ) {
	lease->state = cs_init_reboot_e;
    }
    print_leases();
}

/*****************************************************************************
 * NAME:
 *	dhcpdecline
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 *	*packet		- IN  - link to parsed packet
 *
 * RETURN:
 *	NONE
 *****************************************************************************/
static void
dhcpdecline( struct packet_t *packet )
{
    struct lease_t *lease = (struct lease_t *)0;
    struct hardware_t from_hwaddr;

    note( "DECLINE from %s via %s",
	  print_hw_addr( packet->raw->chaddr, packet->raw->hlen ),
	  packet->raw->giaddr.s_addr ?
	  inet_ntoa( packet->raw->giaddr ) : server_ifinfo.name );

    /* get hardware address we received this packet from */
    from_hwaddr.type = packet->raw->htype;
    from_hwaddr.len = packet->raw->hlen;
    memcpy( &from_hwaddr.u.chaddr, packet->raw->chaddr, from_hwaddr.len );

    /* update administration the lease is unavailable */
    lease = find_offered_lease_by_hwaddr( &from_hwaddr );
    if( lease ) {
	lease->state = cs_uninitialized_e;

	if(find_lease_by_hwaddr(pending_lease_list,&from_hwaddr)==lease) {
	    move_lease( lease, &unavailable_lease_list, &pending_lease_list );
	} else {
	    move_lease( lease, &unavailable_lease_list, &own_lease_list );
	}
	/* update permanent lease administration */
	(void)write_lease_file( lease_file );
    }
    print_leases();
}

/*****************************************************************************
 * NAME:
 *	dhcpinform
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 *	*packet		- IN  - link to parsed packet
 *
 * RETURN:
 *	NONE
 *****************************************************************************/
static void
dhcpinform( struct packet_t *packet )
{
    note( "INFORM from %s via %s",
	  print_hw_addr( packet->raw->chaddr, packet->raw->hlen ),
	  packet->raw->giaddr.s_addr ?
	  inet_ntoa( packet->raw->giaddr ) : server_ifinfo.name );
}

/*****************************************************************************
 * NAME:
 *	dhcpoffer
 *
 * DESCRIPTION:
 *	Send a DHCP offer
 *
 * PARAMETERS:
 *	*packet		- IN  - requesting packet
 *	*lease		- IN  - associated lease
 *
 * RETURN:
 *	NONE
 *****************************************************************************/
static void
dhcpoffer(struct packet_t *packet, struct lease_t *lease, u_int32_t lease_time)
{
    dhcpreply( packet, lease, DHCPOFFER, lease_time );
}

/*****************************************************************************
 * NAME:
 *	dhcpack
 *
 * DESCRIPTION:
 *	Send a DHCP ack
 *
 * PARAMETERS:
 *	*packet		- IN  - requesting packet
 *	*lease		- IN  - associated lease
 *	u_int32_t	- IN  - lease time
 *
 * RETURN:
 *	NONE
 *****************************************************************************/
static void
dhcpack( struct packet_t *packet, struct lease_t *lease, u_int32_t lease_time )
{
    dhcpreply( packet, lease, DHCPACK, lease_time );
}

/*****************************************************************************
 * NAME:
 *	dhcpreply
 *
 * DESCRIPTION:
 *	Send a DHCP reply (ack, or offer)
 *
 * PARAMETERS:
 *	*packet		- IN  - requesting packet
 *	*lease		- IN  - associated lease
 *	unsigned char	- IN  - message type
 *	u_int32_t	- IN  - lease time
 *
 *
 * RETURN:
 *	NONE
 *****************************************************************************/
static void
dhcpreply(
    struct packet_t *packet,
    struct lease_t *lease,
    unsigned char msgtype,
    u_int32_t lease_time)
{
    struct dhcp_packet out;
    struct sockaddr_in to;
    int outlen, flags, result = 0;
    int fd;

    memset( &out, 0, sizeof( out ) );

    out.op = BOOTREPLY;
    out.htype = lease->hw_addr.type;
    out.hlen = lease->hw_addr.len;

    /* out.hops = 0; */
    out.xid = packet->raw->xid;
    /* out.secs = ntohs(0); */
    out.flags = packet->raw->flags;

    /* out.ciaddr.s_addr = ntohl(0); */
    out.yiaddr = lease->ip_addr;
    /* out.siaddr.s_addr = ntohl(0); */
    out.giaddr = packet->raw->giaddr;
    memcpy( out.chaddr, lease->hw_addr.u.chaddr, lease->hw_addr.len );

    /* strncpy( out.sname, server_name, sizeof(out.sname) ); */
    /* strncpy( out.file, boot_file, sizeof(out.file) ); */

    outlen = DHCP_FIXED_NON_UDP;
    outlen += add_mandatory_dhcp_options( &out, msgtype );

    /* only add lease time if > 0 */
    if( lease_time ) {
	outlen += add_dhcp_option( &out,
				   DHO_DHCP_LEASE_TIME,
				   sizeof(lease_time),
				   (unsigned char *)&lease_time);
    }

    /* add subnet mask option */
    if( netmask.s_addr ) {
	outlen += add_dhcp_option( &out,
				   DHO_SUBNET_MASK,
				   sizeof( netmask ),
				   (unsigned char *)&netmask);
    }

    /* add gateway address */
    if( router.s_addr ) {
	outlen += add_dhcp_option( &out,
				   DHO_ROUTERS,
				   sizeof( router ),
				   (unsigned char *)&router);
    }

    /* add dns addresses */
    if( dns_entries > 0 ) {
	outlen += add_dhcp_option( &out,
				   DHO_DOMAIN_NAME_SERVERS,
				   sizeof( struct in_addr ) * dns_entries,
				   (unsigned char *)&dns);
    }

    /* send message to gateway if present, otherwise broadcast on subnet */
    to.sin_family = AF_INET;
    memset( to.sin_zero, 0, sizeof(to.sin_zero) );
    if( packet->raw->giaddr.s_addr ) {
	to.sin_addr = packet->raw->giaddr;
	to.sin_port = htons(LISTEN_PORT); /* listen port because of router */
    } else {
	/* only send a unicast to a renewing client to avoid arping
	 * without the client being bound to an IP address yet */
	if( !( packet->raw->flags & BOOTP_BROADCAST )
	    && ( lease->state != cs_renewing_e ) )
	    to.sin_addr.s_addr = htonl( INADDR_BROADCAST );
	else
	    to.sin_addr = out.yiaddr;
	to.sin_port = htons(SEND_PORT);
    }

    flags = 0;
    if( server_ifinfo.alias_fd != -1)
	fd = server_ifinfo.alias_fd;
    else
	fd = server_ifinfo.main_fd;
    result = sendto( fd, (char *)&out, outlen, flags, &to, sizeof(to) );
    if( result < 0 ) {
	warn( "Failed to send DHCPOFFER/DHCPACK: %m");
    }
}

/*****************************************************************************
 * NAME:
 *	dhcpnak
 *
 * DESCRIPTION:
 *	Send a DHCP offer
 *
 * PARAMETERS:
 *	packet_t *	- IN  - requesting packet
 *	char *		- IN  - error message (zero terminated string)
 *
 * RETURN:
 *	NONE
 *****************************************************************************/
static void
dhcpnak( struct packet_t *packet, char *errmsg )
{
    struct dhcp_packet out;
    struct sockaddr_in to;
    int outlen, flags, result = 0;
    int fd;

    memset( &out, 0, sizeof( out ) );

    out.op = BOOTREPLY;
    out.htype = packet->raw->htype;
    out.hlen = packet->raw->hlen;

    /* out.hops = 0; */
    out.xid = packet->raw->xid;
    /* out.secs = ntohs(0); */
    out.flags = packet->raw->flags | htons( BOOTP_BROADCAST );

    /* out.ciaddr.s_addr = ntohl(0); */
    /* out.yiaddr.s_addr = ntohl(0); */
    /* out.siaddr.s_addr = ntohl(0); */
    out.giaddr = packet->raw->giaddr;
    memcpy( out.chaddr, packet->raw->chaddr, packet->raw->hlen );

    /* strncpy( out.sname, server_name, sizeof(out.sname) ); */
    /* strncpy( out.file, boot_file, sizeof(out.file) ); */

    outlen = DHCP_FIXED_NON_UDP;
    outlen += add_mandatory_dhcp_options( &out, DHCPNAK );

    outlen += add_dhcp_option( &out,
			       DHO_DHCP_MESSAGE,
			       strlen( errmsg ),
			       errmsg );

    /* send message to gateway if present, otherwise broadcast on subnet */
    to.sin_family = AF_INET;
    memset( to.sin_zero, 0, sizeof(to.sin_zero) );
    if( packet->raw->giaddr.s_addr ) {
	to.sin_addr = packet->raw->giaddr;
	to.sin_port = htons(LISTEN_PORT); /* listen port because of router */
    } else {
	to.sin_addr.s_addr = htonl( INADDR_BROADCAST );
	to.sin_port = htons(SEND_PORT);
    }

    flags = 0;
    if( server_ifinfo.alias_fd != -1)
	fd = server_ifinfo.alias_fd;
    else
	fd = server_ifinfo.main_fd;
    result = sendto( fd, (char *)&out, outlen, flags, &to, sizeof(to) );
    if( result < 0 ) {
	warn( "Failed to send DHCPNAK: %m");
    }
}

/*****************************************************************************
 * NAME:
 *	add_mandatory_dhcp_options
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 *
 * RETURN:
 *	total length of option field in DHCP message
 *****************************************************************************/
static int
add_mandatory_dhcp_options(struct dhcp_packet *packet, unsigned char msg_type)
{
    int optlen = 5; /* cookie + end */

    /* add option cookie */
    memcpy( packet->options, DHCP_OPTIONS_COOKIE, 4 );

    /* setup option end */
    packet->options[4] = DHO_END;

    /*
     * we know that the following options fit in the packet
     * so no error check has to be performed
     */
    optlen += add_dhcp_option( packet,
			       DHO_DHCP_MESSAGE_TYPE,
			       sizeof(msg_type),
			       &msg_type );

    /* server identifier is its interface IP address */
    optlen += add_dhcp_option( packet,
			       DHO_DHCP_SERVER_IDENTIFIER,
			       sizeof(server_ifinfo.ip_addr),
			       (unsigned char *)&server_ifinfo.ip_addr );

    return ( optlen );
}

/*****************************************************************************
 * NAME:
 *	add_dhcp_option
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 *
 * RETURN:
 *	total length of option field in DHCP message
 *	-1 on error
 *****************************************************************************/
static int
add_dhcp_option(
    struct dhcp_packet *packet,
    unsigned char type,
    unsigned char len,
    unsigned char *data)
{
    unsigned char *ptr;
    unsigned char *start = packet->options + 4; /* MAGIC */
    unsigned char *end = (unsigned char *)packet + sizeof(*packet);

    /* move to end marker in options field */
    for( ptr = start; *ptr != DHO_END && ptr < end; ) {
	unsigned char optlen;

	/* Pad options don't have a length - just skip them. */
	if (*ptr == DHO_PAD) {
	    ++ptr;
	    continue;
	}
	/* skip complete option */
	optlen = ptr[1];
	ptr += optlen + 2;
    }

    /* check if new option fits in packet */
    if( (ptr + len + 2) > end ) {
	warn( "Option doesn't fit in packet" );
	return -1;
    }

    /* add type, length, option into packet */
    *ptr++ = type;
    *ptr++ = len;
    memcpy( ptr, data, len );
    ptr += len;

    /* mark the new end */
    *ptr++ = DHO_END;

    return( len + 2 );
}

/*****************************************************************************
 * NAME:
 *	get_dhcp_option
 *
 * DESCRIPTION:
 *	extract an option from a dhcp packet and store the option
 *	in the desired location, update the size field with the actual
 *	length of the option.
 *
 * PARAMETERS:
 *	*packet		- INPUT - address of packet to parse
 *	packet_length	- INPUT - length of packet
 *	option		- INPUT - option to extract from messages
 *	*store		- INPUT - location where options can be stored
 *	store_size	- BOTH  - IN  -> max size of store
 *				  OUT -> actual size of store
 *
 * RETURN:
 *	-1 : error
 *	0  : success
 *****************************************************************************/
int
get_dhcp_option(
    struct dhcp_packet *packet,
    int packet_length,
    unsigned char option,
    unsigned char *option_store,
    int *store_size
)
{
    unsigned char *s;
    unsigned char *start = (unsigned char *)packet->options + 4; /* MAGIC */
    unsigned char *end = (unsigned char *)packet + packet_length;
    unsigned char len;
    unsigned char code;

    memset( option_store, 0, *store_size );
    for (s = start; *s != DHO_END && s < end; ) {
	code = s [0];
	/* Pad options don't have a length - just skip them. */
	if (code == DHO_PAD) {
	    ++s;
	    continue;
	}
	/* All other fields (except end, see above) have a
	   one-byte length. */
	len = s [1];

	/* If the length is outrageous, the options are bad. */
	if (s + len + 2 > end) {
	    warn ("Option %d length %d overflows input buffer.",
		  option, len);
	    return -1;
	}

	/* is this the option we're looking for? */
	if( code == option ) {
	    /* see if it fits in option store we received */
	    if( len > *store_size) {
		warn( "Option doesn't fit in store" );
		return -1;
	    }
	    memcpy( option_store, s+2, len );
	    *store_size = len;
	    return 0; /* found it */
	}else{
	     /* adjust pointer including option & length field */
	    s = s + len + 2;
	}
    }
    memset( option_store, 0, *store_size );
    return 0; /* packet ok, but option not present */
}


