/*  $Header$
 *  
 *  sdhcp - simple dhcp server
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <arpa/inet.h>

#include "confparse.h"
#include "debug.h"
#include "sdhcpd.h"

/******************************************************************************
 * local defines
 *****************************************************************************/
#define LOCAL_BUF_SIZE 128

/******************************************************************************
 * static function prototypes
 *****************************************************************************/
static int install_ip_token( const char *token, struct in_addr *in );




int
get_dns_entries( const char *filename, struct in_addr *dns, int max_entries )
{
    int cnt = 0;
    FILE *fp;
    char buffer[LOCAL_BUF_SIZE];
    char ipaddr[IFNAMSIZ];
    int length;

    if(( fp = fopen( filename, "r") ) == NULL) {
	warn( "Failed to open %s: %m", filename );
	return -1;
    }

    while( cnt < max_entries && !feof( fp ) && !ferror( fp )) {
	length = LOCAL_BUF_SIZE;
	read_text_line( buffer, &length, fp );
	if( length == 0 ) {
	    continue;
	}

	if( sscanf( buffer, "nameserver#%s", ipaddr ) == 1) {
	    if( inet_aton( ipaddr, &dns[cnt] )){
		debug( "DNS[%d] = %s", cnt, inet_ntoa( dns[cnt] ) );
		cnt++;
	    }
	}
    }
    if( ferror( fp ) ) {
	warn( "Failed to read %s: %m", filename);
    }
    fclose( fp );
    return cnt;
}

/*
 * DESCRIPTION:
 *	Read the configuration from file with name 'filename'
 *	and store the configuration in the global parameters
 */
int
get_configuration(const char *filename)
{
    int rc = 0;			/* result code == success */
    char buffer[LOCAL_BUF_SIZE];
    char *p = NULL;
    int length;
    FILE *fp;
    int subnet_found = 0;	/* flag to indicate subnet definition */

    /* open configuration file */
    if(( fp = fopen( filename, "r" )) == NULL) {
	warn( "Can't open configuration file '%s': %m", filename );
	return -1;	/* failure */
    }

    /* parse all lines from the configuration file */
    while( !feof( fp ) && !ferror( fp ) ) {
	length = LOCAL_BUF_SIZE;
	read_text_line( buffer, &length, fp );
	if( length == 0 ) {
	    continue;
	}
	/* get token identifier */
	p = strtok( buffer, "#" );
	if( !strcmp( T_SUBNET, p )) {
	    /* read subnet */
	    if( install_ip_token( T_SUBNET, &subnet ) < 0 ){
		rc = -1;
		break;
	    }
	    subnet_found = 1;
	} else if( !strcmp( T_NETMASK, p )) {
	    /* read netmask */
	    if( install_ip_token( T_NETMASK, &netmask ) < 0 ){
		rc = -1;
		break;
	    }
	} else if( !strcmp( T_ROUTER, p )) {
	    /* read router address */
	    if( install_ip_token( T_ROUTER, &router ) < 0 ){
		rc = -1;
		break;
	    }
	} else {
	    warn( "Unknown token: %s", p );
	    rc = -1;
	    break;
	}
    }

    if( ferror( fp ) ) {
	warn( "Failed to read %s: %m", filename);
    }

    /* check for mandatory items */
    if( !subnet_found ){
	warn( "Configuration item '%s' is missing", T_SUBNET );
	rc = -1;
    }

    /* check if own if address part of this subnet */
    if( (server_ifinfo.ip_addr.s_addr & netmask.s_addr) != subnet.s_addr ) {
	warn ( "Server not part of configured subnet" );
	rc = -1;
    }

    /* close configuration file again */
    fclose( fp );

    return( rc );
}

/*
 * install IP address with name token in address 'in'
 * return 0 on success, -1 on failure
 */
static int
install_ip_token( const char *token, struct in_addr *in )
{
    char *p;

    if(( p = strtok( NULL, "#" )) == NULL ){
	warn( "Address of %s definition is missing", token );
	return -1;	/* failure */
    }
    if( !inet_aton( p, in) ) {
	warn( "Incorrect %s address: %s", token, p );
	return -1;
    }
    debug( "Installing %s = %s", token, inet_ntoa( *in ));

    return 0;
}

/*
 * Read and install leases file
 * return 0 on success, -1 on failure
 */
int
get_leases (const char *filename)
{
    int rc = 0; 		/* resultcode = success */
    char buffer[LOCAL_BUF_SIZE];
    char *p = NULL;
    int length;
    FILE *fp;
    struct lease_t *tmp = NULL;

    /* open lease file */
    if(( fp = fopen( filename, "r" )) == NULL) {
	warn( "Can't open lease file '%s': %m", filename );
	return -1;	/* failure */
    }

    /* parse all lines from the configuration file */
    while( !feof( fp ) && !ferror( fp ) ) {
	length = LOCAL_BUF_SIZE;
	read_text_line( buffer, &length, fp );
	if( length == 0 ) {
	    continue;
	}
	/* get IP address */
	p = strtok( buffer, "#" );
	tmp = (struct lease_t *)calloc( 1, sizeof( struct lease_t ));
	if( tmp == NULL ) {
	    warn( "Lease: %m" );
	    rc = -1;
	    break;
	}
	if( !inet_aton( p, &tmp->ip_addr ) ) {
	    warn( "Incorrect IP address '%s' in file '%s", p, filename );
	    free( tmp ); tmp = NULL;
	    rc = -1;
	    continue;
	}
	/* must be in subnet we server */
	if( (tmp->ip_addr.s_addr & netmask.s_addr) != subnet.s_addr) {
	    char nettxt[16], masktxt[16];
	    sprintf( nettxt, "%s", inet_ntoa( subnet ));
	    sprintf( masktxt, "%s", inet_ntoa( netmask ));
	    warn( "IP address %s not part of our subnet %s mask %s" \
		  " -- Skipping entry",
		  inet_ntoa( tmp->ip_addr ), nettxt, masktxt);
	    free( tmp ); tmp = NULL;
	    continue;
	}

	/* get option */
	p = strtok( NULL, "#" );
	switch( *p ){
	case 'H':
	    /* associate raw Client Hardware address with IP address */
	    {
		int i = 0;
		unsigned long uval;

		p = strtok( NULL, "#" );
		debug( "Associating IP %s with Client Hardware address %s",
		       inet_ntoa( tmp->ip_addr ), p );
		p = strtok( p, "/" );
		tmp->hw_addr.type = atoi(p);
		p = strtok( NULL, "/" );
		tmp->hw_addr.len = atoi(p);
		if( tmp->hw_addr.len == 0 || tmp->hw_addr.len > 16 ) {
		    free( tmp ); tmp = NULL;
		    warn( "Incorrect length for Client Hardware address" );
		    continue;
		}
		p = strtok( NULL, "/" );
		p = strtok( p, ":" );
		while( p ) {
		    if( i >= 16 ) {
			/* incorrect if length > 16 */
			i = 0;
			break;
		    }
		    uval = strtoul( p, (char**)0, 16);
		    tmp->hw_addr.u.chaddr[i++] = (u_int8_t)(uval & 0xff);
		    p = strtok( NULL, ":" );
		}
		if( i != tmp->hw_addr.len ) {
		    free( tmp ); tmp = NULL;
		    warn( "Client Hardware address doesn't match length" );
		    continue;
		}
		tmp->state = cs_init_reboot_e;

		/* add lease to internal administration */
		tmp->next = own_lease_list;
		own_lease_list = tmp;
		break;
	    }
	case 'M':
	    /* associate MAC address with IP address */
	    {
		int i = 0;
		unsigned long uval;

		p = strtok( NULL, "#" );
		debug( "Associating IP %s with MAC address %s",
		       inet_ntoa( tmp->ip_addr ), p );
		p = strtok( p, ":" );
		while( p ) {
		    if( i >= 6 ) {
			/* incorrect if length > 6 */
			i=0;
			break;
		    }
		    tmp->hw_addr.type = 1; /* ethernet */
		    tmp->hw_addr.len = 6;
		    uval = strtoul( p, (char**)0, 16);
		    tmp->hw_addr.u.haddr[i++] = (u_int8_t)(uval & 0xff);
		    p = strtok( NULL, ":" );
		}
		if( i < 6 ) {
		    /* MAC address must have size 6 */
		    free( tmp ); tmp = NULL;
		    warn( "Incorrect MAC address -- skipping" );
		    continue;
		}
		tmp->state = cs_init_reboot_e;

		/* add lease to internal administration */
		tmp->next = own_lease_list;
		own_lease_list = tmp;
		break;
	    }
	case 'R':
	    /* IP address reserved */
	    debug( "IP address %s reserved", inet_ntoa( tmp->ip_addr ));

	    /* add lease to internal administration */
	    tmp->next = reserved_lease_list;
	    reserved_lease_list = tmp;
	    break;
	case 'U':
	    /* IP address unavailable */
	    debug( "IP address %s unavailable",
		   inet_ntoa( tmp->ip_addr ));
	    /* add lease to internal administration */
	    tmp->next = unavailable_lease_list;
	    unavailable_lease_list = tmp;
	    break;
	default:
	    warn( "Unsupported option: %s", p );
	    free( tmp ); tmp = NULL;
	    continue;
	}
    }
    if( ferror( fp ) ) {
	warn( "Failed to read %s: %m", filename);
    }

    /* close configuration file again */
    fclose( fp );
    return( rc );
}

/*****************************************************************************
 * NAME:
 *	write_lease_file
 *
 * DESCRIPTION:
 *	Write the internal lease administration to the lease file for permanent
 *	storage. At first the data is written to a temporary file that will
 *	replace the lease file at the end of the operation.
 *
 * PARAMETERS:
 *	const char * filename	- IN  - filename
 *
 * RETURN:
 *	 0 - success
 *	-1 - error
 *****************************************************************************/
int
write_lease_file( const char *filename )
{
    struct lease_t *lease;
    char new_file_name[128];
    char back_file_name[128];
    FILE *fp;

    /* make a temporary lease file */
    sprintf( new_file_name, "%s.tmp", filename );
    if( (fp = fopen( new_file_name, "w" )) == NULL ) {
	warn( "Can't open new lease file: %m" );
	return -1;
    }

    /* write an introduction */
    fprintf( fp, "# Lease file of sdhcpd\n" );
    fprintf( fp, "# This file is automatic generated and describes the" \
	     " relation of IP addresses\n" );
    fprintf( fp, "# in the format:\n" );
    fprintf( fp, "# <IP address> <Action> [Value]\n" );
    fprintf( fp, "#\n# Action:\n" );
    fprintf( fp, "# H - association with client hardware address (type,length" \
	     ",value)\n" );
    fprintf( fp, "# M - association with MAC address\n" );
    fprintf( fp, "# R - IP address is reserved\n" );
    fprintf( fp, "# U - IP address (temporary) unavailable\n" );

    /* start with adding the reserved addresses */
    fprintf( fp, "\n# Reserved addresses\n" );
    if( !reserved_lease_list ) {
	fprintf( fp, "# <none>\n" );
    } else {
	for( lease = reserved_lease_list; lease; lease = lease->next ) {
	    fprintf( fp, "%s\tR\n", inet_ntoa( lease->ip_addr ));
	}
    }

    /* add the own administrated address, try to use MAC notation */
    fprintf( fp, "\n# Associated addresses\n" );
    if( !own_lease_list ) {
	fprintf( fp, "# <none>\n" );
    } else {
	for( lease = own_lease_list; lease; lease = lease->next ) {
	    if( lease->hw_addr.type == 1 && lease->hw_addr.len == 6 ) {
		/* MAC notation for ethernet type address */
		fprintf( fp, "%s\tM\t%s\n", inet_ntoa( lease->ip_addr ),
			 print_hw_addr( lease->hw_addr.u.haddr,
					lease->hw_addr.len ));
	    } else {
		fprintf( fp, "%s\tH\t%d/%d/%s\n", inet_ntoa( lease->ip_addr ),
			 lease->hw_addr.type, lease->hw_addr.len,
			 print_hw_addr( lease->hw_addr.u.chaddr,
					lease->hw_addr.len ));
	    }
	}
    }

    /* and finally add the (temporary) unavailable address */
    fprintf( fp, "\n# Unavailable addresses\n" );
    if( !unavailable_lease_list ) {
	fprintf( fp, "# <none>\n" );
    } else {
	for( lease = unavailable_lease_list; lease; lease = lease->next ) {
	    fprintf( fp, "%s\tU\n", inet_ntoa( lease->ip_addr ));
	}
    }

    /* make a backup of last good lease file */
    sprintf( back_file_name, "%s~", filename );
    if( (unlink(back_file_name) < 0) && errno != ENOENT ) {
	warn( "Can't remove old lease file backup file %s: %m",
	      back_file_name );
	fclose( fp );
	return -1;
    }
    if( link( filename, back_file_name ) < 0 ){
	warn( "Can't backup lease file %s: %m", filename );
	fclose( fp );
	return -1;
    }
    /* make the temp file the current lease file */
    if( rename( new_file_name, filename ) < 0) {
	warn( "Can't install new lease file %s to %s: %m",
	      new_file_name, filename );
	fclose( fp );
	return -1;
    }

    fclose( fp );
    return 0;
}

