/*  $Header$
 *  
 *  sdhcp - simple dhcp server
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdlib.h>
#include <arpa/inet.h>

#include "sdhcpd.h"
#include "debug.h"

/*****************************************************************************
 * NAME:
 *	find_lease
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 *	*packet		- IN  - link to parsed packet
 *
 * RETURN:
 *	NONE
 *****************************************************************************/
struct lease_t *
find_offered_lease_by_hwaddr( struct hardware_t *hw_addr )
{
    struct lease_t *lease = (struct lease_t *)0;

    /* did we already provide a lease */
    lease = find_lease_by_hwaddr( pending_lease_list, hw_addr );
    if( lease ) {
	/* we're already busy with it, return the same lease */
	return lease;
    }

    /* try to find an old lease using the hardware address */
    lease = find_lease_by_hwaddr( own_lease_list, hw_addr );
    if( lease ) {
	/* we know this one, try to use same settings as before */
	return lease;
    }

    return lease;
}

/*****************************************************************************
 * NAME:
 *	find_lease_by_hwaddr
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 *	lease_t *	- IN  - list to search
 *	in_addr		- IN  - IP address
 *
 * RETURN:
 *	lease if found, NULL if not found
 *****************************************************************************/
struct lease_t *
find_lease_by_hwaddr( struct lease_t *lease_list, struct hardware_t *hw_addr )
{
    struct lease_t *lease = (struct lease_t *)0;

    for( lease = lease_list; lease; lease = lease->next ) {
	if( match_hw_addr( &lease->hw_addr, hw_addr ) ) {
	    break;
	}
    }
    return lease;
}

/*****************************************************************************
 * NAME:
 *	find_lease_by_ipaddr
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 *	lease_t *	- IN  - list to search
 *	in_addr		- IN  - IP address
 *
 * RETURN:
 *	lease if found, NULL if not found
 *****************************************************************************/
struct lease_t *
find_lease_by_ipaddr( struct lease_t *lease_list, struct in_addr ipaddr )
{
    struct lease_t *lease = (struct lease_t *)0;

    for( lease = lease_list; lease; lease = lease->next ) {
	if( lease->ip_addr.s_addr == ipaddr.s_addr ) {
	    break;
	}
    }
    return lease;
}

/*****************************************************************************
 * NAME:
 *	add_pending_lease
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 *	*packet		- IN  - link to parsed packet
 *	address		- IN  - suggested address
 *
 * RETURN:
 *	lease_t *
 *****************************************************************************/
struct lease_t *
add_pending_lease( struct packet_t *packet, struct in_addr address )
{
    struct in_addr free_address;
    struct lease_t *lease = (struct lease_t *)0;

    /* see if the suggested address is useable */
    do {
	/* address requested? */
	if( address.s_addr == htonl(0)) {
	    break;
	}

	/* must be in our subnet */
	if( (address.s_addr & netmask.s_addr) != subnet.s_addr ) {
	    debug("Requested address %s is not part of our subnet -- ignoring",
		  inet_ntoa( address ));
	    free_address.s_addr = htonl(0);
	    break;
	}

	/* must be an unknown address */
	if( is_known_ipaddr( address ) ) {
	    debug("Address %s already part of lease administration",
		  inet_ntoa( address ));
	    free_address.s_addr = htonl(0);
	    break;
	}

	if( ping_check( address, PING_TIMEOUT ) != 1 ){
	    debug( "Suggested address %s in use",
		   inet_ntoa( address) );
	    break;
	}

	/* It's ok -- use this address */
	free_address.s_addr = address.s_addr;
    } while( 0 );

    /* try to get a free ip address, if non suggested */
    if( free_address.s_addr == ntohl(0) ) {
	if( get_free_ip_address( &free_address ) < 0 ) {
	    return lease;
	}
    }

    /* fill out the lease info */
    lease = (struct lease_t *)calloc( 1, sizeof( struct lease_t ));
    if( !lease ) {
	warn( "Failed to create lease: %m" );
	return lease;
    }
    lease->ip_addr = free_address;
    lease->hw_addr.type = packet->raw->htype;
    lease->hw_addr.len = packet->raw->hlen;
    memcpy( lease->hw_addr.u.chaddr, packet->raw->chaddr, packet->raw->hlen );
    lease->state = cs_selecting_e;
    lease->next = pending_lease_list;
    pending_lease_list = lease;

    return lease;
}

/*****************************************************************************
 * NAME:
 *	remove_lease
 *
 * DESCRIPTION:
 *	remove a lease from a selected list
 *
 * PARAMETERS:
 *	lease_t **	- IN  - link to address of the selected list
 *	lease_t *	- IN  - link to lease to remove
 *
 * RETURN:
 *	NONE
 *****************************************************************************/
void
remove_lease( struct lease_t **lease_list, struct lease_t *lease )
{
    struct lease_t *tmp;

    if( !lease )
	return;

    /* first entry in the list */
    if( lease == *lease_list ) {
	*lease_list = lease->next;
	free( lease ); lease = NULL;
	return;
    }

    /* part of list */
    for( tmp = *lease_list; tmp; tmp = tmp->next ) {
	if( tmp->next == lease ) {
	    /* found entry */
	    tmp->next = lease->next;
	    free( lease ); lease = NULL;
	    break;
	}
    }
    return;
}

/*****************************************************************************
 * NAME:
 *	move_lease
 *
 * DESCRIPTION:
 *	remove a lease from a selected list
 *
 * PARAMETERS:
 *	lease_t *	- IN  - link to lease to move
 *	lease_t **	- IN  - link to address of the destination list
 *	lease_t **	- IN  - link to address of the source list
 *
 * RETURN:
 *	-1 on error, 0 on success
 *****************************************************************************/
int
move_lease(
    struct lease_t *lease,
    struct lease_t **dst_list,
    struct lease_t **src_list )
{
    struct lease_t *tmp;

    /* make sure lease is not in destination list, to avoid multiple copies */
    for( tmp = *dst_list; tmp; tmp = tmp->next ) {
	if( tmp == lease ) {
	    warn( "move_lease: lease exists in destination");
	    return -1;
	}
    }

    /* first entry in src list */
    if( lease == *src_list ) {
	*src_list = lease->next;
	lease->next = *dst_list;
	*dst_list = lease;
	return 0;
    }

    /* part of list */
    for( tmp = *src_list; tmp; tmp = tmp->next ) {
	if( tmp->next == lease) {
	    tmp->next = lease->next;
	    lease->next = *dst_list;
	    *dst_list = lease;
	}
    }
    return 0;
}

void
remove_list( struct lease_t **lease_list )
{
    struct lease_t *tmp;

    /* clean up pending lease list */
    if( *lease_list)
    {
	do {
	    tmp = (*lease_list)->next;
	    free( *lease_list );
	    *lease_list = tmp;
	} while( *lease_list );
    }
}
/*****************************************************************************
 * NAME:
 *	is_known_ipaddr
 *
 * DESCRIPTION:
 *	Check if IP address is part of the internal lease administration
 *
 * PARAMETERS:
 *	struct in_addr
 *
 * RETURN:
 *	1 - equal
 *	0 - unequal
 *****************************************************************************/
unsigned char
is_known_ipaddr( struct in_addr address )
{
    if( find_lease_by_ipaddr( own_lease_list, address )
	|| find_lease_by_ipaddr( pending_lease_list, address )
	|| find_lease_by_ipaddr( unavailable_lease_list, address )
	|| find_lease_by_ipaddr( reserved_lease_list, address ))
	return 1;
    else
	return 0;
}

/*****************************************************************************
 * NAME:
 *	get_free_ip_address
 *
 * DESCRIPTION:
 *
 * PARAMETERS:
 *	*address	- OUT - location to store in_addr
 *
 * RETURN:
 *	-1 on failure
 *	0 on success
 *****************************************************************************/
int
get_free_ip_address( struct in_addr *address )
{
      /* start is subnet mask + 1, can't give out subnet as host address */
    unsigned long start = ntohl( subnet.s_addr ) + 1;
      /* end is subnet + ~mask, excluding the broadcast */
    unsigned long end = start + ntohl( ~netmask.s_addr ) - 1;
    unsigned long i;

    for( i = start; i < end; i++) {
	address->s_addr = htonl( i );
	/* check if IP address in use in own administration */
	if( is_known_ipaddr( *address ) )
	    continue;

	/* check if IP address is part of our subnet */
	if( ( htonl( i ) & netmask.s_addr ) == subnet.s_addr ){
	    /* check if IP address is in use by ICMP message */
	    if( ping_check( *address, PING_TIMEOUT ) != 1)
		continue;
	    break;
	}
    }
    if( i == end ) {
	/* no free address available */
	return -1;
    }
    return 0;
}

/*****************************************************************************
 * NAME:
 *	match_hw_addr
 *
 * DESCRIPTION:
 *	Compare two hardware addresses and report if they're equal
 *
 * PARAMETERS:
 *	struct hardware_t addr1 & addr2
 *
 * RETURN:
 *	1 - equal
 *	0 - unequal
 *****************************************************************************/
unsigned char
match_hw_addr( struct hardware_t *addr1, struct hardware_t *addr2 )
{
    if( ( addr1->type == addr2->type )
	&& ( addr1->len == addr2->len)
	&& ( memcmp(addr1->u.chaddr, addr2->u.chaddr, addr1->len) == 0 ))
	return 1;
    else
	return 0;
}


