/*  $Header$
 *  
 *  sdhcp - simple dhcp server
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <time.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/ip_icmp.h>

#include "sdhcpd.h"
#include "debug.h"

/*****************************************************************************
 * Function Prototypes
 *****************************************************************************/
static int icmp_echorequest ( int icmp_fd, struct in_addr addr);
static int icmp_echoreply( int icmp_fd, struct in_addr addr );
static u_int32_t checksum (unsigned char *buf, int nbytes, u_int32_t sum);
static u_int32_t wrapsum (u_int32_t sum);


/*****************************************************************************
 * NAME:
 *	icmp_echorequest
 *
 * DESCRIPTION:
 *	Send a ICMP echo request (ping) to a target
 *
 * PARAMETERS:
 *	int		- IN  - file descriptor of ICMP socket to use
 *	in_addr		- IN  - IP address to ping
 *
 * RETURN:
 *	 0 on success
 *	-1 on error
 *****************************************************************************/
static int
icmp_echorequest ( int icmp_fd, struct in_addr addr)
{
    struct sockaddr_in to;
    struct icmp icmp;
    int status;

    to.sin_family = AF_INET;
    to.sin_port = 0;
    to.sin_addr = addr;

    icmp.icmp_type = ICMP_ECHO;
    icmp.icmp_code = 0;
    icmp.icmp_cksum = 0;
    icmp.icmp_seq = 0;
    icmp.icmp_id = addr.s_addr;

    icmp.icmp_cksum = wrapsum (checksum ((unsigned char *)&icmp,
					 sizeof icmp, 0));

    /* Send the ICMP packet... */
    status = sendto (icmp_fd, (char *)&icmp, sizeof icmp, 0,
		     (struct sockaddr *)&to, sizeof to);
    if (status < 0)
	warn ("icmp_echorequest %s: %m", inet_ntoa(to.sin_addr));

    if (status != sizeof icmp)
	return -1;
    return 0;
}

/*****************************************************************************
 * NAME:
 *	icmp_echoreply
 *
 * DESCRIPTION:
 *	Retrieve an ICMP message from the ICMP socket and check it's a reply
 *	from the selected target.
 *
 * PARAMETERS:
 *	int		- IN  - file descriptor of ICMP socket to use
 *	in_addr		- IN  - IP address to ping
 *
 * RETURN:
 *	 0 on match
 *	-1 on error
 *****************************************************************************/
static int
icmp_echoreply( int icmp_fd, struct in_addr addr )
{
    struct icmp *icfrom;
    struct sockaddr_in from;
    u_int8_t icbuf [1500];
    int status;
    int len;

    len = sizeof from;
    status = recvfrom ( icmp_fd, (char *)icbuf, sizeof icbuf, 0,
		       (struct sockaddr *)&from, &len);
    if (status < 0) {
	warn ("icmp_echoreply: %m");
	return -1;
    }
    /* is the message from our target */
    if( from.sin_addr.s_addr != addr.s_addr ) {
	return -1;
    }

    /* Probably not for us. */
    if (status < (sizeof (struct ip)) + (sizeof *icfrom)) {
	return -1;
    }

    len = status - sizeof (struct ip);
    icfrom = (struct icmp *)(icbuf + sizeof (struct ip));

    /* Silently discard ICMP packets that aren't echoreplies. */
    if (icfrom -> icmp_type != ICMP_ECHOREPLY) {
	return -1;
    }
    /* the target did reply */
    return 0;
}

/*****************************************************************************
 * NAME:
 *	checksum
 *
 * DESCRIPTION:
 *	Calculate the checksum over a selected data block
 *
 * PARAMETERS:
 *	*unsigned char	- IN  - pointer to start of data block
 *	int		- IN  - number of bytes
 *	u_int32_t	- OUT - checksum
 *
 * RETURN:
 *	u_int32_t checksum (same as in parameter list)
 *****************************************************************************/
static u_int32_t
checksum (unsigned char *buf, int nbytes, u_int32_t sum)
{
    int i;

    /* Checksum all the pairs of bytes first... */
    for (i = 0; i < (nbytes & ~1); i += 2) {
	sum += (u_int16_t) ntohs(*((u_int16_t *)(buf + i)));
	/* Add carry. */
	if (sum > 0xFFFF)
	    sum -= 0xFFFF;
    }

    /* If there's a single byte left over, checksum it, too.   Network
       byte order is big-endian, so the remaining byte is the high byte. */
    if (i < nbytes) {
	sum += buf [i] << 8;
	/* Add carry. */
	if (sum > 0xFFFF)
	    sum -= 0xFFFF;
    }

    return sum;
}

/* Finish computing the sum, and then put it into network byte order. */
static u_int32_t
wrapsum (u_int32_t sum)
{
    sum = ~sum & 0xFFFF;
    return htons(sum);
}


/*****************************************************************************
 * NAME:
 *	ping_check
 *
 * DESCRIPTION:
 *	Check if an IP address is in use by sending a ping request to it.
 *	If after timeout seconds still no reponse from the target has been
 *	received it's assumed that the address is not in use.
 *
 * PARAMETERS:
 *	in_addr		- IN  - IP address to check
 *	long		- IN  - timeout (sec)
 *
 * RETURN:
 *	 1 - IP address free
 *	 0 - IP address in use
 *	-1 - error
 *****************************************************************************/
int
ping_check( struct in_addr addr, unsigned long timeout )
{
    int icmp_fd;
    fd_set r, w, x;
    struct timeval tm;
    time_t last_time;

    /* open a raw socket for sending a ping */
    if( (icmp_fd = open_icmp_socket()) < 0 ) {
	return -1;
    }

    /* send ping to address */
    if( icmp_echorequest( icmp_fd, addr ) < 0 ) {
	close( icmp_fd );
	return -1;
    }

    FD_ZERO( &w );
    FD_ZERO( &x );

    /* wait for a reply or timeout */
    /* TODO: make it asynchronous */
    tm.tv_usec = 0;
    time( &last_time );
    while ( timeout > 0 ) {
	FD_ZERO( &r );
	FD_SET( icmp_fd, &r);
	tm.tv_sec  = timeout;
	if( select( icmp_fd + 1, &r, &w, &x, &tm ) < 0) {
	    warn( "ICMP select failed: %m" );
	    close( icmp_fd );
	    return -1;
	}

	if( FD_ISSET( icmp_fd, &r ) ) {
	    if( icmp_echoreply( icmp_fd, addr ) == 0 ) {
		close( icmp_fd );
		return 0; /* we received a reply, address is in use*/
	    }
	}
	/* update remaining time to wait */
	timeout -= time(NULL) - last_time;
	time( &last_time );
    }
    close( icmp_fd );
    return 1; /* didn't received a reply, address is free */
}

