/*  $Header$
 *  
 *  sdhcp - simple dhcp server
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <syslog.h>
#include <fcntl.h>
#include <string.h>
#include <sys/param.h>
#include <net/if.h>
#include <netinet/in.h>

#include "dhcp.h"
#include "sdhcpd.h"
#include "debug.h"

/* prototypes */
extern int get_configuration(const char *filename);
extern int get_leases (const char *filename);
extern int get_dns_entries( const char *filename, struct in_addr *dns,
			    int max_entries );

static void usage(char *);
static void handle_message( int sock );
static void dispatch();
static void initialise();

/*****************************************************************************/
/* global variables                                                          */
/*****************************************************************************/
  /* interface on which DHCPD must operate */
char *interface_name = "eth0";
  /* the name of the configuration file */
char *conf_file = DHCPD_CONF_FILE;
char *lease_file = DHCPD_LEASES_FILE;
unsigned char	interface_hwaddr[6];
  /* my interface info */
struct if_info_t server_ifinfo;
  /* the subnet we serve with handing out ip addresses */
struct in_addr subnet;
  /* the netmask on our subnet */
struct in_addr netmask;
  /* the router address we give to clients */
struct in_addr router;
  /* array of Domain Name Servers */
struct in_addr dns[MAX_DNS_ENTRIES];
  /* number of dns entries */
int dns_entries = -1;
  /* list of leases */
struct lease_t *own_lease_list = NULL;
  /* list of pending leases */
struct lease_t *pending_lease_list = NULL;
  /* list of reserved leases (permanent unavailable) */
struct lease_t *reserved_lease_list = NULL;
  /* list of temporary unavailable leases */
struct lease_t *unavailable_lease_list = NULL;
  /* the debug level */
int debug_level = 0;

static void usage( char * appname ) {
    fprintf( stderr, "Usage: %s [-h] [-d level] [-cf config-file]\n", appname);
    fprintf( stderr, "              [-lf lease-file] [interface]\n" );
    exit(1);
}

int
main( int argc, char *argv[] )
{
    int i;
    char *appname;

    /* set all the default values */
    initialise();

    /* determine application name from argumentlist */
    appname = strrchr( argv[0], '/');
    if( !appname )
	appname = argv[0];
    else
	appname++;

    /* parse arguments */
    for( i = 1; i < argc; i++ ) {
	if( !strcmp( argv[i], "-cf" )) {
	    if( ++i == argc )
		usage( appname );
	    conf_file = argv[i];
	} else if( !strcmp( argv[i], "-lf" )) {
	    if( ++i == argc )
		usage( appname );
	    lease_file = argv[i];
	} else if( !strcmp( argv[i], "-d" )) {
	      if( ++i == argc )
		  usage( appname );
	      debug_level = atoi( argv[i] );
	} else if( !strcmp( argv[i], "-h" )) {
	    usage( appname );
	} else {
	    strncpy( server_ifinfo.name, argv[i],
		     sizeof(server_ifinfo.name));
	    interface_name = argv[i];
	}
    }

    /* open syslog and report start of dhcpd */
    openlog( appname, LOG_NDELAY, LOG_DAEMON );
    note( "Simple DHCP Server v%s (%s) started on %s",
	  VER, BT, interface_name );

    /* open interface */
    if( open_interface( &server_ifinfo ) == -1) {
	error( "Failed to open interface" );
    }

    /* make myself the default gateway, this can be overruled with the
     * ROUTER option in the config file */
    router = server_ifinfo.ip_addr;

    /* read configuration */
    if( get_configuration( conf_file ) < 0 ) {
	error( "Configuration file errors encountered" ); /* EXIT */
    }

    /* read the addresses of the domain name server */
    dns_entries = get_dns_entries( RESOLV_FILENAME, (struct in_addr *)&dns,
				   MAX_DNS_ENTRIES );

    /* read and install previous leases */
    get_leases( lease_file );

    /* start dispatching the messages */
    dispatch();

    /* should never executes */
    error( "Fall through" );
    closelog();
    cleanup();
    return 0;
}

static void
dispatch ()
{
    fd_set r, w, x;
    int max = 0;
    int select_count, cnt;
    struct timeval *tvp = NULL;

    FD_ZERO (&w);
    FD_ZERO (&x);

    do {
	/* Set up the read mask. */
	FD_ZERO (&r);

	FD_SET( server_ifinfo.main_fd, &r );
	if( server_ifinfo.alias_fd != -1) {
	    FD_SET( server_ifinfo.alias_fd, &r );
	}
	max = MAX( server_ifinfo.main_fd, server_ifinfo.alias_fd );

	/* Wait for a packet or a timeout... XXX */
	select_count = select (max + 1, &r, &w, &x, tvp);

	/* Not likely to be transitory... */
	if ( select_count < 0 )
	    error ("select: %m");

	for( cnt = 0; cnt <= max; cnt++ ) {
	    if( FD_ISSET( cnt, &r )) {
		handle_message( cnt );
	    }
	}
    } while (1);
}

static void
handle_message( int sock )
{
    struct sockaddr_in from;
    int fromlen, packet_length, size;
    unsigned char dhcp_message_type = 0;
    union {
	unsigned char packbuf[DHCP_MTU_MAX]; /* 1500 */
	struct dhcp_packet packet;
    } u;
    struct packet_t packet;

    /* read the data from the socket */
    if((packet_length = recvfrom( sock, (char *)u.packbuf, sizeof(u), 0,
				  &from, &fromlen)) < 0) {
	warn( "recvfrom failed: %m" );
	return;
    }

    if( packet_length == 0 ) {
	return;
    }

    /* if no MAGIC is present then we can skip it */
    if ( memcmp( u.packet.options, DHCP_OPTIONS_COOKIE, 4 )) {
	debug( "Received a non BOOTP/DHCP message -- Ignoring" );
	return;
    }

    /* check if it's a dhcp message */
    size = sizeof(dhcp_message_type);
    if( get_dhcp_option( &u.packet, packet_length, DHO_DHCP_MESSAGE_TYPE,
			 (unsigned char *)&dhcp_message_type, &size) < 0) {
	debug( "Failed to retrieve option DHCP_MESSAGE_TYPE from packet" );
    }

    /* fill in some fast access stuff parsed from the raw packet */
    memset( &packet, 0, sizeof(packet) );
    packet.raw = &u.packet;
    packet.length = packet_length;
    packet.msg_type = dhcp_message_type;

    /* handle the dhcp message */
    if( dhcp_message_type ) {
	dhcp(&packet);
    }else{
	warn( "Got BOOTP message -- not supported" );
    }
}


/*****************************************************************************
 * initialisation stuff
 *****************************************************************************/
static void
initialise()
{
    memset( &server_ifinfo, 0, sizeof( server_ifinfo ));
    server_ifinfo.main_fd = -1;
    server_ifinfo.alias_fd = -1;
    strncpy( server_ifinfo.name, interface_name, sizeof(server_ifinfo.name));
    netmask.s_addr = htonl(0);
    subnet.s_addr = htonl(0);
    router.s_addr = htonl(0);
}

/*****************************************************************************
 * clean up memory structures
 *****************************************************************************/
void
cleanup()
{
    /* clean up the lease lists */
    remove_list( &own_lease_list );
    remove_list( &pending_lease_list );
    remove_list( &reserved_lease_list );
    remove_list( &unavailable_lease_list );
}


