/*  $Header$
 *  
 *  sdhcp - simple dhcp server
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <errno.h>
#include <strings.h>
#include <syslog.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netinet/in.h>
#include <sys/ioctl.h>

#include "debug.h"
#include "sdhcpd.h"

/* function prototypes */
static int get_dhcp_socket( const char *ifname );

/*****************************************************************************
 * NAME:
 *	open_icmp_socket
 *
 * DESCRIPTION:
 *	Open a raw socket for ICMP messages
 *
 * PARAMETERS:
 *	None
 *
 * RETURN:
 *	filedescriptor >= 0 on success
 *	-1 on error
 *****************************************************************************/
int
open_icmp_socket()
{
    int protocol = ntohs(1);		/* ICMP */
    int state;
    int icmp_fd;

    /*
     * set protocol directly instead of using getprotobyname because of
     * embedded environment
     */
    icmp_fd = socket (AF_INET, SOCK_RAW, protocol);
    if (icmp_fd < 0){
	warn ("unable to create icmp socket: %m");
	return -1;
    }

    /* Make sure it does routing... */
    state = 0;
    if (setsockopt (icmp_fd, SOL_SOCKET, SO_DONTROUTE,
		    (char *)&state, sizeof state) < 0){
	warn ("Unable to disable SO_DONTROUTE on ICMP socket: %m");
	close( icmp_fd );
	return -1;
    }
    return icmp_fd;
}

/*****************************************************************************
 * NAME:
 *	open_interface
 *
 * DESCRIPTION:
 *	Open a raw socket for DHCP messages.
 *	The IP & MAC address are determined from the interface. If the
 *	interface is an aliased interface the primary interface is determined
 *	as well. Eventually a socket is bound to the interface(s).
 *
 * PARAMETERS:
 *	if_info_t*	- IN  - pointer to interface info structure
 *
 * RETURN:
 *	 0 on success
 *	-1 on error
 *****************************************************************************/
int
open_interface(struct if_info_t *if_info)
{
    char localname[IFNAMSIZ];
    char *primary_name;
    int server_socket = -1;
    struct ifreq ifr;
    struct sockaddr_in *sin;

    /* test the usability of this interface be getting the addresses */
    /* make a socket... */
    if( (server_socket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0){
	error( "Can't create socket on interface %s: %m",
	       if_info->name ); /* EXIT */
	return -1;
    }

    /* retrieve primary ip address */
    strncpy(ifr.ifr_name, if_info->name, IFNAMSIZ);
    ifr.ifr_addr.sa_family = AF_INET;
    if( ioctl(server_socket, SIOCGIFADDR, &ifr) < 0 ){
	error( "failed to get server_ip from %s: %m",
	       if_info->name ); /* EXIT */
	return -1;
    }
    sin = (struct sockaddr_in *)&ifr.ifr_addr;
    if_info->ip_addr = sin->sin_addr;
    note( "%s (server_ip) = %s", if_info->name,
	  inet_ntoa(if_info->ip_addr));

    /* retrieve MAC address */
    strncpy(ifr.ifr_name, if_info->name, IFNAMSIZ);
    if( ioctl(server_socket, SIOCGIFHWADDR, &ifr) < 0){
	error( "failed to get hwaddr from %s", if_info->name ); /* EXIT */
	return -1;
    }
    if_info->hw_addr.len = IFHWADDRLEN;
    memcpy(&if_info->hw_addr.u.haddr, &ifr.ifr_hwaddr.sa_data,
	   if_info->hw_addr.len);

    close( server_socket );

    /* check if it's an aliased interface */
    strncpy( localname, if_info->name, sizeof( localname ) );
    primary_name = strtok( localname, ":" );
    if( strcmp( primary_name, if_info->name ) != 0 ) {
	/* it's an aliased interface, bind to primary interface as well */
	if(( if_info->main_fd = get_dhcp_socket( primary_name )) < 0 ) {
	    return -1;
	} else {
	    if(( if_info->alias_fd = get_dhcp_socket( if_info->name )) < 0) {
		return -1;
	    }
	}
    } else {
	/* not an aliased interface */
	if(( if_info->main_fd = get_dhcp_socket( primary_name )) < 0 ) {
	    return -1;
	}
    }
    return 0;
}


/*****************************************************************************
 * NAME:
 *	get_dhcpd_socket
 *
 * DESCRIPTION:
 *	Open and bind the interface to a socket for receiving and sending
 *	DHCP messages. On success the socket is returned.
 *
 * PARAMETERS:
 *	char*		- IN  - interface name
 *
 * RETURN:
 *	 socket >=0 on success
 *	-1 on error
 *****************************************************************************/
int
get_dhcp_socket( const char *ifname )
{
    int sock = -1;
    struct sockaddr_in addr;
    int flag;

#if defined (HAVE_SO_BINDTODEVICE)
    struct ifreq intf;
#endif /* HAVE_SO_BINDTODEVICE */

    /* open an UDP socket */
    if( (sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0){
	error( "Can't create socket on interface %s: %m", ifname );
	return -1;
    }

    bzero(&addr, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(LISTEN_PORT);
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    memset( addr.sin_zero, 0, sizeof( addr.sin_zero ) );

    /*
     * Set the REUSEADDR option so that we don't fail to start
     * if we're being restarted
     */
    flag = 1; /* enable option */
    if( setsockopt( sock, SOL_SOCKET, SO_REUSEADDR,
		    (char *)&flag, sizeof(flag) ) < 0 ){
	error( "Can't set SO_REUSEADDR option on server socket: %m" );
	return -1;
    }

    /* Set the BROADCAST option so that we can broadcast DHCP responses */
    flag = 1; /* enable option */
    if(setsockopt( sock, SOL_SOCKET, SO_BROADCAST,
		  (char *)&flag, sizeof(flag)) < 0){
	error( "Can't set SO_BROADCAST option on server socket: %m");
	return -1;
    }

    /* Bind the socket to this interface's IP address */
    if( bind( sock, (struct sockaddr *)&addr,
	      sizeof( addr )) < 0 ){
	error( "Can't bind to server IP address: %m" ); /* EXIT */
	return -1;
    }

#if defined (HAVE_SO_BINDTODEVICE)
    /* Bind this socket to this interface */
    bzero( &intf, sizeof(intf) );
    strncpy( intf.ifr_name, ifname, IFNAMSIZ );
    if( setsockopt( sock, SOL_SOCKET, SO_BINDTODEVICE,
		    &intf, sizeof(intf)) < 0 ){
	error( "Can't set SO_BINDTODEVICE option on server socket: %m" );
	return -1;
    }
#endif /* HAVE_SO_BINDTODEVICE */

    return sock;
}

