
#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/inet.h>
#include <arpa/tftp.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include "datastore_if.h"
#include "hexstring.h"


//#define DEBUG  1

#ifdef DEBUG
#define DBG(x,y) fprintf(stderr, "%s:%d ", __FILE__, __LINE__) ; fprintf(stderr, x, y)
#else
#define DBG(x,y)
#endif

static int sockfd = 0;
static struct sockaddr_in to;
static struct sockaddr_in from;
static int idcounter = 0;

static char* server_address = "127.0.0.1";
static int   server_port = 1122;

void datastore_set_server_ip(char *ipaddr) {
	server_address = strdup(ipaddr);
}

void datastore_set_server_port(int port) {
	server_port = port;
}

static int socket_open (void);
static int socket_close (void);
static void socket_reset(void);

static int
cmdresp (char *cmd, 
	 char *oid, 
	 u_char *community, size_t community_len,
	 char *in_data, int in_data_len, 
	 char *out_data, int* out_data_len,
	 int in_type, int *out_type)
{
	int len;
	static char buf[1024];	/* static to save stack space */
	static char enc_data[MAX_DATA_LENGTH];
	static char enc_community[MAX_DATA_LENGTH];
	char *part[5];
	int parts;
	char *p;
	int address_len = sizeof (struct sockaddr_in);
	fd_set readfds;
	struct timeval tv;
	int retval;
	char *result;

	FD_ZERO (&readfds);
	
	idcounter++;

	if (!oid) {
		/* example: BEGIN, COMMIT, ROLLBACK */
		snprintf (buf, sizeof (buf), "%s %d\n", cmd, idcounter);
	}
	else {
		if (string_needs_encode(community, community_len, HEXSTRING_FLAGS_NOSPACES)){
			string_encode(community, community_len, enc_community, MAX_DATA_LENGTH);
		}
		else {
			strncpy(enc_community, community, MAX_DATA_LENGTH);
		}

		if (in_data ||  in_type != -1) {
			if (string_needs_encode(in_data, in_data_len, 0)){
				string_encode(in_data, in_data_len, enc_data, MAX_DATA_LENGTH);
			}
			else {
				strncpy(enc_data, in_data, MAX_DATA_LENGTH);
			}
		
			/* example, SET */
			snprintf (buf, sizeof(buf), "%s %d %s %s %d %s", cmd, idcounter, oid, enc_community, in_type, enc_data);
		} 
		else {
			/* example: GET, GETNEXT */
			snprintf (buf, sizeof (buf), "%s %d %s %s\n", cmd, idcounter, oid, enc_community);
		}
	}

	if (sockfd == 0)
		socket_open ();

	DBG ("sending to datastore: \n%s\n", buf);

	sendto (sockfd, buf, strlen (buf) + 1, 0, (struct sockaddr *) &to,
		address_len);

	FD_SET (sockfd, &readfds);

	tv.tv_sec = 4;
	tv.tv_usec = 0;

	retval = select (sockfd + 1, &readfds, NULL, NULL, &tv);

	if (retval) {
		len = recvfrom (sockfd, buf, sizeof (buf), 0,
				(struct sockaddr *) &from, &address_len);
	}
	else {
		/* reset socket */
		socket_reset();
		return -1;
	}

	// Cleanup : add trailing zero and remove newlines
	if (len < sizeof (buf))
		buf[len] = 0;
	p = strchr (buf, '\n');
	if (p)
		*p = 0;

	DBG ("received from datastore: \n%s\n", buf);

	parts = split (buf, ' ', part, 5);

	DBG ("Received %i part(s)\n", parts);

	if (parts < 2) {
		fprintf (stderr,
			 "datastore: Response to short (only %i part(s))\n",
			 parts);
		socket_reset();
		return (-1);
	}

	/* check id */
	if (atol (part[1]) != idcounter) {
		fprintf (stderr, "ID out of sync (ID: %d instead of %d)\n",part[1],idcounter);
		
		socket_reset();
		return (-1);
	}	

	/* copy oid */
	if (parts > 2) {
		strncpy (oid, part[2], MAX_OID_LENGTH);

		if (parts > 3) {
			*out_type = atoi (part[3]);
		
		
			if (parts > 4) {

								/* part[4] is always 0-terminated */
				*out_data_len = string_decode(part[4], out_data, MAX_DATA_LENGTH);
			}
			else {
				strcpy (out_data, "");
				*out_data_len = 0;
			}
		}
	}

	if (part[0]) {
		return (atoi (part[0]));
	}

	fprintf (stderr, "return code empty\n");
	socket_reset();
	return (-1);
}

static void
socket_reset(void)
{
	fprintf(stderr, "Resetting socket \n");
	socket_close();
	socket_open();
}
	

static int
socket_open (void)
{
	if ((sockfd = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
		fprintf (stderr, "Can't open socket : %s\n", strerror (errno));
		return (-1);
	}

	to.sin_family = AF_INET;
	to.sin_port = htons (server_port);
	to.sin_addr.s_addr = inet_addr (server_address);
	bzero (&to.sin_zero, sizeof (to.sin_zero));

	return (0);

}

static int
socket_close (void)
{
	close(sockfd);
	return (0);

}

int
split (char *str, char sep, char **part, int maxparts)
{
	int parts = 0;
	char *p1 = str;
	char *p2;

	part[parts++] = p1;

	while ((p2 = strchr (p1, sep)) && (parts < maxparts)) {
		*p2 = 0;
		p1 = p2 + 1;
		part[parts++] = p1;
	}

	return parts;
}

int
getoid (char *oid, u_char* community, size_t community_len, char *data, int* data_len, int *type)
{
	return(cmdresp("GETOID", oid, community, community_len, NULL, -1, data, data_len, -1, type));
}

int
setoid (char *oid,  u_char* community, size_t community_len, char *data, int* data_len, int *type)
{
	return(cmdresp("SETOID", oid, community, community_len,
			data, *data_len, 
			data, data_len, 
			*type, type));
}

int
getnextoid (char *oid, u_char* community, size_t community_len, char *data, int* data_len, int *type)
{
	return(cmdresp("GETNEXTOID", oid, community, community_len, NULL, -1, data, data_len, -1, type));
}

int 
begin_transaction()
{
	DBG("%s\n", "BEGIN TRANSACTION");
	return(cmdresp("BEGIN", NULL, NULL, -1,
			NULL, -1, 
			NULL, NULL,
			-1, NULL));
}

int 
commit_transaction()
{
	DBG("%s\n", "COMMIT TRANSACTION");
	return(cmdresp("COMMIT", NULL, NULL, -1,
			NULL, -1,
			NULL, NULL,
			-1, NULL));
}

int 
rollback_transaction()
{
	DBG("%s\n", "ROLLBACK TRANSACTION");
	return(cmdresp("ROLLBACK", NULL, NULL, -1,
			NULL, -1, 
			NULL, NULL,
			-1, NULL));
}
