/*
 * multicast.c: This takes care of subscribing the SNMP socket to the
 *		multicast group for SIDED support.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/route.h>
#include <netinet/igmp.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <linux/version.h>
#include <errno.h>
#include <netinet/ip_icmp.h>

#include "multicast.h"

/* local protos */

void set_multicast_route(char *ifname);
static int in_cksum(unsigned short *buf, int sz);

extern int snmp_sock_sd;


/*
 * Walks all interfaces and does the multicast subscribe on
 * every broadcast address found. 
 */

void sided_multicast_init(int boeitniet, void *data) 
{
	int sd, in;
	struct ifconf ifc;
	struct ifreq *ifr;
	struct  sockaddr_in *p;
	struct ip_mreq mreq;
	int mcast_group_joined = 0, mcast_route_set = 0;
	FILE *f;
	char buf[256];
	
	f = fopen("/proc/net/dev_mcast", "r");
	if(f == NULL) {
		fprintf(stderr, "Can't open /proc/net/dev_mcast : %s", strerror(errno));
		return;
	}

	while(fgets(buf, sizeof(buf), f) && !mcast_group_joined) {
		if(strstr(buf, "e00012b") != 0)
			mcast_group_joined = 1;
	}
	fclose(f);
	
	if ( !(mcast_group_joined) && (snmp_sock_sd != 0)) {

	        /* Check in /proc/net/route if the route to the mcast network exists.
	         * ( == 0000000E0 in reverse notwork byte ordeR ??!?!?)
	         */
	        f=fopen("/proc/net/route", "r");
	        if(f == NULL) {
	                fprintf(stderr, "Can't open /proc/net/route : %s", strerror(errno));
	                return;
	        }

	        while(fgets(buf, sizeof(buf), f) && !mcast_route_set) {
			if(strstr(buf, "000000E0") != 0)
				mcast_route_set = 1;
		}
		fclose(f);
		if(!mcast_route_set)
			return;

		/* Get configuration for all interfaces */
		if ( (sd = socket(AF_INET, SOCK_DGRAM, 0)) < 0 ) {
			printf("sided_multicast_init: Could not open socket\n");
		}
		ifc.ifc_len = sizeof(struct ifreq) * 10;
		ifc.ifc_buf = malloc(ifc.ifc_len);
		
		if(!ifc.ifc_buf) {
			close(sd);
			return;
		}

		if(ioctl(sd, SIOCGIFCONF, &ifc) < 0 ) {
			fprintf(stderr, "sided_multicast_init: Could not get interface info\n");
			free(ifc.ifc_buf);
			close(sd);
			return;
		}
	
		ifr = ifc.ifc_req;
		for(in = 0; in < ifc.ifc_len; in += sizeof(struct ifreq)) {
			/* We are not interested in the lo interface */
			if(strcmp(ifr->ifr_name, "lo") == 0) {
				ifr++;
				continue;
			}
	
			if(ioctl(sd, SIOCGIFBRDADDR, ifr) < 0) {
				fprintf(stderr, "sided_multicast_init: Could not get broadcast address %s\n", strerror(errno));
				ifr++;
				continue;
			}

			p = (struct sockaddr_in *)&(ifr->ifr_addr);
			
                        /* Route solicitation should not be sent on apipa/autoip address */
   			if(strncmp(inet_ntoa(p->sin_addr), "169\.254", 7) == 0) {
				ifr++;
				continue;
			}
			
			printf("sided_multicast_init: Send router sollicitation for if %-10s %-15s\n", 
						ifr->ifr_name, 
						inet_ntoa(p->sin_addr));

			send_router_solicitation(p->sin_addr.s_addr);
			ifr++;
		}
		free(ifc.ifc_buf);
	
		igmp_report( htonl(MCAST_INL_DISCOVERY) );

		mreq.imr_interface.s_addr = htonl(INADDR_ANY);
		mreq.imr_multiaddr.s_addr = htonl(MCAST_INL_DISCOVERY);

		/* We first drop the membership, because if we have reinitialized the
		 * IP addresses, the device does not have a Multicast membership anymore
		 * while the socket HAS (problem in the kernel?). Adding a member in this
		 * case would then fail, so we first drop it
		 */
		if( setsockopt(snmp_sock_sd, IPPROTO_IP, IP_DROP_MEMBERSHIP, &mreq, sizeof(mreq)) < 0 ) {
			/* This can occur when a group is joined for the first time */ 
		}

		if( setsockopt(snmp_sock_sd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0 ) {
			printf("sided_multicast_init: Could not add membership : %s\n" , strerror(errno));
		}
		close(sd);
	}
}


void send_router_solicitation(u_long tinaddr)
{
	int icmp_sock, len;
	struct sockaddr_in to;
	struct icmp icmp_msg;
	int optval = 1;
	
//	printf("tinaddr = %lX\n", tinaddr);

	if( (icmp_sock = socket( AF_INET, SOCK_RAW, 1 )) < 0) {	/* 1 = ICMP */
		fprintf(stderr, "socket (sendRouterSolicitation)\n");
	}

	if ( setsockopt(icmp_sock, SOL_SOCKET, SO_BROADCAST,
					&optval, sizeof(optval)) < 0 ) {
		fprintf(stderr, "setsockopt (sendRouterSolicitation)\n");
	}

	memset( &to, 0, sizeof(struct sockaddr_in) );
	to.sin_family = AF_INET;
	to.sin_addr.s_addr = tinaddr;

	memset( &icmp_msg, 0, sizeof(icmp_msg) );
	icmp_msg.icmp_type = ICMP_ROUTERSOLICIT;
	icmp_msg.icmp_code = 0;
	icmp_msg.icmp_void = htonl(0);			/* reserved */
	icmp_msg.icmp_cksum = in_cksum((unsigned short *) &icmp_msg, sizeof(icmp_msg));

	len = sendto(icmp_sock, &icmp_msg, sizeof(icmp_msg), 0,
		 (struct sockaddr *) &to, sizeof(struct sockaddr_in));
	if (len < 0 || len != sizeof(icmp_msg)) {
		fprintf(stderr, "sendto (sendRouterSolicitation) : %s\n", strerror(errno));
	}

	close(icmp_sock);
}

void igmp_report(u_long group)
{
	int igmp_sock, len;
	struct sockaddr_in to;
	struct igmp igmp_msg;

	if( (igmp_sock = socket( AF_INET, SOCK_RAW, 2 )) < 0) {	/* 2 = IGMP */
		fprintf(stderr, "socket (igmp_report)\n");
	}

	memset( &to, 0, sizeof(struct sockaddr_in) );
	to.sin_family = AF_INET;
	to.sin_addr.s_addr = group;

	memset( &igmp_msg, 0, sizeof(igmp_msg) );
	igmp_msg.igmp_type = IGMP_V1_MEMBERSHIP_REPORT;
	igmp_msg.igmp_code = 0;
	igmp_msg.igmp_group.s_addr = group;
	igmp_msg.igmp_cksum = in_cksum((unsigned short *) &igmp_msg, sizeof(igmp_msg));

	len = sendto(igmp_sock, &igmp_msg, sizeof(igmp_msg), 0,
		 (struct sockaddr *) &to, sizeof(struct sockaddr_in));
	if (len < 0 || len != sizeof(igmp_msg)) {
		fprintf(stderr, "sendto (igmp_report) : %s\n", strerror(errno));
	}

	close(igmp_sock);
}

static int in_cksum(unsigned short *buf, int sz)
{
	int nleft = sz;
	int sum = 0;
	unsigned short *w = buf;
	unsigned short ans = 0;

	while (nleft > 1) {
	sum += *w++;
	nleft -= 2;
	}

	if (nleft == 1) {
	*(unsigned char *) (&ans) = *(unsigned char *) w;
	sum += ans;
	}

	sum = (sum >> 16) + (sum & 0xFFFF);
	sum += (sum >> 16);
	ans = ~sum;
	return (ans);
}


