/* dhcpd.c
 *
 * Moreton Bay DHCP Server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <time.h>
#include <sys/time.h>

#include "aplog.h"
#include "debug.h"
#include "dhcpd.h"
#include "arpping.h"
#include "socket.h"
#include "options.h"
#include "files.h"
#include "leases.h"
#include "packet.h"
#include "serverpacket.h"
#include "pidfile.h"


/* globals */
struct dhcpOfferedAddr *leases;
struct server_config_t server_config;

/* Delete temporary lease file. */
void temp_leasefile_delete(char *temp_leasefile)
{
    if (temp_leasefile)
	unlink(temp_leasefile);
}

/* Exit and cleanup */
static void exit_server(int retval)
{
    /* write out lease file to file */
    DEBUG(LOG_INFO, "Write out leases to file");
    if (kill(getpid(), SIGUSR1))
        DEBUG(LOG_INFO, "Unable to write out leases to file");

    pidfile_delete(server_config.pidfile);
    temp_leasefile_delete(server_config.temp_lease_file);
    CLOSE_LOG();
    exit(retval);
}


/* SIGTERM handler */
static void udhcpd_killed(int sig)
{
    sig = 0;
    DEBUG(LOG_INFO, "Received SIGTERM");
    exit_server(0);
}


#ifdef COMBINED_BINARY
int udhcpd(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{
    fd_set rfds;
    struct timeval tv;
    int server_socket;
    int bytes, retval;
    struct dhcpMessage packet;
    unsigned char *state;
    char *server_id, *requested;
    u_int32_t server_id_align, requested_align;
    unsigned long timeout_end;
    //struct option_set *option;
    struct dhcpOfferedAddr *lease;
    struct sockaddr_in *sin;
    FILE *pid_fd;
    char localname[IFNAMSIZ];

    /* server ip addr */
    int fd = -1;
    struct ifreq ifr;

    argc = argv[0][0];		/* get rid of some warnings */

    OPEN_LOG("udhcpd");
    DEBUG(LOG_INFO, "udhcp server (v%s) started", VERSION);

    memset(&server_config, 0, sizeof(struct server_config_t));

    /* This file contains the configuration options that are not settable through SNMP.
     * Read the defaults only once.
     */
    read_config(DHCPD_CONF_FILE, 1);

    /* This file contains the SNMP-settable configuration options */
    read_config(DHCPD_SNMP_CONF_FILE, 0);

    /* Check if the options that are read in, are valid combinations for giving out leases,
     * if their not, the server exits! */
    if (check_config()) {
        DEBUG(LOG_ERR, "Config isn't valid, exit.");
        aplog(EV_DHCPS_CONFIG_FAILED, NULL, NULL);
        exit_server(1);
    }

    pid_fd = pidfile_acquire(server_config.pidfile);
    pidfile_write_release(pid_fd);

    leases =
	malloc(sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
    memset(leases, 0,
	   sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
    read_leases(server_config.lease_file, server_config.stat_lease_file);

    if ((fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) >= 0) {
	ifr.ifr_addr.sa_family = AF_INET;
	strcpy(ifr.ifr_name, server_config.interface);

        /* Loop if we can't get the IF ADDR. */
        while (ioctl(fd, SIOCGIFADDR, &ifr) != 0)
                sleep(10);

	    sin = (struct sockaddr_in *) &ifr.ifr_addr;
	    server_config.server = sin->sin_addr.s_addr;
	    DEBUG(LOG_INFO, "%s (server_ip) = %s", ifr.ifr_name,
		  inet_ntoa(sin->sin_addr));

	if (ioctl(fd, SIOCGIFINDEX, &ifr) == 0) {
	    DEBUG(LOG_INFO, "adapter index %d", ifr.ifr_ifindex);
	    server_config.ifindex = ifr.ifr_ifindex;
	} else {
	    DEBUG(LOG_ERR, "SIOCGIFINDEX failed!");
            aplog(EV_DHCPS_SOCK_FAILED, "SIOCGIFINDEX", NULL);
	    exit_server(1);
	}
	if (ioctl(fd, SIOCGIFHWADDR, &ifr) == 0) {
	    memcpy(server_config.arp, ifr.ifr_hwaddr.sa_data, 6);
	    DEBUG(LOG_INFO,
		  "adapter hardware address %02x:%02x:%02x:%02x:%02x:%02x",
		  server_config.arp[0], server_config.arp[1],
		  server_config.arp[2], server_config.arp[3],
		  server_config.arp[4], server_config.arp[5]);
	} else {
	    DEBUG(LOG_ERR, "SIOCGIFHWADDR failed!");
            aplog(EV_DHCPS_SOCK_FAILED, "SIOCGIFHWADDR", NULL);
	    exit_server(1);
	}
    } else {
	DEBUG(LOG_ERR, "socket failed!");
        aplog(EV_DHCPS_SOCK_FAILED, "raw socket", NULL);
	exit_server(1);
    }
    close(fd);

#if 0 /* Not needed */
#ifndef DEBUGGING
    pid_fd = pidfile_acquire(server_config.pidfile);	/* hold lock during fork. */
    switch (vfork()) {
    case -1:
	perror("vfork");
	exit_server(1);
     /*NOTREACHED*/ case 0:
	break;			/* child continues */
    default:
	exit(0);		/* parent exits */
     /*NOTREACHED*/}
    close(0);
    setsid();
    pidfile_write_release(pid_fd);
#endif
#endif /* if 0 */

    signal(SIGUSR1, write_leases);
    signal(SIGUSR2, write_temp_leases);
    signal(SIGTERM, udhcpd_killed);

    /* check if it's an aliased interface, if the interface is aliased
     * never use it anymore.
     */
    strncpy(localname, server_config.interface, sizeof(localname));
    server_config.interface = strtok(localname, ":");


    timeout_end = time(0) + server_config.auto_time;
    while (1) {			/* loop until universe collapses */

        server_socket =
	    listen_socket(INADDR_ANY, SERVER_PORT,
			  server_config.interface);
	if (server_socket == -1) {
	    DEBUG(LOG_ERR, "couldn't create server socket -- au revoir");
            aplog(EV_DHCPS_SOCK_FAILED, "server socket", NULL);
	    exit_server(0);
	}

        FD_ZERO(&rfds);
	FD_SET(server_socket, &rfds);
	if (server_config.auto_time) {
	    tv.tv_sec = timeout_end - time(0);
	    if (tv.tv_sec <= 0) {
		tv.tv_sec = server_config.auto_time;
		timeout_end = time(0) + server_config.auto_time;
		write_leases(0);
	    }
	    tv.tv_usec = 0;
	}
	retval =
	    select(server_socket + 1, &rfds, NULL, NULL,
		   server_config.auto_time ? &tv : NULL);
	if (retval == 0) {
	    write_leases(0);
	    timeout_end = time(0) + server_config.auto_time;
	    close(server_socket);
	    continue;
	} else if (retval < 0) {
	    DEBUG(LOG_INFO, "error on select");
	    close(server_socket);
	    continue;
	}

	bytes = get_packet(&packet, server_socket);	/* this waits for a packet - idle */

        /* Check if the server' IP address changed */
        check_for_server_ip(NULL, NULL);

        close(server_socket);
	if (bytes < 0)
	    continue;

        if ((state = get_option(&packet, DHCP_MESSAGE_TYPE)) == NULL) {
	    DEBUG(LOG_INFO, "couldnt get option from packet, could be a BOOTP REQUEST");

            /*  If the packet does not contain a DHCP_MESSAGE_TYPE, it could be a BOOTP message
                We check if the minimum size is 300 bytes. Silently discard smaller packets. (RFC1542)
            */
            if (bytes < 300) {
                DEBUG(LOG_INFO, "Message length is less than 300 bytes --ignoring");
                continue;
            }

            DEBUG(LOG_INFO, "received BOOTP REQUEST");
            sendBootpReply(&packet);
            continue;
	}

	lease = find_lease_by_chaddr(packet.chaddr);
	switch (state[0]) {

        case DHCPDISCOVER:
	    DEBUG(LOG_INFO, "received DISCOVER");

	    if (sendOffer(&packet) < 0) {
		DEBUG(LOG_ERR, "send OFFER failed -- ignoring");
	    }
	    break;
	case DHCPREQUEST:
	    DEBUG(LOG_INFO, "received REQUEST");

	    requested = get_option(&packet, DHCP_REQUESTED_IP);
	    server_id = get_option(&packet, DHCP_SERVER_ID);

	    if (requested)
		memcpy(&requested_align, requested, 4);
	    if (server_id)
		memcpy(&server_id_align, server_id, 4);

            if (lease) {
                if (server_id) {
		    /* SELECTING State */
		    DEBUG(LOG_INFO, "server_id = %08x",
			  ntohl(server_id_align));
		    if (server_id_align == server_config.server
			&& requested && requested_align == lease->yiaddr) {

                        /* Check if an outstanding lease still complies with the (new) configuation.
                         * Since (theoretically) a server config change can happen while a client is
                         * still in SELECTING state, we check it also here. */
                        if (check_outstanding(requested_align, lease)) {
                            sendNAK(&packet);
			    lease->state = STATE_CLIENTREFUSED;
		        }
                        else {
                            sendACK(&packet, lease->yiaddr);
		        }
		    }
		} else {
		    if (requested) {
			/* INIT-REBOOT State */
			/* Make sure client is (still) on correct network before extending lease */
                        if (check_outstanding(requested_align, lease)) {
			    lease->state = STATE_CLIENTREFUSED;
                            sendNAK(&packet);
		        }

                        if (lease->yiaddr == requested_align) {
                            sendACK(&packet, lease->yiaddr);
		        }
                        else {
			    lease->state = STATE_CLIENTREFUSED;
                            sendNAK(&packet);
		        }
		    } else {
			/* RENEWING or REBINDING State */
			/* Make sure client doesn't have to move to another subnet before extending lease */
                        if (check_outstanding(packet.ciaddr, lease)) {
			    lease->state = STATE_CLIENTREFUSED;
                            sendNAK(&packet);
		        }

                        if (lease->yiaddr == packet.ciaddr) {
			    sendACK(&packet, lease->yiaddr);
		        }
			else {
			    /* don't know what to do!!!! */
   			    lease->state = STATE_CLIENTREFUSED;
			    sendNAK(&packet);
			}
		    }
		}
	    } else                              	/* else remain silent */
            DEBUG(LOG_INFO, "Don't know lease. Remain silent");

            break;
        case DHCPDECLINE:
	    DEBUG(LOG_INFO, "received DECLINE");
	    if (lease) {
		memset(lease->chaddr, 0, 16);
		lease->expires = time(0) + server_config.decline_time;
                lease->state = STATE_REFUSEDBYCLIENT; 
	    }
	    break;
	case DHCPRELEASE:
	    DEBUG(LOG_INFO, "received RELEASE");
	    if (lease)
		lease->expires = time(0);
	    break;
	case DHCPINFORM:
	    DEBUG(LOG_INFO, "received INFORM");
	    send_inform(&packet);
	    break;
	default:
	    DEBUG(LOG_WARNING, "unsupported DHCP message (%02x) -- ignoring",
		state[0]);
	}
    }

    return 0;
}
