/*
 * files.c -- DHCP server file manipulation *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * Modifications are made for reading in the static leases file
 * generated by ucd-snmp deamon January 2002
 *
 */

#include <stdio.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <linux/if.h>
#include <sys/stat.h>

#include "aplog.h"
#include "debug.h"
#include "dhcpd.h"
#include "files.h"
#include "options.h"
#include "leases.h"

/* on these functions, make sure you datatype matches */
static int read_ip(char *line, void *arg)
{
    int res;
    struct in_addr addr;
    res = inet_aton(line, &addr);
    if (res == 0) {
	DEBUG(LOG_ERR,
	      "read_ip: Line %s couldn't be parsed to network address",
	      line);
	return 0;
    }
    DEBUG(LOG_INFO, "read_ip: value %s", inet_ntoa(addr));
    *((u_int32_t *) arg) = addr.s_addr;
    return 1;
}


static int read_str(char *line, void *arg)
{
    char **dest = arg;
    int i;

    if (*dest)
	free(*dest);
    *dest = strdup(line);

    /* elimate trailing whitespace */
    for (i = strlen(*dest) - 1; i > 0 && isspace((*dest)[i]); i--);
    (*dest)[i > 0 ? i + 1 : 0] = '\0';

    DEBUG(LOG_INFO, "read_str: value %s", *dest);

    return 1;
}


static int read_u32(char *line, void *arg)
{
    u_int32_t *dest = arg;
    *dest = strtoul(line, NULL, 0);

    DEBUG(LOG_INFO, "read_u32: value %d", *dest);

    return 1;
}


static int read_yn(char *line, void *arg)
{
    char *dest = arg;
    if (!strcasecmp("yes", line) || !strcmp("1", line)
	|| !strcasecmp("true", line))
	*dest = 1;
    else if (!strcasecmp("no", line) || !strcmp("0", line)
	     || !strcasecmp("false", line))
	*dest = 0;
    else
	return 0;

    return 1;
}


/* read a dhcp option and add it to opt_list */
static int read_opt(char *line, void *arg)
{
    struct option_set **opt_list = arg;
    char *opt, *val;
    char fail;
    struct dhcp_option *option = NULL;
    int length = 0;
    char buffer[255];
    u_int16_t result_u16;
    int16_t result_s16;
    u_int32_t result_u32;
    int32_t result_s32;

    int i;
    
    if (!(opt = strtok(line, " \t=")))
	return 0;

    for (i = 0; options[i].code; i++)
	if (!strcmp(options[i].name, opt)) {
	    option = &(options[i]);
	    break;
	}

    if (!option)
	return 0;

    do {
	val = strtok(NULL, ", \t");
	if (val) {
	    fail = 0;
	    length = 0;
	    switch (option->flags & TYPE_MASK) {
	    case OPTION_IP:
		read_ip(val, buffer);
		break;
	    case OPTION_IP_PAIR:
		read_ip(val, buffer);
		if ((val = strtok(NULL, ", \t/-")))
		    read_ip(val, buffer + 4);
		else
		    fail = 1;
		break;
	    case OPTION_STRING:
		length = strlen(val);
		if (length > 254)
		    length = 254;
		memcpy(buffer, val, length);
		break;
	    case OPTION_BOOLEAN:
		if (!read_yn(val, buffer))
		    fail = 1;
		break;
	    case OPTION_U8:
		buffer[0] = strtoul(val, NULL, 0);
		break;
	    case OPTION_U16:
		result_u16 = htons(strtoul(val, NULL, 0));
		memcpy(buffer, &result_u16, 2);
		break;
	    case OPTION_S16:
		result_s16 = htons(strtol(val, NULL, 0));
		memcpy(buffer, &result_s16, 2);
		break;
	    case OPTION_U32:
		result_u32 = htonl(strtoul(val, NULL, 0));
		memcpy(buffer, &result_u32, 4);
		break;
	    case OPTION_S32:
		result_s32 = htonl(strtol(val, NULL, 0));
		memcpy(buffer, &result_s32, 4);
		break;
	    default:
    DEBUG(LOG_INFO, "read_opt: invalid option? %d",option->flags);
		break;
	    }
	    length += option_lengths[option->flags & TYPE_MASK];
	    if (!fail)
		attach_option(opt_list, option, buffer, length);
	} else {
	    fail = 1;
	}
    } while (!fail && option->flags & OPTION_LIST);
    return 1;
}


static struct config_keyword keywords[] = {
    /* keyword          handler   variable address                      default */
    {"start", read_ip, &(server_config.start), "192.168.1.1"},
    {"end", read_ip, &(server_config.end), "192.168.1.254"},
    {"interface", read_str, &(server_config.interface), "eth0"},
    {"option", read_opt, &(server_config.options), ""},
    {"opt", read_opt, &(server_config.options), ""},
    {"max_leases", read_u32, &(server_config.max_leases), "254"},
    {"remaining", read_yn, &(server_config.remaining), "yes"},
    {"arpping", read_yn, &(server_config.arpping), "yes"},
    {"auto_time", read_u32, &(server_config.auto_time), "0"},
    {"decline_time", read_u32, &(server_config.decline_time), "3600"},
    {"conflict_time", read_u32, &(server_config.conflict_time), "3600"},
    {"offer_time", read_u32, &(server_config.offer_time), "60"},
    {"min_lease", read_u32, &(server_config.min_lease), "60"},
    {"lease", read_u32, &(server_config.lease), "864000"},
    {"lease_file", read_str, &(server_config.lease_file),
     "/var/etc/udhcpd.leases.conf"},
    {"temp_lease_file", read_str, &(server_config.temp_lease_file),
     "/tmp/udhcpd.leases"},
    {"stat_lease_file", read_str, &(server_config.stat_lease_file),
     "/var/etc/dhcpServer.conf"},
    {"pidfile", read_str, &(server_config.pidfile), "/var/run/udhcpd.pid"},
    {"notify_file", read_str, &(server_config.notify_file), ""},
    {"siaddr", read_ip, &(server_config.siaddr), "192.168.1.254"},
    {"sname", read_str, &(server_config.sname), ""},
    {"boot_file", read_str, &(server_config.boot_file), ""},
    {"dynamic_bootp", read_yn, &(server_config.dynamic_bootp), "no"},
    {"server_subnet", read_ip, &(server_config.server_subnet), "192.168.1.0"},
    {"", NULL, NULL, ""}
};


int read_config(char *file, int read_defaults)
{
    FILE *in;
    char buffer[80], *token, *line, *endofline;
    int i;
    DEBUG(LOG_INFO, "read_config: Set to defaults");
    for (i = 0; (read_defaults == 1 && strlen(keywords[i].keyword)); i++)
	if (strlen(keywords[i].def))
	    keywords[i].handler(keywords[i].def, keywords[i].var);

    if (!(in = fopen(file, "r"))) {
	DEBUG(LOG_ERR, "unable to open config file: %s", file);
        aplog(EV_DHCPS_NO_CONF_FILE, file, NULL);
	return 0;
    }
    DEBUG(LOG_INFO, "read_config: Lets read in the config file");
    while (fgets(buffer, 80, in)) {
	if (strchr(buffer, '\n'))
	    *(strchr(buffer, '\n')) = '\0';
	if (strchr(buffer, '#'))
	    *(strchr(buffer, '#')) = '\0';
	token = buffer + strspn(buffer, " \t");
	if (*token == '\0')
	    continue;
	line = token + strcspn(token, " \t=");
	if (*line == '\0')
	    continue;
	*line = '\0';
	line++;
	line = line + strspn(line, " \t=");
	if (*line == '\0')
	    continue;

	endofline = line + strcspn(line, " \0\t=");
	*endofline = '\0';

	for (i = 0; strlen(keywords[i].keyword); i++) {
	    if (!strcasecmp(token, keywords[i].keyword)) {
		keywords[i].handler(line, keywords[i].var);
	    }
	}
    }
    fclose(in);
    
    cmp_file_time();
    
    return 1;
}


/* the dummy var is here so this can be a signal handler */
void write_leases(int dummy)
{
    FILE *fp;
    unsigned int i;
    char buf[255];
    time_t curr = time(0);
    unsigned long lease_time;
    char empty_mac[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
    
    dummy = 0;
    
    if (!(fp = fopen(server_config.lease_file, "w"))) {
	DEBUG(LOG_ERR, "Unable to open %s for writing",
	    server_config.lease_file);
	return;
    }
    
    for (i = 0; i < server_config.max_leases; i++) {

	if ((leases[i].yiaddr != 0) 
		|| (memcmp(leases[i].chaddr, empty_mac, 6) != 0)
		)  {
		
	    /* Do not write leases for which an static lease is configured, but
	     * are not the static leases itself */
	    if(!leases[i].static_lease && find_lease_by_chaddr(&(leases[i].chaddr[0]))->static_lease) {
		    continue;
	    }
		
	    if (server_config.remaining) {
		if (lease_expired(&(leases[i])))
		    lease_time = 0;
		else
		    lease_time = leases[i].expires - curr;
	    } else {
		if(leases[i].expires > 0) {
			lease_time = leases[i].expires;
		}
		else {
			lease_time = 0;
		}
	    }
	    
	    /* determine state if state is not refused */
	    if((leases[i].state != STATE_CLIENTREFUSED)
			    && (leases[i].state != STATE_REFUSEDBYCLIENT)) {
		    if(leases[i].static_lease) {
			    leases[i].state = STATE_CLIENTCONFIGURED;
		    }
		    else if (lease_time == 0) {
			    leases[i].state = STATE_CLIENTEXPIRED;
		    }
		    else {
			    leases[i].state = STATE_CLIENTACTIVE;
		    }
	    }
    
    	    lease_time = htonl(lease_time);
	    
	    fwrite(leases[i].chaddr, 6, 1, fp); /* optimization for ISL, write only MAC part */
            fwrite(&(leases[i].yiaddr), 4, 1, fp);
	    fwrite(&lease_time, 4, 1, fp);
	    fwrite(&leases[i].state, 1, 1, fp);
	}
    }
    fclose(fp);

    if (server_config.notify_file) {
	sprintf(buf, "%s %s", server_config.notify_file,
		server_config.lease_file);
	system(buf);
    }
}


/* 
 * This function writes out an snmp-parsable file to RAM, 
 * that can be re-read by the ucd-snmp deamon. 
 *
 * the dummy var is here so this can be a signal handler 
 */  
void write_temp_leases(int dummy)
{
    FILE *fp;
    unsigned int i;
    char buf[255];
    time_t curr = time(0);
    unsigned long lease_time;
                          
    struct in_addr ipaddr;
    char *ch_ipaddr;
    
    dummy = 0;

    if (!(fp = fopen(server_config.temp_lease_file, "w"))) {
	DEBUG(LOG_ERR, "Unable to open %s for writing",
	    server_config.temp_lease_file);
	return;
    }
    
    for (i = 0; i < server_config.max_leases; i++) {
	
        int state= STATE_CLIENTNOTINUSE;
        
        ipaddr.s_addr = leases[i].yiaddr;
        /* Convert ipaddr from binary data into string */ 
        ch_ipaddr = inet_ntoa(ipaddr);

        /* Check for static leases */
	if (leases[i].static_lease)
            state = STATE_CLIENTCONFIGURED;
           //continue;

	if (leases[i].yiaddr != 0) {
	    if (server_config.remaining) {
		if (lease_expired(&(leases[i])))
                {
                    lease_time = 0;
                    state      = STATE_CLIENTEXPIRED; 
                    //FIXME, check what to do for preconfigured leases, fall back to STATE_CLIENTCONFIGURED?
                }  
		else
                {
		    lease_time = leases[i].expires - curr;
                    
                    if (state == STATE_CLIENTCONFIGURED && leases[i].expires == server_config.lease)
                        lease_time = leases[i].expires;
                    else
                        state      = STATE_CLIENTACTIVE;
                }

                if (leases[i].state == STATE_CLIENTREFUSED)
                    state = STATE_CLIENTREFUSED;

	    } else {
		lease_time = leases[i].expires;
                //FIXME, check what to do for !config.remaining optie
	    }
	    lease_time = htonl(lease_time);
            
	    fprintf(fp, "clientMacAddress %d %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x\n", i,
	           	  leases[i].chaddr[0], leases[i].chaddr[1],
			  leases[i].chaddr[2], leases[i].chaddr[3],
			  leases[i].chaddr[4], leases[i].chaddr[5]);
            
            ipaddr.s_addr = leases[i].yiaddr;
            /* Convert ipaddr from binary data into string */ 
            ch_ipaddr = inet_ntoa(ipaddr);
            fprintf(fp, "clientIpAddress %d %s\n", i, ch_ipaddr);
            
            fprintf(fp, "clientState %d %d\n", i, state);

            lease_time = ntohl(lease_time);
	    fprintf(fp, "clientExpiration %d %lu\n", i, lease_time);           
            
	}       
    }
    fclose(fp);

    if (server_config.notify_file) {
	sprintf(buf, "%s %s", server_config.notify_file,
		server_config.lease_file);
	system(buf);
    }
}

int parse_macaddress(unsigned char *input, unsigned char *output)
{
    unsigned int i = 0;
    unsigned char *p;

    for (p = strtok(input, "-:"); p; p = strtok(NULL, "-:\r\n\t")) {
	output[i] = (int) strtol(p, (char **) NULL, 16);       
	i++;
    }
    if (i == 6) {
	return 0;
    } 
    return -1;
}

void parse_reservationIpAddress(char *cptr)
{
    char *tok;
    int index;
    time_t curr = time(0);
    struct in_addr ipaddr;
    
    tok = (char *)strtok(cptr, "reservationIpAddress ");
    
    if ( tok != NULL )
    {
            index = atoi(tok) - 1;
            
            if ( index < 0 || index > MAX_STATIC_LEASES )
            {
                    fprintf(stderr, "reservationIpAddress = %s: Parse error index too large\n",tok);
            }
            else
            {
                    tok = (char *)strtok(NULL, " \n");
                    
                    if ( tok != NULL )
                    {
                        /* Convert string ipaddress (tok) to binary ip address */
                        if (!(inet_aton(tok, &ipaddr)))
                        {
                            DEBUG(LOG_ERR, "udhcpd: parse_reservationIpAddress: unable to convert %s to binary ip-address\n", tok);
                        }
                        
                        leases[index].yiaddr = ipaddr.s_addr;
                        leases[index].static_lease = TRUE;
                        leases[index].expires = server_config.lease;
			
		        if (server_config.remaining)
				leases[index].expires += curr;

			
			
                    }
            }
    }

    if ( tok == NULL )
    {
         fprintf(stderr, "reservationIpAddress = %s: Parse error\n",cptr);
    }
}

void parse_reservationMacAddress(char *cptr)
{
    char *tok;
    int index;
    
    tok = (char *)strtok(cptr, "reservationMacAddress ");

    if ( tok != NULL )
    {
            index = atoi(tok) - 1;
            
            if ( index < 0 || index > MAX_STATIC_LEASES )
            {
                    fprintf(stderr, "reservationMacAddress = %s: Parse error index too large\n",tok);
            }
            else
            {
                    tok = (char *)strtok(NULL, " \n");
                    
                    if ( tok != NULL )
                    {   
                        if ( parse_macaddress(tok, leases[index].chaddr) != 0 )
                        {
                            fprintf(stderr, "reservationMacAddress = %s: Parse error MacAddress not OK\n", tok);
                        }
                    }
            }
    }

    if ( tok == NULL )
    {
         fprintf(stderr, "reservationMacAddress = %s: Parse error\n",cptr);
    }
}

void parse_reservationRowStatus(const char *cptr)
{
    char *tok;
    int index;
    int i;

    tok = strtok(cptr, "reservationRowStatus ");

    if (tok != NULL)
	{
            index = atoi(tok) - 1;              

            if (index < 0 || index > MAX_STATIC_LEASES )
		{
			fprintf(stderr, "reservationRowStatus = %s: Parse error index too large\n",cptr);                     
		}
		else
		{
			(int)tok = strtok(NULL, " \n");
			if (tok != NULL)
			{
				leases[index].rowstatus = atol(tok);

                                /* Only allow leases that have their rowstatus active */
                                if (leases[index].rowstatus != 1 /*ROW_ACTIVE*/)
                                {   
						/* clear entry with non-active lease */
						for(i=0;i<16;i++)
					         leases[index].chaddr[i] = (u_int8_t)0; 
					         leases[index].yiaddr = (u_int32_t)0; 
                                                 leases[index].state = (u_int8_t)0; /* STATE_CLIENTNOTINUSE */
						 leases[index].rowstatus = (long)0;
						 leases[index].booting = (u_int8_t)0;                                          
                                                 leases[index].expires = (u_int32_t)0;
                                }
			}                
                }
	}

	if (tok == NULL)
	{
		fprintf(stderr, "parse_reservationRowStatus = %s: Parse error\n",cptr);              
	}           
}

/*
 * read_static_snmp_leases()
 *
 * Altered this function, to be able to read in SNMP parseable format.
 *
 */
int read_static_snmp_leases(char *file)
{
    FILE *fp;
    char buffer[80];
    unsigned int i = 0;

    if (!(fp = fopen(file, "r"))) {
        DEBUG(LOG_ERR, "udhcp: Unable to open file %s for reading\n", file);
        return -1;
    }

    while (fgets(buffer, 80, fp) && i < server_config.max_leases) 
    {
        /* 
         * Find the tokens reservationIpAddress, reservationMacAddress and reservationRowStatus
         * Skip empty lines and comments(#)
         */
        if (strchr(buffer, '\n'))
	    *(strchr(buffer, '\n')) = '\0';
	if (strchr(buffer, '#'))
	    *(strchr(buffer, '#')) = '\0';

        if (strstr(buffer, "reservationIpAddress")) 
            {   
                parse_reservationIpAddress(buffer);
                i++;
            }

        if (strstr(buffer, "reservationMacAddress")) 
            {   
                parse_reservationMacAddress(buffer);
            }

        if (strstr(buffer, "reservationRowStatus")) 
            {   
                parse_reservationRowStatus(buffer);
            }
    }

    if (i >= server_config.max_leases) {
	if (fgetc(fp) == EOF)       
	    DEBUG(LOG_WARNING, "Too many leases while loading %s\n", file);
    }
    
    fclose(fp);
    return i;
}

/*
 *  Replaced this function for read_static_snmp_leases, to be able to read in
 *  file in snmp parseable format
 */
int read_static_leases(char *file)
{
    FILE *fp;
    char buffer[80], *mac_addr, *ip_addr, *allow, *endofline;
    unsigned int i = 0;
    u_int32_t yiaddr;

    if (!(fp = fopen(file, "r"))) {
	DEBUG(LOG_ERR, "Unable to open static leases file %s for reading",
	    file);
	return 0;
    }

    while (fgets(buffer, 80, fp) && i < server_config.max_leases) {

	/* Find the token MAC_ADDRESS and IP_ADDRESS */
	if (strchr(buffer, '\n'))
	    *(strchr(buffer, '\n')) = '\0';
	if (strchr(buffer, '#'))
	    *(strchr(buffer, '#')) = '\0';
	mac_addr = buffer + strspn(buffer, " \t");
	if (*mac_addr == '\0')
	    continue;

	allow = mac_addr + strcspn(mac_addr, " \t=");
	if (*allow == '\0')
	    continue;
	*allow = '\0';
	allow++;
	allow = allow + strspn(allow, " \t=");
	if (*allow == '\0')
	    continue;
	if (strcmp(allow, "DENY") == 0) {
	    parse_macaddress(mac_addr, leases[i].chaddr);
	    leases[i].expires = -1;
	    leases[i].state = 0;
            leases[i].static_lease = TRUE;
	    leases[i].booting = FALSE;
            DEBUG(LOG_INFO, "Found lease:");
	    DEBUG(LOG_INFO,
		  "-MAC Addr: %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x",
		  leases[i].chaddr[0], leases[i].chaddr[1],
		  leases[i].chaddr[2], leases[i].chaddr[3],
		  leases[i].chaddr[4], leases[i].chaddr[5]);
	    DEBUG(LOG_INFO, "-  Static: TRUE");
	    DEBUG(LOG_INFO, "- Booting: DENY");
	    i++;
	    continue;
	}

	ip_addr = allow + strcspn(allow, " \t=");
	if (*ip_addr == '\0')
	    continue;
	*ip_addr = '\0';
	ip_addr++;
	ip_addr = ip_addr + strspn(ip_addr, " \t=");
	if (*ip_addr == '\0')
	    continue;
	endofline = ip_addr + strcspn(ip_addr, " \t=");
	*endofline = '\0';

	/* Check and convert the IP address */
	if (read_ip(ip_addr, &yiaddr)) {
	    if (ntohl(yiaddr) >= ntohl(server_config.start)
		&& ntohl(yiaddr) <= ntohl(server_config.end)) {
		/* The ip address is in our service area. */
		leases[i].yiaddr = yiaddr;
		parse_macaddress(mac_addr, leases[i].chaddr);
		leases[i].expires = -1;
		leases[i].state = 0;
                leases[i].static_lease = TRUE;
		leases[i].booting = TRUE;
                DEBUG(LOG_INFO, "Found lease:");
		DEBUG(LOG_INFO,
		      "-MAC Addr: %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x",
		      leases[i].chaddr[0], leases[i].chaddr[1],
		      leases[i].chaddr[2], leases[i].chaddr[3],
		      leases[i].chaddr[4], leases[i].chaddr[5]);
		DEBUG(LOG_INFO, "-  Static: TRUE");
		DEBUG(LOG_INFO, "- Booting: ALLOW");
		i++;
	    }
	}
    }

    DEBUG(LOG_INFO, "Read %d static leases", i);

    if (i >= server_config.max_leases) {
	if (fgetc(fp) == EOF)
	    /* might be helpfull to drop expired leases */
	    DEBUG(LOG_WARNING, "Too many leases while loading %s\n", file);
    }

    fclose(fp);
    return i;
}


void read_leases(char *file, char *static_file)
{
    FILE *fp;
    unsigned int i = 0;
    time_t curr = time(0);
    struct dhcpOfferedAddr lease;
    unsigned long lease_time;
        
    memset(&(lease.chaddr), 0x00, 16);

    //i = read_static_leases(static_file);
    i = read_static_snmp_leases(static_file);
    
    if (!(fp = fopen(file, "r"))) {
	DEBUG(LOG_ERR, "Unable to open %s for reading", file);
	return;
    }
    
    while (i < server_config.max_leases
	   && (fread(&(lease.chaddr), 6, 1, fp) == 1)) {
	    
	fread(&(lease.yiaddr), 4, 1, fp);
	fread(&(lease_time), 4, 1, fp);
	lease_time = ntohl(lease_time);
	fread(&(lease.state), 1, 1, fp);
	
	if(lease.state == STATE_CLIENTCONFIGURED) {
		continue;
	}
	    
	if (ntohl(lease.yiaddr) >= ntohl(server_config.start)
	    && ntohl(lease.yiaddr) <= ntohl(server_config.end)) {
            
            leases[i].yiaddr = lease.yiaddr;
	    leases[i].expires = lease_time;
            leases[i].state = lease.state;
            
            if (server_config.remaining)
		leases[i].expires += curr;
            
            memcpy(leases[i].chaddr, lease.chaddr, 6);
	    leases[i].static_lease = FALSE;
	    leases[i].booting = TRUE;
            DEBUG(LOG_INFO, "Found lease:");
	    DEBUG(LOG_INFO,
		  "-MAC Addr: %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x\n",
		  leases[i].chaddr[0], leases[i].chaddr[1],
		  leases[i].chaddr[2], leases[i].chaddr[3],
		  leases[i].chaddr[4], leases[i].chaddr[5]);
	    //DEBUG(LOG_INFO, "-  Static: FALSE");
	    //DEBUG(LOG_INFO, "- Booting: ALLOW");
	    i++;
	}
        else
            DEBUG(LOG_INFO, "Lease %lu no longer falls in subnet range.", lease.yiaddr);
    }

    DEBUG(LOG_INFO, "Read total of %d leases", i);

    if (i == server_config.max_leases) {
	if (fgetc(fp) == EOF)
	    /* might be helpfull to drop expired leases */
	    DEBUG(LOG_WARNING, "Too many leases while loading %s\n", file);
    }
    fclose(fp);
}


/* When the server is started, it should check if the options that are read in,
 * are valid combinations for giving out leases
 *
 * Returns 0 on success.
 * Returns -1 on error.
 */
int check_config()
{
    unsigned int subnet_mask;
    unsigned int router;
    struct option_set *opt_list;
    
    if ((opt_list = find_option(server_config.options, DHCP_SUBNET))) 
    {
        memcpy(&subnet_mask, opt_list->data+2, 4);
    }
    else
    {
        DEBUG(LOG_ERR, "Couldn't get option subnet from option list.");
        return -1;
    }

    if ((opt_list = find_option(server_config.options, DHCP_ROUTER))) 
    {
        memcpy(&router, opt_list->data+2, 4);
    }
    else
    {
	router = 0;
    }
    
    /* All addresses should unequal 0.0.0.0.*/
    if (server_config.start == 0 || server_config.end == 0 || 
        server_config.server_subnet == 0 || subnet_mask == 0 )
    {   
        DEBUG(LOG_ERR, "One of the checked addresses equals 0.0.0.0");
        return -1;
    }
    
    /* Minimum, maximum address, and gateway address must be class A, B, or C */
    if ( !( IN_CLASSA(ntohl(server_config.start))  || IN_CLASSB(ntohl(server_config.start)) || IN_CLASSC(ntohl(server_config.start)) ) ||
         !( IN_CLASSA(ntohl(server_config.end))    || IN_CLASSB(ntohl(server_config.end))   || IN_CLASSC(ntohl(server_config.end))   ) ||
         !( IN_CLASSA(ntohl(router))               || IN_CLASSB(ntohl(router))              || IN_CLASSC(ntohl(router)) )            )
    {
        fprintf(stderr, "Min, max and/or gateway address does not fall in class A, B or C network range.\n");
        return -1;
    }

    /* Check if end address is greater than start address.*/
    if (ntohl(server_config.start) > ntohl(server_config.end))
    {   
        DEBUG(LOG_ERR, "Start address is greater than End address");
        return -1;
    }

    /* Check if the bitwise logical-AND operation of one of the addresses with the network mask
     * results in a class A, B or C address and is the same class for all addresses.
     */
    if (!(
        ( IN_CLASSA((server_config.start & subnet_mask)) && IN_CLASSA((server_config.end & subnet_mask)) && 
          IN_CLASSA((router & subnet_mask)) ) ||
        ( IN_CLASSB((server_config.start & subnet_mask)) && IN_CLASSB((server_config.end & subnet_mask)) && 
          IN_CLASSB((router & subnet_mask)) ) ||
        ( IN_CLASSC((server_config.start & subnet_mask)) && IN_CLASSC((server_config.end & subnet_mask)) && 
          IN_CLASSC((router & subnet_mask)) ) 
       ) )
    {
        fprintf(stderr, "address in combination with subnet mask does not fall in class A, B or C network range.\n");
        return -1;
    }

    /* All went well, return 0 */
    return 0;

}

int read_resolv_file()
{
    FILE *f;
    char buffer[80], output[80];
    int length = 0, i = 0;
    int ip[4];
    
    memset(output, 0, 80);
    
    f = fopen(RESOLV_CONF, "r");
    
    if (f == NULL) 
        return -1;

    while (fgets(buffer, 80, f) ) 
    {
        if (sscanf(buffer, "nameserver %d.%d.%d.%d", &ip[0], &ip[1], &ip[2], &ip[3]) != 0 )
        {   
	    for (i = 0; i < 4; i++)
            {
                output[i+length] = (char)ip[i];
            }
            length += 4;
        }   
    }

    /* minimum sanity check on length */
    if ((length == 4) || (length == 8) || (length == 12)) {
    	/* quick hack. We know that DNS is on 5 */
    	replace_option(&(options[5]), output, length);
    }

    fclose(f);
    return 0;
}

void cmp_file_time()
{   
    struct stat stat_buf;
    static time_t mod_time = -1;
    
    if (stat(RESOLV_CONF, &stat_buf) == 0)
    {   
        /* File exists... */
        if (stat_buf.st_mtime != mod_time)
        {   
            read_resolv_file();
            mod_time = stat_buf.st_mtime;
        }
    }
    else {
	    /* file does not exist. TODO remove the DNS option */
    }
}


/* Options, such as subnet and subnet mask can be changed.
 * When the server is requested to extend an outstanding lease,
 * it should be checked if the lease still complies with the configuration.
 *
 * This function checks "addr" with the subnet and subnet mask.
 * Returns 0 when it matches and -1 when it doesn't.
 */
int check_outstanding(u_int32_t addr, struct dhcpOfferedAddr *lease)
{           
   u_int32_t subnet, subnet_mask, start, end, router;
   struct option_set *opt_list;
   
   /* Check if the resolv file isn't changed. */
   cmp_file_time();
   
   /* Check if it's not the server IP address */
   if (check_for_server_ip(addr, lease))
       return -1;
   
   /* We don't check static leases. */
   if (lease->static_lease)
       return 0;
   
   if ((opt_list = find_option(server_config.options, DHCP_SUBNET))) 
   {
        memcpy(&subnet_mask, opt_list->data+2, 4);
   }
   else
   {
       DEBUG(LOG_ERR, "Couldn't get option subnet from option list.");
       return -1;
   }

   if ((opt_list = find_option(server_config.options, DHCP_ROUTER))) 
   {
        memcpy(&router, opt_list->data+2, 4);
   }
   else
   {
	/* router is optional */
	router = 0;
   }
   
   addr         = ntohl(addr);
   router       = ntohl(router);
   subnet_mask  = ntohl(subnet_mask);
   start        = ntohl(server_config.start);
   end          = ntohl(server_config.end);
   subnet       = ntohl(server_config.server_subnet); 
   
   if ( (addr & subnet_mask) != subnet )
   {
       DEBUG(LOG_WARNING, "DHCP Server: Client IP address does not comply with subnet(mask)");
       return -1;
   }
   
   if ( addr < start || addr > end )
   {
       DEBUG(LOG_WARNING, "DHCP Server: Client IP address does not comply with serverSubnet range");
       return -1;
   }

   /* Check if requested IP address doesn't belong to static lease. */
   if (check_for_static(lease))
   {
       return -1;
   }
   
   return 0;
}

