/*
 * leases.c -- tools to manage DHCP leases
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 */

#include <time.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "aplog.h"
#include "debug.h"
#include "dhcpd.h"
#include "files.h"
#include "options.h"
#include "leases.h"
#include "arpping.h"

/* clear every lease out that chaddr OR yiaddr matches and is nonzero */
void clear_lease(u_int8_t * chaddr, u_int32_t yiaddr)
{
    unsigned int i, blank_chaddr = 0, blank_yiaddr = 0;

    for (i = 0; i < 16 && !chaddr[i]; i++);
    if (i == 16)
	blank_chaddr = 1;
    blank_yiaddr = (yiaddr == 0);

    for (i = 0; i < server_config.max_leases; i++)
	if ((!blank_chaddr && !memcmp(leases[i].chaddr, chaddr, 16)) ||
	    (!blank_yiaddr && leases[i].yiaddr == yiaddr)) {
	    /* Only clear leases which are not static */
	    if (leases[i].static_lease == FALSE)
		memset(&(leases[i]), 0, sizeof(struct dhcpOfferedAddr));
	}
}


/* add a lease into the table, clearing out any old ones */
struct dhcpOfferedAddr *add_lease(u_int8_t * chaddr, u_int32_t yiaddr,
				  unsigned long lease)
{
    struct dhcpOfferedAddr *oldest;

    /* clean out any old ones */
    clear_lease(chaddr, yiaddr);

    oldest = oldest_expired_lease();

    if (oldest) {
	memcpy(oldest->chaddr, chaddr, 16);
	oldest->yiaddr = yiaddr;
	oldest->expires = time(0) + lease;
	oldest->static_lease = FALSE;
        oldest->booting = TRUE;
        oldest->state = 0;
    }

    return oldest;
}


/* true if a lease has expired */
int lease_expired(struct dhcpOfferedAddr *lease)
{
    if (lease->static_lease)
	return 0;
    return (lease->expires < (unsigned long) time(0));
}


/* Find the oldest expired lease, NULL if there are no expired leases */
struct dhcpOfferedAddr *oldest_expired_lease(void)
{
    struct dhcpOfferedAddr *oldest = NULL;
    unsigned long oldest_lease = time(0);
    unsigned int i;


    for (i = 0; i < server_config.max_leases; i++)
	/* Find the oldest not static lease */
	if ((oldest_lease > leases[i].expires) &&
	    (leases[i].static_lease == FALSE)) {
	    oldest_lease = leases[i].expires;
	    oldest = &(leases[i]);
	}
    return oldest;
}


/* Find the first lease that matches chaddr, NULL if no match */
struct dhcpOfferedAddr *find_lease_by_chaddr(u_int8_t * chaddr)
{
    unsigned int i;

    for (i = 0; i < server_config.max_leases; i++)
	if (!memcmp(leases[i].chaddr, chaddr, 16))
	    return &(leases[i]);

    return NULL;
}


/* Find the first lease that matches yiaddr, NULL is no match */
struct dhcpOfferedAddr *find_lease_by_yiaddr(u_int32_t yiaddr)
{
    unsigned int i;

    for (i = 0; i < server_config.max_leases; i++)
	if (leases[i].yiaddr == yiaddr)
	    return &(leases[i]);

    return NULL;
}


/* find an assignable address, it check_expired is true, we check all the expired leases as well.
 * Maybe this should try expired leases by age... */
u_int32_t find_address(int check_expired)
{
    u_int32_t addr, ret = 0;
    struct dhcpOfferedAddr *lease = NULL;

    addr = server_config.start;
    for (; ntohl(addr) <= ntohl(server_config.end);
	 addr = htonl(ntohl(addr) + 1)) {

	/* Check if it's not the server IP address */
        if (check_for_server_ip(addr, NULL))
            continue;
        
        /* ie, 192.168.55.0 */
	if (!(ntohl(addr) & 0xFF))
	    continue;

	/* ie, 192.168.55.255 */
	if ((ntohl(addr) & 0xFF) == 0xFF)
	    continue;

	/* lease is not taken */
	if ((!(lease = find_lease_by_yiaddr(addr)) ||
	     /* or it expired and we are checking for expired leases */
	     (check_expired && lease_expired(lease))) &&
	    /* and it isn't on the network */
	    !check_ip(addr)) {
	    ret = addr;
	    break;
	}
    }
    return ret;
}


/* check is an IP is taken, if it is, add it to the lease table */
int check_ip(u_int32_t addr)
{
    char blank_chaddr[] = {[0 ... 15] = 0 };
    struct in_addr temp;
    int ret_code;

    /* Check if we want to arpping for this system,
     * if not just return is it would be a valid address
     */
    if (!server_config.arpping)
	return 0;

    ret_code =
	arpping(addr, server_config.server, server_config.arp,
		server_config.interface);
    if (ret_code == -1) {
	/* So the arpping failed don't use this/any lease */
	return 1;
    }
    if (!ret_code) {
	temp.s_addr = addr;
	LOG(LOG_INFO,
	    "%s belongs to someone, reserving it for %ld seconds",
	    inet_ntoa(temp), server_config.conflict_time);
	add_lease(blank_chaddr, addr, server_config.conflict_time);
	return 1;
    } else
	return 0;
}

/* Check if IP is not IP of a static lease */
int check_for_static(struct dhcpOfferedAddr *lease)
{
   struct dhcpOfferedAddr *temp_lease = find_lease_by_yiaddr(lease->yiaddr);
   
   if (temp_lease != NULL && temp_lease->static_lease == TRUE &&
       memcmp(temp_lease->chaddr, lease->chaddr, 6) )
   {
       return -1;
   }
   
   return 0;

}

/* Check if the IP is not the IP of the server */
int check_for_server_ip(u_int32_t addr, struct dhcpOfferedAddr *lease)
{
    int fd = -1;
    char *ip_addr;
    struct ifreq ifr;
    struct sockaddr_in *sin;
    struct in_addr in;

    /* Get the server IP address. */
    if ((fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) >= 0) 
    {
	ifr.ifr_addr.sa_family = AF_INET;
        
        /* Use alias br0:0 for bridge device. */
        if (strncmp("br0", server_config.interface, 3) == 0 )
                strncpy(ifr.ifr_name, "br0:0", IFNAMSIZ);
        
        if (ioctl(fd, SIOCGIFADDR, &ifr) == 0) 
        {
	    sin = (struct sockaddr_in *) &ifr.ifr_addr;
	    server_config.server = sin->sin_addr.s_addr;             
	}
        else 
          aplog(EV_DHCPS_SOCK_FAILED, "SIOCGIFADDR", NULL);
     
    }
    else 
      aplog(EV_DHCPS_SOCK_FAILED, "raw socket", NULL);
    

    close(fd);

    if (addr != NULL && addr == server_config.server)
    {
        if (lease != NULL && lease->static_lease)
        {
            in.s_addr = addr;
            ip_addr = inet_ntoa(in);
            aplog(EV_DHCPS_RES_ERROR, ip_addr, NULL);
        }
        
        return -1;
    }
        
    return 0;
}
