/* 
 * options.c -- DHCP server option packet tools 
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "debug.h"
#include "dhcpd.h"
#include "files.h"
#include "options.h"
#include "leases.h"


/* supported options are easily added here */
struct dhcp_option options[] = {
    /* name[10]     flags                           code */
    {"subnet", OPTION_IP, 0x01},
    {"timezone", OPTION_S32, 0x02},
    {"router", OPTION_IP | OPTION_LIST, 0x03},
    {"timesvr", OPTION_IP | OPTION_LIST, 0x04},
    {"namesvr", OPTION_IP | OPTION_LIST, 0x05},
    {"dns", OPTION_IP | OPTION_LIST, 0x06},
    {"logsvr", OPTION_IP | OPTION_LIST, 0x07},
    {"cookiesvr", OPTION_IP | OPTION_LIST, 0x08},
    {"lprsvr", OPTION_IP | OPTION_LIST, 0x09},
    {"hostname", OPTION_STRING, 0x0c},
    {"bootsize", OPTION_U16, 0x0d},
    {"domain", OPTION_STRING, 0x0f},
    {"swapsvr", OPTION_IP, 0x10},
    {"rootpath", OPTION_STRING, 0x11},
    {"ipttl", OPTION_U8, 0x17},
    {"mtu", OPTION_U16, 0x1a},
    {"broadcast", OPTION_IP, 0x1c},
    {"ntpsrv", OPTION_IP | OPTION_LIST, 0x2a},
    {"wins", OPTION_IP | OPTION_LIST, 0x2c},
    {"requestip", OPTION_IP, 0x32},
    {"lease", OPTION_U32, 0x33},
    {"dhcptype", OPTION_U8, 0x35},
    {"serverid", OPTION_IP, 0x36},
    {"tftp", OPTION_STRING, 0x42},
    {"bootfile", OPTION_STRING, 0x43},
    {"", 0x00, 0x00}
};

/* Lengths of the different option types */
int option_lengths[] = {
    [OPTION_IP] = 4,
    [OPTION_IP_PAIR] = 8,
    [OPTION_BOOLEAN] = 1,
    [OPTION_STRING] = 0,
    [OPTION_U8] = 1,
    [OPTION_U16] = 2,
    [OPTION_S16] = 2,
    [OPTION_U32] = 4,
    [OPTION_S32] = 4
};


/* get an option with bounds checking (warning, not aligned). */
unsigned char *get_option(struct dhcpMessage *packet, int code)
{
    int i, length;
    static char err[] = "bogus packet, option fields too long.";	/* save a few bytes */
    unsigned char *optionptr;
    int over = 0, done = 0, curr = OPTION_FIELD;

    optionptr = packet->options;
    i = 0;
    length = 308;
    while (!done) {
	if (i >= length) {
	    LOG(LOG_WARNING, err);
	    return NULL;
	}
	if (optionptr[i + OPT_CODE] == code) {
	    if (i + 1 + optionptr[i + OPT_LEN] >= length) {
		LOG(LOG_WARNING, err);
		return NULL;
	    }
	    return optionptr + i + 2;
	}
	switch (optionptr[i + OPT_CODE]) {
	case DHCP_PADDING:
	    i++;
	    break;
	case DHCP_OPTION_OVER:
	    if (i + 1 + optionptr[i + OPT_LEN] >= length) {
		LOG(LOG_WARNING, err);
		return NULL;
	    }
	    over = optionptr[i + 3];
	    i += optionptr[OPT_LEN] + 2;
	    break;
	case DHCP_END:
	    if (curr == OPTION_FIELD && over & FILE_FIELD) {
		optionptr = packet->file;
		i = 0;
		length = 128;
		curr = FILE_FIELD;
	    } else if (curr == FILE_FIELD && over & SNAME_FIELD) {
		optionptr = packet->sname;
		i = 0;
		length = 64;
		curr = SNAME_FIELD;
	    } else
		done = 1;
	    break;
	default:
	    i += optionptr[OPT_LEN + i] + 2;
	}
    }
    return NULL;
}


/* return the position of the 'end' option (no bounds checking) */
int end_option(unsigned char *optionptr)
{
    int i = 0;

    while (optionptr[i] != DHCP_END) {
	if (optionptr[i] == DHCP_PADDING)
	    i++;
	else
	    i += optionptr[i + OPT_LEN] + 2;
    }
    return i;
}


/* add an option string to the options (an option string contains an option code,
 * length, then data) */
int add_option_string(unsigned char *optionptr, unsigned char *string)
{
    int i, end = end_option(optionptr);

    /* end position + string length + option code/length + end option */
    if (end + string[OPT_LEN] + 2 + 1 >= 308) {
	for (i = 0; options[i].code && options[i].code != string[OPT_CODE];
	     i++);
	LOG(LOG_ERR, "Option %s (0x%02x) did not fit into the packet!",
	    options[i].code ? options[i].name : "unknown",
	    string[OPT_CODE]);
	return 0;
    }
    DEBUG(LOG_INFO, "adding option 0x%02x", string[OPT_CODE]);
    memcpy(optionptr + end, string, string[OPT_LEN] + 2);
    optionptr[end + string[OPT_LEN] + 2] = DHCP_END;
    return string[OPT_LEN] + 2;
}


/* add a one to four byte option to a packet */
int add_simple_option(unsigned char *optionptr, unsigned char code,
		      u_int32_t data)
{
    char length = 0;
    int i, end;
    char buffer[4];		/* Cant copy straight to optionptr, it might not be aligned */

    for (i = 0; options[i].code; i++)
	if (options[i].code == code) {
	    length = option_lengths[options[i].flags & TYPE_MASK];
	    break;
	}

    if (!length) {
	DEBUG(LOG_ERR, "Could not add option 0x%02x", code);
	return 0;
    }

    DEBUG(LOG_INFO, "adding option 0x%02x", code);
    end = end_option(optionptr);
    optionptr[end + OPT_CODE] = code;
    optionptr[end + OPT_LEN] = length;

    switch (length) {
    case 1:
	buffer[0] = (char) data;
	break;
    case 2:
	*((u_int16_t *) buffer) = htons(data);
	break;
    case 4:
	*((u_int32_t *) buffer) = htonl(data);
	break;
    }
    memcpy(&optionptr[end + 2], buffer, length);
    optionptr[end + length + 2] = DHCP_END;
    return length;
}


/* find option 'code' in opt_list */
struct option_set *find_option(struct option_set *opt_list, char code)
{
    while (opt_list && opt_list->data[OPT_CODE] < code)
	opt_list = opt_list->next;

    if (opt_list && opt_list->data[OPT_CODE] == code)
	return opt_list;
    else
	return NULL;
}

/* add an option to the opt_list */
void attach_option(struct option_set **opt_list,
		   struct dhcp_option *option, char *buffer, int length)
{
    struct option_set *existing, *new, **curr;

    /* add it to an existing option */
    if ((existing = find_option(*opt_list, option->code))) {
	DEBUG(LOG_INFO, "Attaching option %s to existing member of list",
	      option->name);
	if (option->flags & OPTION_LIST) {
	    if (existing->data[OPT_LEN] + length <= 255) {
		existing->data = realloc(existing->data,
					 existing->data[OPT_LEN] + length +
					 2);
		memcpy(existing->data + existing->data[OPT_LEN] + 2,
		       buffer, length);
		existing->data[OPT_LEN] += length;
	    }			/* else, ignore the data, we could put this in a second option in the future */
	}			/* else, ignore the new data */
    } else {
	DEBUG(LOG_INFO, "Attaching option %s to list", option->name);

	/* make a new option */
	new = malloc(sizeof(struct option_set));
	new->data = malloc(length + 2);
	new->data[OPT_CODE] = option->code;
	new->data[OPT_LEN] = length;
	memcpy(new->data + 2, buffer, length);

	curr = opt_list;
	while (*curr && (*curr)->data[OPT_CODE] < option->code)
	    curr = &(*curr)->next;

	new->next = *curr;
	*curr = new;
    }
}


/*
 *  This function replaces option(s) for new one(s).
 */
void replace_option(struct dhcp_option *option, char *buffer, int length)
{   
    struct option_set *opt_list;
    
    opt_list = find_option(server_config.options, option->code);
    
    if (opt_list == NULL)
    {
        /* The option doesn't exist yet, so simply attach it. */
        attach_option((&server_config.options), option , buffer, length);
    }
    else
    {
        /* The option does exist, so we have to replace it.
           If the new situation has less options then the old one, the remaining options must be deleted.
           If the new situation has more options then the old one, options have to be added. */ 
        if (length == opt_list->data[OPT_LEN])
        {   /* sizes are equal, so overwrite old data with new one. */
            memcpy(opt_list->data + 2, buffer, length);
        }
        else if (length < opt_list->data[OPT_LEN])
        {   /* new size is less then old size, so we have to remove some old data. */
            opt_list->data = realloc(opt_list->data, length + 2);
            memcpy(opt_list->data + 2, buffer, length);
            opt_list->data[OPT_LEN] = length;
        }
        else
        {   /* new size is more then old size, so we have to add some new data. */
            opt_list->data = realloc(opt_list->data, length + 2);
            memcpy(opt_list->data + 2, buffer, length);
            opt_list->data[OPT_LEN] = length;
        } 
    }
}

