/*  $Header$
 *
 *  upgraded - specialized tftp upgrade daemon
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <arpa/inet.h>
#include <arpa/tftp.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "identification/image_header.h"
#include "identification/components.h"
#include "inl_debug.h"
#include "inl_misc.h"
#include "image.h"
#include "aplog.h"
#include "rockhopper.h"

#include "crc32.h"

#define TFTP_SERVER_PORT	69
#define TFTP_HEADER_SIZE	4
#define TIME_OUT		30

#define UPLOAD_FILENAME         "upload.img"

extern char *get_password(void);
extern int get_lock_status(const char *filename);

/* function prototypes */
static void usage( char *appname );
static void open_interface();
static void dispatch();
static void tftp_fsm( struct tftphdr *tp, unsigned int length,
		      struct sockaddr_in *from );
static void tftp_error( int error_code, char *error_msg,
			struct sockaddr_in *to );
static void tftp_send( int opcode, int block, struct sockaddr_in *to );
static upg_image_check_t image_check( image_header_t *hdr );
static void sig_alarm(int sig_no);
static int get_max_size_for_component(unsigned long comp_id);

  /* local variables */
						/* send and receive buffer = */
static char tftp_buffer[SEGSIZE + TFTP_HEADER_SIZE];

static int sd;					/* socket descriptor */
static struct sockaddr_in server_sockaddr;	/* server socket addr */
static struct sockaddr_in wrq_sockaddr;		/* requestor socket addr */

static char *snmp_write_community = "private";	/* default password */
static char *upload_dir = "/tftp";		/* default upload directory */
static char *flash_programmer = NULL;

/* The next 3 variables are module-global, because the sig_alarm handler needs
 * them to cleanup files and filedescriptors
 */
static FILE *file = NULL;                       /* pointer to uploadfile */
static int fd = 0;                              /* filedescriptor of uploadfile */
static char *filename;                          /* filename of uploaded file */

static void usage( char * appname )
{
    fprintf( stderr, "Usage: %s [-d tftp-dir] [-f flash_programmer]\n", appname);
    fprintf( stderr, "\t\t[-h] Shows this help\n" );
    exit(1);
}

int
main( int argc, char *argv[] )
{
    int i;
    char *appname;

    signal(SIGALRM, sig_alarm);

    /* determine application name from argumentlist */
    appname = strrchr( argv[0], '/');
    if( !appname )
	appname = argv[0];
    else
	appname++;

    /* parse arguments */
    for (i = 1; i < argc; i++ ) {
	if (!strcmp( argv[i], "-h" )) {
		usage( appname );
	} else if (!strcmp( argv[i], "-d" )) {
	    if (++i == argc )
		usage( appname );
	    upload_dir = argv[i];
	} else if (!strcmp( argv[i], "-f" )) {
	    if (++i == argc )
		usage( appname );
	    flash_programmer = argv[i];
        } else {
	    usage( appname );
	}
    }

    /* open syslog and report start of the tftp upgrade server */
    openlog( appname, LOG_NDELAY, LOG_DAEMON );
    note( "TFTP Upgrade Server v%s (%s) started", VER, BT );

    /* open interface */
    open_interface();

    /* handle messages */
    dispatch();

    /* should never executes */
    error( "Fall through" );
    closelog();
    return 0;
}

static void
sig_alarm(int signo)
{
    static char fullname[128];

    /* Alarm trigged, cleanup uploaded file */
    snprintf(fullname, sizeof(fullname), "%s/%s", upload_dir, filename);

    unlink(fullname);
    if(file)
        fclose(file);

    if(fd)
        close(fd);

    signal(SIGALRM, sig_alarm);
}


static void
open_interface()
{
    /* create an UDP socket to receive requests */
    if(( sd = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP )) == -1 ) {
	error_and_die( "Failed to create socket: %m" );
    }

    /* bind the socket to tftp port */
    server_sockaddr.sin_family = AF_INET;
    server_sockaddr.sin_port = htons(TFTP_SERVER_PORT);
    server_sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    bzero( &server_sockaddr.sin_zero, sizeof( server_sockaddr.sin_zero ) );
    if( bind( sd, (struct sockaddr *)&server_sockaddr, sizeof( server_sockaddr ) ) == -1 ) {
	close( sd );
	error_and_die( "Failed to bind to socket: %m" );
    }

    /* invalidate requestor address */
    bzero( &wrq_sockaddr, sizeof(wrq_sockaddr) );
}

static char *tftp_image_check(image_header_t *img_hdr)
{
    char *mode;

    switch( image_check(img_hdr) ){
    case upg_image_ok:
	mode = NULL;
	break;

    case upg_image_unsupported_oem_id:
	mode = "Unsupported OEM, aborting upgrade.";
	break;

    case upg_image_unsupported_product:
	mode = "Unsupported product, aborting upgrade.";
	break;

    case upg_image_unsupported_hardware_revs:
	mode = "Unsupported hardware revision, aborting upgrade.";
	break;

    case upg_image_unsupported_tftp_if:
	mode = "Unsupported TFTP interface, aborting upgrade.";
	break;

    case upg_image_unsupported_boot:
	mode = "Unsupported Bootloader revision, aborting upgrade.";
	break;

    case upg_image_unsupported_uapfw:
	mode = "Unsupported microAP revision, aborting upgrade.";
	break;

    case upg_image_unsupported_webfileset:
	mode = "Unsupported Web Interface Fileset, aborting upgrade.";
	break;

    case upg_image_invalid:
    default:
	mode = "Invalid image, aborting upgrade.";
	break;
    }

    return mode;
}

static void
dispatch()
{
    struct tftphdr *tp;
    struct sockaddr_in from;
    int address_len, recv_length, flags;

    /* setup from sockaddr */
    from.sin_family = AF_INET;
    from.sin_port = htons(0);
    from.sin_addr.s_addr = htonl(0);
    bzero( &from.sin_zero, sizeof( from.sin_zero ) );

    /* keep on waiting for messages and handle them */
    while( 1 ) {
	address_len = sizeof( from );
	flags = 0;
	recv_length = recvfrom( sd, tftp_buffer, sizeof(tftp_buffer),
				flags, (struct sockaddr *)&from, &address_len );
	if( recv_length == -1 ) {
	    if(errno != EINTR)
		warn( "recvfrom failed: %m" );
	} else {
#if 0
	    debug( "received %d bytes from %s:%d", recv_length,
		   inet_ntoa( from.sin_addr ), ntohs( from.sin_port ) );
#endif

	    tp = (struct tftphdr *)tftp_buffer;
	    tp->th_opcode = ntohs(tp->th_opcode);
	    tftp_fsm( tp, recv_length, &from );
	}
    }
}

static void
tftp_fsm( struct tftphdr *tp, unsigned int length, struct sockaddr_in *from )
{
    static int last_sync_block;

    static unsigned long crc_start;
    static unsigned long crc_end;
    static unsigned long crc_code;
    static int max_img_size = -1;
    static int bytes_received = 0;

    size_t cnt;
    size_t rcnt;

    unsigned long crc_code_calc = 0L;
    unsigned char buf[512];
   
    unsigned int block;

    char *mode, *password, *cp;

    int write_length;
    static char fullname[128];

    switch( tp->th_opcode ) {
    case WRQ:
	password = tp->th_stuff;
	mode = &password[strlen(password) + 1];

#if 0
	if(( filename = strchr( password, '/' )) != NULL ) {
	    *filename = '\0';
	    filename++;
	}
#else
	filename = UPLOAD_FILENAME;
#endif

	debug( "received WRQ, filename = %s, password = %s",
	       filename, password );

	if(snmp_get_aplockstate() != 0) {
		note("Upgrade not possible: AP is locked");
		tftp_error( EACCESS, "Access violation, access point is locked", from);
		break;
	}

	/* Retrieve password */
	if((snmp_write_community = get_password()) == NULL ) {
		warn("Upgrade not possible: Password file not present or corrupt");
		tftp_error( EACCESS, "Access violation", from);
		break;
	}

	/* verify that the password is correct and mode == octet */
	for (cp = mode; *cp; cp++){
	    if (isupper(*cp)){
		*cp = tolower(*cp);
	    }
	}

	if( strcmp( password, snmp_write_community )) {
	    warn("Upgrade not possible: Incorrect Password %s", password);
	    tftp_error( EACCESS, "Access violation", from);
	    break;
	}

        if( strcmp( mode, "octet" )) {
            warn("Upgrade not possible: No binary mode used");
	    tftp_error( EACCESS, "Access violation", from);
	    break;
        }

	/* Set alarm, so file will be removed when no DATA packets come */
	alarm(TIME_OUT);
	/*
	 * first the file is stored on a RAM disk, if it already exists
	 * the upgrade is performed by another method and access must be
	 * denied
	 */
	snprintf( fullname, sizeof(fullname), "%s/%s", upload_dir, filename );
	fd = open( fullname, O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
	if( fd < 0 ) {
	    warn( "Access violation: %m" );
	    tftp_error( EACCESS, "Access violation", from);
	    break;
	}
	file = fdopen( fd, "w" );
	if( file == NULL ) {
	    warn( "fdopen: %m" );
	    tftp_error( EUNDEF, "fdopen failed", from );
	    break;
	}

	/* valid write request, send ACK */
	tftp_send( ACK, 0, from );

	last_sync_block = 0;
	wrq_sockaddr.sin_addr.s_addr = from->sin_addr.s_addr;
	wrq_sockaddr.sin_port = from->sin_port;
	break;

    case DATA:

        /* received data, set alarm */
	alarm(TIME_OUT);

      	if( ( from->sin_addr.s_addr != wrq_sockaddr.sin_addr.s_addr )
	    || ( from->sin_port != wrq_sockaddr.sin_port ) ) {
	    warn( "Received data from a source that didn't send the WRQ" );
	    tftp_error( EACCESS, "Access violation", from );
	    goto tftp_data_err;
	}

	block = ntohs(tp->th_block);

	length -= TFTP_HEADER_SIZE;

	/* validate the image by checking the header except for combi */

	if (block == 1 ) {
	
	   bytes_received = 0;
		
	   /*------------------------------------------------*
	    | Fetch some crc stuff for crc-checking later... |
	    *------------------------------------------------*/
            crc_start = ((image_header_t *)tp->th_msg)->crc_start;
	    crc_end = ((image_header_t *)tp->th_msg)->crc_end;
	    crc_code = ((image_header_t *)tp->th_msg)->crc_code;

	    mode = tftp_image_check((image_header_t *)tp->th_data);

	    if( mode ) {
		warn( mode );
		tftp_error( EACCESS, mode, from );
		goto tftp_data_err;
	    }
	    max_img_size = get_max_size_for_component(((image_header_ext*)tp->th_msg)->component_id);
	}

	bytes_received += length;
	
	/* check size */
	if(bytes_received > max_img_size) {
               tftp_error( EACCESS, "Image size too big.", from );
               warn( "Image is too big. aborting upgrade.");
	       goto tftp_data_err;
	}
	
	if( block == last_sync_block + 1) {
	    last_sync_block = block;

	    write_length = fwrite( tp->th_data, 1, length, file );
	    if( write_length != length ) {
		warn( "fwrite: %m" );
		tftp_error( EUNDEF, "fwrite failed", from );
		goto tftp_data_err;
	    }
	} else {
	    /* out of sync */
	    warn( "Blocks out of sync (prev %d, curr %d)", last_sync_block, block);
	    break;
	}

	/* last data block is indicated with length < max size */
	if( length < SEGSIZE ) {
	    debug( "upload completed" );
	    alarm(0);
	    fclose( file );
	    close( fd );
	    
	    /*---------------------*
	     | Check the image CRC |
	     *---------------------*/
	    file = fopen(fullname, "r");
            fseek(file, (size_t)crc_start, SEEK_SET);
	    crc_code_calc = 0xFFFFFFFFL;
            while ((cnt = fread((void *)buf, (size_t)1, (size_t)512, file))) {
	       rcnt = ftell( file);
	       if (rcnt > crc_end) {
		  /*-----------------------------*
		   | We read passed crc_code_end |
		   *-----------------------------*/
                  crc_code_calc = crc32( crc_code_calc, (const void *)buf, (int)(rcnt - crc_end - 1));
		  break;
               } else {
                  crc_code_calc = crc32( crc_code_calc, (const void *)buf, (int)cnt);
               }
            }
	    fclose( file);
	    /*--------------------------------------------------------------*
	     | Is the calculated crc equal to the crc of the image header ? |
	     *--------------------------------------------------------------*/
	    if (crc_code_calc != crc_code) {
               tftp_error( EACCESS, "Invalid image CRC", from );
               warn( "Image CRC is not correct, aborting upgrade.");
	       goto tftp_data_err;
            }
	    debug("Image CRC is okay. Starting Flash Programmer...");
	    aplog(EV_TFTP_INFO, "Upgrade successful" , NULL);

	    /* ACK last block */
	    tftp_send( ACK, block, from );
	    
	    /* start flash programmer if specified */
	    if( flash_programmer ) {
		execlp( flash_programmer, flash_programmer, fullname, NULL );
		error( "Failed to start flash programmer %s: %m",
		       flash_programmer );
	    }else{
		exit(0);
	    }
	}else {
	    /* send ACK to last block that was in-sequence */
	    tftp_send( ACK, block, from );
	}
	break;

tftp_data_err:
	if( unlink(fullname) < 0 ) {
	    warn( "Can't remove false uploaded file %s: %m", fullname );
	    fclose( file );
	    close( fd );
	    alarm(0);
	}
	break;

    default:
	tftp_error( EBADOP, "Illegal TFTP operation", from );
	if( (unlink(fullname) < 0) && errno != ENOENT ) {
	    warn( "Can't remove false uploaded file %s: %m", fullname );
	    fclose( file );
	    close( fd );
	    alarm(0);
	}
	break;
    }

    return;
}

static void
tftp_error( int error_code, char *error_msg, struct sockaddr_in *to )
{
    struct tftphdr *tp;
    int length;

    if(error_code == EACCESS) {
	aplog(EV_TFTP_ACCESS_VIOLATION, error_msg, NULL);
    } else {
	aplog(EV_TFTP_ERROR, error_msg, NULL);
    }

    tp = (struct tftphdr *)tftp_buffer;
    tp->th_opcode = htons((u_short)ERROR);
    tp->th_code = htons((u_short)error_code);
    strcpy( tp->th_msg, error_msg );
    length = strlen( error_msg );
    tp->th_msg[length] = '\0';
    length += (TFTP_HEADER_SIZE + 1);		/* count trailing zero */
    if( sendto( sd, tftp_buffer, length, 0, (struct sockaddr *)to, sizeof(*to) ) != length ) {
	warn( "sendto failed: %m" );
    }

    /* invalidate requestor address */
    bzero( &wrq_sockaddr, sizeof(wrq_sockaddr) );
}

static void
tftp_send( int opcode, int block, struct sockaddr_in *to )
{
    struct tftphdr *tp;
    int length;

    tp = (struct tftphdr *)tftp_buffer;
    tp->th_opcode = htons((u_short)opcode);
    tp->th_block = htons((u_short)block);

    length = TFTP_HEADER_SIZE;

    if( sendto( sd, tftp_buffer, length, 0, (struct sockaddr *)to, sizeof(*to) ) != length ) {
	warn( "sendto failed: %m" );
    }
}

static upg_image_check_t
image_check( image_header_t *hdr )
{
    image_header_t *hdr_list[2];

    if (hdr->magic1 == MAGIC_OKAY) {
       hdr_list[0] = hdr;
       hdr_list[1] = NULL;
       return image_file_check (hdr_list);
    } else {
       return upg_image_invalid;
    }
}

static int 
get_max_size_for_component(unsigned long comp_id)
{
	switch(comp_id) {
		case BR_COMP_ID_DAP1:
			return(CONFIG_ROCKHOPPER_APFW_END_ADDR - CONFIG_ROCKHOPPER_APFW_START_ADDR);
			break;
		case BR_COMP_ID_DFS1:
			return(CONFIG_ROCKHOPPER_MAX_WFS_SIZE);
			break;
		case BR_COMP_ID_LTM1: 
		case BR_COMP_ID_ITM1:
			return(CONFIG_ROCKHOPPER_MAX_UAP_SIZE);
			break;
		case BR_COMP_ID_DBCO:
			return((CONFIG_ROCKHOPPER_APFW_END_ADDR - CONFIG_ROCKHOPPER_APFW_START_ADDR) 
					+ CONFIG_ROCKHOPPER_MAX_WFS_SIZE 
					+ CONFIG_ROCKHOPPER_MAX_UAP_SIZE);
			break;
		default:
			return -1;
			break;
	}
}
