/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>

#include "types.h"
#include "table.h"
#include "object.h"
#include "objflag.h"
#include "serialize.h"
#include "debug.h"
#include "debug.h"
#include "schema.h"
#include "misc.h"
#include "configfile.h"
#include "aplog.h"

#define WRITE_TIMEOUT   5

#define CFGFILE_HEADER 	"#\n" \
			"# AP Configuration file\n" \
			"#\n" \
			"# This file is automaticly written by datastore, do not edit\n" \
			"#\n" \
			"\n"
 
#define CONFIGFILE_DATA_DEFAULT  1
#define CONFIGFILE_DATA_USER     2
#define CONFIGFILE_DATA_UPGRADE  3

extern gAVLTree *schematree;
extern gAVLTree *objtree;

static int write_timer = -1;

static void addoid(char *oid, char *data, int configtype)
{
	struct object_t *obj;

	
	/* Create object */
	switch(configtype) {
		case CONFIGFILE_DATA_DEFAULT:
			obj = object_create(oid);
			assert(obj);
			break;
		case CONFIGFILE_DATA_USER:
			obj = object_create(oid);
			
			if(!obj) {
				/* user config is corrupted. Log to aplog and
				   ignore this oid */
				aplog(EV_SNMP_CONFIGFILE_ERROR, oid, data, NULL);
				return;
			}
			break;
		case CONFIGFILE_DATA_UPGRADE:
			obj = object_get(oid);
			if(!obj) {
				/* object is not created by default config 
				 * It doesn't exist in new configuration, drop it. */
				aplog(EV_SNMP_INFO_UPGRADE, oid, data, NULL);
				return;
			}
			break;
		default:
			die();
	}
	
	assert(obj->sch);

	if(deserialize(obj, data) != R_NOERR) {
		/* Parse error in configfile, react depending on configtype */
		switch(configtype) {
			case CONFIGFILE_DATA_DEFAULT:
				/* should never happen in default data */
				die();
				break;
			case CONFIGFILE_DATA_USER:
				/* User datafile corrupted, report but ignore */
				aplog(EV_SNMP_CONFIGFILE_ERROR, oid, data, NULL);
				
				/* The read data is invalid, but the object
				 * should be persistant. The next write of the
				 * data should write the, hopefully correct, 
				 * value for the object 
				 *
				 * Mark the schema entry of this object as 
				 * persistant 
				 */
				obj->sch->persistant = 1;
				break;
			case CONFIGFILE_DATA_UPGRADE:
				/* New firmware couldn't parse old datafile.
				 * report and ignore */
				aplog(EV_SNMP_INFO_UPGRADE, oid, data, NULL);
				
				/* Don't mark the object persistant. If the object
				 * should still be persistant after upgrade, the
				 * reading of the default data has enabled this
				 * flag. If not, the object is not persistant in
				 * the new firmware
				 */
				break;
			default:
				die();
		}
	}
	else {
		/* Mark the schema entry of this object as persistant */
		obj->sch->persistant = 1;
	}
	
	/* Mark object dirty */
	object_flag_set(obj, OF_DIRTY);

}

static void addrow(char *oid, char **indexdata, int indices,  char **data, int parts, int configtype)
{
	struct index_t *index;
	struct row_t *row;
	struct object_t *obj;
	int i;
	int user_row_created = B_FALSE;
	struct schent_t* sch;
	
	sch = schema_get(oid);
	assert(sch);
	
	index = index_new(sch);

	for(i=0; i<indices; i++) {
		deserialize(index_get_obj(index, i), indexdata[i]);
	}

	switch(configtype) {
		case CONFIGFILE_DATA_DEFAULT:
			row = row_create_by_index(index);
			assert(row);
			break;
		case CONFIGFILE_DATA_USER:
			row = row_create_by_index(index);
			if(!row) {
				/* user config is corrupted. Log to aplog, 
				   cleanup and leave function */
				aplog(EV_SNMP_CONFIGFILE_ERROR, oid, "", NULL);
				index_done(index);
				return;
			}
			break;
		case CONFIGFILE_DATA_UPGRADE:
			/* check if the row already exists */
			row = row_get_by_index(index);
			if(!row) {
				/* row doesn't exist yet -> create it */
				row = row_create_by_index(index);
				if(row) {
					user_row_created = B_TRUE;
				}
				else {
					/* this firmware couldn't parse the old
					 * row. Report, but ignore */
					aplog(EV_SNMP_INFO_UPGRADE, oid, "", NULL);
					index_done(index);
					return;
				}
			}
			break;
		default:
			die();
			break;
	}

	index_done(index);
	
	if(parts > row->columns) {
		switch(configtype) {
			case CONFIGFILE_DATA_DEFAULT:
				die();
				break;
			case CONFIGFILE_DATA_USER:
				aplog(EV_SNMP_CONFIGFILE_ERROR, oid, "", NULL);
				return;
				break;
			case CONFIGFILE_DATA_UPGRADE:
				if(user_row_created) {
					row_delete(row);
				}
				return;
				break;
			default:
				die();
				break;
		}
	}

	for(i=0; i<parts; i++) {
		obj = row_get_obj(row, i);
		assert(obj);
		if (strcmp(data[i], "NULL") != 0) {
			if (deserialize(obj, data[i]) != R_NOERR) {
				switch(configtype) {
					case CONFIGFILE_DATA_DEFAULT:
						printf("Inconsisent values in datastore.dat\n");
						printf("Value  = %s\n", data[i]);
						printf("Object = %s\n", obj->sch->oid);
						printf("Rowoidp= %s\n", row->oidpart);
						printf("Range  = %s\n", obj->sch->range);
						die();
						break;
					case CONFIGFILE_DATA_USER:
						aplog(EV_SNMP_CONFIGFILE_ERROR, oid, data[i], NULL);
						break;
					case CONFIGFILE_DATA_UPGRADE:
						aplog(EV_SNMP_INFO_UPGRADE, oid, data[i], NULL);
						break;
					default:
						die();
				}
			}
		}
		
		/* Mark object dirty */
		object_flag_set(obj, OF_DIRTY);
	}
	
 }


 
 
 
static int read_config(char *fname, int configtype)
{
	char buf[256];

	char *part[MAX_NO_COLUMNS + MAX_NO_INDICES + 2]; /* +2 = "row" and "/" */  
	int parts;

	char *index[MAX_NO_INDICES];
	int indices;

	char tableoid[MAX_OID_LENGTH + 1];

	int linenum=0;

	struct schent_t *sch = NULL;

	FILE *f;
	char *cmd;
	char *oid;
	char *data;
	char *p;
	
	int table_read = B_FALSE;

	dbg(DBG_CFGFILE, "Opening configfile '%s'", fname);
	f = fopen(fname, "r");
	
	if(f==NULL) {
		dbg(DBG_CFGFILE, "Can't open cfgfile for reading");
		return(-1);
	}

	while(fgets(buf, sizeof(buf), f)) {
		
		linenum++;

		// Skip comments and empty lines, chomp off newline

		if(buf[0] == '#') continue;
		p=strchr(buf, '\n'); if(p) *p=0;
		p=strchr(buf, '\r'); if(p) *p=0;
		if(buf[0] == 0) continue;
	
		parts = splitws(buf, part, 2);

		if(parts < 2) {
			fprintf(stderr, "Parse error in datastore.dat on line %i\n", linenum);
			continue;
		}
	
		cmd = part[0];
		data = part[1];

		dbg(DBG_CFGFILE, "Read %s %s", cmd, data);
	
		switch(part[0][0]) {
			case 'o':	// obj

				parts = splitws(data, part, 2);
				if(parts == 1) {
					part[1] = "";
				}
					
				oid = part[0];
				data = part[1];
	
				addoid(oid, data, configtype);
				sch = NULL;
				table_read = B_FALSE;
				break;
		
			case 't':	// tbl
	
				oid = data;
				table_read = B_TRUE;
				
				sch = schema_get(oid);
				if(!sch || (sch && (sch->type != T_TABLE))) {
					/* schema for this oid does not exist or is not a table*/
					switch(configtype) {
						case CONFIGFILE_DATA_DEFAULT:
							fprintf(stderr, "Parse error on line %i, oid is not a table oid\n", linenum);						
							die();
							break;
						case CONFIGFILE_DATA_USER:
							aplog(EV_SNMP_CONFIGFILE_ERROR, oid, "" , NULL);
							break;
						case CONFIGFILE_DATA_UPGRADE:
							aplog(EV_SNMP_INFO_UPGRADE, oid, "", NULL);
							break;
						default:
							die();
							break;
					}
					break;
				}

				strcpy(tableoid, oid);
				sch->persistant = 1;
		
				break;
		
			case 'r':	// row

				if(!table_read) {
					printf("Row in configfile before tbl defined at linenum %d\n", linenum);
					goto row_error;
				}
		
				if(!sch) {
					/* this row corresponds with a table for which no
					 * oid could be found. Just ignore this row 
					 */
					break;
				}
				
				p=strchr(data, '/');
		
				if (!p) {
					printf("Parse error in line %i, '/' not found in row\n", linenum);
row_error:
					switch(configtype) {
						case CONFIGFILE_DATA_DEFAULT:
							die();
							break;
						case CONFIGFILE_DATA_USER:
							aplog(EV_SNMP_CONFIGFILE_ERROR, tableoid, "", NULL);
							break;
						case CONFIGFILE_DATA_UPGRADE:
							aplog(EV_SNMP_INFO_UPGRADE, tableoid, "", NULL);
							break;
						default:
							die();
							break;
					}
					break;

				}
				
				*p=0;
				p++;
	
				indices = splitws(data, index, MAX_NO_INDICES);
				dbg(DBG_CFGFILE, "Read %d parts for index", indices);
	
				parts = splitws(p, part, MAX_NO_COLUMNS);
				dbg(DBG_CFGFILE, "Read %d parts for data", parts);
	
				addrow(tableoid, index, indices, part, parts, configtype);
		
				break;
		
			default:
				printf("Syntax error in configfile line %d\n", linenum);
		}
	
	}

	fclose(f);

	dbg(DBG_CFGFILE, "Done reading configuration file");

	return(0);
}


/**
 * Read in the configuration file
 *
 * Format :
 *
 * Set 'standalone' oid :  <oid> = <value>
 * Add row to table     :  <oid> + <value> <value> <value> <...>
 *
 * If the user_fname-file does not exist, the default_fname-file is
 * read.
 * If the user_fname-file does exist, the current build date in /etc/BUILD
 * is compared with the build date in the user_fname file. If they match
 * only the user_fname is read. If they don't match, it is assumed that
 * the firmware was upgraded: First the default file is read and after 
 * that the user_fname is read. 
 *
 */
int configfile_read(char* default_fname, char* user_fname) {

	FILE* tmp_fd;
	char buf1[80];
	char buf2[80];
	char* p1, *p2;
	
	tmp_fd = fopen(user_fname, "r");
	
	if(!tmp_fd) {
		/* no user file, just read default */
		aplog(EV_SNMP_INFO, "Reading default config\n", NULL);
		return(read_config(default_fname, CONFIGFILE_DATA_DEFAULT));
	}
	else {
		p1 = NULL;
		p2 = NULL;
		/* read until the "# BUILDTIME" tag */
		while(fgets(buf1, sizeof(buf1), tmp_fd)) {
			if(strncmp(buf1, "# BUILDTIME", sizeof("# BUILDTIME") - 1) == 0) {
				/* found tag */
				p1 = buf1 + 2; /* skip # and space */	
				break;
			}
		}						
		fclose(tmp_fd);
		if(p1) {
			/* read /etc/BUILD */
			tmp_fd = fopen("/etc/BUILD", "r");
			assert(tmp_fd);
			
			/* read until the BUILDTIME tag */
			while(fgets(buf2, sizeof(buf2), tmp_fd)) {
				if(strncmp(buf2, "BUILDTIME", sizeof("BUILDTIME") - 1) == 0) {
					/* found it */
					p2 = buf2;
					break;
				}
			}
			
			fclose(tmp_fd);
		}
		
		if(p1 && p2 && (strcmp(p1, p2) == 0)) {
			/* BUILDTIME's match, just read user_data */
			aplog(EV_SNMP_INFO, "Reading user config\n", NULL);
			return(read_config(user_fname, CONFIGFILE_DATA_USER));
			
		}
		else {
			/* no match, assume upgrade */
			aplog(EV_SNMP_INFO, "Reading default and user config (upgrade)\n", NULL);
			read_config(default_fname, CONFIGFILE_DATA_DEFAULT);
			
			return(read_config(user_fname, CONFIGFILE_DATA_UPGRADE));
		}
	}
}

static int writetable(struct schent_t *sch, FILE *f)
{
	struct indexoid_t *ioid;
	struct object_t *obj;
	struct schent_t *sch2;
	struct row_t *row;
	struct index_t *idx;
	char buf[256];
	char buf2[512];
	char *buf_p, *buf2_p;
	int i;

	// We should only be called for persistant tables

	assert(sch->persistant);

	// And only for tables !

	assert(sch->type = T_TABLE);

	// If one of our objects has already the OF_CFGWRITTEN flag,
	// this table is already written, and should not be done again

	row = sch->rowlist;
	if(row) {
		obj = &(row->objs[0]);
		if(obj) {
			if(object_flag_isset(obj, OF_CFGWRITTEN)) {	
				return(0);
			}
		}
	}



	ioid = sch->indexlist;

	// First find other tables this table depends on by index, and do these
	// first

	while(ioid) {

		sch2 = schema_get(ioid->oid);
	
		assert(sch2);
		assert(sch2->prev);
		while(sch2->prev) sch2=sch2->prev;
		assert(sch2->type == T_TABLE);

		if(sch2 != sch) {
			writetable(sch2, f);
		}
	
		ioid = ioid->next;
	}


	row = sch->rowlist;

	fprintf(f, "\n");
	fprintf(f, "tbl %s\n", sch->oid);
	
	while(row) {
	
		fprintf(f, "row");
		
		idx = index_get(row);
		assert(idx);
	
		for(i=0; i<idx->num; i++) {
			serialize(index_get_obj(idx, i), buf, sizeof(buf));
			fprintf(f," %s", buf);
		}
	
		index_done(idx);
	
		fprintf(f," /");
		
		for(i=0; i<row->columns; i++) {
			obj = row_get_obj(row, i);
			assert(obj);
			
			if(obj->sch->type != T_OCTETSTR) {
				serialize(row_get_obj(row, i), buf, sizeof(buf));
			}
			else {
				serialize(row_get_obj(row, i), buf2, sizeof(buf2));
				
				/* escape spaces */
				buf_p = buf;
				buf2_p = buf2;
				
				while((*buf2_p) != '\0') {
					
					if(*buf2_p == ' ') {
						*buf_p = '\\'; /* \\ makes one backslash */
						buf_p++;
					}								
					
					*buf_p = *buf2_p;
					
					buf_p++;
					buf2_p++;
				}
				
				*buf_p = '\0';
			}
				
			fprintf(f," %s", buf);
	
			object_flag_set(obj, OF_CFGWRITTEN);
		}
		
		fprintf(f, "\n");
/*		
		printf("f = %x\n", f);
		printf("f->fd = %i\n", f->fd);
		printf("f->bufpos = %x\n", f->bufpos);
		printf("f->bufread = %x\n", f->bufread);		
		printf("f->bufwrite = %x\n", f->bufwrite);
		printf("f->bufstart = %x\n", f->bufstart);
		printf("f->bufend = %x\n", f->bufend);
		printf("f->next = %x\n", f->next);
		printf("f->mode = %c\n", f->mode);
		printf("f->ungot = %c\n", f->ungot);
		printf("buf = %s\n", buf);
*/		
		row = row->next;
	}

	fprintf(f, "\n");

	return(0);
}



static int writeobjects(FILE *f)
{
	struct object_t *obj;
	struct schent_t *sch;
	gAVLCursor avlcursor;
	char buf[256];

	obj = (struct object_t *) gAVLFirst(&avlcursor, objtree);

	while(obj) {
		object_flag_clear(obj, OF_CFGWRITTEN);
		obj = (struct object_t *) gAVLNext(&avlcursor);
	}


	obj = (struct object_t *) gAVLFirst(&avlcursor, objtree);

	while(obj) {
		dbg(DBG_CFGFILE, "Writing %s.X", obj->sch->oid);
		
		sch = obj->sch;
		assert(sch);
		
		// If this object is part of a table, call the
		// writetable function for it's sch
	
		if(! object_flag_isset(obj, OF_CFGWRITTEN)) {

			if((sch->prev) || (sch->type == T_TABLE)) {

				while(sch->prev) sch=sch->prev;
		
				if(sch->persistant) {
					writetable(sch, f);
				}
	
			} else {
				/* This is a leaf object */
				if(sch->persistant) {
					serialize(obj, buf, sizeof(buf));
					fprintf(f, "obj %s.0 %s\n", obj->sch->oid, buf);
					object_flag_set(obj, OF_CFGWRITTEN);
				}
			}
		}
	
		obj = (struct object_t *) gAVLNext(&avlcursor);
	}
	
	return(0);

}	

static void writebuilddate(FILE* fp) {
	
	FILE* bfp;
	char buf[80];
	
	bfp = fopen("/etc/BUILD", "r");
	assert(bfp);
	
	while(fgets(buf, sizeof(buf), bfp)) {
		if (strncmp(buf, "BUILDTIME", sizeof("BUILDTIME") - 1) == 0) {
			fprintf(fp, "# %s", buf);
			break;
		}
	}

	fclose(bfp);	
	
}

int configfile_write(char *fname)
{
	FILE *f;
	struct schent_t *sch;
	gAVLCursor avlcursor;

	dbg(DBG_CFGFILE, "Writing configuration file");

	f = fopen(fname, "w");
	setvbuf(f, NULL, _IONBF, 0); /* disable buffering (to workaround buffer bug) */

	fprintf(f, CFGFILE_HEADER);
	
	writebuilddate(f);

	writeobjects(f);
	
	sch = (struct schent_t *) gAVLFirst(&avlcursor, schematree);

	/* write empty but persistant tables */
	while(sch) {
		if((sch->type == T_TABLE) 
				&& (sch->rowlist == NULL)
				&& (sch->persistant)) {
			writetable(sch,f);		
		}
	
		sch = (struct schent_t *) gAVLNext(&avlcursor);
	}
	
	fclose(f);

	return(0);
}

void configfile_schedule_write(){
	write_timer = WRITE_TIMEOUT;
}

void configfile_scheduler_check() {
	
	if(write_timer != -1) {
		dbg(DBG_CFGFILE, "Writing configfile in %i second(s)", write_timer);
		write_timer--;
		
		if(write_timer < 1) {
			char *cmd[4];
			configfile_write("/tmp/datastore.dat");
			//unlink("/usr/etc/datastore.dat");
			
			//cmd[0] = "/bin/mv";
			cmd[0] = "/bin/cp";
			cmd[1] = "/tmp/datastore.dat";
			cmd[2] = "/usr/etc/datastore.dat";
			cmd[3] = NULL;
			
			exec(cmd);			
			/* Stardate 10404.26 Max Yin, note from BARGTMN
                                In order to prevent datastore.dat lost, not delete /usr/etc/datastore.dat
                                and not remove /tmp/datastore.dat
                         */
			write_timer = -1;
		}
	}		
}
