/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>

#include "schema.h"
#include "types.h"
#include "object.h"
#include "debug.h"

/** 
 * Linked list of handlers that are scheduled for execution at time 'when'
 */
struct hsched_t {
	struct handler_t *hnd;
	char *oid;
	time_t when;
	struct hsched_t *next;
};

struct hsched_t *hsched_list = NULL;

/**
 * Add a handler to be scheduled in 'seconds' seconds 
 */
static int handler_scheduler_add(struct handler_t *hnd, char *oid, int seconds)
{
	struct hsched_t *hs;
	
	hs = hsched_list;

	/* See if this handler is already scheduled */
	while(hs) {
		if(hs->hnd == hnd) {
			dbg(DBG_HANDLER, "Handler for %s already in scheduler, updating", oid);
			break;
		}
		hs = hs->next;
	}	

	/* If the handler was not found in the list, add it */
	if(hs == NULL) {
		hs = malloc(sizeof(struct hsched_t));
		assert(hs);
		hs->next = hsched_list;
		hsched_list = hs;

		hs->hnd  = hnd;
		hs->oid  = strdup(oid);
	}
	
	/* 'when' must be updated for handlers that were already in the list */
	hs->when = time(NULL) + seconds;

	dbg(DBG_HANDLER, "Scheduling handler for %s in %d seconds", oid, seconds);

	return(0);
}

int handler_scheduler_check(void)
{
        struct hsched_t *hs;
        struct hsched_t *hsprev;
        struct hsched_t *hsnext;
        time_t now;

	now = time(NULL);
	hs = hsched_list;
	hsprev = NULL;
	
	while(hs) {
		hsnext = hs->next;
	
		if(now >= hs->when) {
	
			/* Call the handler */
			dbg(DBG_HANDLER, "Calling scheduled handler for %s", hs->oid);

			assert(hs->hnd->func); /* Paranoia */
			
			hs->hnd->func();
	
			/* Remove it from the list */
			if(hsprev) {
				hsprev->next = hs->next;
			}
			else {
				/* first in list */
				hsched_list = hs->next;
			}

			assert(hs->oid);
			free(hs->oid);
			free(hs);
		}
		else {
			hsprev = hs;
		}
		hs = hsnext;
	}

	return(0);
}


/**
 * Dump a list of scheduled handlers
 */
int handler_scheduler_dump(void)
{
        struct hsched_t *hs;
        time_t now;

	now = time(NULL);
	hs = hsched_list;

	while(hs) {
		printf("oid:%s  in %ld seconds\n", 
			hs->oid, (hs->when - now));
			
		hs = hs->next;
	}

	return(0);
}


int handler_call(char *oid, int type, int seconds)
{
	struct schent_t *sch;
	struct handlerlist_t *hlist;
	int exptime;

	sch = schema_get(oid);

	if(sch == NULL) {
		sch = schema_find_schent_for_oid(oid);
	}

	if(sch == NULL) {
		return(0);
	}

	/* If the oid is in a table, find the table schema entry.
	 * The handler for all oids in a table is always connected
	 * to the table entry 
	 */
	while(sch->prev) {
		sch = sch->prev;
	}

	hlist = sch->handlerlist;

	while(hlist) {
		if(hlist->hnd->type == type) {			
			if(seconds == 0) {
				if(type == H_GET) {
					if((hlist->hnd->expiretime != 0) && (time(NULL) > hlist->hnd->expiretime)) {
						hlist->hnd->expiretime = 0;
						dbg(DBG_HANDLER, "Calling GET handler for %s", sch->oid);

						assert(hlist->hnd->func); /* Paranoia */
						exptime = hlist->hnd->func();
						dbg(DBG_HANDLER, "Returned from GET handler for %s", sch->oid);
			
						if(exptime == 0) {
							dbg(DBG_HANDLER, "Expire time disabled, function is never called again");
						} else {
							dbg(DBG_HANDLER, "Expire time set to %d seconds", exptime);
							hlist->hnd->expiretime = time(NULL) + exptime;
						}
					}
				}
				if(type == H_CHK) 
				{
					int ret_val;
					
					dbg(DBG_HANDLER, "Calling CHK handler for %s", sch->oid);
					
					assert(hlist->hnd->func); /* Paranoia */
					ret_val = hlist->hnd->func();
					
					dbg(DBG_HANDLER, "CHK handler returned %i\n", ret_val);
					
					if (ret_val != R_NOERR) {
						return ret_val;
					}									
				}
				if(type == H_SET) {
					dbg(DBG_HANDLER, "Calling SET handler for %s", sch->oid);
					
					assert(hlist->hnd->func);
					hlist->hnd->func();
				}
				
			} else {
				handler_scheduler_add(hlist->hnd, oid, seconds);
			}		

		}
		hlist = hlist->next;
	}
	return R_NOERR;
}

struct handler_t *handler_add(int type, int(*func)(void))
{
	struct handler_t *new;

	assert(func);
	
	new = malloc(sizeof(struct handler_t));
	assert(new);

	new->type = type;
	new->func = func;
	new->expiretime = time(NULL);

	return(new);
}


/*
 * Connect a handler to the schent_t object with the given oid
 */
 
int handler_connect(struct handler_t *hnd, char *oid)
{
	struct schent_t *sch;
	struct handlerlist_t *hlist;

	sch = schema_get(oid);

	dbg(DBG_HANDLER, "connecting handler to %s", oid);

	if(sch == NULL) {
		printf("Can't connect handler to unknown oid %s\n", oid);
		return(-1);
	}

	hlist = malloc(sizeof(struct handlerlist_t));
	assert(hlist);

	hlist->hnd  = hnd;
	hlist->next = sch->handlerlist;
	
	sch->handlerlist = hlist;

	/* create a 'ghost' table object, if needed. Table oid's are not "real"
	 * objects, but the handlers should connect to them. Therefore a dummy
	 * "ghost" object is created
	 */
	if((sch->type == T_TABLE) && (!object_get(sch->oid))) {
		dbg(DBG_HANDLER, "Creating ghost object %s for table handler", sch->oid);
		object_create(sch->oid);
	}

	return(0);
}

