/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <assert.h>

#include "journal.h"
#include "types.h"
#include "object.h"
#include "objdata.h"
#include "schema.h"
#include "table.h"
#include "debug.h"
#include "misc.h"

#define JF_DATAPRESENT   (1 << 0)
#define JF_DATAMALLOCED  (1 << 1) 

static int journal_active = 0;
static int journal_size = 0;

static struct jent_t *journal = NULL;

struct jent_t *journal_getfirstentry()
{
	return journal;
}

static int data_size(struct object_t *obj) 
{
	assert(obj);
	assert(obj->sch);
	
	switch(obj->sch->type) {
		case T_INTEGER:
		case T_UNSIGNED:
		case T_ROWSTAT:
		case T_COUNTER:
		case T_GAUGE:			
		case T_TICKS:
			return(sizeof(int));
			break;
		case T_COUNTER64: 
			return(sizeof(unsigned long long));
			break;
		case T_OCTETSTR:
		case T_OPAQUE:
		case T_OBJECTID:
			assert(obj->flags & OF_DATAPRESENT);
			return(((string*)obj->value)->len + 1 + sizeof(int));
			break;
		case T_IPADDR:
			return (4); /* TODO */
			break;
		case T_BITS:
			return((object_get_max_bit(obj) / 8) + 1);
			break;

		default:
			printf("Unsupported type for journaling\n");
			die();
			break;
		
	}	
}


static void journal_add(int action, void *thing)
{
	struct jent_t *j;
	int ds;

	if(journal_active == 0) return;

	j = malloc(sizeof(struct jent_t));
	assert(j);
	journal_size += sizeof(struct jent_t);

	j->action = action;
	j->value = 0;
	j->flags = 0;

	switch(j->action) {
		case J_ROWADD:
		case J_ROWDEL:
			j->row  = (struct row_t *)thing;
			break;
	
		case J_OBJSET:
			j->obj  = (struct object_t *)thing;
			if(j->obj->flags & OF_DATAPRESENT) {
				/* duplicate data */
				if(object_has_malloced_data(j->obj)) {
					ds = data_size(j->obj);
					j->value = (unsigned long) malloc(ds);
					assert(j->value);
					memcpy((void*)j->value, (void*) j->obj->value, ds);
					journal_size += ds;
					j->flags |= JF_DATAMALLOCED;
				}
				else {
					/* embedded data */
					j->value = j->obj->value;
				}
				
				j->flags |= JF_DATAPRESENT;
			}
			break;
	}

	j->next = journal;
	journal = j;
}




static void restore_data(struct object_t *obj, unsigned long value) {

	if((obj->flags & OF_DATAPRESENT) && (object_has_malloced_data(obj))) {
		free((void*)obj->value);
	}
	obj->value = value;
}

void journal_begin(void)
{	
	if(journal_active == 1) {
		printf("Journal already active\n");
	}
	dbg(DBG_JOURNAL, "journal begin");
	journal_active = 1;
}

void journal_rowadd(struct row_t *row)
{
	journal_add(J_ROWADD, (void *)row);
}

void journal_rowdel(struct row_t *row)
{
	journal_add(J_ROWDEL, (void *)row);
}

void journal_objset(struct object_t *obj)
{
	journal_add(J_OBJSET, (void *)obj);
}


void journal_commit(void)
{
	struct jent_t *j;

	if(journal_active != 1) {
		printf("Can't commit : journal is not active\n");
		return;
	}

	j=journal;

	while(journal) {
		j = journal;
		switch(j->action) {
			case J_ROWADD:
				break;
			case J_ROWDEL:
				row_delete(j->row);				
				break;
				
			case J_OBJSET:
				if((j->flags & JF_DATAMALLOCED)) {
					free((void*) j->value);
					j->value = 0;
				}
				break;
			default:
				die();
				break;	
		}
		journal = journal->next;
		
		free(j);
	}

	journal_size = 0;
	journal_active = 0;

}


void journal_rollback(void)
{
	struct jent_t *j;

	if(journal_active != 1) {
		printf("Can't rollback : journal is not active\n");
		return;
	}


	dbg(DBG_JOURNAL, "journal rollback");

	journal_active = 0;

	while(journal) {
		
		j = journal;

		switch(j->action) {
	
			case J_ROWADD:
				dbg(DBG_JOURNAL, "- deleting added row on %s", j->row->sch->oid);
				row_delete(j->row);
				j->row = NULL;
				break;
			case J_ROWDEL:
				dbg(DBG_JOURNAL, "- adding deleted row on %s", j->row->sch->oid);
				j->row->next = j->row->sch->rowlist;
				if(j->row->next) {
					j->row->next->prev = j->row;
				}
				j->row->sch->rowlist = j->row;
				j->row->prev = NULL;
				break;
			case J_OBJSET:
				dbg(DBG_JOURNAL, "- reverting %s%s", j->obj->sch->oid, j->obj->row ? j->obj->row->oidpart : ".0");
				restore_data(j->obj, j->value);

				//object_flag_clear(j->obj, OF_DIRTY);
				/* Don't clear flag, because : 
				 * - SET of OID X with correct value -> set handler scheduled 
				 * - Within 3 seconds, SET of OID X with incorrect value.
				 * The set handler of the first handler will be called, the 
				 * object should still be 'dirty' 
				 */
				j->value = 0;
				break;
		}
		
		journal = journal->next;

		free(j);
		
	}
	
	journal_active = 0;
}


void journal_dump(void)
{
	struct jent_t *j;

	j = journal;

	while(j) {
		switch(j->action) {
			case J_ROWADD:	
				printf("ROWADD p=%p row->sch->oid=%s\n", j->row, j->row->sch->oid); 
				break;
		
			case J_ROWDEL:	
	
				printf("ROWDEL p=%p row->sch->oid=%s\n", j->row, j->row->sch->oid); 
				break;
		
			case J_OBJSET:	
	
				printf("OBJSET p=%p OID=%s%s \n", j->obj, j->obj->sch->oid, j->obj->row ? j->obj->row->oidpart : ".0"); 
				break;
		}
	
		j = j->next;
	
	}

	printf("\nJournal size : %.2f kb\n", (journal_size/1024.0));
}
