/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

#include "misc.h"
#include "types.h"
#include "gAVLTree.h"
#include "schema.h"
#include "debug.h"
#include "object.h"
#include "journal.h"
#include "handler.h"

gAVLTree *schematree;

static char *hex2ipstr(char *str);

gAVLKey schema_getavlkey(void *item)
{
	struct schent_t *obj = (struct schent_t *)item;

	return((gAVLKey)obj->oid);
}


int schema_inittree(void)
{
        schematree = gAVLAllocTree((gAVLKey) schema_getavlkey, (gAVLCompare) oidcmp);

	return(0);
}


struct schent_t *schema_get(char *oid)
{
	struct schent_t *sch;
        
	sch = (struct schent_t *)gAVLSearch(schematree, (gAVLKey)oid);

	return (sch);
}

struct row_t* schema_get_first_row(struct schent_t* sch) {
	
	assert(sch);
	
	handler_call(sch->oid, H_GET, 0);
	
	return(sch->rowlist);
}


/*
 * Helper functions for schema_read
 */
 
static int schema_type2int(char *p)
{
	if(strcmp(p, "INTEGER") == 0) return(T_INTEGER);
	if(strcmp(p, "INTEGER32") == 0) return(T_INTEGER);
	if(strcmp(p, "UNSIGNED32") == 0) return(T_UNSIGNED);
	if(strcmp(p, "OCTETSTR") == 0) return(T_OCTETSTR);
	if(strcmp(p, "DISPLAYSTR") == 0) return(T_OCTETSTR);
	if(strcmp(p, "IPADDR") == 0) return(T_IPADDR);
	if(strcmp(p, "TABLE") == 0) return(T_TABLE);
	if(strcmp(p, "ROWSTAT") == 0) return(T_ROWSTAT);
	if(strcmp(p, "COUNTER") == 0) return(T_COUNTER);
	if(strcmp(p, "COUNTER64") == 0) return(T_COUNTER64);
	if(strcmp(p, "GAUGE") == 0) return(T_GAUGE);
	if(strcmp(p, "OBJECTID") == 0) return(T_OBJECTID);
	if(strcmp(p, "OPAQUE") == 0) return(T_OPAQUE);	
	if(strcmp(p, "TICKS") == 0) return(T_TICKS);		
	if(strcmp(p, "NOTIF") == 0) return(T_NOTIFY);
	if(strcmp(p, "BITS") == 0) return (T_BITS);
	if(strcmp(p, "AUGMENT") == 0) return (T_AUGMENT);		

	printf("Unknown TYPE in schema: %s\n", p);
	die();
}


static int schema_access2int(char *p)
{
	int a = 0;

	if(strchr(p, 'R')) a |= A_READ;
	if(strchr(p, 'W')) a |= A_WRITE;
	if(strchr(p, 'C')) a |= A_CREATE;
	if(strchr(p, 'N')) a |= A_NOTIFY;

	return(a);
}

/*
 * Add all indices in oidlist ("idx1 idx2 idx3 ...")
 * to schema entry 'sch'
 */
static void addindices(struct schent_t *sch, char *oidlist)
{
	char *part[MAX_NO_INDICES];
	int parts;
	int i;
	struct indexoid_t *ind;
	struct indexoid_t *pind = NULL;

	dbg(DBG_SCHEMA, "addindices to %s", sch->oid);

	parts = splitws(oidlist, part, MAX_NO_INDICES);

	/* we expect at least one index */
	assert(parts > 0);
	
	assert(parts <= MAX_NO_INDICES);
	
	for(i = 0; i < parts; i++) {
		dbg(DBG_SCHEMA, "  + %s", part[i]);
		
		ind = malloc(sizeof(struct indexoid_t));
		assert(ind);
		
		ind->oid = strdup(part[i]);
		assert(ind->oid);
		
		ind->next = NULL;
		if(pind) {
			pind->next = ind;
		} else {
			sch->indexlist = ind;
		}
		pind = ind;
	}
}
		
static void schema_free(struct schent_t* sch) {
	if(sch->oid) {
		free(sch->oid);
	}
	if(sch->defval) {
		free(sch->defval);
	}
	if(sch->range) {
		free(sch->range);
	}
	
	/* some things should still be empty, because we can only free
	 * schema entries during schema read */
	assert(sch->rowlist == NULL);
	assert(sch->handlerlist == NULL);
	assert(sch->indexlist == NULL);
	
	free(sch);
}	

int schema_read(char *fname)
{
	char buf[1024];
	FILE *f;
	char *p;
	char *type;
	int linenum = 0;
	struct schent_t *sch = NULL;
	struct schent_t *prevsch = NULL;
	int initial = 1;

	f = fopen(fname, "r");

	if(f == NULL) {
		fprintf(stderr, "Can't open %s : %s\n", fname, strerror(errno));
		return -1;
	}

	while(fgets(buf, sizeof(buf), f)) {

		linenum++;
		
		/* ignore comment lines */
		if(*buf == '#') {
			continue;
	        }

		/* Each schema entry is ended with an empty line */
		if((*buf == '\n') || initial) {
			if(sch) {
				/* check if we just read an AUGMENT */
				if(sch->type == T_AUGMENT) {
					prevsch = sch;
					sch = schema_get(prevsch->oid);
					assert(sch);

					/* Add the augments to the last column */
					while(sch->next) {
						sch = sch->next;
					}
					
					dbg(DBG_SCHEMA, "\tAUGMENT after column %s", sch->oid);
					
					/* cleanup augment schent */
					schema_free(prevsch);				
				}
				else {
					gAVLInsert(schematree, (void *)sch);
				}
		 	}
	
			prevsch = sch;
	
			sch = malloc(sizeof(struct schent_t));
			assert(sch);
			memset(sch, 0, sizeof(struct schent_t));
	
			initial = 0;

			if(*buf == '\n') {
				continue;
			}
		}

		// chomp
		p = strchr(buf, '\n');

		/* If there is no '\n' in our buf, the bufsize is too small to
		 * read a complete line for this schema. Increase the size of 
		 * the buf array to solve this.
		 */
		assert(p);
		
		*p = 0;
		p = buf;

		/* Columns in a table are indented. Add the column to the
		 * previous schema entry
		 */
		if(isblank(*p)) {
			assert(prevsch);
				
			if(prevsch) {
				prevsch->next = sch;
			}
	
			sch->prev = prevsch;
	
			while((*p == ' ') || (*p == '\t'))
				p++;
		}
	
		type = p++;
		while(isblank(*p))
			p++;
		
		switch(*type) {
			case 'O':		// oid
				sch->oid = strdup(p);
				dbg(DBG_SCHEMA, "Read %s", sch->oid);
				break;
			case 'T':		// type
 				sch->type = schema_type2int(p);
				break;
			case 'A':		// access
				sch->access = schema_access2int(p);
				break;
			case 'L': /* we don't use that */
				break;
			case 'R':		// range
				sch->range = strdup(p);
				break;
			case 'I':		// index
				if (sch->type != T_AUGMENT) {
					addindices(sch, p);
				}
				else {
					sch->oid = strdup(p);
					dbg(DBG_SCHEMA, "\tRead AUGMENT for table %s", sch->oid);
				}
				break;
			case 'D': /* Default */
				if (sch->type == T_IPADDR) {
					sch->defval = hex2ipstr(p);
				}
				else {
					sch->defval = strdup(p);
				}
				break;
			default:
				printf("Syntax error in schema line %d : %s\n", linenum, buf);
				break;
		}				

	}

	fclose(f);
	
	return(0);
}

struct schent_t *schema_find_schent_for_oid(char *oid)
{
	char *p;
	struct schent_t *sch;
	char *oidcopy;

	oidcopy = strdup(oid);

	sch = NULL;

	p = oidcopy + strlen(oidcopy);
	while((sch == NULL) && (p > oidcopy)) {
		p = oidcopy + strlen(oidcopy);
		while((p > oidcopy) && (*p != '.')) 
			p--;
		if(*p == '.') 
			*p=0;
		sch = schema_get(oidcopy);
	}

	free(oidcopy);

	return(sch);
}

/*
 * Generate a ASCII dump of the schema
 */
 
void schema_dump(void)
{
	struct schent_t *sch;
	gAVLCursor avlcursor;
	struct handler_t *hnd;
	struct handlerlist_t *hlist;
	int count=0;

	sch = (struct schent_t *) gAVLFirst(&avlcursor, schematree);

	while(sch) {
	
		printf("%s %d\n", sch->oid, sch->type);
	
		hlist = sch->handlerlist;
	
		while(hlist) {
			hnd = hlist->hnd;
			printf("  handler %d\n", hlist->hnd->type);
			hlist = hlist->next;
		}

		sch = (struct schent_t *) gAVLNext(&avlcursor);
		count++;
	}

	printf("\nCount : %d\n", count);
}

int schema_are_related(struct schent_t* sch1, struct schent_t* sch2) {

	struct schent_t* s1;
	struct schent_t* s2;
	
	s1 = sch1;
	while(s1 && s1->prev) {
		s1 = s1->prev;
	}
	
	s2 = sch2;
	while(s2 && s2->prev) {
		s2 = s2->prev;
	}	

	return(s1 && (s1 == s2));
}				

		
static char* hex2ipstr(char* str) {
	
	char ipbuf[16]; 
	int p1, p2, p3, p4;
	char* ret_str;
	
	assert(strlen(str) == 12);
	assert(strncmp(str, "HEX:", 4) == 0);
	
	if (sscanf(str, "HEX:%02X%02X%02X%02X", &p1, &p2, &p3, &p4) != 4) {
		printf("%s: Error parsing IP-address %s\n", __FUNCTION__ , str);
		return NULL;
	}
	
	snprintf(ipbuf, sizeof(ipbuf), "%d.%d.%d.%d", p1, p2, p3, p4);
	
	ret_str = strdup(ipbuf);
	assert(ret_str);
	
	return (ret_str);		
}
	
