/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */


#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
                     
#include "misc.h"
#include "object.h"
#include "types.h"
#include "objdata.h"

static int char2hex(char *in, int inlen, char *out, int maxoutlen);
static int hex2char(char *in, char *out, int maxoutlen);



/*
 * Serializes the data of the given object into a human-readable string,
 * depending on the type of the data. The length of the serialized string
 * is returned.
 */
int serialize(struct object_t *obj, char *buf, int maxlen)
{
	struct schent_t *sch;
	string *str;
	char* byte_arr;
	int len;
	int i;
	unsigned int ui;
	unsigned long long ull;
	struct in_addr *addr;

	sch = obj->sch;
	assert(sch);

	if(!(obj->flags & OF_DATAPRESENT)) {
		len = snprintf(buf, maxlen, "NULL");
		return(len);
	}

	switch(sch->type) {
		case T_INTEGER:
			i = object_get_int(obj);
			len = snprintf(buf, maxlen, "%d", i);
			break;

		case T_UNSIGNED:
			ui = object_get_uint(obj);
			len = snprintf(buf, maxlen, "%u", ui);
			break;
	
		case T_ROWSTAT:
			i = object_get_rowstat(obj);
			len = snprintf(buf, maxlen, "%d", i);
			break;
	
		case T_OCTETSTR:
			str = object_get_string(obj);
			len = char2hex(str->data, str->len, buf, maxlen);
			break;
			
		case T_BITS:
			byte_arr = object_get_bits(obj);
			len = char2hex(byte_arr, (object_get_max_bit(obj) / 8) + 1, buf, maxlen);
			break;
			
		case T_IPADDR:
	
			addr = object_get_ipaddr(obj);
	
			len = snprintf(buf, maxlen, "%s", inet_ntoa(*addr));
			break;
	
		case T_COUNTER:
			
			ui = object_get_counter(obj);
			len = snprintf(buf, maxlen, "%u", ui);
			break;
			
		case T_COUNTER64:
			
			ull = object_get_counter64(obj);
			len = snprintf(buf, maxlen, "%llu", ull);
			break;
			
		case T_GAUGE:
			
			ui = object_get_gauge(obj);
			len = snprintf(buf, maxlen, "%u", ui);
			break;
			
		case T_OBJECTID:
			
			str = object_get_objectid(obj);
			len = char2hex(str->data, str->len, buf, maxlen);
			break;
			
		case T_OPAQUE:
			
			str = object_get_opaque(obj);
			len = char2hex(str->data, str->len, buf, maxlen);
			break;
			
		case T_TICKS:
			
			ui = object_get_ticks(obj);
			len = snprintf(buf, maxlen, "%u", ui);
			break;
			
		case T_NOTIFY:
			
			str = object_get_notify(obj);
			len = char2hex(str->data, str->len, buf, maxlen);
			break;
			
		default:
	
			printf("Unhandled type %d\n", sch->type);
			die();
	
	}

	return(len);
}


/*
 *
 * de-serialize the human-readable string into the object's data.
 *
 */

int deserialize(struct object_t *obj, char *buf)
{
	struct schent_t *sch;
	char *p1;
	int len;
	int max_bit;
	int byte;
	int bit_in_byte; 
	int bit;
	int value;
	unsigned char mask;
	struct in_addr addr;
	int ret_val;
	
	sch = obj->sch;
	assert(sch);

	switch(sch->type) {

		case T_INTEGER:
			return(object_set_int(obj, atoi(buf)));
			break;

		case T_UNSIGNED:
			return(object_set_uint(obj, atoi(buf)));
			break;
	
		case T_ROWSTAT:
			return(object_set_rowstat(obj, atoi(buf)));
			break;
	
		case T_OCTETSTR:

			p1 = strdup(buf);
			assert(p1);
			len = hex2char(p1, p1, strlen(buf));
			
			ret_val = object_set_string(obj, p1, len);
			free(p1);
			return(ret_val);
			break;
			
		case T_BITS:
			/* TODO check this. It will probably always return OK */
			p1 = strdup(buf);
			assert(p1);
			len = hex2char(p1, p1, strlen(buf));
			max_bit = object_get_max_bit(obj);
			ret_val = R_NOERR;
			for(bit=0; (bit <= max_bit) && (ret_val == R_NOERR); bit++) {
				byte = bit / 8;
				bit_in_byte = bit % 8;
				mask = 1 << (7-bit_in_byte);
				value = ((p1[byte] & mask)>0)?1:0;
				ret_val = object_set_bit(obj, bit, value);
			}
			free(p1);
			
			return(ret_val);			
			break;
			
		case T_IPADDR:
			if (!inet_aton(buf, &addr)) {
				printf("%s: Error parsing IP address (%s)", __FUNCTION__, buf);
			}
			return(object_set_ipaddr(obj, addr));
			break;
			
		case  T_COUNTER:
			
			return(object_set_counter(obj, atoi(buf)));
			break;
			
		case T_COUNTER64:
			
			return(object_set_counter64(obj, atoll(buf)));
			break;
		
		case T_GAUGE:
			
			return(object_set_gauge(obj, atoi(buf)));
			break;
			
		case T_OBJECTID:
			return(object_set_objectid(obj, buf));
			break;
			
		case T_OPAQUE:
			p1 = strdup(buf);
			len = hex2char(p1, p1, strlen(buf));
			ret_val = object_set_opaque(obj, p1, len);
			free(p1);
			return(ret_val);
			break;
			
		case T_TICKS:
			return(object_set_ticks(obj, atoi(buf)));
			break;
			
		case T_NOTIFY:
			p1 = strdup(buf);
			len = hex2char(p1, p1, strlen(buf));
			ret_val = object_set_notify(obj, p1, len);
			free(p1);
			return(ret_val);
			break;
	
		default:
			
			printf("Unhandled type %d\n", sch->type);
			die();
	}

	return(R_WRONGTYPE);
}




/* Encode a array of characters to "HEX:...." format */
static int char2hex(char *in, int inlen,  char *out, int maxoutlen)
{
	char *list = "0123456789ABCDEF";
	char *p;
	unsigned char *q;
	int i;
	int l;

	i = 0;
	q = in;

	/* Only encode non-readable strings to HEX */
	for(l=0; l<inlen; l++) {
		if((*q<32) || (*q>127)) i=1;
		q++;
	}

        if((i==0) && (inlen > 0)) {
		memcpy(out, in, (inlen>maxoutlen) ? maxoutlen : inlen);
		if(inlen<=maxoutlen) out[inlen]=0;
		return(strlen(out));
	}

	l=0;
	p=out;

	p += sprintf(out, "HEX:");

	for(i=0; i<inlen; i++) {
		if((p-out) > (maxoutlen-1)) break;
		p += sprintf(p, "%c%c", list[(in[i]>>4)&0xf], list[in[i]&0xf]);
	}

	return(p-out);	
}



// decode "HEX:...." format to a array of characters

int hex2char(char *in, char *out, int maxoutlen)
{
	char *p, *q;
	int a;
	int j;

	/* Check if the input is HEX encoded. If not, just return a copy */
	if(strncmp(in, "HEX:", 4) != 0) {
		strncpy(out, in, maxoutlen);
		return(strlen(out));
	}

	p = out;
	q = in+4;

	while(*q && *q+1 && (p-out < maxoutlen-1)) {
		*p=0;
		for(j=0; j<2; j++) {
			a=0;
			if((*q >= '0') && (*q<='9'))
				 a = (*q) - '0';
			else if((*q >= 'A') && (*q<='F')) 
				a = *q - 'A' + 10;
			else if((*q >= 'a') && (*q<='f')) 
				a = *q - 'a' + 10;
			else {
				printf("Invalid HEX char %c\n", *q);
				die();
			}
			
			*p = (*p<<4) + a;
			q++;
		}
		p++;
	}
	
	*p=0;

	return(p-out);
}

// EOT
