/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */
#ifndef _DATASTORE_TYPES_H_
#define _DATASTORE_TYPES_H_
/** @file types.h
 *
 * @brief Definition of the basic types and defines used in the datastore
 */


#include <time.h>

#define T_UNKNOWN	0
#define	T_INTEGER 	1
#define T_UNSIGNED	2
//#define	T_STRING  	3	// not used any more
#define	T_IPADDR  	4
#define	T_TABLE   	5
#define T_ROWSTAT	6
#define T_OCTETSTR	7
#define T_COUNTER       8
#define T_COUNTER64     9
#define T_GAUGE        10
#define T_OBJECTID     11
#define T_OPAQUE       12
#define T_TICKS        13
#define T_NOTIFY       14  /* TBD: what to do with this ? */
#define T_BITS         15
#define T_AUGMENT      99  /* ONLY USED DURING SCHEMA_READ */

#define H_GET     	1
#define H_SET     	2
#define H_CHK   	3

#define A_NOACCESS      0x00
#define A_READ 		0x01
#define A_WRITE		0x02
#define A_CREATE	0x04
#define A_NOTIFY	0x08

#define R_NOERR		0
#define R_NOSUCHOID	1
#define R_NOACCESS	2
#define R_OUTOFRANGE	3
#define R_NONEXIST	4
#define R_WRONGTYPE     5
#define R_CHECKFAILED	6
#define R_ENDOFTREE     7
#define R_MEMORYFAIL    8

#define RS_ACTIVE	1
#define RS_NOTINSERVICE	2
#define RS_NOTREADY	3
#define RS_CREATEGO	4
#define RS_CREATEWAIT	5
#define RS_DESTROY	6

#define OF_DIRTY	(1 << 0)
#define OF_CFGWRITTEN	(1 << 1)
#define OF_FLOATING     (1 << 2) /* floating objects are objects not in the AVL tree */
#define OF_DATAPRESENT  (1 << 3) /* determines if the object has data */

#define J_ROWADD	1
#define J_ROWDEL	2
#define J_OBJSET	3

#define B_FALSE         0
#define B_TRUE          (!B_FALSE)

/** The max length an OID string can be */
#define MAX_OID_LENGTH   64

/** The maximum number of indices. (counted in objects, 
 *  so a ip-address in the index counts as 1, not 4 
 */
#define MAX_NO_INDICES   10

/** The maximum number of digits in the index part of an OID */
#define MAX_INDEX_LENGTH 16

/** The maximum number of columns in a row */
#define MAX_NO_COLUMNS 64

struct table_t;
struct handler_t;
struct schent_t;
struct object_t;
struct row_t;

/** @struct index_t 
 * @brief Holds information about an index from a table. 
 *
 * This object (should) only exist during the creation procedure of a
 * row in a table. It can also be derived from an existing row to 
 * access the (external) index objects.
 *
 * @warning Don't forget to clean up this object with index_done().
 *
 * @see table.h
 */
struct index_t {
	struct schent_t *sch;		/**< Pointer the the corresponding (Table) schent_t */
	int num;			/**< The number of objects in the index */
	struct object_t **obj;		/**< Pointer to an array (of size num) of the objects
					     which define the index */
};


/** @struct handler_t 
 *  @brief Holds information about a handler function
 */
struct handler_t {
	int type;			/**< type is H_GET, H_SET or H_CHK */
	int (*func)(void);		/**< pointer to handler function */
	time_t expiretime;		/**< defines when handler expires */

};


/** @struct handlerlist_t
  *  @brief List of handler_t structs 
  */
struct handlerlist_t {
	struct handler_t *hnd;		/**< Pointer to a handler_t */
	struct handlerlist_t *next;	/**< Linked list next-pointer */
};


/** @struct indexoid_t 
 *  @brief List of index-oid's for a table schent_t 
 */
struct indexoid_t {
	char *oid;			/**< Pointer to OID */
	struct indexoid_t *next;	/**< Linked list next-pointer */
};


/** @struct schent_t
  * @brief Schema entry. Holds all properties of underlying objects 
  *
  * A schema entry (@c schent_t) holds the information from a single OID
  * as defined in a MIB. It contains information about default values,
  * type of access, OID, type, range, etc. \n
  * It also contains a flag @c persistant to indicate if a object from
  * this type should be stored in the persistant datafile.
  * The @c handlerlist attribute holds a linked list of handlercalls 
  * which should be called when certain actions on this OID are
  * done. i.e. SET, GET. \n
  * If the @c type of this schent_t is T_TABLE, then this @c schent_t is
  * the parent schent_t of a table. The @c next attribute then will point 
  * to the first column (also a @c schent_t) of this table.\n
  * If this @c schent_t is a column then the @c next attribute will point
  * to an other column schema entry or to NULL if it is the last column. 
  * To determine if a schent_t is a column, just follow the @c prev attribute.
  * The last @c schent_t should be a @c T_TABLE schema entry.
  */
struct schent_t {
	
	char *oid;				/**< oid, dottet numeric */

	char *defval;				/**< Pointer to a default value, or NULL if not defined */
	char *range;				/**< String describing the allowed ranges for child objects */

	struct row_t *rowlist;			/**< Pointer to a linked list of table_t rows, or NULL if no table attached */
	int rowcount;				/**< counts the number of rows. This is redundant information. */

	struct handlerlist_t *handlerlist;	/**< Pointer to a linked list of handler_t, or NULL if no handlers attached */

	struct indexoid_t *indexlist;		/**< @todo Document this */
	
	struct schent_t *prev;			/**< Linked list previous-pointer (only used for tables) */
	struct schent_t *next;			/**< Linked list next-pointer (only used for tables) */

	unsigned char persistant;		/**< If set, this record is written by configfile_write */
	unsigned char access;			/**< Bitmap of access lists A_... @see types.h */
	unsigned char type;			/**< Type. Is one of T_.... @see types.h*/
	
	
};


/** @struct object_t 
  * @brief Object that holds the actual 'live' data.
  *
  * This struct holds the actual data of an OID. It can be created by the 
  * functions as defined in object.h. The data in the object can be 
  * manipulated by the set_object_XXX and get_object_XXX functions as 
  * defined in objdata.h. \n
  * The @c data attribute points to the actual data. It should never 
  * be accessed directly (other then checking if it is there)\n
  * The @c flags attribute should be manipulated/retreived with the
  * functions as defined in objflag.h
  * If this object is part of a row in a table, the @c row attribute
  * points to a struct row_t. Through this structure the other columns
  * of a row can be accessed. \n
  * The @c sch attribute points to the correspondig schema entry.
  * There is no OID attribute. The exacte OID of the object can be
  * derived from the sch->oid and the oidpart in the row. This this
  * object doesn't have a row pointer, the OID is the sch->oid + ".0"
  */		  		  
struct object_t {
	unsigned long value;			/**< The value of the object. Only valid if the OF_DATAPRESENT is set. The
	                                             meaning of this value depends on the type of the object. */
	struct row_t *row;			/**< Pointer to the object's table row, or NULL if it's not in a table */
	struct schent_t *sch;			/**< Pointer to the object's schema_entry. should NEVER be NULL */
	char flags;				/**< A bitmask of OF_... values */	
};


/** @struct row_t types.h
 *  @brief Holds information about a single row in a table. 
 * 
 * The row_t struct holds information about a single row in a table. It is
 * connected to the corresponding Table schema entry. This object is generated
 * with the row_create_by_XXXX functions defined in table.h
 */
struct row_t {
	struct schent_t *sch;		/**< Pointer to the corresponding table entry. */
	char *oidpart;			/**< The index-part of the oid. */
	struct row_t *prev;		/**< A pointer to the previous row. NULL is this is the first row. */
	struct row_t *next;		/**< A pointer to the next row. NULL if this is the last row. */
	int columns;			/**< The number of columns. (== size of array obj ). */	
	struct object_t objs[1];        /**< Array of objects in this row */
};



/** @struct moddescr_t
  * @brief Module description
  */
struct moddescr_t {
	char *name;				/**< The name of the module */
	int(*fn_ini)(void);			/**< Pointer to initialization function, or NULL if none */
	int(*fn_get)(void);			/**< Pointer to GET function, or NULL if none */
	int(*fn_set)(void);			/**< Pointer to SET function, or NULL if none */
	int(*fn_chk)(void);			/**< Pointer to CHK function, or NULL if none */
	int(*fn_pin)(void);			/**< Pointer to Post INit function , or NULL if none */	
	char *oidlist;				/**< List of OID's that should be attached to this handler. Seperated by spaces */
};


/** @struct jent_t 
  * @brief Journal entry 
  */
struct jent_t {
	int action;				/**< The journalized action (J_ROWDEL, J_ROWADD or J_OBJSET) */
	struct row_t *row;                      /**< The journalized row (only valid if action is J_ROWDEL or J_ROWADD) */
	struct object_t *obj;                   /**< Pointer to the journalized object (only valid if action is J_OBJSET */
	unsigned long value;	 		/**< Old value of the journalized object (only valid if action is J_OBJSET) */ 
	unsigned long flags;			/**< Old flags of the journalized object */
	struct jent_t *next;			/**< Pointer to the next journal entry */
};

/** @struct string_t
  * @brief Struct that holds a (non NULL-terminated) string 
  *
  * This struct is used internally in the datastore to hold the SNMP octet strings.
  * Structs of this type should be generated by the mkstring function. 
  */
struct string_t {
	int len;
	char data[1];
};

typedef struct string_t string;

#endif
