/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "regulatory.h"
#include "types.h"
#include "pda.h"
#include "aplog.h"
#include "misc.h"

#define PARSE_ERROR(msg) fprintf(stderr, "Parse error in line %i: %s\n", linenum, msg)

struct pda_freq_t {
	int freq;
	struct pda_freq_t *next;
};

static struct country_domain_t *country_domains = NULL;
static struct country_domain_t *default_country_domain = NULL;
static struct country_domain_t *current_country_domain3 = NULL;
static struct country_domain_t *current_country_domain5 = NULL;

static struct pda_freq_t *pda_freqs = NULL; /* linked list of HW supported freqs */

static int linenum = 0;

static struct country_domain_t* get_country_domain(char* id, int type);
static struct country_domain_t* new_country_domain(char* id, int type);
static struct country_domain_t* new_country(char* id);
static struct country_domain_t* new_domain(char* id);
static char* read_line(FILE* fp, char**p);
static char parse_cs_token(char** p, char** token);
static void parse_freq(char*p, int* no_freqs, unsigned int *frequencies);
static void parse_countries(char*p, int* no_countries, char **countries);
static void parse_countries_in_domain(FILE*fp, char *p);
static struct power_group_t* parse_max_output_power(FILE* fp, char* p);
static void parse_country_domains(FILE* fp, char* p, int type);
static void parse_default(char* p);
static int freq_in_pda(int freq);

/* exported functions */
struct country_domain_t* regulatory_get_country(char* id) {
	return(get_country_domain(id, REGULATORY_TYPE_COUNTRY));
} 

struct country_domain_t* regulatory_get_domain(char* id) {
	return(get_country_domain(id, REGULATORY_TYPE_DOMAIN));
}

struct country_domain_t* regulatory_get_country_domains(void) {
	return (country_domains);
}

struct country_domain_t* regulatory_get_default_country_domain(void) {
	return (default_country_domain);
}

struct country_domain_t* regulatory_get_current_country_domain(int ifnum) {

	switch(ifnum) {
	
	case 3:
		return(current_country_domain3);
		break;
	case 5:
		return(current_country_domain5);
		break;
	default:
		die();
		return(NULL);
	}
}

void regulatory_set_current_country_domain(int ifnum, struct country_domain_t* cd) {

	switch(ifnum) {
	
	case 3:
		current_country_domain3 = cd;
		break;
	case 5:
		current_country_domain5 = cd;
		break;
	default:
		die();
	}
}

/* this functions tries to find the given country code in one of
   the supported regdomains or countries and returns the corresponding 
   country_domain struct */
struct country_domain_t* regulatory_get_country_domain_by_country_code(char* id) {
		
	struct country_domain_t* cd;
	int i;
	
	cd = country_domains;
	
	while(cd) {
		for(i = 0; i < cd->no_countries; i++) {
			if(strcmp(id, cd->countries[i]) == 0) {
				return cd;
			} 
		}							
		
		cd = cd->next;
	}
	
	return NULL;
	
}

int regulatory_domainid2value(char* id){
	
	if(strcmp(id, "FCC") == 0) {
		return (0x10);
	} 
	else if(strcmp(id, "IC") == 0) {
		return (0x20);
	}	
	else if(strcmp(id, "ETSI") == 0) {
		return (0x30);
	}	
	else if(strcmp(id, "Spain") == 0) {
		return (0x31);
	}	
	else if(strcmp(id, "France") == 0) {
		return (0x32);
	}	
	else if(strcmp(id, "MPHPT") == 0) {
		return (0x40);
	}	
	else {
		return (0xFF);
	}
}

char* regulatory_value2domainid(int value) {
	
	switch(value) {
		case 0x10:
			return("FCC");
			break;
		case 0x20:
			return("IC");
			break;
		case 0x30:
			return("ETSI");
			break;
		case 0x31:
			return("Spain");
			break;
		case 0x32:
			return("France");
			break;
		case 0x40:
			return("MPHPT");
			break;
		default:
		 	return("");
			break;
	}	
}

int regulatory_country_in_domain(struct country_domain_t* cd, char* country) {
	int i;
	
	assert(cd);
	assert(country);
	
	for(i=0; i < cd->no_countries; i++) {
		
		if(strncmp(country, cd->countries[i], 3) == 0) {
			return R_NOERR;
		}
	}
	
	return R_CHECKFAILED;
}


/* static functions */




static struct country_domain_t* get_country_domain(char* id, int type) {
	
	struct country_domain_t* cd;
	
	cd = country_domains;
	
	while(cd) {
		if (cd->type == type && (strcmp(cd->id, id) == 0)) {
			return(cd);
		}		
		cd = cd->next;
	}
	return NULL;
}

static struct country_domain_t* new_country_domain(char* id, int type) {
	
	struct country_domain_t* cd;
	
	cd = calloc(sizeof(struct country_domain_t), 1);
	
	cd->id = strdup(id);
	cd->type = type;
	cd->next = country_domains;
	country_domains = cd;
	
	return(cd);
}

static struct country_domain_t* new_country(char* id) {
	struct country_domain_t* cd;

	cd = new_country_domain(id, REGULATORY_TYPE_COUNTRY);
	
	cd->countries = malloc(sizeof(char*));
	
	cd->countries[0] = strdup(id);
	cd->no_countries = 1;
	
	return(cd);
}

static struct country_domain_t* new_domain(char* id) {
	
	return(new_country_domain(id, REGULATORY_TYPE_DOMAIN));
}

static char* read_line(FILE* fp, char**p) {
	static char buf[256];
	
	*p = buf;
	
	linenum++;
	return(fgets(buf, sizeof(buf), fp));	
}


/* parse a Comma Separated token. 
 * 
 * - skips leading blanks
 * - replaces the character after the token with a \0
 * - returns the character that was found after the token (ie. ';', '}' or ' ')
 * - *token will point to the token
 * - places *p after the \0 places after the token
 */
static char parse_cs_token(char** p, char** token) {
	char *p1, *p2;
	char sep;
	
	p1 = *p;
	
	while(isblank(*p1)) {
		p1++;
	}
	
	p2 = p1;
	while((*p1 != ';') 
			&& (*p1 != '{') 
			&& (*p1 != ',') 
			&& (*p1 != '\n')
			&& (*p1 != ' ')
			&& (*p1 != ';')
			) {
		p1++;
	}
	
	sep = *p1;
	*p1 = '\0';
	
	*token = p2;
	*p = p1+1;
	
	return(sep);
			
}

static void parse_freq(char*p, int* no_freqs, unsigned int *frequencies) {
	
	char sep;
	char* id;
	char* p1;
	int freq;
	
	p1 = p;
	
	do {
		sep = parse_cs_token(&p1, &id);
		
		freq = atoi(id);
		
		/* If the HW supports this freq, add it */
		if(freq_in_pda(freq)){
			frequencies[*no_freqs] = atoi(id);
			(*no_freqs)++;
		}
		
	} while (sep == ',');
	
}

static void parse_countries(char*p, int* no_countries, char **countries) {
	
	char sep;
	char* id;
	char* p1;
	char strbuf[4];
		
	p1 = p;
	
	do {
		sep = parse_cs_token(&p1, &id);
		snprintf(strbuf, sizeof(strbuf), "%s  ", id);
		countries[*no_countries] = strdup(strbuf);
		(*no_countries)++;
		
	} while (sep == ',');
	
}



static void parse_countries_in_domain(FILE*fp, char *p) {
	char *p2, *p1;
	struct country_domain_t* cd;
	char* countries[REGULATORY_MAX_COUNTRIES_IN_DOMAIN];
	char strbuf[4];
	
	p2 = strchr(p, ' ');
	
	if(!p2) {
		PARSE_ERROR("No domain found after countries-in-domain token");
		return;
	}
	
	*p2 = '\0';
	
	cd = regulatory_get_domain(p);	
	
	if(cd->countries) {
		PARSE_ERROR("Error. Multiple countries-in-domain definition for domain");
		return;
	}
	
	/* ignore rest of line, assume it is "{" */
	while(read_line(fp, &p1)) {
		/* skip empty and comment lines */
		if(*p1 == '#' || *p1 == '\n') {
			continue;
		}		
		
		/* skip first blanks */
		while(isblank(*p1)) {
			p1++;
		}
		
		if(*p1 == '}') {
			/* create countries array */
			cd->countries = malloc(sizeof(char*) * cd->no_countries);
			memcpy(cd->countries, countries, sizeof(char*) * cd->no_countries);
			return;
		}
		
		p2 = strchr(p1, ' ');
		
		if(!p2) {
			PARSE_ERROR("No countries or domain token found");
			continue;
		}
		
		*p2 = '\0';
		
		if(strcmp(p1, "countries") == 0) {
			parse_countries(p2+1, &(cd->no_countries), countries);
		}
		else if(strcmp(p1, "default") == 0) {
			if (cd->default_country) {

				PARSE_ERROR("Error multiple definition for default country");
				free(cd->default_country);
			}			

			p1 = p2 + 1;
			p2 = strchr(p1, ';');
			if(!p2) {
				PARSE_ERROR("Parse error in default clause");
				return;
			}
			*p2 = '\0';
			
			/* pad with spaces */
			snprintf(strbuf, sizeof(strbuf), "%s  ", p1);
			cd->default_country = strdup(strbuf);
		}
		else {
			PARSE_ERROR("Parse error in countries-in-domain group");
		}
		
	}
	
}
	
static struct power_group_t* parse_max_output_power(FILE* fp, char* p) {
	
	char* p1, *p2;
	unsigned int frequencies[REGULATORY_MAX_FREQS_IN_GROUP];
	struct power_group_t* pg;
	
	/* create new power_group */
	pg = calloc(sizeof(struct power_group_t), 1);
	pg->frequencies = NULL;
	pg->no_freqs = 0;
	pg->next = NULL;
	
	/* p starts with the maxpower */
	pg->max_power = atoi(p);
			
	p1 = p;
	
	/* ignore rest of line, assume it is "{" */
	while(read_line(fp, &p1)) {
		/* skip empty and comment lines */
		if(*p1 == '#' || *p1 == '\n') {
			continue;
		}		
		
		/* skip first blanks */
		while(isblank(*p1)) {
			p1++;
		}
		
		if(*p1 == '}') {
			/* create frequencies array */
			if(pg->no_freqs == 0) {
				/* no frequencies in this power group, free it */
				free(pg);
				return (NULL);
			}
			else {
				pg->frequencies = malloc(sizeof(unsigned int) * pg->no_freqs);
				memcpy(pg->frequencies, frequencies, sizeof(unsigned int) * pg->no_freqs);
				return (pg);
			}
		}
		
		p2 = strchr(p1, ' ');
		
		if(!p2) {
			PARSE_ERROR("No token found");
			continue;
		}
		
		*p2 = '\0';
		
		if(strcmp(p1, "freq") == 0) {
			parse_freq(p2+1, &(pg->no_freqs), frequencies);
		}
		else {
			PARSE_ERROR("Parse error in power group");
		}
	}
	
	PARSE_ERROR("missing '}'");
		
	return(NULL);
	
}

static void parse_country_domains(FILE* fp, char* p, int type) {
	
	char* id;
	char* p1, *p2;
	char sep;
	struct regulatory_group_t* regulgrp;
	struct country_domain_t* cd;
	struct regulatory_group_list_t* rgl;
	struct power_group_t* pgr;
	char country_buf[4];
	
	/* create new regulatory_group */
	regulgrp = calloc(sizeof(struct regulatory_group_t), 1);
	regulgrp->eirp = 0;
	regulgrp->default_freq2 = 0;
	regulgrp->default_freq5 = 0;
	regulgrp->power_groups = NULL;
	
	p1 = p;
	
	
	do {
		do {
			sep = parse_cs_token(&p1, &id);

			cd = NULL;
			if(type == REGULATORY_TYPE_DOMAIN) {
				cd = regulatory_get_domain(id);
				if(!cd) {
					cd = new_domain(id);
				}
				
								
			}
			else {
				/* make sure the 'id' is 3 chars */
				strncpy(country_buf, id, sizeof(country_buf));
				
				if(country_buf[2] == 0x00) {
					country_buf[2] = ' ';
					country_buf[3] = 0x00;
				}
				
				cd = regulatory_get_country(country_buf);
				if(!cd) {
					cd = new_country(country_buf);
				}
			}
			
			if(cd) {
				/* add the current freqgrp to this domain/country */
				rgl = calloc(sizeof(struct regulatory_group_list_t), 1);
				rgl->regul_grp = regulgrp;
				rgl->next = cd->regul_groups;
				cd->regul_groups = rgl;
			}
			
		} while (sep == ',');
		
		/* read next token */
		sep = parse_cs_token(&p1, &id);
		
		if (sep == ' ') {
			if(strcmp(id, "countries") == 0) {
				type = REGULATORY_TYPE_COUNTRY;
			}
			else if(strcmp(id, "domains") == 0) {
				type = REGULATORY_TYPE_DOMAIN;
			}
			else {
				PARSE_ERROR("Unknown token");
			}
		}
				
	} while (sep == ' ');
	
	/* read lines in this group */
	while(read_line(fp, &p1)) {
		/* skip first blanks */
		while(isblank(*p1)) {
			p1++;
		}
		
		/* skip empty and comment lines */
		if(*p1 == '#' || *p1 == '\n') {
			continue;
		}
		
		if(*p1 == '}') {
			break;
		}
		
		p2 = strchr(p1, ' ');
		
		if(!p2) {
			PARSE_ERROR("Nothing behind token");
			continue;
		}
		
		*p2 = '\0';
		
		if(strcmp(p1, "max-output-power") == 0) {
			pgr = parse_max_output_power(fp, p2+1);
			
			if(pgr) {
				pgr->next = regulgrp->power_groups;
				regulgrp->power_groups = pgr;
			}
		}
		else if(strcmp(p1, "EIRP") == 0) {
			if(*(p2+1) == 'y'){
				regulgrp->eirp = 1;
			}
			else {
				regulgrp->eirp = 0;
			}
		}
		else if(strcmp(p1, "default-freq-2") == 0) {
			regulgrp->default_freq2 = atoi(p2+1);
		}
		else if(strcmp(p1, "default-freq-5") == 0) {
			regulgrp->default_freq5 = atoi(p2+1);
		}
		else {
			PARSE_ERROR("Unknown token");
		}
	}
		
}

static void parse_default(char* p) {
	
	char *p1, *p2, *p3;
	
	if(default_country_domain) {
		PARSE_ERROR("Parse error. Multiple default (country/domain) definitions");
		return;
	}
	
	p1 = p;
	p2 = strchr(p1, ' ');
	if(!p2) {
		PARSE_ERROR("Parse error in default clause");
		return;
	}
	
	*p2 = '\0';
	
	p3 = strchr(p2+1, ';');
	if(!p3) {
		PARSE_ERROR("Parse error in default clause. Missing ';'");
		return;
	}
	*p3 = '\0';
	
	if(strcmp(p1, "domain") == 0) {
		default_country_domain = regulatory_get_domain(p2+1);
	}
	else if(strcmp(p1, "country") == 0) {
		default_country_domain = regulatory_get_country(p2+1);
	}
	else {
		PARSE_ERROR("Unknown token");
		return;
	}
	
	return;
}

static int freq_in_pda(int freq) {
	struct pda_freq_t* freq_p;
	
	if(!pda_freqs) {
		/* No frequencies in PDA -> support all frequencies */
		return B_TRUE;
	}
	else {
		freq_p = pda_freqs;
		while(freq_p) {
			if(freq == freq_p->freq) {
				return B_TRUE;
			}
			freq_p = freq_p->next;
		}
		return B_FALSE;
	}
}

static void read_pda(void) {
	
	struct Pdr_Struct *pdrPtr;
	u_char* p;
	int no_channels;
	struct pda_freq_t *freq_p = NULL;
	
	pdrPtr = pda_Find_Pdr(PDR_PRISM_PA_CAL_OUTPUT_POWER_LIMITS, NULL);
	if(!pdrPtr) {
		printf("PDR_PRISM_PA_CAL_OUTPUT_POWER_LIMITS not found in PDA\n");
		return;
	}
	
	p = pdrPtr->PdrData;
	
	p++; /* skip calibration method revision */
	no_channels = *p;
	p++;
	
	while(no_channels > 0) {
		freq_p = (struct pda_freq_t*) calloc(sizeof(struct pda_freq_t), 1);
		freq_p->freq = *p + ((*(p+1)) << 8);
		freq_p->next = pda_freqs;
		pda_freqs = freq_p;
		p = p + 8; /* skip other fields, goto next frequency */
		no_channels--;

	}
}

static void read_config(void){ 
	
	FILE* fp;
	char* p1, *p2;
	linenum = 0;
	
	fp = fopen(REGULATORY_CONF_FILE, "r");
	assert(fp);
	
	while(read_line(fp, &p1)) {
		
		/* skip empty and comment lines */
		if(*p1 == '#' || *p1 == '\n') {
			continue;
		}		
		
		/* skip first blanks */
		while(isblank(*p1)) {
			p1++;
		}
		
		p2 = strchr(p1, ' ');
		
		if(!p2) {
			PARSE_ERROR("Nothing behind token\n");
			continue;
		}
		
		*p2 = '\0';
		
		if (strcmp(p1, "domains") == 0) {
			parse_country_domains(fp, p2+1, REGULATORY_TYPE_DOMAIN);
		}
		else if (strcmp(p1, "countries") == 0) {
			parse_country_domains(fp, p2+1, REGULATORY_TYPE_COUNTRY);
		}
		else if (strcmp(p1, "countries-in-domain") == 0) {
			parse_countries_in_domain(fp, p2+1);
		}
		else if (strcmp(p1, "default") == 0) {
			parse_default(p2+1);
		}		
		else {
			PARSE_ERROR("Unknown token\n");
		}
		
	}
}


static void dump_regul_group(struct regulatory_group_t* regul_grp) {
	
	struct power_group_t* pg;
	int i;
	
	if(regul_grp->eirp) {
		printf("EIRP yes\n");
	}
	else {
		printf("EIRP no\n");
	}
	
	printf("Default frequency in 2.4Ghz band: %i\n", regul_grp->default_freq2);
	printf("Default frequency in  5 Ghz band: %i\n", regul_grp->default_freq5);
	

	
	pg = regul_grp->power_groups;
	
	while(pg) {
		printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
		printf("Max power   : %i\n", pg->max_power);
		printf("Frequencies : ");
		
		for(i = 0; i < pg->no_freqs; i++) {
			printf("%i ", pg->frequencies[i]);
		}
		printf("\n");
		pg = pg->next;
		printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
	}
			


	
}

void regulatory_dump(void) {
	struct country_domain_t* cd;
	struct regulatory_group_list_t* rgl;
	int i;
	
	cd = country_domains;
	
	while(cd) {
		
		printf("====================================\n");
		if(cd->type == REGULATORY_TYPE_COUNTRY) {
			printf("Country: %s\n", cd->id);
		}
		else if(cd->type == REGULATORY_TYPE_DOMAIN) {
			printf("Domain: %s\n", cd->id);
		}
		else {
			printf("Unknown type (%i): %s\n", cd->type, cd->id);
		}
		printf("------------------------------------\n");
		printf("Supported countries: ");
		
		for(i = 0; i < cd->no_countries; i++) {
			printf("%s ", cd->countries[i]);
		}
		
		printf("\n");
		
		printf("------------------------------------\n");
		
		rgl = cd->regul_groups;
		
		while(rgl) {
			dump_regul_group(rgl->regul_grp);
			rgl = rgl->next;
		}		
		
		
		printf("====================================\n");
		cd = cd->next;
	}	

	if(default_country_domain) {
		printf("Default country/domain: %s\n", default_country_domain->id);
	}
	else {
		printf("NO DEFAULT COUNTRY/DOMAIN\n");
	}
	
}

void regulatory_init() {
	
	read_pda();
	read_config();
//	regulatory_dump();
}

int regulatory_freq2channel(unsigned int freq) {
	
	
	if((freq >= 2412) && (freq <= 2472)) {
		return( (freq - 2407) / 5);
	}
	else if(freq == 2484) {
		return(14); /* Japan channel */
	}
	else if((freq >= 5000) && (freq <= 5995)) {
		return( (((int)freq) - 5000) / 5);
	}
	else if((freq >= 4910) && (freq <= 4995)) {
		return( ((freq-4910) / 5) + 238);
	}
	else {
		return(0);
	}
}

unsigned int regulatory_channel2freq(int channel, int band) {
	
	switch(band) {
		case REGULATORY_BAND_5:
			if(channel < 238) {
				return(5000 + (channel *5));	
			}
			else {
				return(4910+((channel-238)*5));
			}
			break;
			
		case REGULATORY_BAND_2DOT4:
			if(channel != 14) {
				return(2407 + (channel * 5));
			}
			else {
				return(2462);
			}
			break;
			
		default:
			return (0);
	}
	
	
}

double mW2dBm(int milliwatts) {
	return(10 * log10(milliwatts));
}

int regulatory_freq2band(int freq) {
	return(freq < 3000 ? REGULATORY_BAND_2DOT4 : REGULATORY_BAND_5);
}
