/*
 * misc-bzip2.c
 * 
 * bzip2 kernel decompression for uCLinux by Michiel Thuys
 * Copyright (C) 2002 Intersil Americas Inc.
 *
 * malloc by Hannu Savolainen 1993 and Matthias Urlichs 1994
 *
 * Modified for ARM Linux by Russell King
 *
 * Nicolas Pitre <nico@visuaide.com>  1999/04/14 :
 *  For this code to run directly from Flash, all constant variables must
 *  be marked with 'const' and all other variables initialized at run-time 
 *  only.  This way all non constant variables will end up in the bss segment,
 *  which should point to addresses in RAM and cleared to 0 on start.
 *  This allows for a much quicker boot time.
 */

#include <asm/uaccess.h>
#include <asm/arch/uncompress.h>
#include <asm/proc/uncompress.h>

#include <bzlib.h>
#include "../../../../include/asm-armnommu/arch-isl3893/gpio.h"

/* #define DEBUG */

/* The start and end of the compressed kernel. These are set in head.S */
char *input_data;
char *input_data_end;

/* The malloc area. */
static unsigned long free_mem_ptr;
static unsigned long free_mem_ptr_end;

static void error(char *m);

void *malloc(int size)
{
	void *p;

#ifdef DEBUG
	puts("Malloc: size 0x");
	puthex(size);
	puts("\n");
#endif

	if (size <0) error("Malloc error\n");
	if (free_mem_ptr <= 0) error("Memory error\n");

	free_mem_ptr = (free_mem_ptr + 3) & ~3;	/* Align */

	if (free_mem_ptr + size >= free_mem_ptr_end)
		return NULL;

	p = (void *)free_mem_ptr;
	free_mem_ptr += size;

	return p;
}

void free(void *where)
{
	/* We do nothing here, because after decompression, we leave anyway */
}

void bz_internal_error(int errcode)
{
	puts("bzip internal error ");
	puthex(errcode);
	error("");
}

static void error(char *x)
{
	puts("\n\n");
	puts(x);
	puts("\n\n -- System halted");

	while(1);	/* Halt */
}

unsigned long
decompress_kernel(unsigned long output_start, unsigned long free_mem_ptr_p,
		  unsigned long free_mem_ptr_end_p, unsigned long max_output_end)
{
	int ret;
	int destLen		= max_output_end - output_start;
	free_mem_ptr		= free_mem_ptr_p;	/* This is our malloc area */
	free_mem_ptr_end	= free_mem_ptr_end_p;

	//********** light on (green) power LED / modify by marcel.liao
		#define PWR_ORANGE	0
		#define PWR_GREEN	4

		uGPIO3->DDR = gpio_set((1<<PWR_GREEN));
		uGPIO3->DR = gpio_set((1<<PWR_GREEN));

	proc_decomp_setup();
	arch_decomp_setup();

	puts("\nUncompressing Linux (bzip2)...");
	/* We first try fast decompressing that uses a lot of memory. If that
	 * fails in a small RAM configuration, we retry with the slower
	 * decompression that uses a bit less memory
	 */
	ret = BZ2_bzBuffToBuffDecompress((unsigned char *)output_start,
					&destLen,
					input_data,
					(unsigned int)(&input_data_end[0] - &input_data[0]),
					0,
					0);
	if(ret == BZ_MEM_ERROR) {
		free_mem_ptr		= free_mem_ptr_p;	/* re-init malloc area */
		free_mem_ptr_end	= free_mem_ptr_end_p;

		ret = BZ2_bzBuffToBuffDecompress((unsigned char *)output_start,
					&destLen,
					input_data,
					(unsigned int)(&input_data_end[0] - &input_data[0]),
					1,
					0);
	}
	if(ret != BZ_OK) {
		//********** light off (green) power LED
			uGPIO3->DDR = gpio_set((1<<PWR_GREEN));
			uGPIO3->DR = gpio_clr((1<<PWR_GREEN));

		//********** light on (orange) power LED
			uGPIO3->DDR = gpio_set((1<<PWR_ORANGE));
			uGPIO3->DR = gpio_set((1<<PWR_ORANGE));

		puts("Decompression failed, bunzip2 returned error code -");
		puthex(-ret);
		if(ret == BZ_MEM_ERROR) {
			puts("\n\nNot enough memory (0x");
			puthex(free_mem_ptr_p);
			puts(" - 0x");
			puthex(free_mem_ptr_end_p);
			puts(") available");
		}
		error("\n"); //it'll be halted inside while-loop
	}
	puts(" done, booting the kernel.\n");

	//********** light on (green) power LED
		uGPIO3->DDR = gpio_set((1<<PWR_GREEN));
		uGPIO3->DR = gpio_set((1<<PWR_GREEN));

	return destLen;		/* size of the decompressed kernel */
}
