/*
 *  linux/arch/armnommu/mm/arm946.S: PU functions for ARM946
 *  Copyright (C) 2002 Intersil Americas Inc.
 */ 
#include <linux/linkage.h>
#include <linux/config.h>
#include <asm/assembler.h>
#include <asm/constants.h>
#include <asm/procinfo.h>
#include <asm/hardware.h>

/*
 * This is the maximum size of an area which will be invalidated
 * using the single invalidate entry instructions.  Anything larger
 * than this, and we go for the whole cache.
 *
 * This value should be chosen such that we choose the cheapest
 * alternative.
 */
#define MAX_AREA_SIZE	16384

/*
 * the cache line size of the I and D cache
 */
#define DCACHELINESIZE	32
#define ICACHELINESIZE	32

/*
 * and the page size
 */
#define PAGESIZE	4096

	.text

/*
 * cpu_arm946_data_abort()
 *
 * obtain information about current aborted instruction
 *
 * r0 = address of aborted instruction
 *
 * Returns:
 *  r0 = address of abort
 *  r1 != 0 if writing
 *  r3 = FSR
 */
	.align	5
ENTRY(cpu_arm946_data_abort)
	ldr	r1, [r0]			@ read aborted instruction
	tst	r1, r1, lsr #21			@ C = bit 20
	sbc	r1, r1, r1			@ r1 = C - 1
	mov	r0, #0
	mov	r3, #13				@ We map a data abort on a section fault (FSR=13)
	mov	pc, lr				@ Note that the abort address (r0) is not determined here

/* NOTE: All functions below just return, because all low-level
 * operations are performed by the MVC.

/*
 * cpu_arm946_check_bugs()
 */
ENTRY(cpu_arm946_check_bugs)
	mov	pc, lr

/*
 * cpu_arm946_proc_init()
 */
ENTRY(cpu_arm946_proc_init)
	mov	pc, lr

/*
 * cpu_arm946_proc_fin()
 */
ENTRY(cpu_arm946_proc_fin)
/* We don't need to do anything here.
 * The MVC will do it for us
 */
	mov	pc, lr

/*
 * cpu_arm946_reset(loc)
 *
 * Perform a soft reset of the system.  Put the CPU into the
 * same state as it would be if it had been reset, and branch
 * to what would be the reset vector.
 *
 * loc: location to jump to for soft reset
 */
	.align	5
ENTRY(cpu_arm946_reset)
/* We don't need to do anything here.
 * The MVC will do it for us
 */
	mov	pc, lr

/*
 * cpu_arm946_do_idle()
 */
	.align	5
ENTRY(cpu_arm946_do_idle)  
	mov	pc, lr

/* ================================= CACHE ================================ */


/*
 * cpu_arm946_cache_clean_invalidate_all()
 *
 * clean and invalidate all cache lines
 *
 * Note:
 *  1. we should preserve r0 at all times
 */
	.align	5
ENTRY(cpu_arm946_cache_clean_invalidate_all)
	mov	r3, #0
	mcr	p15, 0, r3, c7, c5, 0			@ flush ICache
	mcr	p15, 0, r3, c7, c6, 0			@ flush DCache
	mov	pc, lr

/*
 * cpu_arm946_cache_clean_invalidate_range(start, end, flags)
 *
 * clean and invalidate all cache lines associated with this area of memory
 *
 * start: Area start address
 * end:   Area end address
 * flags: nonzero for I cache as well
 */
	.align	5
ENTRY(cpu_arm946_cache_clean_invalidate_range)
	mov	r3, #0
	mcr	p15, 0, r3, c7, c6, 0			@ flush DCache
	mov	pc, lr

/*
 * cpu_arm946_flush_ram_page(page)
 *
 * clean and invalidate all cache lines associated with this area of memory
 *
 * page: page to clean and invalidate
 */
	.align	5
ENTRY(cpu_arm946_flush_ram_page)
	mov	pc, lr
	

/* ================================ D-CACHE =============================== */

/*
 * cpu_arm946_dcache_invalidate_range(start, end)
 *
 * throw away all D-cached data in specified region without an obligation
 * to write them back.  Note however that we must clean the D-cached entries
 * around the boundaries if the start and/or end address are not cache
 * aligned.
 *
 * start: virtual start address
 * end:   virtual end address
 */
	.align	5
ENTRY(cpu_arm946_dcache_invalidate_range)
//	mov	r3, #0
//	mcr	p15, 0, r3, c7, c6, 0			@ flush DCache

	tst	r0, #DCACHELINESIZE - 1
	bic	r0, r0, #DCACHELINESIZE - 1
	mcrne	p15, 0, r0, c7, c10, 1		@ clean D entry
	tst	r1, #DCACHELINESIZE - 1
	mcrne	p15, 0, r1, c7, c10, 1		@ clean D entry
1:	mcr	p15, 0, r0, c7, c6, 1		@ invalidate D entry
	add	r0, r0, #DCACHELINESIZE
	cmp	r0, r1
	blt	1b
	mov	pc, lr

/*
 * cpu_arm946_dcache_clean_range(start, end)
 *
 * For the specified virtual address range, ensure that all caches contain
 * clean data, such that peripheral accesses to the physical RAM fetch
 * correct data.
 *
 * start: virtual start address
 * end:   virtual end address
 */
	.align	5
ENTRY(cpu_arm946_dcache_clean_range)
	mov	r3, #0
	mcr	p15, 0, r3, c7, c10, 4			@ drain write buffer
	mov	pc, lr

/*
 * cpu_arm946_dcache_clean_page(page)
 *
 * Cleans a single page of dcache so that if we have any future aliased
 * mappings, they will be consistent at the time that they are created.
 *
 * page: virtual address of page to clean from dcache
 *
 * Note:
 *  1. we don't need to flush the write buffer in this case.
 *  2. we don't invalidate the entries since when we write the page
 *     out to disk, the entries may get reloaded into the cache.
 */
	.align	5
ENTRY(cpu_arm946_dcache_clean_page)
	mov	pc, lr @ Just return

/*
 * cpu_arm946_dcache_clean_entry(addr)
 *
 * Clean the specified entry of any caches such that the MMU
 * translation fetches will obtain correct data.
 *
 * addr: cache-unaligned virtual address
 */
	.align	5
ENTRY(cpu_arm946_dcache_clean_entry)
	mov	pc, lr @ Just return

/* ================================ I-CACHE =============================== */

/*
 * cpu_arm946_icache_invalidate_range(start, end)
 *
 * invalidate a range of virtual addresses from the Icache
 *
 * start: virtual start address
 * end:   virtual end address
 */
	.align	5
ENTRY(cpu_arm946_icache_invalidate_range)
	mov	pc, lr @ Just return

ENTRY(cpu_arm946_icache_invalidate_page)
	mov	pc, lr @ Just return


/* ================================== TLB ================================= */

/*
 * cpu_arm946_tlb_invalidate_all()
 *
 * Invalidate all TLB entries
 */
	.align	5
ENTRY(cpu_arm946_tlb_invalidate_all)
	mov	pc, lr
	

/*
 * cpu_arm946_tlb_invalidate_range(start, end)
 *
 * invalidate TLB entries covering the specified range
 *
 * start: range start address
 * end:   range end address
 */
	.align	5
ENTRY(cpu_arm946_tlb_invalidate_range)
	mov	pc, lr

/*
 * cpu_arm946_tlb_invalidate_page(page, flags)
 *
 * invalidate the TLB entries for the specified page.
 *
 * page:  page to invalidate
 * flags: non-zero if we include the I TLB
 */
	.align	5
ENTRY(cpu_arm946_tlb_invalidate_page)
	mov	pc, lr

/* =============================== PageTable ============================== */

/*
 * cpu_arm946_set_pgd(pgd)
 *
 * Set the translation base pointer to be as described by pgd.
 *
 * pgd: new page tables
 */
	.align	5
ENTRY(cpu_arm946_set_pgd)
	mov	pc, lr

/*
 * cpu_arm946_set_pmd(pmdp, pmd)
 *
 * Set a level 1 translation table entry, and clean it out of
 * any caches such that the MMUs can load it correctly.
 *
 * pmdp: pointer to PMD entry
 * pmd:  PMD value to store
 */
	.align	5
ENTRY(cpu_arm946_set_pmd)
	mov	pc, lr

/*
 * cpu_arm946_set_pte(ptep, pte)
 *
 * Set a PTE and flush it out
 */
	.align	5
ENTRY(cpu_arm946_set_pte)
	mov	pc, lr


cpu_manu_name:
	.asciz	"ARM"
ENTRY(cpu_arm946_name)
	.asciz	"ARM946"
	.align

	.section ".text.init", #alloc, #execinstr

__arm946_setup:
	mov	pc, lr @ Just return

	.text

/*
 * Purpose : Function pointers used to access above functions - all calls
 *	     come through these
 */
	.type	arm946_processor_functions, #object
arm946_processor_functions:
	.word	cpu_arm946_data_abort
	.word	cpu_arm946_check_bugs
	.word	cpu_arm946_proc_init
	.word	cpu_arm946_proc_fin
	.word	cpu_arm946_reset
	.word   cpu_arm946_do_idle

	/* cache */
	.word	cpu_arm946_cache_clean_invalidate_all
	.word	cpu_arm946_cache_clean_invalidate_range
	.word	cpu_arm946_flush_ram_page

	/* dcache */
	.word	cpu_arm946_dcache_invalidate_range
	.word	cpu_arm946_dcache_clean_range
	.word	cpu_arm946_dcache_clean_page
	.word	cpu_arm946_dcache_clean_entry

	/* icache */
	.word	cpu_arm946_icache_invalidate_range
	.word	cpu_arm946_icache_invalidate_page

	/* tlb */
	.word	cpu_arm946_tlb_invalidate_all
	.word	cpu_arm946_tlb_invalidate_range
	.word	cpu_arm946_tlb_invalidate_page

	/* pgtable */
	.word	cpu_arm946_set_pgd
	.word	cpu_arm946_set_pmd
	.word	cpu_arm946_set_pte
	.size	arm946_processor_functions, . - arm946_processor_functions

	.type	cpu_arm946_info, #object
cpu_arm946_info:
	.long	cpu_manu_name
	.long	cpu_arm946_name
	.size	cpu_arm946_info, . - cpu_arm946_info

	.type	cpu_arch_name, #object
cpu_arch_name:
	.asciz	"armv5"
	.size	cpu_arch_name, . - cpu_arch_name

	.type	cpu_elf_name, #object
cpu_elf_name:
	.asciz	"v5"
	.size	cpu_elf_name, . - cpu_elf_name
	.align

	.section ".proc.info", #alloc, #execinstr

	.type	__arm946_proc_info,#object
__arm946_proc_info:
	.long	0x41009460			@ ID Code
	.long	0xff00fff0			@ ID Code mask
	.long	0x00000c1e			@ mmuflags copied from arm920 (NOT USED FOR NOMMU!)
	b	__arm946_setup
	.long	cpu_arch_name
	.long	cpu_elf_name
	.long	HWCAP_SWP | HWCAP_HALF | HWCAP_26BIT
	.long	cpu_arm946_info
	.long	arm946_processor_functions
	.size	__arm946_proc_info, . - __arm946_proc_info




