/*
 * MTD map driver for ATMEL flash chips (non-CFI)
 *
 * Author: R. van Brenk + M. Thuys <rbrenk@intersil.com>
 *
 * $Id: atmel_flash.c, Exp $
 *
 * Copyright (c) 2001 Axis Communications AB
 *
 * This file is "not realy"/"Not yet" under GPL.
 *
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/mtd/map.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/flashchip.h>

#define DEVICE_TYPE_X8	(8 / 8)
#define DEVICE_TYPE_X16	(16 / 8)
#define DEVICE_TYPE_X32	(32 / 8)

/* Toggle Bit to determine end of erase/write */
#define D6_MASK 0x40
#define D7_MASK	0x80

/* Addresses */
#define ADDR_MANUFACTURER               0x0000
#define ADDR_DEVICE_ID                  0x0002  /* 0x0001 * 2 */
#define ADDR_UNLOCK_1                   0xAAAA  /* 0x5555 * 2 */
#define ADDR_UNLOCK_2                   0x5554  /* 0x2AAA * 2 */

/* Commands */
#define CMD_UNLOCK_DATA_1               0x00AA
#define CMD_UNLOCK_DATA_2               0x0055
#define CMD_MANUFACTURER_UNLOCK_DATA    0x0090
#define CMD_UNLOCK_BYPASS_MODE          0x00A0	/* Bypass unlock */
#define CMD_PROGRAM_UNLOCK_DATA         0x00A0
#define CMD_RESET_DATA                  0x00F0
#define CMD_SECTOR_ERASE_UNLOCK_DATA    0x0080
#define CMD_SECTOR_ERASE_UNLOCK_DATA_2  0x0030


/* Manufacturers */
#define MANUFACTURER_ATMEL      		0x001F

#define AT49BV1614						0x00C0
#define AT49BV1614T     				0x00C2
#define AT49BV321                       0x00C8
#define AT49BV321T                      0x00C9


struct atmel_flash_private
{
	int device_type;
	int interleave;
	int numchips;
	struct flchip chips[0];
};

struct atmel_flash_info
{
	const __u16 mfr_id;
	const __u16 dev_id;
	const char *name;
	const u_long size;
	const int numeraseregions;
	const struct mtd_erase_region_info regions[4];
};



static int atmel_flash_read(struct mtd_info *, loff_t, size_t, size_t *,
							u_char *);
static int atmel_flash_write(struct mtd_info *, loff_t, size_t, size_t *,
							 const u_char *);
static int atmel_flash_erase(struct mtd_info *, struct erase_info *);
static void atmel_flash_sync(struct mtd_info *);
static int atmel_flash_suspend(struct mtd_info *);
static void atmel_flash_resume(struct mtd_info *);

//static void atmel_flash_destroy(struct mtd_info *);

static const char im_name[] = "atmel_flash";



static inline __u32 wide_read(struct map_info *map, __u32 addr)
{
	if ( map->buswidth == 1 )
	{
		return map->read8(map, addr);
	}
	else if ( map->buswidth == 2 )
	{
		return map->read16(map, addr);
	}
	else if ( map->buswidth == 4 )
	{
		return map->read32(map, addr);
	}

	return 0;
}

static inline void wide_write(struct map_info *map, __u32 val, __u32 addr)
{

	if ( map->buswidth == 1 )
	{
		map->write8(map, val, addr);
	}
	else if ( map->buswidth == 2 )
	{
		map->write16(map, val, addr);
	}
	else if ( map->buswidth == 4 )
	{
		map->write32(map, val, addr);
	}
}

static inline __u32 make_cmd(struct atmel_flash_private *private, __u32 cmd)
{
	if ( (private->interleave == 2) &&
		 (private->device_type == DEVICE_TYPE_X16) )
	{
		cmd |= (cmd << 16);
	}

	return cmd;
}

static inline void send_unlock(struct map_info *map)
{
	wide_write(map, (CMD_UNLOCK_DATA_1 << 16) | CMD_UNLOCK_DATA_1,
			   map->buswidth * ADDR_UNLOCK_1);
	wide_write(map, (CMD_UNLOCK_DATA_2 << 16) | CMD_UNLOCK_DATA_2,
			   map->buswidth * ADDR_UNLOCK_2);
}

static inline void send_cmd(struct map_info *map, __u32 cmd)
{

	send_unlock(map);
	cmd = make_cmd((struct atmel_flash_private *)map->fldrv_priv, cmd);
	wide_write(map, cmd, map->buswidth * ADDR_UNLOCK_1);
}

static inline void send_cmd_to_addr(struct map_info *map, __u32 cmd,
									unsigned long addr)
{
	send_unlock(map);
	cmd = make_cmd((struct atmel_flash_private *)map->fldrv_priv, cmd);
	wide_write(map, cmd, addr);
}



struct mtd_info *atmel_flash_probe(struct map_info *map)
{
	const struct atmel_flash_info table[] = {
		{
			mfr_id: MANUFACTURER_ATMEL,
			dev_id: AT49BV1614,
			name: "ATMEL AT49BV1614",
			size: 0x00200000,
			numeraseregions: 3,
			regions: {
				{ offset: 0x000000, erasesize: 0x02000, numblocks:  8},
				{ offset: 0x010000, erasesize: 0x08000, numblocks:  2},
				{ offset: 0x020000, erasesize: 0x10000, numblocks: 30}
			}
		},
		{
			mfr_id: MANUFACTURER_ATMEL,
			dev_id: AT49BV1614T,
			name: "ATMEL AT49BV1614T",
			size: 0x00200000,
			numeraseregions: 3,
			regions: {
				{ offset: 0x000000, erasesize: 0x10000, numblocks: 30},
				{ offset: 0x01E000, erasesize: 0x08000, numblocks:  2},
				{ offset: 0x01F000, erasesize: 0x02000, numblocks:  8}
			}
		},
        {
			mfr_id: MANUFACTURER_ATMEL,
			dev_id: AT49BV321,
			name: "ATMEL AT49BV321",
			size: 0x00400000,
			numeraseregions: 2,
			regions: {
				{ offset: 0x000000, erasesize: 0x02000, numblocks:  8},
                { offset: 0x010000, erasesize: 0x10000, numblocks: 63}
			}
		},
        {
			mfr_id: MANUFACTURER_ATMEL,
			dev_id: AT49BV321T,
			name: "ATMEL AT49BV321T",
			size: 0x00400000,
			numeraseregions: 2,
			regions: {
				{ offset: 0x000000, erasesize: 0x10000, numblocks: 63},
                { offset: 0x3F0000, erasesize: 0x02000, numblocks:  8}
			}
		}
    };
	struct mtd_info *mtd;
	__u32 mfr_id, dev_id;
	struct flchip *chip;
	struct atmel_flash_private temp;
	int i;
	int found;
	struct atmel_flash_private *private;

	printk("%s: Probing for ATMEL flash...\n", map->name);

	temp.device_type = DEVICE_TYPE_X16;	// Assume X16 (FIXME)
	temp.interleave = 2;
	map->fldrv_priv = &temp;

	send_cmd(map, CMD_RESET_DATA);
	send_cmd(map, CMD_MANUFACTURER_UNLOCK_DATA);

	mfr_id = wide_read(map, map->buswidth * ADDR_MANUFACTURER);
	printk("Manufacturer ID (probing at address %x): %x\n", map->buswidth * ADDR_MANUFACTURER, mfr_id);
	printk("Map info: buswidth: %d\n    map_priv_1 %ld\n", map->buswidth, map->map_priv_1);

	send_cmd(map, CMD_RESET_DATA);
	send_cmd(map, CMD_MANUFACTURER_UNLOCK_DATA);
	dev_id = wide_read(map, map->buswidth * ADDR_DEVICE_ID);

	printk("Device ID (probing at address %x): %x\n", map->buswidth * ADDR_DEVICE_ID, dev_id);

	send_cmd(map, CMD_RESET_DATA);

	if ( (map->buswidth == 4) && ((mfr_id >> 16) == (mfr_id & 0xffff)) &&
		 ((dev_id >> 16) == (dev_id & 0xffff)) )
	{
		mfr_id = mfr_id & 0xffff;
		dev_id = dev_id & 0xffff;
	}
	else
	{
		temp.interleave = 1;
	}

	mtd = (struct mtd_info*)kmalloc(sizeof(*mtd), GFP_KERNEL);
	if ( !mtd )
	{
		printk(KERN_WARNING
			   "%s: kmalloc failed for info structure\n", map->name);
		map->fldrv_priv = NULL;
		return NULL;
	}
	memset(mtd, 0, sizeof(*mtd));

	found = 0;
	for ( i = 0; i < (sizeof(table)/sizeof(table[0])); i++ )
	{
		if ( (mfr_id == table[i].mfr_id) &&
			 (dev_id == table[i].dev_id) )
		{
			int j;

			printk("%s: Found %d x %s %ldMb\n", map->name,
				   temp.interleave, table[i].name,
				   (table[i].size*8)/(1024*1024));

			mtd->size = table[i].size * temp.interleave;
			mtd->numeraseregions = table[i].numeraseregions;
			mtd->eraseregions = kmalloc(
									   sizeof(struct mtd_erase_region_info) *
									   mtd->numeraseregions, GFP_KERNEL);
			if ( !mtd->eraseregions )
			{
				printk(KERN_WARNING "%s: Failed to allocate "
					   "memory for MTD erase region info\n",
					   map->name);
				kfree(mtd);
				map->fldrv_priv = NULL;
				return NULL;
			}
			for ( j = 0; j < mtd->numeraseregions; j++ )
			{
				mtd->eraseregions[j].offset =
				table[i].regions[j].offset *
				temp.interleave;
				mtd->eraseregions[j].erasesize =
				table[i].regions[j].erasesize *
				temp.interleave;
				mtd->eraseregions[j].numblocks =
				table[i].regions[j].numblocks;
				if ( mtd->erasesize <
					 mtd->eraseregions[j].erasesize )
				{
					mtd->erasesize =
					mtd->eraseregions[j].erasesize;
				}
			}
			found = 1;
			break;
		}
	}
	if ( !found )
	{
		printk(KERN_WARNING "%s: unknown flash device\n", map->name);
		kfree(mtd);
		map->fldrv_priv = NULL;
		return NULL;
	}

	mtd->type = MTD_NORFLASH;
	mtd->flags = MTD_CAP_NORFLASH;
	mtd->name = map->name;
	mtd->erase = atmel_flash_erase;
	mtd->read = atmel_flash_read;
	mtd->write = atmel_flash_write;
	mtd->sync = atmel_flash_sync;
	mtd->suspend = atmel_flash_suspend;
	mtd->resume = atmel_flash_resume;
	//map->fldrv_destroy = atmel_flash_destroy;
	//mtd->priv = map;

private = kmalloc(sizeof(struct atmel_flash_private) +
				  sizeof(struct flchip),
				  GFP_KERNEL);
	if ( !private )
	{
		printk(KERN_WARNING
			   "%s: kmalloc failed for private structure\n", map->name);
		kfree(mtd);
		map->fldrv_priv = NULL;
		return NULL;
	}
	memset(private, 0, sizeof(struct atmel_flash_private));private->device_type = temp.device_type;private->interleave = temp.interleave;

	chip = private->chips;
	memset(chip, 0, sizeof(*chip));

	map->fldrv_priv = private;

	init_waitqueue_head(&chip->wq);
	spin_lock_init(&chips->_spinlock);
	chip->state = FL_READY;
	chip->mutex = &chip->_spinlock;

	return mtd;
}



static int atmel_flash_read(struct mtd_info *mtd, loff_t from, size_t len,
							size_t *retlen, u_char *buf)
{
	DECLARE_WAITQUEUE(wait, current);
	unsigned long timeo = jiffies + HZ;
	struct map_info *map = mtd->priv;
	struct atmel_flash_private *private = map->fldrv_priv;
	struct flchip *chip = &private->chips[0];

	*retlen = 0;

	if ( (from + len) > mtd->size )
	{
		printk(KERN_WARNING "%s: read request past end of device "
			   "(0x%lx)\n", map->name, (long) (from + len));

		return -EINVAL;
	}

	retry:
	spin_lock_bh(chip->mutex);

	if ( chip->state != FL_READY )
	{
		printk(KERN_INFO "%s: waiting for chip to read, state = %d\n",
			   map->name, chip->state);
		set_current_state(TASK_UNINTERRUPTIBLE);
		add_wait_queue(&chip->wq, &wait);

		spin_unlock_bh(chip->mutex);

		schedule();
		remove_wait_queue(&chip->wq, &wait);

		if ( signal_pending(current) )
			return -EINTR;

		timeo = jiffies + HZ;

		goto retry;
	}

//	from += chip->start;

	chip->state = FL_READY;

	map->copy_from(map, buf, from, len);

	wake_up(&chip->wq);
	spin_unlock_bh(chip->mutex);

	*retlen = len;

	return 0;
}



static inline int atmel_flash_is_busy(struct map_info *map, unsigned long addr, __u32 data,
									  int interleave)
{
	return( (wide_read(map, addr) & D7_MASK) !=
			(data & D7_MASK) );
}

static int write_one_word(struct map_info *map, struct flchip *chip,
						  unsigned long adr, __u32 datum)
{
	unsigned long timeo = jiffies + HZ;
	struct atmel_flash_private *private = map->fldrv_priv;
	DECLARE_WAITQUEUE(wait, current);
	int ret = 0;
	int times_left;
// printk("write_one_word(): 0x%x\n", adr);

	retry:
	spin_lock_bh(chip->mutex);

	if ( chip->state != FL_READY )
	{
		printk("%s: waiting for chip to write, state = %d\n",
			   map->name, chip->state);
		set_current_state(TASK_UNINTERRUPTIBLE);
		add_wait_queue(&chip->wq, &wait);

		spin_unlock_bh(chip->mutex);

		schedule();
		remove_wait_queue(&chip->wq, &wait);
		printk(KERN_INFO "%s: woke up to write\n", map->name);
		if ( signal_pending(current) )
			return -EINTR;

		timeo = jiffies + HZ;

		goto retry;
	}

	chip->state = FL_WRITING;

//	adr += chip->start;
	ENABLE_VPP(map);

	send_cmd(map, CMD_PROGRAM_UNLOCK_DATA);
	wide_write(map, datum, adr);

	spin_unlock_bh(chip->mutex);
	times_left = 500000;

	while ( times_left-- && atmel_flash_is_busy(map, adr, datum,
												private->interleave) )
	{

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
		if ( current->need_resched )
#else
		if ( need_resched )
#endif
			schedule();
	}
	if ( !times_left )
	{
		printk(KERN_WARNING "%s: write to 0x%lx timed out!\n",
			   map->name, adr);
		ret = -EIO;
	}
	else
	{
		__u32 verify;
		if ( (verify = wide_read(map, adr)) != datum )
		{
			printk(KERN_WARNING "%s: write failed at addr %lx. datum = %x, "
				   "verify = %x, times = %d\n", map->name, adr, datum, verify, times_left);
			ret = -EIO;
		}
// printk("Read, %x at 0x%x\n", verify, adr);
	}
	spin_lock_bh(chip->mutex);

	DISABLE_VPP(map);
	chip->state = FL_READY;
	wake_up(&chip->wq);
	spin_unlock_bh(chip->mutex);

	return ret;
}



static int atmel_flash_write(struct mtd_info *mtd, loff_t to , size_t len,
							 size_t *retlen, const u_char *buf)
{
	struct map_info *map = mtd->priv;
	struct atmel_flash_private *private = map->fldrv_priv;
	int ret = 0;
	unsigned long ofs;
// printk("atmel_flash_write(): 0x%x\n", to);

	*retlen = 0;
	if ( !len )
		return 0;

	ofs = to;

	/* If it's not bus-aligned, do the first byte write */
	if ( ofs & (map->buswidth - 1) )
	{
		unsigned long bus_ofs = ofs & ~(map->buswidth - 1);
		int i = ofs - bus_ofs;
		int n = 0;
		u_char tmp_buf[2];
		__u32 datum;

		map->copy_from(map, tmp_buf, bus_ofs + private->chips[0].start,
					   map->buswidth);
		while ( len && i < map->buswidth )
			tmp_buf[i++] = buf[n++], len--;

		if ( map->buswidth == 2 )
		{
			datum = *(__u16*)tmp_buf;
		}
		else if ( map->buswidth == 4 )
		{
			datum = *(__u32*)tmp_buf;
		}
		else
		{
			return -EINVAL;	 /* should never happen, but be safe */
		}

		ret = write_one_word(map, &private->chips[0], bus_ofs, datum);
		if ( ret )
			return ret;

		ofs += n;
		buf += n;
		(*retlen) += n;

		if ( ofs )
		{
			return 0;
		}
	}

	/* We are now aligned, write as much as possible */
	while ( len >= map->buswidth )
	{
		__u32 datum;

		if ( map->buswidth == 1 )
		{
			datum = *(__u8*)buf;
		}
		else if ( map->buswidth == 2 )
		{
			datum = *(__u16*)buf;
		}
		else if ( map->buswidth == 4 )
		{
			datum = *(__u32*)buf;
		}
		else
		{
			return -EINVAL;
		}

		ret = write_one_word(map, &private->chips[0], ofs, datum);

		if ( ret )
		{
			return ret;
		}

		ofs += map->buswidth;
		buf += map->buswidth;
		(*retlen) += map->buswidth;
		len -= map->buswidth;
	}

	if ( len & (map->buswidth - 1) )
	{
		int i = 0, n = 0;
		u_char tmp_buf[2];
		__u32 datum;

		map->copy_from(map, tmp_buf, ofs + private->chips[0].start,
					   map->buswidth);
		while ( len-- )
		{
			tmp_buf[i++] = buf[n++];
		}

		if ( map->buswidth == 2 )
		{
			datum = *(__u16*)tmp_buf;
		}
		else if ( map->buswidth == 4 )
		{
			datum = *(__u32*)tmp_buf;
		}
		else
		{
			return -EINVAL;	 /* should never happen, but be safe */
		}

		ret = write_one_word(map, &private->chips[0], ofs, datum);

		if ( ret )
		{
			return ret;
		}

		(*retlen) += n;
	}

	return 0;
}



static inline int erase_one_block(struct map_info *map, struct flchip *chip,
								  unsigned long adr, u_long size)
{
	unsigned long timeo = jiffies + HZ;
	struct atmel_flash_private *private = map->fldrv_priv;
	DECLARE_WAITQUEUE(wait, current);
// printk("erase_one_block(): adr: 0x%lx, size: 0x%lx\n", adr, size);

	retry:
	spin_lock_bh(chip->mutex);

	if ( chip->state != FL_READY )
	{
		printk("erase_one_block(): chip not ready, state = %d\n", chip->state);
		set_current_state(TASK_UNINTERRUPTIBLE);
		add_wait_queue(&chip->wq, &wait);

		spin_unlock_bh(chip->mutex);

		schedule();
		remove_wait_queue(&chip->wq, &wait);

		if ( signal_pending(current) )
		{
			return -EINTR;
		}

		timeo = jiffies + HZ;

		goto retry;
	}

	chip->state = FL_ERASING;

//	adr += chip->start;
	ENABLE_VPP(map);
	send_cmd(map, CMD_SECTOR_ERASE_UNLOCK_DATA);
	send_cmd_to_addr(map, CMD_SECTOR_ERASE_UNLOCK_DATA_2, adr);

	timeo = jiffies + (HZ*20);

	spin_unlock_bh(chip->mutex);
	schedule_timeout(HZ/20);
	spin_lock_bh(chip->mutex);

	while ( atmel_flash_is_busy(map, adr, 0xFF, private->interleave) )
	{

#if 0
		if ( chip->state != FL_ERASING )
		{
			/* Someone's suspended the erase. Sleep */
			set_current_state(TASK_UNINTERRUPTIBLE);
			add_wait_queue(&chip->wq, &wait);

			spin_unlock_bh(chip->mutex);
			printk(KERN_INFO "%s: erase suspended. Sleeping\n",
				   map->name);
			schedule();
			remove_wait_queue(&chip->wq, &wait);

			if ( signal_pending(current) )
			{
				return -EINTR;
			}

			timeo = jiffies + (HZ*2); /* FIXME */
			spin_lock_bh(chip->mutex);
			continue;
		}

		/* OK Still waiting */
		if ( time_after(jiffies, timeo) )
		{
			chip->state = FL_READY;
			spin_unlock_bh(chip->mutex);
			printk(KERN_WARNING "%s: waiting for erase to complete "
				   "timed out.\n", map->name);
			DISABLE_VPP(map);

			return -EIO;
		}
#endif
		/* Latency issues. Drop the lock, wait a while and retry */
		spin_unlock_bh(chip->mutex);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
		if ( current->need_resched )
#else
		if ( need_resched )
#endif
			schedule();
		else
			udelay(1);

		spin_lock_bh(chip->mutex);
		continue;
	}

	/* Verify every single word */
	{
		int address;
		int error = 0;
		__u8 verify;

		for ( address = adr; address < (adr + size); address++ )
		{
			if ( (verify = map->read8(map, address)) != 0xFF )
			{
				error = 1;
				break;
			}
// printk("%04x ", verify);
		}
		if ( error )
		{
			chip->state = FL_READY;
			spin_unlock_bh(chip->mutex);
			printk(KERN_WARNING
				   "%s: verify error at 0x%x, size %ld.\n",
				   map->name, address, size);
			DISABLE_VPP(map);

			return -EIO;
		}
	}

	DISABLE_VPP(map);
	chip->state = FL_READY;
	wake_up(&chip->wq);
	spin_unlock_bh(chip->mutex);

	return 0;
}



static int atmel_flash_erase(struct mtd_info *mtd, struct erase_info *instr)
{
	struct map_info *map = mtd->priv;
	struct atmel_flash_private *private = map->fldrv_priv;
	unsigned long adr, len;
	int ret = 0;
	int i;
	int first;
	struct mtd_erase_region_info *regions = mtd->eraseregions;

// printk("atmel_flash_erase()\n");
// printk("  instr->len = %ld\n", instr->len);
// printk("  instr->addr = 0x%lx\n", instr->addr);

	if ( instr->addr > mtd->size )
	{
		return -EINVAL;
	}

	if ( (instr->len + instr->addr) > mtd->size )
	{
		return -EINVAL;
	}

	/* Check that both start and end of the requested erase are
	 * aligned with the erasesize at the appropriate addresses.
	 */

	i = 0;

	/* Skip all erase regions which are ended before the start of
	   the requested erase. Actually, to save on the calculations,
	   we skip to the first erase region which starts after the
	   start of the requested erase, and then go back one.
	*/

	while ( (i < mtd->numeraseregions) &&
			(instr->addr >= regions[i].offset) )
	{
		i++;
	}
	i--;

	/* OK, now i is pointing at the erase region in which this
	 * erase request starts. Check the start of the requested
	 * erase range is aligned with the erase size which is in
	 * effect here.
	 */

	if ( instr->addr & (regions[i].erasesize-1) )
	{
		return -EINVAL;
	}

	/* Remember the erase region we start on. */

	first = i;

	/* Next, check that the end of the requested erase is aligned
	 * with the erase region at that address.
	 */

	while ( (i < mtd->numeraseregions) &&
			((instr->addr + instr->len) >= regions[i].offset) )
	{
		i++;
	}

	/* As before, drop back one to point at the region in which
	 * the address actually falls.
	 */

	i--;

	if ( (instr->addr + instr->len) & (regions[i].erasesize-1) )
	{
		return -EINVAL;
	}

	adr = instr->addr;
	len = instr->len;

	i = first;

	while ( len )
	{
		ret = erase_one_block(map, &private->chips[0], adr,
							  regions[i].erasesize);

		if ( ret )
		{
			return ret;
		}

		adr += regions[i].erasesize;
		len -= regions[i].erasesize;

		if ( adr ==
			 (regions[i].offset +
			  (regions[i].erasesize * regions[i].numblocks)) )
		{
			i++;
		}
	}

	if ( instr->callback )
	{
		instr->callback(instr);
	}

	return 0;
}



static void atmel_flash_sync(struct mtd_info *mtd)
{
	struct map_info *map = mtd->priv;
	struct atmel_flash_private *private = map->fldrv_priv;
	int i;
	struct flchip *chip;
	int ret = 0;
	DECLARE_WAITQUEUE(wait, current);

	for ( i=0; !ret && (i < private->numchips); i++ )
	{
		chip = &private->chips[i];

		retry:
		spin_lock_bh(chip->mutex);

		switch ( chip->state )
		{
		case FL_READY:
		case FL_STATUS:
		case FL_CFI_QUERY:
		case FL_JEDEC_QUERY:
			chip->oldstate = chip->state;
			chip->state = FL_SYNCING;
			/* No need to wake_up() on this state change -
			 * as the whole point is that nobody can do anything
			 * with the chip now anyway.
			 */
		case FL_SYNCING:
			spin_unlock_bh(chip->mutex);
			break;

		default:
			/* Not an idle state */
			add_wait_queue(&chip->wq, &wait);

			spin_unlock_bh(chip->mutex);

			schedule();

			remove_wait_queue(&chip->wq, &wait);

			goto retry;
		}
	}

	/* Unlock the chips again */

	for ( i--; i >= 0; i-- )
	{
		chip = &private->chips[i];

		spin_lock_bh(chip->mutex);

		if ( chip->state == FL_SYNCING )
		{
			chip->state = chip->oldstate;
			wake_up(&chip->wq);
		}
		spin_unlock_bh(chip->mutex);
	}
}



static int atmel_flash_suspend(struct mtd_info *mtd)
{
	printk("atmel_flash_suspend(): not implemented!\n");
	return -EINVAL;
}



static void atmel_flash_resume(struct mtd_info *mtd)
{
	printk("atmel_flash_resume(): not implemented!\n");
}



/*static void atmel_flash_destroy(struct mtd_info *mtd)
{
	struct map_info *map = mtd->priv;
	struct atmel_flash_private *private = map->fldrv_priv;
	kfree(private);
}*/


int __init atmel_flash_init(void)
{
	inter_module_register(im_name, THIS_MODULE, &atmel_flash_probe);

	return 0;
}

void __exit atmel_flash_exit(void)
{
	inter_module_unregister(im_name);
}

module_init(atmel_flash_init);
module_exit(atmel_flash_exit);

