/*
 *
 * Flash mapping for Intersil uClinux based Access Point Reference Designs
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>


#define WINDOW_ADDR CONFIG_MTD_ISLMAP_START
#define WINDOW_SIZE CONFIG_MTD_ISLMAP_LEN
#define BUSWIDTH CONFIG_MTD_ISLMAP_BUSWIDTH

#ifdef CONFIG_MTD_AMDSTD
struct mtd_info *amd_flash_probe(struct map_info *map);
#endif
#ifdef CONFIG_MTD_ATMEL
struct mtd_info *atmel_flash_probe(struct map_info *map);
#endif
#ifdef CONFIG_MTD_CFI
struct mtd_info *cfi_probe(struct map_info *map);
#endif
#ifdef CONFIG_MTD_JEDECPROBE
struct mtd_info *jedec_probe(struct map_info *map);
#endif
      
#include <linux/mtd/partitions.h>
#include <rockhopper.h>
/* This include file contains the flash partitioning */
#include <flash_layout.h>

static struct mtd_info *mymtd;

static __u8 islmap_read8(struct map_info *map, unsigned long ofs)
{
    return *(volatile __u8 *)(map->map_priv_1 + ofs);
}

static __u16 islmap_read16(struct map_info *map, unsigned long ofs)
{
    return *(volatile __u16 *)(map->map_priv_1 + ofs);
}

static __u32 islmap_read32(struct map_info *map, unsigned long ofs)
{
    return *(volatile __u32 *)(map->map_priv_1 + ofs);
}

static void islmap_write8(struct map_info *map, __u8 d, unsigned long adr)
{
    *(volatile __u8 *)(map->map_priv_1 + adr) = d;
}

static void islmap_write16(struct map_info *map, __u16 d, unsigned long adr)
{
    *(volatile __u16 *)(map->map_priv_1 + adr) = d;
}

static void islmap_write32(struct map_info *map, __u32 d, unsigned long adr)
{
    *(volatile __u32 *)(map->map_priv_1 + adr) = d;
}

/*
 * We need to access the flash byte by byte, so don't use memcpy for the copy_from/to
 * because that may be optimised.
 */
static void islmap_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
    while (len) {
            len--;
            *(volatile __u8 *) to = islmap_read8(map, from);
            ((__u8 *) to)++;
            from++;
    }
}

static void islmap_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
    while (len) {
            len--;
            islmap_write8(map, *(volatile __u8 *) from, to);
            ((volatile __u8 *) from)++;
            to++;
    }
}

struct map_info islmap_map = {
	name: "Intersil uClinux Flash mapping",
	size: WINDOW_SIZE,
	buswidth: BUSWIDTH,
    read8: islmap_read8,
    read16: islmap_read16,
    read32: islmap_read32,
    copy_from: islmap_copy_from,
    write8: islmap_write8,
    write16: islmap_write16,
    write32: islmap_write32,
    copy_to: islmap_copy_to
};

int __init init_islmap(void)
{
	unsigned long cum_partitionsize;
	int i;

    printk(KERN_NOTICE "Intersil mapped flash device: %x at %x\n", WINDOW_SIZE, WINDOW_ADDR);
    islmap_map.map_priv_1 = (unsigned long)ioremap(WINDOW_ADDR, WINDOW_SIZE);

	/* Extra check to make sure our
	 * no-mmu implementation will not fail
	 */
    if (!islmap_map.map_priv_1 && WINDOW_ADDR != islmap_map.map_priv_1 ) {
		printk("Failed to ioremap\n");
		return -EIO;
	}
	/* Try probing different devices. */
#ifdef CONFIG_MTD_CFI
    mymtd = cfi_probe(&islmap_map);
#endif

#ifdef CONFIG_MTD_AMDSTD
	if (!mymtd)
	{
        mymtd = amd_flash_probe(&islmap_map);
	}
#endif
#ifdef CONFIG_MTD_ATMEL
	if (!mymtd)
	{
        mymtd = atmel_flash_probe(&islmap_map);
	}
#endif
#ifdef CONFIG_MTD_JEDECPROBE
    if (!mymtd)
    {
        mymtd = jedec_probe(&islmap_map);
    }
#endif

//  mymtd = do_map_probe("cfi", &islmap_map);
	if (mymtd) {
		mymtd->module = THIS_MODULE;

    	/* partition fixup */
		cum_partitionsize = 0;
		partition_info[0].offset = 0;

		/* The first partition (Bootcode) is read only */
		partition_info[0].mask_flags = MTD_WRITEABLE;

		for (i = 0; i < NUM_PARTITIONS - 1; i++) {
			cum_partitionsize += partition_info[i].size;
			partition_info[i + 1].offset = cum_partitionsize;
		}
		partition_info[NUM_PARTITIONS - 1].size = mymtd->size - cum_partitionsize;

		if (add_mtd_partitions(mymtd, partition_info, NUM_PARTITIONS)) {
		  printk("Error adding MTD partitions\n");
		}

		return 0;
	}

    iounmap((void *)islmap_map.map_priv_1);
	return -ENXIO;
}

static void __exit cleanup_islmap(void)
{
	if (mymtd) {
		del_mtd_device(mymtd);
		map_destroy(mymtd);
	}
    if (islmap_map.map_priv_1) {
        iounmap((void *)islmap_map.map_priv_1);
        islmap_map.map_priv_1 = 0;
	}
}

module_init(init_islmap);
module_exit(cleanup_islmap);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("David Woodhouse <dwmw2@infradead.org>");
MODULE_DESCRIPTION("Generic configurable MTD map driver");
