/*
 * $Id: physmap.c,v 1.15 2001/10/02 15:05:14 dwmw2 Exp $
 *
 * Normal mappings of chips in physical memory
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>


#define WINDOW_ADDR CONFIG_MTD_PHYSMAP_START
#define WINDOW_SIZE CONFIG_MTD_PHYSMAP_LEN
#define BUSWIDTH CONFIG_MTD_PHYSMAP_BUSWIDTH

#ifdef CONFIG_MTD_AMDSTD
struct mtd_info *amd_flash_probe(struct map_info *map);
#endif
#ifdef CONFIG_MTD_ATMEL
struct mtd_info *atmel_flash_probe(struct map_info *map);
#endif
#ifdef CONFIG_MTD_CFI
struct mtd_info *cfi_probe(struct map_info *map);
#endif
#ifdef CONFIG_MTD_JEDECPROBE
struct mtd_info *jedec_probe(struct map_info *map);
#endif
      
#define MYPARTITIONS

#ifdef MYPARTITIONS

#include <linux/mtd/partitions.h>

/*
 * Define partitions for flash device
 */

#ifdef CONFIG_ARCH_INL
#include <rockhopper.h>
#include <flash_layout.h>
#endif

#endif /* MYPARTITIONS */

static struct mtd_info *mymtd;

__u8 physmap_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 physmap_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 physmap_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

void physmap_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void physmap_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void physmap_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void physmap_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

void physmap_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

struct map_info physmap_map = {
	name: "Physically mapped flash",
	size: WINDOW_SIZE,
	buswidth: BUSWIDTH,
	read8: physmap_read8,
	read16: physmap_read16,
	read32: physmap_read32,
	copy_from: physmap_copy_from,
	write8: physmap_write8,
	write16: physmap_write16,
	write32: physmap_write32,
	copy_to: physmap_copy_to
};

int __init init_physmap(void)
{
	unsigned long cum_partitionsize;
	int i;

       	printk(KERN_NOTICE "physmap flash device: %x at %x\n", WINDOW_SIZE, WINDOW_ADDR);
	physmap_map.map_priv_1 = (unsigned long)ioremap(WINDOW_ADDR, WINDOW_SIZE);

	/* Extra check to make sure our
	 * no-mmu implementation will not fail
	 */
	if (!physmap_map.map_priv_1 && WINDOW_ADDR != physmap_map.map_priv_1 ) {
		printk("Failed to ioremap\n");
		return -EIO;
	}
	/* Try probing different devices. */
#ifdef CONFIG_MTD_CFI
	mymtd = cfi_probe(&physmap_map);
#endif

#ifdef CONFIG_MTD_AMDSTD
	if (!mymtd)
	{
		mymtd = amd_flash_probe(&physmap_map);
	}
#endif
#ifdef CONFIG_MTD_ATMEL
	if (!mymtd)
	{
		mymtd = atmel_flash_probe(&physmap_map);
	}
#endif
#ifdef CONFIG_MTD_JEDECPROBE
    if (!mymtd)
    {
        mymtd = jedec_probe(&physmap_map);
    }
#endif

//	mymtd = do_map_probe("cfi", &physmap_map);
	if (mymtd) {
		mymtd->module = THIS_MODULE;
#ifdef MYPARTITIONS
    	/* partition fixup */
		cum_partitionsize = 0;
		partition_info[0].offset = 0;
		/* The first partition (Bootcode) is read only */
		partition_info[0].mask_flags = MTD_WRITEABLE;

		for (i = 0; i < NUM_PARTITIONS - 1; i++)
		{
			cum_partitionsize += partition_info[i].size;
			partition_info[i + 1].offset = cum_partitionsize;
		}
		partition_info[NUM_PARTITIONS - 1].size = mymtd->size - cum_partitionsize;

		if (add_mtd_partitions(mymtd, partition_info, NUM_PARTITIONS))
		{
		  printk("Error adding MTD partitions\n");
		}
#else
		add_mtd_device(mymtd);
#endif
		return 0;
	}

	iounmap((void *)physmap_map.map_priv_1);
	return -ENXIO;
}

static void __exit cleanup_physmap(void)
{
	if (mymtd) {
		del_mtd_device(mymtd);
		map_destroy(mymtd);
	}
	if (physmap_map.map_priv_1) {
		iounmap((void *)physmap_map.map_priv_1);
		physmap_map.map_priv_1 = 0;
	}
}

module_init(init_physmap);
module_exit(cleanup_physmap);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("David Woodhouse <dwmw2@infradead.org>");
MODULE_DESCRIPTION("Generic configurable MTD map driver");
