/*********************************************************************

    File:               islmgt.c
    Author:             Rob Bastings
    Version:            0.0.1
    Created:            Nov 22, 2002
    Updated:            
    Company:            Intersil Americas Inc.
    Description:        Management data dispatcher for the Intersil
                        Wireless LAN adapter based on the 38xx chips.
    History:
    version 0.0.1:      Nov 22, 2002

*********************************************************************/

#include <linux/skbuff.h>
#include <linux/rtnetlink.h>
#include <linux/isil_netlink.h>
#include <net/sock.h>
#include <blob.h>          /* Device types */
	 	
#include "prism_mgt.h"

#define IND_TABLE_SZ    10
#define RESP_TABLE_SZ   10

#ifndef NULL
#   define  NULL ((void *)0)
#endif

static struct ind_rec {
    unsigned int   dev_type;  
    unsigned int   dev_seq;  
    unsigned int   oid;
    mgt_indication_t ind;
} ind_table[IND_TABLE_SZ];

static struct resp_rec {
    unsigned int   src_id;  
    unsigned int   src_seq;  
    unsigned int   oid;
    mgt_response_t resp;
} resp_table[RESP_TABLE_SZ];

static mgt_confirm_t cnf_hndl = NULL;

static struct sock *nl_sock = NULL; /* netlink socket */

static int ntl_confirm( unsigned int src_id, unsigned int src_seq,
                        unsigned int dev_type, unsigned int dev_seq, 
                        unsigned int op, unsigned int oid, void *data, unsigned long len); 
static int ntl_transmit(pid_t pid, int seq, int operation, 
                        int oid, long *data, unsigned long data_len, 
                        int dev_id ); 
static void ntl_receive(struct sock *sk, int len);

static int pimfor_encode_header(int operation, unsigned long oid, int device_id, 
                                int flags, int length, struct pimfor_hdr *header);
static int pimfor_decode_header(struct pimfor_hdr *header, int *version,
                                int *operation, unsigned long *oid,
                                int *device_id, int *flags, unsigned long *length);


/********************************************************************
*                      Management interfaces
*********************************************************************/

int mgt_request( unsigned int dev_type, unsigned int dev_seq,              
                 unsigned int src_id,   unsigned int src_seq,              
                 unsigned int op, int oid, void *data, unsigned long len)
{
    int i;
    int err = 0;
    struct ind_rec *pi;
    
    printk (KERN_ERR "mgt_request(%d, %d, %d, %d, %d, %x, <data>, %d)\n", 
            dev_type, dev_seq, src_id, src_seq, op, oid, len );
    for ( i = 0, pi = ind_table; i < IND_TABLE_SZ; i++, pi++ ) {
        if ( ( pi->dev_type == dev_type ) && /* ( pi->dev_seq == dev_seq ) && */
             ( ( pi->oid == oid ) || pi->oid == 0 ) && ( pi->ind != NULL ) ) {
            if ( (*pi->ind)( dev_type, dev_seq, src_id, src_seq, op, oid, data, len ) < 0 )
                err= -1;
        }
    }

    return err;
}

int mgt_response( unsigned int src_id, unsigned int src_seq,
                  unsigned int dev_type, unsigned int dev_seq, 
                  unsigned int op, int oid, void *data, unsigned long len)
{
    int i;
    int err = 0;
    struct resp_rec *pr;
    
    printk (KERN_ERR "mgt_response(%d, %d, %d, %d, %d, %x, <data>, %d)\n", 
            src_id, src_seq, dev_type, dev_seq, op, oid, len );
    for ( i = 0, pr = resp_table; i < RESP_TABLE_SZ; i++, pr++ ) {
        if ( ( pr->src_id == src_id ) && /* ( pr->src_seq == src_seq ) && */
             ( ( pr->oid == oid ) || pr->oid == 0 ) && ( pr->resp != NULL ) ) {
            if ( (*pr->resp)( src_id, src_seq, dev_type, dev_seq, op, oid, data, len ) < 0)
                err = -1;
        }
    }

    if ( cnf_hndl ) {
        if ( ( *cnf_hndl )( src_id, src_seq, dev_type, dev_seq, op, oid, data, len ) < 0 )
            err = -1;
    }

    return err;
}

int mgt_indication_handler( unsigned int dev_type, unsigned int dev_seq, 
                            unsigned int oid, mgt_indication_t ind )
{
    int i;
    struct ind_rec *pi;
    
    if ( ind ) {
        for ( i = 0, pi = ind_table; i < IND_TABLE_SZ; i++, pi++ ) {
           if ( pi->ind == 0 ) { 
              pi->dev_type = dev_type;
              pi->dev_seq  = dev_seq;
              pi->oid      = oid;
              pi->ind      = ind;
              return 0;
           }
        }
    }

    return -1;
}

int mgt_response_handler( unsigned int src_id, unsigned int src_seq, 
                            unsigned int oid, mgt_response_t resp )
{
    int i;
    struct resp_rec *pr;
    
    if ( resp ) {
        for ( i = 0, pr = resp_table; i < RESP_TABLE_SZ; i++, pr++ ) {
           if ( pr->resp == 0 ) { 
              pr->src_id   = src_id;
              pr->src_seq  = src_seq;
              pr->oid      = oid;
              pr->resp     = resp;
              return 0;
           }
        }
    }

    return -1;
}

int mgt_confirm_handler( mgt_confirm_t cnf )
{
    if ( cnf ) {
        cnf_hndl = cnf;
        return 0;
    }

    return -1;
}

int mgt_init( void )
{
    int i;
    struct ind_rec  *ip;
    struct resp_rec *pr;

    for ( i = 0, ip = ind_table; i < IND_TABLE_SZ; i++, ip++ ) {
        ip->dev_type = 0;  
        ip->dev_seq  = 0;  
        ip->oid      = 0;
        ip->ind      = NULL;
    }

    for ( i = 0, pr = resp_table; i < RESP_TABLE_SZ; i++, pr++ ) {
        pr->src_id   = 0;  
        pr->src_seq  = 0;  
        pr->oid      = 0;
        pr->resp     = NULL;
    }
    
    cnf_hndl = NULL;

    /* Netlink socket */
    if (nl_sock == NULL)
    {
        nl_sock = netlink_kernel_create(NETLINK_ISIL, ntl_receive);
        if (nl_sock == NULL)
            printk("netlink_kernel_create failed\n");
    }

    mgt_confirm_handler( ntl_confirm );
    
    return 0;
}

int mgt_cleanup( void )
{
    release_sock( nl_sock );
    
    return 0;
}

/*********************************************************************
*                        Netlink interface
*********************************************************************/

static int ntl_confirm( unsigned int src_id, unsigned int src_seq,
                        unsigned int dev_type, unsigned int dev_seq, 
                        unsigned int op, unsigned int oid, void *data, unsigned long len)
{
    return ntl_transmit(src_id, src_seq, op, oid, (long*)data, len, dev_seq );
}

static int ntl_transmit(pid_t pid, int seq, int operation, int oid, long *data, unsigned long data_len, 
                        int dev_id )
{
    unsigned char *old_tail;
    size_t size = 0;
    struct sk_buff *skb;
    struct nlmsghdr *nlh;
    struct pimfor_hdr *pimfor_hdr;
    int err = 0;
    unsigned int trpgrp = TRAPGRP_ETH_TRAPS | TRAPGRP_WLAN_TRAPS |TRAPGRP_BLOBDEV_TRAPS;

/*	
    blob.h              isil_netlink.h
    BIS_DT_802DOT11     TRAPGRP_ETH_TRAPS      
    BIS_DT_802DOT3      TRAPGRP_WLAN_TRAPS       
    DEVICE_BLOB         TRAPGRP_BLOBDEV_TRAPS          
*/

    size = NLMSG_LENGTH(data_len + sizeof(struct pimfor_hdr));
    skb = alloc_skb(size, GFP_ATOMIC);

    if ( skb ) {

        old_tail = skb->tail;
        nlh = NLMSG_PUT(skb, pid, seq, NETLINK_TYPE_PIMFOR, size - NLMSG_ALIGN(sizeof(struct nlmsghdr)));
        pimfor_hdr = (struct pimfor_hdr *)NLMSG_DATA(nlh);
        
        pimfor_encode_header(operation, oid, dev_id, 0 /*flags*/, data_len, pimfor_hdr);
        memcpy(PIMFOR_DATA(pimfor_hdr), data, data_len);
        
        nlh->nlmsg_len = skb->tail - old_tail;
        NETLINK_CB(skb).dst_groups = 0;
    } else {

        printk(KERN_ERR "ntl_transmit: alloc_skb failed\n");
        return -1;
    }

    if (pid != 0 ) { 
        err = netlink_unicast(nl_sock, skb, pid, MSG_DONTWAIT);
    } else { 
        netlink_broadcast(nl_sock, skb, pid, trpgrp ,GFP_ATOMIC );
    }
    
    return err;

nlmsg_failure:
    if ( skb )
            kfree_skb(skb);
    printk(KERN_ERR "ntl_transmit: NLMSG_PUT failed\n");
    return -1;
}

static void ntl_receive(struct sock *sk, int len)
{
    int err;
	struct sk_buff *skb;
    pid_t pid;
    struct nlmsghdr *nl_header; 
    __u32 seq;
    struct pimfor_hdr *header;
    char *data;
    int version;
    int operation;
    unsigned long oid;
    int dev_id;
    int flags;
    unsigned long length;
    
	do 
    {
        if (rtnl_shlock_nowait())
                return;
        
        while ( (skb = skb_dequeue(&sk->receive_queue)) != NULL ) 
        {
            pid = NETLINK_CB(skb).pid;
            nl_header = (struct nlmsghdr*) skb->data;
            seq = nl_header->nlmsg_seq;
            header = (struct pimfor_hdr*)(skb->data+(sizeof(struct nlmsghdr)));
            data = PIMFOR_DATA(header);

            if ( nl_header->nlmsg_type == NETLINK_TYPE_PIMFOR )    
            {   
                pimfor_decode_header(header, &version, &operation, &oid, &dev_id, &flags, &length);

                if (version == PIMFOR_VERSION_1) {
                    
                    err = mgt_request( DEV_NETWORK, dev_id, pid, seq, operation, oid, data, length );

                    if ( err < 0 ) {
                        netlink_ack(skb, nl_header, -1 );
                    }
    
                } else { 

                    netlink_ack(skb, nl_header, -1 );
                } 
            } else {

                netlink_ack(skb, nl_header, -1 );
            }

            kfree_skb(skb);
        }
        up(&rtnl_sem);
    
    } while (nl_sock && nl_sock->receive_queue.qlen);
}


/*********************************************************************
*                        PimFor header
*********************************************************************/

static int pimfor_encode_header(int operation, unsigned long oid, int device_id, 
                                int flags, int length, struct pimfor_hdr *header)
{
    if( !header )
            return -1;

    memset(header, 0, sizeof(struct pimfor_hdr));

    // byte oriented members
    header->version = PIMFOR_VERSION_1;
    header->operation = operation;
    header->device_id = device_id;
    header->flags = flags;

    // word oriented members with byte order depending on the flags
    if (flags & PIMFOR_FLAG_LITTLE_ENDIAN) {
            // use little endian coding
            header->oid = cpu_to_le32(oid);
            header->length = cpu_to_le32(length);
    } else {
            // use big endian coding
            header->oid = cpu_to_be32(oid);
            header->length = cpu_to_be32(length);
    }
    return 0;
}


static int pimfor_decode_header(struct pimfor_hdr *header, int *version,
                                int *operation, unsigned long *oid,
                                int *device_id, int *flags, unsigned long *length)
{
    if(!header)
            return -1;

    // byte oriented members
    *version   = header->version;
    *operation = header->operation;
    *device_id = header->device_id;
    *flags = header->flags;

    // word oriented members with byte order depending on the flags
    if (*flags & PIMFOR_FLAG_LITTLE_ENDIAN) {
            // use little endian coding
            *oid = le32_to_cpu(header->oid);
            *length = le32_to_cpu(header->length);
    } else {
            // use big endian coding
            *oid = be32_to_cpu(header->oid);
            *length = be32_to_cpu(header->length);
    }
    
    return 0;
}


