/******************************************************************************

    File:               isl38xx_pci.c
    Author:             W.Termorshuizen
    Version:            0.3.0.1
    Created:            June 19th, 2002
    Updated:            November 5th, 2002
    Company:            Intersil Americas Inc.
    Description:        This file contains the linux driver for the Intersil
                        Wireless LAN adapters.
    History:	    	Maintained in file isl38xx.h

******************************************************************************/
#define __KERNEL_SYSCALLS__

#define BLOB_V2

/* 
 * Mod. 1: define BLOB_V2 added
 * Mod. 2: isl38xx_upload_firmware: static buffer (stack overflow)
 * Mod. 3: pimfor_de(en)code_header: static functions 
 * Mod. 4: wait_for_oid_response: response copied to user buffer
 * 
*/

#include <linux/version.h>
#ifdef MODULE
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/wait.h>

#include <asm/processor.h>      // Processor type for cache alignment.
#include <asm/bitops.h>
#include <asm/io.h>

#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

#include <pcmcia/driver_ops.h>
#include <asm/uaccess.h>
#include <linux/unistd.h>
#include <linux/wireless.h>

#include <bloboid.h>      // additional types and defs for isl38xx fw
#include "prism_pci.h"
#include "prism_mgt.h"


#define DEBUG( f, args... ) 	K_DEBUG( f, pc_debug, args )

static void pimfor_encode_header( int, unsigned long, int, int, int, pimfor_header * );
static void pimfor_decode_header( pimfor_header *, int *, unsigned long *, int *,
    	    	    	    	int *, int * );

/******************************************************************************
        Global variable definition section
******************************************************************************/
#ifdef MODULE
MODULE_AUTHOR("Wim Termorshuizen <wim.termorshuizen@intersil.com>");
MODULE_DESCRIPTION("Intersil ISL38xx 802.11 Wireless LAN adapter");
MODULE_LICENSE("GPL");
#endif

const char *version =
    "version 0.3.0.1 05/11/2002 W.Termorshuizen, Intersil Americas Inc.\n";

int pc_debug = VERBOSE;
int init_mode = CARD_DEFAULT_MODE;
int init_channel = CARD_DEFAULT_CHANNEL;
int init_bsstype = CARD_DEFAULT_BSSTYPE;
int init_wep = CARD_DEFAULT_WEP;
int init_filter = CARD_DEFAULT_FILTER;
int init_wds = CARD_DEFAULT_WDS;
int init_authen = CARD_DEFAULT_AUTHEN;
int init_dot1x = CARD_DEFAULT_DOT1X;
char *init_ssid = NULL;

#ifdef MODULE
MODULE_PARM( pc_debug, "i" );
MODULE_PARM( init_mode, "i" );
MODULE_PARM( init_channel, "i" );
MODULE_PARM( init_bsstype, "i" );
MODULE_PARM( init_ssid, "s" );
MODULE_PARM( init_wep, "i" );
MODULE_PARM( init_filter, "i" );
MODULE_PARM( init_wds, "i" );
MODULE_PARM( init_authen, "i" );
MODULE_PARM( init_dot1x, "i" );
#endif

struct net_device *root_isl38xx_device = NULL;

// temporary variables
int errno;

#if TARGET_SYSTEM == CONFIG_PCMCIA

struct driver_operations isl38xx_ops =
{
    DRIVER_NAME, isl38xx_attach, isl38xx_suspend, isl38xx_resume,
        isl38xx_detach
};

#else

#define PCI_TYPE        (PCI_USES_MEM | PCI_ADDR0 | PCI_NO_ACPI_WAKE)

// PCI Class & Sub-Class code, Network-'Other controller'
#define PCI_CLASS_NETWORK_OTHERS 0x280

struct pci_device_id pci_id_table[] =
{
    {
        PCIVENDOR_INTERSIL, PCIDEVICE_ISL3877,
        PCI_ANY_ID, PCI_ANY_ID,
        0, 0,
        // Driver data, we just put the name here
        (unsigned long) "Intersil PRISM Indigo Wireless LAN adapter"
    },
    {
        PCIVENDOR_INTERSIL, PCIDEVICE_ISL3890,
        PCI_ANY_ID, PCI_ANY_ID,
        0, 0,
        // Driver data, we just put the name here
        (unsigned long) "Intersil PRISM Duette Wireless LAN adapter"
    },
    {
        0, 0, 0, 0, 0, 0, 0
    }
};

#if TARGET_SYSTEM == CONFIG_HOTPLUG

// register the device with the Hotplug facilities of the kernel
MODULE_DEVICE_TABLE( pci, pci_id_table );

#endif	// CONFIG_HOTPLUG


struct pci_driver isl38xx_pci_drv_id =
{
    {},
    DRIVER_NAME,                // Driver name
    pci_id_table,               // id table
    isl38xx_probe_pci,          // probe function
    NULL,                       // remove function
    NULL,                       // save_state function
    NULL,                       // suspend function
    NULL,                       // resume function
    NULL,                       // enable_wake function
};

#endif

static int mgtc_indication( unsigned int dev_type, unsigned int dev_seq, 
                            unsigned int src_id, unsigned int src_seq,
                            unsigned int op, unsigned int oid, void *data, unsigned long len);


/******************************************************************************
    Upper data link 
******************************************************************************/

int isl38xx_transmit(struct sk_buff *skb, struct net_device *nwdev)
{
    isl38xx_private *private_config = nwdev->priv;
    isl38xx_control_block *control_block = private_config->control_block;
    isl38xx_fragment *fragment;
    queue_entry *entry;
    u32 in_queue, index, reg, counter;
    int fragments;
    int frame_size;
    struct sk_buff *newskb;
    int newskb_offset;


    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_transmit \n");

    // check whether the deviec is running
//      if( !netif_running( nwdev ) )
//      {
//              DEBUG( SHOW_ERROR_MESSAGES, "%s: Tx on stopped device!\n",
//                     nwdev->name);
//              return 1;
//      }

//      tx_timeout_check( nwdev, isl38xx_tx_timeout );
//      skb_tx_check( nwdev, skb );

    // determine the amount of fragments needed to store the frame
    frame_size = skb->len < ETH_ZLEN ? ETH_ZLEN : skb->len;
    if( init_wds ) frame_size += 6;
    fragments = frame_size / MAX_FRAGMENT_SIZE;
    fragments += frame_size % MAX_FRAGMENT_SIZE == 0 ? 0 : 1;

    DEBUG(SHOW_TRACING, "Fragments needed for frame %i\n", fragments);

    // check whether shadow queue entries may be returned to the freeq and
    // hosted sk_buff structures can be freed
    while (1)
    {
        // whouou, an endless loop, make sure I can get out !
        if (queue_size(private_config->remapped_device_base,
            &private_config->data_low_tx_shadowq) == 0)
            // nothing in the shadow queue
            break;

        // an entry in the shadow queue, get a pointer to it through root
        entry = &private_config->data_low_tx_shadowq;
        entry = (queue_entry *) entry->next;

        // get the number of control block fragments not transmitted yet
        in_queue = isl38xx_in_queue(control_block, ISL38XX_CB_TX_DATA_LQ);

        // check whether the number of transmitted fragments exceed the number
        // of fragments in the first queue entry
        if (private_config->data_fragments_in_wait - in_queue >=
            entry->fragments)
        {
            // the fragments of the entry have been transmitted
            // get the entry from the shadow queue with the skb
            get_queue(private_config->remapped_device_base,
                &private_config->data_low_tx_shadowq, &entry);

            // delete the streaming DMA mapping before freeing the skb
            pci_unmap_single( private_config->pci_device,
                (dma_addr_t) entry->dev_address, entry->size,
                PCI_DMA_TODEVICE );

            // free the skbuf structure hosted by the queue entry
            dev_kfree_skb((struct sk_buff *) entry->host_address);

            // put the queue entry back in the free queue
            put_queue(private_config->remapped_device_base,
                &private_config->data_tx_freeq, entry);

            // update the fragments in wait counter
            private_config->data_fragments_in_wait -= fragments;
        }
        else
        {
            // not all fragments of this entry have been transmitted yet
            DEBUG(SHOW_TRACING, "Outstanding fragments %i\n",
                private_config->data_fragments_in_wait - in_queue);
            break;
        }
    }

    // check whether the destination queue has enough fragments for the frame
    in_queue = isl38xx_in_queue(control_block, ISL38XX_CB_TX_DATA_LQ);
    if (ISL38XX_CB_TX_QSIZE - in_queue < fragments)
    {
        // Error, cannot add the frame because the queue is full
        DEBUG(SHOW_ERROR_MESSAGES, "Error: Queue [%i] not enough room\n",
        ISL38XX_CB_TX_DATA_LQ);
        return -EBUSY;
    }

    // get a free entry from the data tx queue for skbuf hosting
    if (get_queue(private_config->remapped_device_base,
        &private_config->data_tx_freeq, &entry) != 0)
    {
        // could not get a header, the queue is full
        return -EBUSY;
    }

    // Check alignment and WDS frame formatting. The start of the packet should
    // be aligned on a 4-byte boundary. If WDS is enabled add another 6 bytes
    // and add WDS address information
    if( ((int) skb->data & 0x03) | init_wds )
    {
/*
        // recalculate the number of bytes to add and re-allign
        offset += init_wds ? 6 : 0;
        offset += (((int) skb->data) + offset) & 0x03;

        DEBUG(SHOW_TRACING, "skb offset %i wds %i\n", offset, init_wds );

        if (!skb_cloned(skb) && (skb_tailroom(skb) >= offset))
        {
            unsigned char *src = skb->data;

            skb_reserve(skb, offset);
            if( init_wds )
            {
                memmove( skb->data+6, src, skb->len );
                memcpy( skb->data,
                    private_config->wds_link_table[0].address, 6 );
            }
            else
                memmove( skb->data, src, skb->len );

            DEBUG(SHOW_TRACING, "memmove %p %p %i \n", skb->data, src,
                skb->len );
        }
        else
        {
*/
            newskb = dev_alloc_skb( init_wds ? skb->len+6 : skb->len );
            newskb_offset = (int) newskb->data & 0x03;

            // Check if newskb->data is aligned
            if( newskb_offset )
                skb_reserve(newskb, newskb_offset);

            skb_put(newskb, init_wds ? skb->len+6 : skb->len );
            if( init_wds )
            {
                memcpy( newskb->data+6, skb->data, skb->len );
                memcpy( newskb->data,
                    private_config->wds_link_table[0].address, 6 );
            }
            else
                memcpy( newskb->data, skb->data, skb->len );

            DEBUG(SHOW_TRACING, "memcpy %p %p %i wds %i\n", newskb->data,
                skb->data, skb->len, init_wds );

            newskb->dev = skb->dev;
            dev_kfree_skb(skb);
            skb = newskb;
//        }
    }

    // set the host address to skb start address for freeing it later, the
    // dev address to the start of actual data and the size indicator to the
    // number of fragments in the skb
    entry->host_address = (u32) skb;
        entry->size = skb->len;
    entry->fragments = fragments;

    // set the streaming DMA mapping for proper PCI bus operation
    entry->dev_address = (u32) pci_map_single( private_config->pci_device,
        (void *) skb->data, skb->len, PCI_DMA_TODEVICE );
    if( entry->dev_address == (u32) NULL )
    {
        // error mapping the buffer to device accessable memory address
        DEBUG(SHOW_ERROR_MESSAGES, "Error mapping DMA address, data lost\n");

        // free the skbuf structure before aborting
        dev_kfree_skb((struct sk_buff *) entry->host_address);
            return -EIO;
    }

    // display the buffer contents for debugging
    DEBUG( SHOW_BUFFER_CONTENTS, "\ntx %p ", skb->data );
    display_buffer((char *) skb->data, skb->len );

    // setup the fragments to point to the separate frame parts
    for (counter = 0; counter < fragments; counter++,
        frame_size -= MAX_FRAGMENT_SIZE)
    {
        // get a pointer to the control block destination fragment using the
        // frame counter and link it with the skbuf data block
        index = le32_to_cpu(
            control_block->driver_curr_frag[ISL38XX_CB_TX_DATA_LQ]) %
            ISL38XX_CB_TX_QSIZE;
        fragment = &control_block->tx_data_low[index];
        fragment->address = cpu_to_le32(entry->dev_address + counter *
            MAX_FRAGMENT_SIZE);
        fragment->size = cpu_to_le16( frame_size > MAX_FRAGMENT_SIZE ?
            MAX_FRAGMENT_SIZE : frame_size );
    }

    // store the entry in the tx shadow queue and increment the index pointer
    put_queue(private_config->remapped_device_base,
        &private_config->data_low_tx_shadowq, entry);
    add_le32p(&control_block->driver_curr_frag[ISL38XX_CB_TX_DATA_LQ],
        fragments);
    private_config->data_fragments_in_wait += fragments;

    // check whether the data tx queue is full now
    in_queue = isl38xx_in_queue(control_block, ISL38XX_CB_TX_DATA_LQ);
    if (ISL38XX_CB_TX_QSIZE - in_queue < ISL38XX_MIN_QTHRESHOLD)
    {
        // stop the transmission of network frames to the driver
        netif_stop_queue(nwdev);

        // set the full flag for the data low transmission queue
        private_config->data_low_tx_full = 1;
    }

    // trigger the device by setting the Update bit in the Device Int register
    reg = readl(private_config->remapped_device_base + ISL38XX_DEV_INT_REG);
    reg |= ISL38XX_DEV_INT_UPDATE;
    writel(reg, private_config->remapped_device_base + ISL38XX_DEV_INT_REG);
    udelay(ISL38XX_WRITEIO_DELAY);

    // set the transmission time
    nwdev->trans_start = jiffies;
    private_config->statistics.tx_packets++;
    private_config->statistics.tx_bytes += skb->len;
    return 0;
}

int isl38xx_receive(isl38xx_private * private_config)
{
    struct net_device *nw_device = private_config->my_module;
    isl38xx_control_block *control_block = private_config->control_block;
    queue_entry *entry;
    struct sk_buff *skb;
    u16 size;
    u32 index, reg, offset;
    unsigned char *src;

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_receive \n");

    // get the first entry in the queue hosting the sk_buff
    get_queue(private_config->remapped_device_base,
      &private_config->data_low_rx_shadowq, &entry);

    // the device has written an Ethernet frame in the data area
    // of the sk_buff without updating the structure, do it now
    size = le16_to_cpu(control_block->rx_data_low[private_config->index_data_low].size);
    skb = (struct sk_buff *) entry->host_address;
    offset = ((u32)
    	control_block->rx_data_low[private_config->index_data_low].address -
	(u32) skb->data ) & 3;

    DEBUG( SHOW_TRACING, "frq->addr %x skb->data %p offset %u truesize %u\n ",
        control_block->rx_data_low[private_config->index_data_low].address,
	skb->data, offset, skb->truesize );

    skb_put( skb, size );
    if( offset )
    {
	// shift the buffer allocation offset bytes to get the right frame
        skb_pull( skb, 2 );
	skb_put( skb, 2 );
    }

    // display the buffer contents for debugging
    DEBUG( SHOW_BUFFER_CONTENTS, "\nrx %p ", skb->data );
    display_buffer((char *) skb->data, skb->len );

    // check whether WDS is enabled and whether the data frame is a WDS frame
    if( init_wds )
    {
        // WDS enabled, check for the wds address on the first 6 bytes of the buffer
        if( address_compare( (unsigned char *) skb->data,
            (unsigned char *) private_config->wds_link_table[0].address ))
        {
            // addresses match, remove the WDS address from the frame
            src = skb->data+6;
            memmove( skb->data, src, skb->len-6 );
            skb_trim( skb, skb->len-6 );
        }
    }

    // delete the streaming DMA mapping before passing the skb
    pci_unmap_single( private_config->pci_device, (dma_addr_t) entry->dev_address,
        entry->size, PCI_DMA_FROMDEVICE );

    DEBUG(SHOW_TRACING, "Fragment size %i in skb at %p\n", size, skb);
    DEBUG(SHOW_TRACING, "Skb data at %p, length %i\n", skb->data, skb->len);

    // display the buffer contents for debugging
    DEBUG( SHOW_BUFFER_CONTENTS, "\nrx %p ", skb->data );
    display_buffer((char *) skb->data, skb->len );

    // do some additional sk_buff and network layer parameters
    skb->dev = nw_device;
    skb->protocol = eth_type_trans(skb, nw_device);
    skb->ip_summed = CHECKSUM_NONE;
    private_config->statistics.rx_packets++;
    private_config->statistics.rx_bytes += size;

    // deliver the skb to the network layer
    netif_rx(skb); 

    // free the entry by returning it to the freeq
    put_queue(private_config->remapped_device_base,
        &private_config->data_rx_freeq, entry);

    // increment and wrap the read index for the rx data low queue
    private_config->index_data_low = (private_config->index_data_low + 1) %
        ISL38XX_CB_RX_QSIZE;

    // add one or more entries hosting an sk_buff structure
    while (queue_size(private_config->remapped_device_base,
        &private_config->data_low_rx_shadowq) != ISL38XX_CB_RX_QSIZE)
    {
        // get an entry from the freeq and put it in the shadow queue
        if (get_queue(private_config->remapped_device_base,
            &private_config->data_rx_freeq, &entry))
        {
            // Oops, no more entries in the freeq ?
            DEBUG(SHOW_ERROR_MESSAGES, "Wow: data_rx_freeq empty !\n");
            break;
        }
        else
        {
            // allocate an sk_buff for received data frames storage
	    // include any required allignment operations
            if (skb = dev_alloc_skb(MAX_FRAGMENT_SIZE+2), skb == NULL)
            {
                // error allocating an sk_buff structure elements
                DEBUG(SHOW_ERROR_MESSAGES, "Error allocating skb \n");

                // free the entry by returning it to the freeq
                put_queue(private_config->remapped_device_base,
                    &private_config->data_rx_freeq, entry);

                break;
            }

            // now set the members in the entry to the skbuf structure
            entry->host_address = (u32) skb;
            entry->size = skb->len;
            entry->fragments = 1;

            // set the streaming DMA mapping for proper PCI bus operation
            entry->dev_address = (u32) pci_map_single(
                private_config->pci_device,     (void *) skb->data, skb->len,
                PCI_DMA_FROMDEVICE );
            if( entry->dev_address == (u32) NULL )
            {
                // error mapping the buffer to device accessable memory address
                DEBUG(SHOW_ERROR_MESSAGES, "Error mapping DMA address\n");

                // free the entry by returning it to the freeq
                put_queue(private_config->remapped_device_base,
                    &private_config->data_rx_freeq, entry);

                // free the skbuf structure before aborting
                dev_kfree_skb((struct sk_buff *) entry->host_address);

                break;
            }

            // use the entry for the device interface management rx queue
            // these should always be inline !
            index = le32_to_cpu(
                control_block->driver_curr_frag[ISL38XX_CB_RX_DATA_LQ]) %
                ISL38XX_CB_RX_QSIZE;
            control_block->rx_data_low[index].address =
                cpu_to_le32(entry->dev_address);

            // put the entry got from the freeq in the shadow queue
            put_queue(private_config->remapped_device_base,
                &private_config->data_low_rx_shadowq, entry);

            // increment the driver read pointer
            add_le32p(&control_block->driver_curr_frag[ISL38XX_CB_RX_DATA_LQ],
                1 );
        }
    }

    // trigger the device by setting the Update bit in the Device Int register
    reg = readl(private_config->remapped_device_base + ISL38XX_DEV_INT_REG);
    reg |= ISL38XX_DEV_INT_UPDATE;
    writel(reg, private_config->remapped_device_base + ISL38XX_DEV_INT_REG);
    udelay(ISL38XX_WRITEIO_DELAY);

    return 0;
}

void isl38xx_tx_timeout(struct net_device *nwdev)
{
    isl38xx_private *private_config = nwdev->priv;
    struct net_device_stats *statistics = &private_config->statistics;
    int error = 0;

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_tx_timeout \n");

    // increment the transmit error counter
    statistics->tx_errors++;

    // reset the device and restore its configuration
    if (error = isl38xx_reset(private_config), error != 0)
    {
        DEBUG(SHOW_ERROR_MESSAGES, "Error %d on TX timeout card reset!\n",
            error);
    }
    else
    {
        DEBUG(SHOW_ERROR_MESSAGES, "Driver stopped TX/RX ......\n");

//        nwdev->trans_start = jiffies;
//        netif_wake_queue( nwdev );
    }

    return;
}


/******************************************************************************
    Upper management link
******************************************************************************/

DECLARE_WAIT_QUEUE_HEAD(mgmt_queue_wait);

static int mgtc_indication( unsigned int dev_type, unsigned int dev_seq, 
                            unsigned int src_id, unsigned int src_seq,
                            unsigned int op, unsigned int oid, void *data, unsigned long len)
{
    struct net_device *dev;
	isl38xx_private *private_config;
    unsigned long dlen;

    printk("mgtc_indication(%d, %d, %d, %d, %d, %x, %x, %d)\n",
       (int)dev_type, (int)dev_seq, (int)src_id, (int)src_seq, (int)op, (int)oid, (int)data, (int)len );

    if ( ! (dev = dev_get_by_index(dev_seq)) ) 
        return -1;

    private_config = dev->priv;

    if ( (isl38xx_queue_pimfor(private_config, op, oid, 0, data, len, 0)) < 0 ) 
        return -1;

    if ( (wait_for_oid_response(private_config, oid, (void *)data, &dlen )) < 0 )
        return -1;
    
    if ( (mgt_response( src_id, src_seq, dev_type, dev_seq, op, oid, data, dlen)) < 0)
        return -1;

    return 0;
}

int isl38xx_transmit_pimfor(isl38xx_private * private_config)
{
    int fragments, counter;
    isl38xx_control_block *control_block = private_config->control_block;
    isl38xx_fragment *fragment;
    u32 index, reg, in_queue, in_shadowq;
    queue_entry *entry;

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_transmit_pimfor \n");

    // check whether there is a pending response in the receive queue which
    // needs to be handled first before applying a new frame
    if( in_queue = isl38xx_in_queue(control_block, ISL38XX_CB_RX_MGMTQ),
        in_queue != 0 )
    {
        // receive queue not empty
        DEBUG(SHOW_TRACING, "receive queue not empty\n");
        return 0;
    }

        // read the number of entries in the transmit queue and check whether the
        // transmit queue is empty for receiving the next pimfor frame
    if( in_queue = isl38xx_in_queue(control_block, ISL38XX_CB_TX_MGMTQ),
        in_queue != 0 )
    {
        // transmit queue not empty
        DEBUG(SHOW_TRACING, "transmit queue not empty\n");
        return 0;
    }

        // first check whether a clean up should take place on the shadow queue
    while( private_config->index_mgmt_tx > in_queue )
    {
        // free the first entry in the shadow queue which contains the
        // PIMFOR frame transmitted to the device
        get_queue(private_config->remapped_device_base,
                &private_config->mgmt_tx_shadowq, &entry);
        put_queue(private_config->remapped_device_base,
                &private_config->mgmt_tx_freeq, entry);

        // decrement the real management index
        private_config->index_mgmt_tx--;
    }

    // check whether there is at least one frame in the shadowq
    if( in_shadowq = queue_size( private_config->remapped_device_base,
        &private_config->mgmt_tx_shadowq ), !in_shadowq )
    {
        // no entries in the shadowq, quit
        DEBUG(SHOW_TRACING, "shadowq empty, no more frames\n");
        return 0;
    }

    // there is a frame in the shadowq, get its fragments count
    entry = (queue_entry *) &private_config->mgmt_tx_shadowq;
    entry = (queue_entry *) entry->next;
    fragments = entry->fragments;

    // transmit the fragment(s) by placing them in the transmit queue
    for (counter = 0; counter < fragments; counter++)
    {
        u16 fragment_flags;

        // determine the fragment flags depending on the nr of fragments
        fragment_flags = fragments - counter > 1 ? FRAGMENT_FLAG_MF : 0;

        // get a pointer to the destination fragment using the frame counter
        // and link it with the free queue entry data block
        index = le32_to_cpu(control_block->driver_curr_frag[ISL38XX_CB_TX_MGMTQ]) %
                ISL38XX_CB_MGMT_QSIZE;
        fragment = &control_block->tx_data_mgmt[index];

        DEBUG(SHOW_TRACING, "Entry address %p\n", entry);
        DEBUG(SHOW_TRACING, "Index in transmit queue %i\n", index);
        DEBUG(SHOW_TRACING, "Fragment address %p\n", fragment);

        fragment->address = cpu_to_le32(entry->dev_address);
        fragment->size = cpu_to_le16(entry->size);
        fragment->flags = cpu_to_le16(fragment_flags);

        // increment the index pointer, wrapping is not done on queue but
        // on variable type range
        add_le32p(&control_block->driver_curr_frag[ISL38XX_CB_TX_MGMTQ], 1);

        // increment the queue entry pointer
        entry = (queue_entry *) entry->next;
    }

    DEBUG(SHOW_TRACING, "Frame with [%i] fragments added to transmit queue \n",
      fragments);

    // increment the read management queue index
    private_config->index_mgmt_tx += fragments;

    // trigger the device by setting the Update bit in the Device Int register
    reg = readl(private_config->remapped_device_base + ISL38XX_DEV_INT_REG);
    reg |= ISL38XX_DEV_INT_UPDATE;
    writel(reg, private_config->remapped_device_base + ISL38XX_DEV_INT_REG);
    udelay(ISL38XX_WRITEIO_DELAY);

    return 0;
}

int isl38xx_queue_pimfor(isl38xx_private * private_config,
                                  int operation, unsigned long oid,
                                  int flags, void *data, int length,
                                  int use_tunnel_oid)
{
    int fragments, counter;
    u32 source, destination;
    queue_entry *entry;

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_queue_pimfor \n");

    // determine the amount of fragments needed for the frame in case of a SET
    if (use_tunnel_oid)
    {
        fragments = (length + 2 * PIMFOR_HEADER_SIZE) / MGMT_FRAME_SIZE;
        if ((length + 2 * PIMFOR_HEADER_SIZE) % MGMT_FRAME_SIZE != 0)
            fragments++;
    }
    else
    {
        fragments = (length + PIMFOR_HEADER_SIZE) / MGMT_FRAME_SIZE;
        if ((length + PIMFOR_HEADER_SIZE) % MGMT_FRAME_SIZE != 0)
        fragments++;
    }

    // check whether data is actually written to the device
    if (operation == PIMFOR_OP_GET)
    {
        // operation is get, no data is actually written, reset the fragments
        // counter to send a single fragment
        fragments = 1;
    }

    // check whether the number of fragments doesn't exceed the maximal possible
    // amount of frames in the transmit queue
    if( fragments > ISL38XX_CB_MGMT_QSIZE )
    {
        // Oops, the fragment count exceeds the maximal value
        DEBUG(SHOW_ERROR_MESSAGES, "Error: Frame with [%i] fragments is too "
            "large \n", fragments );
        return -1;
    }

    // check whether there is enough room for the fragment(s) in the freeq
    if( queue_size( private_config->remapped_device_base,
        &private_config->mgmt_tx_freeq ) < fragments )
    {
        // Oops, the management freeq does not have enough entries
        DEBUG(SHOW_ERROR_MESSAGES, "Error: Management FreeQ not enough entries "
            "for [%i] fragments \n", fragments );
        return -1;
    }

    // queue the frame(s) in the transmit shadow queue
    for (counter = 0, source = (u32) data; counter < fragments; counter++)
    {
        u16 fragment_size = 0;

        // get an entry from the free queue for hosting the frame
        get_queue(private_config->remapped_device_base,
            &private_config->mgmt_tx_freeq, &entry);

        // set the destination address to the entry data block for building
        // the PIMFOR frame there
        destination = entry->host_address;

        DEBUG(SHOW_TRACING, "Entry address %p\n", entry);
        DEBUG(SHOW_TRACING, "Destination address %x\n", destination);
        DEBUG(SHOW_TRACING, "Destination for device %x\n", entry->dev_address);

        // place the PIMFOR header in the first fragment
        if (counter == 0)
        {
            // check whether the frame should have a tunnel oid
            if (use_tunnel_oid)
            {
                // start with a tunnel oid header
                pimfor_encode_header(PIMFOR_OP_SET, OID_INL_TUNNEL,
                     PIMFOR_DEV_ID_MHLI_MIB, flags,
                     length + PIMFOR_HEADER_SIZE,
                     (pimfor_header *) destination);

                // set the fragment size and increment the destin pointer
                fragment_size += PIMFOR_HEADER_SIZE;
                destination += PIMFOR_HEADER_SIZE;
            }

            // create the header directly in the fragment data area
            pimfor_encode_header(operation, oid, PIMFOR_DEV_ID_MHLI_MIB,
                 flags, length, (pimfor_header *) destination);

            // display the buffer contents for debugging
            display_buffer((char *) destination, 12);

            // set the fragment size and increment the destin pointer
            fragment_size += PIMFOR_HEADER_SIZE;
            destination += PIMFOR_HEADER_SIZE;
        }

        // add the data to the fragment in case of a set or for some
		// specific objects
        if( (operation == PIMFOR_OP_SET) || (oid == DOT11_OID_STAKEY) )
        {
            int size = (length > (MGMT_FRAME_SIZE - fragment_size)) ?
                MGMT_FRAME_SIZE - fragment_size : length;
            memcpy((void *) destination, (void *) source, size);

            // set the fragment size and flag members
            fragment_size += size;

            // display the buffer contents for debugging
            display_buffer((char *) destination, 12);

            // update the source pointer and the data length
            source += size;
            length -= size;
        }

        // save the number of fragments still to go in the frame together with
        // the fragment size
        entry->fragments = fragments - counter;
        entry->size = fragment_size;

        // store the entry in the tx shadow queue
        put_queue(private_config->remapped_device_base,
            &private_config->mgmt_tx_shadowq, entry);
    }

    DEBUG(SHOW_TRACING, "Frame with [%i] fragments added to shadow tx queue \n",
        fragments);

    return 0;
}

int isl38xx_receive_pimfor(isl38xx_private * private_config)
{
    struct net_device *nw_device = private_config->my_module;
    isl38xx_control_block *control_block = private_config->control_block;
    pimfor_header *received_header;
    char *received_data;
    int device_id, operation, flags, length;
    unsigned long oid;
    queue_entry *entry, *new_entry;
    u32 reg;

#if EVENT_SUPPORT == WE_EVENTS
    union iwreq_data wrq;
#endif

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_receive_pimfor \n");

    // get the received response from the shadow queue
    get_queue(private_config->remapped_device_base,
                &private_config->mgmt_rx_shadowq, &entry);

    DEBUG(SHOW_TRACING, "Entry address 0x%p\n", entry);
    DEBUG(SHOW_TRACING, "Entry device address 0x%x\n", entry->dev_address);

    // determine how the process the fragment depending on the flags in the
    // PIMFOR header, check for tunnel oid PIMFOR frames
    received_header = (pimfor_header *) entry->host_address;
    if (pimfor_decode_header(received_header, &operation, &oid, &device_id,
        &flags, &length), oid == OID_INL_TUNNEL)
    {
        DEBUG(SHOW_PIMFOR_FRAMES, "PIMFOR: Tunnel OID frame received \n");

        // PMFOR frame in fragment has the tunnel oid, redo the header decode
        received_header = (pimfor_header *) (entry->host_address +
            PIMFOR_HEADER_SIZE);
        pimfor_decode_header(received_header, &operation, &oid, &device_id,
            &flags, &length);
    }

    DEBUG(SHOW_PIMFOR_FRAMES,
        "PIMFOR: op %i, oid 0x%lx, device %i, flags 0x%x length 0x%x \n",
        operation, oid, device_id, flags, length);

    // display the buffer contents for debugging
    display_buffer((char *) received_header, sizeof(pimfor_header));
    display_buffer(((char *) received_header) + PIMFOR_HEADER_SIZE, length );

    // PIMFOR frame created by this driver, handle its contents
    // check whether the frame contains driver initialization data
    if (oid == GEN_OID_MACADDRESS)
    {
        // a MAC address response message, store its contents
        if ((operation == PIMFOR_OP_RESPONSE) && (length == 6))
        {
            // copy the MAC address received
            received_data = (char *) received_header + PIMFOR_HEADER_SIZE;
            memcpy(nw_device->dev_addr, received_data, 6);
        }
    }
    else if (oid == OID_INL_MODE)
    {
        // the mode response message, store its contents
        if ((operation == PIMFOR_OP_RESPONSE) && (length == 4))
        {
            // copy the low byte which is at the first byte location
            received_data = (char *) received_header + PIMFOR_HEADER_SIZE;
            private_config->mode = (int) *received_data;
        }
    }
    else if (oid == GEN_OID_LINKSTATE)
    {
        // the mode response message, store its contents
        if ((operation == PIMFOR_OP_RESPONSE) && (length == 4))
        {
            // copy the low byte which is at the first byte location
            received_data = (char *) received_header + PIMFOR_HEADER_SIZE;
            private_config->linkstate = (int) *received_data;
        }
    }

    // perform a check on the freeq whether the entry can be queued to other
    // queues or must be hold to make sure there are enough
    if( get_queue(private_config->remapped_device_base,
        &private_config->mgmt_rx_freeq, &new_entry) == 0)
    {
        // succeeded in getting a new entry from the freeq
        // refresh the device mgmt queue entry
        control_block->rx_data_mgmt[private_config->index_mgmt_rx].address =
            cpu_to_le32(new_entry->dev_address);

        // put the new entry in the mgmt rx shadow queue
        put_queue(private_config->remapped_device_base,
                &private_config->mgmt_rx_shadowq, new_entry);

        // deliver the held entry to the appropriate queue
        if (flags & PIMFOR_FLAG_APPLIC_ORIGIN)
        {
            // PIMFOR frame originated by higher application, send the entire
            // fragment up without processing

            DEBUG(SHOW_TRACING, "PIMFOR Application Originated \n");

            // send the entry to the PIMFOR receive queue for further processing
            put_queue(private_config->remapped_device_base,
                &private_config->pimfor_rx_queue, entry);
        }
        else
        {
            // check whether the frame concerns a response, error or trap
            if (operation == PIMFOR_OP_TRAP)
            {
                // received a trap frame
                received_data = (char *) received_header + PIMFOR_HEADER_SIZE;

                DEBUG(SHOW_TRAPS,
                    "TRAP: oid 0x%lx, device %i, flags 0x%x length %i\n",
                    oid, device_id, flags, length);

                // check whether the TRAP should be deleted or queued depending
                // on the queue size and whether the queue is read
                if( queue_size( private_config->remapped_device_base,
                    &private_config->trap_rx_queue ) > MAX_TRAP_RX_QUEUE )
                {
                    // free the entry to the freeq
                    put_queue(private_config->remapped_device_base,
                            &private_config->mgmt_rx_freeq, entry);
                }
                else
                {
                    // send the entry to the TRAP receive q
                    put_queue(private_config->remapped_device_base,
                            &private_config->trap_rx_queue, entry);

#if EVENT_SUPPORT == WE_EVENTS
    	    	    // signal user space application(s) that a TRAP is queued
		    // use the AP MAC address get ioctl for signalling
		    wrq.ap_addr.sa_family = ARPHRD_ETHER;
		    memcpy( wrq.ap_addr.sa_data, nw_device->dev_addr, 6 );
		    wireless_send_event( nw_device, SIOCGIWAP, &wrq, "ISIL" );
#elif EVENT_SUPPORT == ISL_EVENTS
			mgmt_handle_trap();

#endif // EVENT_SUPPORT
                }
            }
            else
            {
                // received either a repsonse or an error
                // send the entry to the IOCTL receive q for further processing
                put_queue(private_config->remapped_device_base,
                        &private_config->ioctl_rx_queue, entry);

                // signal the waiting procs that a response has been received
                wake_up_interruptible(&mgmt_queue_wait);
            }
        }
    }
    else
    {
        // not succeeded, use the held one and drop its contents
        // refresh the device mgmt queue entry
        control_block->rx_data_mgmt[private_config->index_mgmt_rx].address =
            cpu_to_le32(entry->dev_address);

        // put the old entry back in the mgmt rx shadow queue
        put_queue(private_config->remapped_device_base,
            &private_config->mgmt_rx_shadowq, entry);
    }

    // increment the driver read pointer
    add_le32p(&control_block->driver_curr_frag[ISL38XX_CB_RX_MGMTQ], 1);

    // trigger the device by setting the Update bit in the Device Int register
    reg = readl(private_config->remapped_device_base + ISL38XX_DEV_INT_REG);
    reg |= ISL38XX_DEV_INT_UPDATE;
    writel(reg, private_config->remapped_device_base + ISL38XX_DEV_INT_REG);
    udelay(ISL38XX_WRITEIO_DELAY);

    // increment and wrap the read index for the rx management queue
    private_config->index_mgmt_rx = (private_config->index_mgmt_rx + 1) %
        ISL38XX_CB_MGMT_QSIZE;

    return 0;
}

int wait_for_oid_response(isl38xx_private * private_config, unsigned long oid,
                                 void *data, unsigned long *len )
{
    queue_entry *entry;
    pimfor_header *received_header;
    int q_size, counter, device_id, operation, flags, rlength, rvalue = 0;
    unsigned long received_oid;

    DECLARE_WAITQUEUE(wait, current);

    // check for other processes already waiting for response
    if (private_config->processes_in_wait == 0)
    {
        // no process in the wait queue, flush the queue
        while (queue_size(private_config->remapped_device_base,
            &private_config->ioctl_rx_queue))
        {
            // flush each entry remaining in the queue
            get_queue(private_config->remapped_device_base,
                &private_config->ioctl_rx_queue, &entry);
            put_queue(private_config->remapped_device_base,
                &private_config->mgmt_rx_freeq, entry);
        }
    }

    // trigger the transmission of the management shadowq before going to sleep
    isl38xx_transmit_pimfor( private_config );

    // add the process to the wait queue and increment the process counter
    add_wait_queue(&mgmt_queue_wait, &wait);
    current->state = TASK_INTERRUPTIBLE;
    private_config->processes_in_wait++;

    DEBUG(SHOW_TRACING, "Process enter sleep mode\n");

    while (1)
    {
        // go to sleep with the alarm bells on sharp
        interruptible_sleep_on_timeout(&mgmt_queue_wait, ISL38XX_WAIT_CYCLE);

        // hello there, I got woken up because there is either a response out
        // there or the bells went off, check whether the lock is up
        while (private_config->ioctl_queue_lock)
        {
            // the lock is up, another process is reading the ioctl rx queue
            schedule_timeout(1 * HZ / 10);
        }

        // It's my turn, set the lock rightaway
        private_config->ioctl_queue_lock = 1;

        // check the IOCTL receive queue for responses
        if( q_size = queue_size(private_config->remapped_device_base,
            &private_config->ioctl_rx_queue), q_size != 0 )
        {
            // something in the queue, check every entry in the queue
            for (counter = 0; counter < q_size; counter++)
            {
                // get the entry from the queue and read the header
                get_queue(private_config->remapped_device_base,
                    &private_config->ioctl_rx_queue, &entry);
                received_header = (pimfor_header *) entry->host_address;

                // check whether the oid response is for me
                if( pimfor_decode_header(received_header, &operation,
                    &received_oid, &device_id, &flags,&rlength ),
                    oid == received_oid )
                {
                    DEBUG(SHOW_TRACING, "IOCTL response found in queue.\n");

                    // found my response, check the response type
                    if( operation == PIMFOR_OP_RESPONSE )
                    {
                        // return the address of the data field
                        if (rlength < *len ) 
                            *len = rlength;
                        memcpy( data, ((void *) received_header + PIMFOR_HEADER_SIZE), *len );

                        DEBUG(SHOW_TRACING, "Response source 0x%p\n", data);
                    }
                    else
                    {
                        // error or not supported operation
                        rvalue = -ENOTSUPP;
                    }

                    // free the entry
                    put_queue(private_config->remapped_device_base,
                        &private_config->mgmt_rx_freeq, entry);

                    // continue and leave the sleepingroom
                    goto leave;
                }

                // put the entry back in the queue
                put_queue(private_config->remapped_device_base,
                    &private_config->ioctl_rx_queue, entry);
            }

            // checked all entries in the queue without success
            rvalue = -ETIMEDOUT;
            goto leave;
        }
        else
        {
            // nothing in the queue, that's got to be a timeout
            rvalue = -EBUSY;
            goto leave;
        }
    }

leave:
    private_config->processes_in_wait--;
    private_config->ioctl_queue_lock = 0;
    current->state = TASK_RUNNING;
    remove_wait_queue(&mgmt_queue_wait, &wait);
    return rvalue;
}

int isl38xx_ioctl(struct net_device *nwdev, struct ifreq *rq, int cmd )
{
    int error = 0;

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_ioctl cmd %x \n", cmd);

    switch (cmd)
    {
        default:
            error = -EOPNOTSUPP;
    }

    return error;
}

/******************************************************************************
    Device initialisation  
******************************************************************************/

void isl38xx_initialize( isl38xx_private *private_config )
{
    char configmode[4] = { INL_CONFIG_MANUALRUN, 0, 0, 0 };
    char cardmode[4] = { 0, 0, 0, 0 };
    char bsstype[4] =  { 0, 0, 0, 0 };
    char channel[4] =  { 0, 0, 0, 0 };
    char authen[4] = { 0, 0, 0, 0 };
    char invoke[4] = { 0, 0, 0, 0 };
    char exunencrypt[4] = { 0, 0, 0, 0 };
    char dot1xen[4] = { 0, 0, 0, 0 };
    struct obj_key key = { DOT11_PRIV_WEP, 0, "" };
    struct obj_ssid essid_obj = { sizeof( CARD_DEFAULT_SSID ) - 1,
                                CARD_DEFAULT_SSID };
    struct obj_buffer psm_buffer = { PSM_BUFFER_SIZE, 0 };

    // store all device initialization management frames in the
    // management transmit shadow queue
    cardmode[0] = (char) init_mode;
    channel[0] = (char) init_channel;
    channel[1] = (char) (init_channel >> 8);
    bsstype[0] = (char) init_bsstype;
    authen[0] = (char) init_authen;
    invoke[0] = (char) init_wep;
    exunencrypt[0] = (char) init_filter;
    dot1xen[0] = (char) init_dot1x;
    if( init_ssid != NULL )
    {
        // parameter specified when loading the module
        essid_obj.length = strlen( init_ssid );
        strcpy( essid_obj.octets, init_ssid );
    }
    psm_buffer.base = (void *) private_config->device_psm_buffer;
    if( init_wds )
    {
        // enable wds mode in the host interface
        configmode[0] |= INL_CONFIG_WDS;
    }

    DEBUG(SHOW_TRACING, "cardmode: %i \n", init_mode );
    DEBUG(SHOW_TRACING, "channel: %i \n", init_channel );
    DEBUG(SHOW_TRACING, "bsstype: %i \n", init_bsstype );
    DEBUG(SHOW_TRACING, "wep: %i \n", init_wep );
    DEBUG(SHOW_TRACING, "authen: %i \n", init_authen );
    DEBUG(SHOW_TRACING, "filter: %i \n", init_filter );
    DEBUG(SHOW_TRACING, "wds: %i \n", init_wds );
    DEBUG(SHOW_TRACING, "ssid: %i %s \n", essid_obj.length, essid_obj.octets );
    DEBUG(SHOW_TRACING, "psm: %p %li\n", psm_buffer.base, psm_buffer.size );

    // store all device initialization management frames in the
    // management transmit shadow queue

    //  get firmware configuration data
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_GET,
        GEN_OID_MACADDRESS, 0, NULL, 6, 0);

    // set the configuration mode to manual run & wds option
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET,
        OID_INL_CONFIG, 0, &configmode[0], 4, 0);

    // set the mode for stopping the device
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET,
        OID_INL_MODE, 0, &cardmode[0], 4, 0);

    // set the bsstype
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET,
        DOT11_OID_BSSTYPE, 0, &bsstype[0], 4, 0);

    // set the channel
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET,
        DOT11_OID_CHANNEL, 0, &channel[0], 4, 0);

    // set the SSID
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET,
        DOT11_OID_SSID, 0, (void *) &essid_obj, sizeof(struct obj_ssid), 0);

    // set the PSM buffer object
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET,
        DOT11_OID_PSMBUFFER, 0, (void *) &psm_buffer,
        sizeof(struct obj_buffer), 0);

	// set the authentication enable option
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET,
		DOT11_OID_AUTHENABLE, 0, &authen[0], 4, 0);

    // set the privacy invoked mode
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET,
        DOT11_OID_PRIVACYINVOKED, 0, &invoke[0], 4, 0);

    // set the exunencrypted object
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET,
        DOT11_OID_EXUNENCRYPTED, 0, &exunencrypt[0], 4, 0);

    // set default key 1
    strcpy( key.key, CARD_DEFAULT_KEY1 );
    key.length = strlen(CARD_DEFAULT_KEY1) > sizeof(key.key) ?
        sizeof(key.key) : strlen(CARD_DEFAULT_KEY1);
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET, DOT11_OID_DEFKEY1, 0,
        &key, sizeof(struct obj_key), 0);

    // set default key 2
    strcpy( key.key, CARD_DEFAULT_KEY2 );
    key.length = strlen(CARD_DEFAULT_KEY2) > sizeof(key.key) ?
        sizeof(key.key) : strlen(CARD_DEFAULT_KEY2);
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET, DOT11_OID_DEFKEY2, 0,
        &key, sizeof(struct obj_key), 0);

    // set default key 3
    strcpy( key.key, CARD_DEFAULT_KEY3 );
    key.length = strlen(CARD_DEFAULT_KEY3) > sizeof(key.key) ?
        sizeof(key.key) : strlen(CARD_DEFAULT_KEY3);
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET, DOT11_OID_DEFKEY3, 0,
        &key, sizeof(struct obj_key), 0);

    // set default key 4
    strcpy( key.key, CARD_DEFAULT_KEY4 );
    key.length = strlen(CARD_DEFAULT_KEY4) > sizeof(key.key) ?
        sizeof(key.key) : strlen(CARD_DEFAULT_KEY4);
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET, DOT11_OID_DEFKEY4, 0,
        &key, sizeof(struct obj_key), 0);

    // set the dot1x enable object
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET,
        DOT11_OID_DOT1XENABLE, 0, &dot1xen[0], 4, 0);

    // set the mode again for starting the device
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_SET,
        OID_INL_MODE, 0, &cardmode[0], 4, 0);

    // call the transmit function for triggering the transmitting of the
    // first frame in the shadow queue
    isl38xx_transmit_pimfor( private_config );
}

int isl38xx_ioctl_getrange(struct net_device *nwdev,
    struct iw_range *range )
{
    isl38xx_private *private_config = nwdev->priv;
    struct iw_range range_obj;
    char data[IW_MAX_BITRATES], bitrate;
    unsigned long len = IW_MAX_BITRATES;
    int rvalue, counter;

    // request the device for the rates
    isl38xx_queue_pimfor(private_config, PIMFOR_OP_GET,
       DOT11_OID_RATES, 0, NULL, IW_MAX_BITRATES+1, 0);

    // enter sleep mode after supplying oid and local buffers for reception
    // to the management queue receive mechanism
    if (rvalue = wait_for_oid_response(private_config, DOT11_OID_RATES,
        (void *)data, &len ), rvalue != 0)
    {
        // error in getting the response, return the value
        return rvalue;
    }

    DEBUG(SHOW_TRACING, "getrates: bufaddr %p \n", range );

    // set the wireless extension version number
    range_obj.we_version_compiled = WIRELESS_EXT;

    for( counter=0; counter < IW_MAX_BITRATES; counter++ )
    {
        bitrate = data[counter];
        // check whether the rate value is set
        if( bitrate )
        {
            DEBUG( SHOW_TRACING, "rate: basic %i value %x \n",
                (bitrate & 0x80) >> 7, bitrate & 0x7F );

            // convert the bitrates to Mbps
            range_obj.bitrate[counter] = (double)(bitrate & 0x7F);
            range_obj.bitrate[counter] *= 1000000;

            // increment the number of rates counter
            range_obj.num_bitrates++;
        }
    }

    // return the data if the destination buffer is defined
    if (copy_to_user( (void *) range, (void *) &range_obj,
        sizeof(struct iw_range)))
        return -EFAULT;

    return 0;
}

/******************************************************************************
    Device Control functions
******************************************************************************/
void enable_update_interrupt(void *device)
{
    u32 reg;

    // set the update queue status bit in the Interrupt Enable Register
    reg = readl(device + ISL38XX_INT_EN_REG);
    reg |= ISL38XX_INT_IDENT_UPDATE;
    writel(reg, device + ISL38XX_INT_EN_REG);
}

void disable_update_interrupt(void *device)
{
    u32 reg;

    // clear the update queue status bit in the Interrupt Enable Register
    reg = readl(device + ISL38XX_INT_EN_REG);
    reg &= ~ISL38XX_INT_IDENT_UPDATE;
    writel(reg, device + ISL38XX_INT_EN_REG);
}


/******************************************************************************
    Device Interface functions
******************************************************************************/
void isl38xx_interrupt(int irq, void *config, struct pt_regs *regs)
{
    u32 reg;
    isl38xx_private *private_config = config;
    void *device = private_config->device_id;

    // received an interrupt request on a shared IRQ line
    // check whether there is any source of interrupt on the device
    reg = readl(device + ISL38XX_INT_IDENT_REG);

    // also check the contents of the Interrupt Enable Register, because this
    // will filter out interrupt sources from other devices on the same irq !
    reg &= readl(device + ISL38XX_INT_EN_REG);

    if (reg &= ISL38XX_INT_SOURCES, reg != 0)
    {
        // a bit is set identifying an interrupt request from our device
        // reset the request bits in the Identification register
        writel(reg, device + ISL38XX_INT_ACK_REG);

        DEBUG(SHOW_FUNCTION_CALLS,
        "IRQ: Identification register 0x%p 0x%x \n", device, reg);

        // check for each bit in the register separately
        if (reg & ISL38XX_INT_IDENT_UPDATE)
        {
            // Queue has been updated
            DEBUG(SHOW_TRACING, "IRQ: Update flag \n");

            DEBUG(SHOW_QUEUE_INDEXES,
                "CB drv Qs: [%i][%i][%i][%i][%i][%i]\n",
                le32_to_cpu(private_config->control_block->driver_curr_frag[0]),
                le32_to_cpu(private_config->control_block->driver_curr_frag[1]),
                le32_to_cpu(private_config->control_block->driver_curr_frag[2]),
                le32_to_cpu(private_config->control_block->driver_curr_frag[3]),
                le32_to_cpu(private_config->control_block->driver_curr_frag[4]),
                le32_to_cpu(private_config->control_block->driver_curr_frag[5])
                );

            DEBUG(SHOW_QUEUE_INDEXES,
                "CB dev Qs: [%i][%i][%i][%i][%i][%i]\n",
                le32_to_cpu(private_config->control_block->device_curr_frag[0]),
                le32_to_cpu(private_config->control_block->device_curr_frag[1]),
                le32_to_cpu(private_config->control_block->device_curr_frag[2]),
                le32_to_cpu(private_config->control_block->device_curr_frag[3]),
                le32_to_cpu(private_config->control_block->device_curr_frag[4]),
                le32_to_cpu(private_config->control_block->device_curr_frag[5])
                );

            // check all three queues in priority order
            // call the PIMFOR receive function until the queue is empty
            while (isl38xx_in_queue(private_config->control_block,
                ISL38XX_CB_RX_MGMTQ) != 0)
            {
                DEBUG(SHOW_TRACING, "Received frame in Management Queue\n");
                isl38xx_receive_pimfor(private_config);

                // call the pimfor transmit function for processing the next
                // management frame in the shadow queue
                isl38xx_transmit_pimfor( private_config );

            }
/*
            while( isl38xx_in_queue( private_config->control_block,
                    ISL38XX_CB_RX_DATA_HQ ) != 0 )
            {
                    DEBUG( SHOW_TRACING, "Received frame in Data High Queue \n");
            }
*/
            while (isl38xx_in_queue(private_config->control_block,
                ISL38XX_CB_RX_DATA_LQ) != 0)
            {
                DEBUG(SHOW_TRACING, "Received frame in Data Low Queue \n");
                isl38xx_receive(private_config);
            }

            // check whether the data transmit queues were full
            if (private_config->data_low_tx_full)
            {
                // check whether the transmit is not full anymore
                if (isl38xx_in_queue(private_config->control_block,
                    ISL38XX_CB_TX_DATA_LQ) != ISL38XX_CB_TX_QSIZE)
                {
                    // nope, the driver is ready for more network frames
                    netif_wake_queue(private_config->my_module);

                    // reset the full flag
                    private_config->data_low_tx_full = 0;
                }
            }
/*
            // check whether the data transmit queues were full
            if( private_config->data_high_tx_full )
            {
                // check whether the transmit is not full anymore
                if ( isl38xx_in_queue( private_config->control_block,
                    ISL38XX_CB_TX_DATA_HQ ) != ISL38XX_CB_TX_QSIZE )
                {
                    // nope, the driver is ready for more network frames
                    netif_wake_queue( private_config->my_module );

                    // reset the full flag
                    private_config->data_high_tx_full = 0;
                }
            }
*/
        }

        if (reg & ISL38XX_INT_IDENT_INIT)
        {
            // Device has been initialized
            DEBUG(SHOW_TRACING, "IRQ: Init flag, device initialized \n");

            // perform card initlialization by sending objects
            isl38xx_initialize( private_config );
        }

        if (reg & ISL38XX_INT_IDENT_WAKEUP)
        {
            // Device has been woken up to active state
            DEBUG(SHOW_TRACING, "IRQ: Wakeup flag \n");
        }

        if (reg & ISL38XX_INT_IDENT_SLEEP)
        {
            // Device intends to move to powersave state
            DEBUG(SHOW_TRACING, "IRQ: Sleep flag \n");
        }
    }

    return;
}


/******************************************************************************
    Network Interface Control & Statistical functions
******************************************************************************/
int isl38xx_open(struct net_device *nwdev)
{
//      isl38xx_private *private_config = nwdev->priv;

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_open \n");

    MOD_INC_USE_COUNT;

    netif_start_queue(nwdev);
//      netif_mark_up( nwdev );


    return 0;
}

int isl38xx_close(struct net_device *nwdev)
{
//      isl38xx_private *private_config = nwdev->priv;

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_close \n");

    netif_stop_queue(nwdev);
//      netif_mark_down( nwdev );

    MOD_DEC_USE_COUNT;
    return 0;
}

void isl38xx_set_multicast_list(struct net_device *dev)
{
    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_set_mc_list \n");
}

struct net_device_stats *isl38xx_statistics(struct net_device *nwdev)
{
    isl38xx_private *private_config = nwdev->priv;

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_statistics \n");

    return &private_config->statistics;
}

int isl38xx_reset(isl38xx_private * private_config)
{

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_reset \n");

    // reset the device interface first ????????????
    isl38xx_interface_reset(private_config);

    // restore the configuration in the device


    // ........

    return 0;
}

void isl38xx_interface_reset(isl38xx_private * private_config)
{
    u32 reg;
    void *device = private_config->remapped_device_base;

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_interface_reset \n");

    // load the address of the control block in the device
    writel(private_config->device_host_address,
           device + ISL38XX_CTRL_BLK_BASE_REG);
    udelay(ISL38XX_WRITEIO_DELAY);

    // set the reset bit in the Device Interrupt Register
    reg = readl(device + ISL38XX_DEV_INT_REG);
    reg |= ISL38XX_DEV_INT_RESET;
    writel(reg, device + ISL38XX_DEV_INT_REG);
    udelay(ISL38XX_WRITEIO_DELAY);

    // reset the reset bit in the Device Interrupt Register ?
    reg &= ~ISL38XX_DEV_INT_RESET;
    writel(reg, device + ISL38XX_DEV_INT_REG);
    udelay(ISL38XX_WRITEIO_DELAY);

    // enable the interrupt for detecting initialization
    reg = ISL38XX_INT_IDENT_INIT | ISL38XX_INT_IDENT_UPDATE |
        ISL38XX_INT_IDENT_SLEEP | ISL38XX_INT_IDENT_WAKEUP;
    writel(reg, device + ISL38XX_INT_EN_REG);
    udelay(ISL38XX_WRITEIO_DELAY);
}

/******************************************************************************
    Device configuration functions
******************************************************************************/
int isl38xx_upload_firmware(struct pci_dev *pci_device, char *filename,
                                  isl38xx_private * private_config)
{
    u32 reg;
    int ifp;
    long bcount, length;
    mm_segment_t fs;
    void *device;
    static char buffer[ISL38XX_MEMORY_WINDOW_SIZE];

    // firmware uploading starts with resetting the device
    device = private_config->remapped_device_base;

    // clear the RAMBoot and the Reset bit
    reg = readl(device + ISL38XX_CTRL_STAT_REG);
    reg &= ~ISL38XX_CTRL_STAT_RESET;
    reg &= ~ISL38XX_CTRL_STAT_RAMBOOT;
    writel(reg, device + ISL38XX_CTRL_STAT_REG);
    udelay(ISL38XX_WRITEIO_DELAY);

    // set the Reset bit without reading the register !
    reg |= ISL38XX_CTRL_STAT_RESET;
    writel(reg, device + ISL38XX_CTRL_STAT_REG);
    udelay(ISL38XX_WRITEIO_DELAY);

    // clear the Reset bit
    reg &= ~ISL38XX_CTRL_STAT_RESET;
    writel(reg, device + ISL38XX_CTRL_STAT_REG);

    // wait a while for the device to reboot
    mdelay(50);

    // prepare the Direct Memory Base register
    reg = ISL38XX_DEV_FIRMWARE_ADDRES;

    // for file opening temporarily tell the kernel I am not a user for
    // memory management segment access
    fs = get_fs();
    set_fs(KERNEL_DS);

    // open the file with the firmware for uploading
    if (ifp = open( filename, O_RDONLY, 0 ), ifp == 0)
    {
        // error opening the file
        DEBUG(SHOW_ERROR_MESSAGES, "ERROR: File opening did not success \n");
        set_fs(fs);
        return -1;
    }
    // enter a loop which reads data blocks from the file and writes them
    // to the Direct Memory Windows
    do
    {
        // set the cards base address for writting the data
        writel(reg, device + ISL38XX_DIR_MEM_BASE_REG);
        length = 0;

        // read a block of data until buffer is full or end of file
        do
        {
            bcount = read(ifp, &buffer[length], sizeof(buffer) - length);
            length += bcount;
        }
        while ((length != ISL38XX_MEMORY_WINDOW_SIZE) && (bcount != 0));

        // write the data to the Direct Memory Window
        memcpy_toio(device + ISL38XX_DIRECT_MEM_WIN, buffer, length);

        // increment the write address
        reg += ISL38XX_MEMORY_WINDOW_SIZE;
    }
    while (bcount != 0);

    // close the file
    close(ifp);

    // switch back the segment setting
    set_fs(fs);

    // now reset the device
    // clear the Reset bit set the RAMBoot bit
    reg = readl(device + ISL38XX_CTRL_STAT_REG);
    reg &= ~ISL38XX_CTRL_STAT_RESET;
    reg |= ISL38XX_CTRL_STAT_RAMBOOT;
    writel(reg, device + ISL38XX_CTRL_STAT_REG);
    udelay(ISL38XX_WRITEIO_DELAY);

    // set the reset bit latches the host override and RAMBoot bits
    // into the device for operation when the reset bit is reset
    reg |= ISL38XX_CTRL_STAT_RESET;
    writel(reg, device + ISL38XX_CTRL_STAT_REG);
    udelay(ISL38XX_WRITEIO_DELAY);

    // clear the reset bit should start the whole circus
    reg &= ~ISL38XX_CTRL_STAT_RESET;
    writel(reg, device + ISL38XX_CTRL_STAT_REG);
    udelay(ISL38XX_WRITEIO_DELAY);

    // now the last step is to reset the interface
    isl38xx_interface_reset(private_config);

    return 0;
}

struct net_device *isl38xx_probe(struct net_device *nwdev,
                                       long ioaddr, int irq)
{
    // initially setup an ethernet device
    nwdev = init_etherdev(nwdev, 0);

    // setup the structure members
    nwdev->base_addr = ioaddr;
    nwdev->irq = irq;

    // initialize the function pointers
    nwdev->open = &isl38xx_open;
    nwdev->stop = &isl38xx_close;
    nwdev->get_stats = &isl38xx_statistics;
    nwdev->do_ioctl = &isl38xx_ioctl;
    nwdev->hard_start_xmit = &isl38xx_transmit;

//      nwdev->set_multicast_list = &set_rx_mode;

#ifdef HAVE_TX_TIMEOUT
    nwdev->watchdog_timeo = ISL38XX_TX_TIMEOUT;
    nwdev->tx_timeout = &isl38xx_tx_timeout;
#endif

    return nwdev;
}

int isl38xx_config(struct pci_dev *pci_device,
                         struct net_device *nw_device)
{
    void *mapped_address;
    void *driver_address;
    void *queue_base;
    dma_addr_t device_address;
    long counter;
    long offset;
    queue_entry *pointerq;
    isl38xx_private *private_config;
    isl38xx_control_block *control_block;
    struct sk_buff *skb;

    // remap the PCI device base address to accessable
    if (mapped_address = ioremap(nw_device->base_addr, ISL38XX_PCI_MEM_SIZE),
        mapped_address == NULL)
    {
        // error in remapping the PCI device memory address range
        DEBUG(SHOW_ERROR_MESSAGES, "ERROR: PCI memory remapping failed \n");
        return -1;
    }

    // save the start and end address of the PCI memory area
    nw_device->mem_start = (unsigned long) mapped_address;
    nw_device->mem_end = (unsigned long) mapped_address + ISL38XX_PCI_MEM_SIZE;

    DEBUG(SHOW_TRACING, "PCI Memory remapped to 0x%p \n", mapped_address);

        // register the net device
//      if( register_netdev( nw_device ))
//      {
//              DEBUG(SHOW_ERROR_MESSAGES, "ERROR: Net device not registered \n");
//              iounmap(mapped_address);
//              return -1;
//      }

    // allocate a block of memory on the host for the control block
    if (driver_address = pci_alloc_consistent(pci_device,
        ISL38XX_HOST_MEM_BLOCK, &device_address), driver_address == NULL)
    {
        // error allocating the bblock of PCI memory
        DEBUG(SHOW_ERROR_MESSAGES,
                "ERROR: Could not allocate DMA accessable " "memory \n");
        iounmap(mapped_address);
        return -1;
    }

    DEBUG(SHOW_TRACING, "Device memory block at 0x%x \n", device_address);
    DEBUG(SHOW_TRACING, "Driver memory block at 0x%p \n", driver_address);
    DEBUG(SHOW_TRACING, "Memory size 0x%x bytes\n", ISL38XX_HOST_MEM_BLOCK);

//      pci_free_consistent(pci_device, ISL38XX_HOST_MEM_BLOCK, driver_address,
//              device_address);
//      iounmap(mapped_address);
//      DEBUG( SHOW_TRACING, "Breakpoint for testing\n" );
//      return -1;

    // add a private device structure to the network device for private
    // definitions needed by the driver
    if (private_config = kmalloc(sizeof(isl38xx_private), GFP_KERNEL |
        GFP_DMA), private_config == NULL)
    {
        // error allocating the DMA accessable memory area
        DEBUG(SHOW_ERROR_MESSAGES, "ERROR: Could not allocate additional "
        "memory \n");
        pci_free_consistent(pci_device, ISL38XX_HOST_MEM_BLOCK, driver_address,
            device_address);
        iounmap(mapped_address);
        return -1;
    }

    memset(private_config, 0, sizeof(isl38xx_private));
    nw_device->priv = private_config;
    private_config->next_module = root_isl38xx_device;
    root_isl38xx_device = nw_device;
    private_config->my_module = nw_device;

    // assign the Control Block pointer to the allocated area
    control_block = (isl38xx_control_block *) driver_address;
    private_config->control_block = control_block;

    // clear the indexes in the frame pointer
    for (counter = 0; counter < ISL38XX_CB_QCOUNT; counter++)
    {
        control_block->driver_curr_frag[counter] = cpu_to_le32(0);
        control_block->device_curr_frag[counter] = cpu_to_le32(0);
    }

    // save the device configurations in the device private configuration
    private_config->remapped_device_base = mapped_address;
    private_config->device_host_address = device_address;
    private_config->device_psm_buffer = device_address + CONTROL_BLOCK_SIZE;
    private_config->driver_mem_address = driver_address;

    DEBUG(SHOW_TRACING, "PSM Buffer at 0x%p size %i\n",
        (void *) private_config->device_psm_buffer, PSM_BUFFER_SIZE );

    // configure the allocated area for PSM buffer and queueing mechanism
    // with the following layout
    // offset 0x0000:       Control block,  size = 1024 bytes
    //      ,  ,  0x0400:   PSM Buffer, size = entries * 0x0600
    //      ,  ,  0x????:   Queue memory

    // set the queue base address behind the control block
    queue_base = driver_address + CONTROL_BLOCK_SIZE + PSM_BUFFER_SIZE;

    DEBUG(SHOW_TRACING, "Queue memory base at 0x%p \n", queue_base);

    // initialize all the queues in the private configuration structure
    init_queue(&private_config->mgmt_tx_freeq);
    init_queue(&private_config->mgmt_rx_freeq);
    init_queue(&private_config->data_tx_freeq);
    init_queue(&private_config->data_rx_freeq);

    init_queue(&private_config->mgmt_tx_shadowq);
    init_queue(&private_config->mgmt_rx_shadowq);
    init_queue(&private_config->ioctl_rx_queue);
    init_queue(&private_config->trap_rx_queue);
    init_queue(&private_config->pimfor_rx_queue);
    init_queue(&private_config->data_low_tx_shadowq);
    init_queue(&private_config->data_low_rx_shadowq);
    init_queue(&private_config->data_low_rx_queue);
//      init_queue( &private_config->data_high_tx_shadowq );
//      init_queue( &private_config->data_high_rx_shadowq );
//      init_queue( &private_config->data_high_rx_queue );

    // determine the offset for each queue entry configuration
    // allign on 16 byte boundary
    offset = sizeof(queue_entry) / 16;
    offset *= 16;
    offset += ((sizeof(queue_entry) % 16) != 0 ? 16 : 0);

    // configure all management queue(s) entries and add them to the freeqs
    for (counter = 0; counter < MGMT_TX_FRAME_COUNT; counter++)
    {
        // configure the queue entry at the beginning of the memory block
        pointerq = (queue_entry *) queue_base;
        pointerq->host_address = (u32) queue_base + offset;
        pointerq->size = MGMT_FRAME_SIZE - offset;
        pointerq->fragments = 1;

        // also translate the host address for the device
        pointerq->dev_address = (u32) virt_to_bus((void *)
                pointerq->host_address);

        // add the configured entry to the management tx root queue
        put_queue(mapped_address, &private_config->mgmt_tx_freeq, pointerq);

        // increment the queue base address pointer to the next entry
        queue_base += MGMT_FRAME_SIZE;

        // perform a check on the destination address in PCI memory
        if (queue_base > driver_address + ISL38XX_HOST_MEM_BLOCK)
        {
            // Oops, end address of fragment will be out of bound
            DEBUG(SHOW_ERROR_MESSAGES, "ERROR: Queue memory out of bound, "
                "addres 0x%p, enlarge host memory \n", queue_base);
            pci_free_consistent(pci_device, ISL38XX_HOST_MEM_BLOCK,
                driver_address, device_address);
            iounmap(mapped_address);
            return -1;
        }
    }

    for (counter = 0; counter < MGMT_RX_FRAME_COUNT; counter++)
    {
        // configure the queue entry at the beginning of the memory block
        pointerq = (queue_entry *) queue_base;
        pointerq->host_address = (u32) queue_base + offset;
        pointerq->size = MGMT_FRAME_SIZE - offset;
        pointerq->fragments = 1;

        // also translate the host address for the device
        pointerq->dev_address = (u32) virt_to_bus((void *)
            pointerq->host_address);

        // add the configured entry to the management rx root queue
        put_queue(mapped_address, &private_config->mgmt_rx_freeq, pointerq);

        // increment the queue base address pointer to the next entry
        queue_base += MGMT_FRAME_SIZE;

        // perform a check on the destination address in PCI memory
        if (queue_base > driver_address + ISL38XX_HOST_MEM_BLOCK)
        {
            // Oops, end address of fragment will be out of bound
            DEBUG(SHOW_ERROR_MESSAGES, "ERROR: Queue memory out of bound, "
                "addres 0x%p, enlarge host memory \n", queue_base);
            pci_free_consistent(pci_device, ISL38XX_HOST_MEM_BLOCK,
                driver_address, device_address);
            iounmap(mapped_address);
            return -1;
        }
    }

    for (counter = 0; counter < DATA_TX_FRAME_COUNT; counter++)
    {
        // add only the entry pointer without data memory allocation
        pointerq = (queue_entry *) queue_base;
        put_queue(mapped_address, &private_config->data_tx_freeq, pointerq);

        // increment the queue base address pointer to the next entry
        queue_base += DATA_FRAME_SIZE;

        // perform a check on the destination address in PCI memory
        if (queue_base > driver_address + ISL38XX_HOST_MEM_BLOCK)
        {
            // Oops, end address of fragment will be out of bound
            DEBUG(SHOW_ERROR_MESSAGES, "ERROR: Queue memory out of bound, "
                "addres 0x%p, enlarge host memory \n", queue_base);
            pci_free_consistent(pci_device, ISL38XX_HOST_MEM_BLOCK,
                driver_address, device_address);
            iounmap(mapped_address);
            return -1;
        }
    }

    for (counter = 0; counter < DATA_RX_FRAME_COUNT; counter++)
    {
        // add only the entry pointer without data memory allocation
        pointerq = (queue_entry *) queue_base;
        put_queue(mapped_address, &private_config->data_rx_freeq, pointerq);

        // increment the queue base address pointer to the next entry
        queue_base += DATA_FRAME_SIZE;

        // perform a check on the destination address in PCI memory
        if (queue_base > driver_address + ISL38XX_HOST_MEM_BLOCK)
        {
            // Oops, end address of fragment will be out of bound
            DEBUG(SHOW_ERROR_MESSAGES, "ERROR: Queue memory out of bound, "
                "addres 0x%p, enlarge host memory \n", queue_base);
            pci_free_consistent(pci_device, ISL38XX_HOST_MEM_BLOCK,
                driver_address, device_address);
            iounmap(mapped_address);
            return -1;
        }
    }

    DEBUG(SHOW_TRACING, "Queue memory end at 0x%p \n", queue_base);

    // Joho, ready setting up the queueing stuff, now fill the shadow rx queues
    // of both the management and data queues
    for (counter = 0; counter < ISL38XX_CB_MGMT_QSIZE; counter++)
    {
        // get an entry from the freeq and put it in the shadow queue
        if (get_queue(mapped_address, &private_config->mgmt_rx_freeq,
            &pointerq))
        {
            // Oops, no more entries in the freeq ?
            DEBUG(SHOW_ERROR_MESSAGES, "Error: No more entries in freeq\n");
            pci_free_consistent(pci_device, ISL38XX_HOST_MEM_BLOCK,
                driver_address, device_address);
            iounmap(mapped_address);
            return -1;
        }
        else
        {
            // use the entry for the device interface management rx queue
            // these should always be inline !
            control_block->rx_data_mgmt[counter].address =
                cpu_to_le32(pointerq->dev_address);

            // put the entry got from the freeq in the shadow queue
            put_queue(mapped_address, &private_config->mgmt_rx_shadowq,
                pointerq);
        }
    }

    for (counter = 0; counter < ISL38XX_CB_RX_QSIZE; counter++)
    {
        // get an entry from the freeq and put it in the shadow queue
        if (get_queue(mapped_address, &private_config->data_rx_freeq,
              &pointerq))
        {
            // Oops, no more entries in the freeq ?
            DEBUG(SHOW_ERROR_MESSAGES, "Error: No more entries in freeq\n");
            pci_free_consistent(pci_device, ISL38XX_HOST_MEM_BLOCK,
                driver_address, device_address);
            iounmap(mapped_address);
            return -1;
        }
        else
        {
            // allocate an sk_buff for received data frames storage
	    // include any required allignment operations
            if (skb = dev_alloc_skb(MAX_FRAGMENT_SIZE+2), skb == NULL)
            {
                // error allocating an sk_buff structure elements
                DEBUG(SHOW_ERROR_MESSAGES, "Error allocating skb \n");
                pci_free_consistent(pci_device, ISL38XX_HOST_MEM_BLOCK,
                    driver_address, device_address);
                iounmap(mapped_address);
                return -1;
            }

            // now set the members in the entry to the skbuf structure
            pointerq->host_address = (u32) skb;
            pointerq->dev_address = (u32) virt_to_bus((void *) skb->data);
            pointerq->size = sizeof(struct sk_buff);
            pointerq->fragments = 1;

            // use the entry for the device interface management rx queue
            // these should always be inline !
            control_block->rx_data_low[counter].address =
                cpu_to_le32(pointerq->dev_address);

            // put the entry got from the freeq in the shadow queue
            put_queue(mapped_address, &private_config->data_low_rx_shadowq,
                pointerq);
        }
/*
        // do it again for the data high queue
        if( get_queue( mapped_address, &private_config->data_rx_freeq,
            &pointerq ))
        {
            // Oops, no more entries in the freeq ?
            DEBUG( SHOW_ERROR_MESSAGES, "Error: No more entries in freeq\n" );
            pci_free_consistent( pci_device, ISL38XX_HOST_MEM_BLOCK,
                driver_address, device_address );
            iounmap( mapped_address );
            return -1;
        }
        else
        {
            // allocate an sk_buff for received data frames storage
	    // include any required allignment operations
            if (skb = dev_alloc_skb(MAX_FRAGMENT_SIZE+2), skb == NULL)
            {
                // error allocating an sk_buff structure elements
                DEBUG( SHOW_ERROR_MESSAGES, "Error allocating skb \n" );
                pci_free_consistent( pci_device, ISL38XX_HOST_MEM_BLOCK,
                    driver_address, device_address );
                iounmap( mapped_address );
                return -1;
            }

            // now set the members in the entry to the skbuf structure
            pointerq->host_address = (u32) skb;
            pointerq->dev_address = (u32) virt_to_bus( (void *) skb->data );
            pointerq->size = sizeof( struct sk_buff );
            pointerq->fragments = 1;

            // use the entry for the device interface management rx queue
            // these should always be inline !
            control_block->rx_data_high[counter].address =
                 cpu_to_le32(pointerq->dev_address);

            // put the entry got from the freeq in the shadow queue
            put_queue( mapped_address, &private_config->data_high_rx_shadowq,
                pointerq );
        }
*/
    }

    // since the receive queues are filled with empty fragments, now we can
    // set the corresponding indexes in the Control Block
    control_block->driver_curr_frag[ISL38XX_CB_RX_DATA_LQ] =
        cpu_to_le32(ISL38XX_CB_RX_QSIZE);
//    control_block->driver_curr_frag[ISL38XX_CB_RX_DATA_HQ] =
//      cpu_to_le32(ISL38XX_CB_RX_QSIZE);
    control_block->driver_curr_frag[ISL38XX_CB_RX_MGMTQ] =
        cpu_to_le32(ISL38XX_CB_MGMT_QSIZE);

    // reset the real index registers and full flags
    private_config->index_mgmt_rx = 0;
        private_config->index_mgmt_tx = 0;
    private_config->index_data_low = 0;
    private_config->index_data_high = 0;
    private_config->data_low_tx_full = 0;
    private_config->data_high_tx_full = 0;

    // reset the queue read locks, process wait counter
    private_config->ioctl_queue_lock = 0;
    private_config->processes_in_wait = 0;
    private_config->data_fragments_in_wait = 0;

    return 0;
}


/******************************************************************************
    Module initialization functions
******************************************************************************/
#if TARGET_SYSTEM == CONFIG_PCMCIA

dev_node_t *isl38xx_attach(dev_locator_t * loc)
{
    u32 io;
    u16 dev_id;
    u8 bus, devfn, irq;
    struct pci_dev *pci_device;
    struct net_device *nw_device;
    dev_node_t *node;
    isl38xx_private *private_config;
    int rvalue;
    int dma_mask = 0xffffffff;
    char firmware[256];

    // perform some initial setting checks
    if (loc->bus != LOC_PCI)
        return NULL;
    bus = loc->b.pci.bus;
    devfn = loc->b.pci.devfn;

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_attach(bus %d, function %d)\n",
        bus, devfn);

    // get some pci settings for verification
    pcibios_read_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, &io);
    pcibios_read_config_byte(bus, devfn, PCI_INTERRUPT_LINE, &irq);
    pcibios_read_config_word(bus, devfn, PCI_DEVICE_ID, &dev_id);
    if (io &= ~3, io == 0 || irq == 0)
    {
        DEBUG(SHOW_ERROR_MESSAGES, "The ISL38XX Ethernet interface was not "
        "assigned an %s.\n" KERN_ERR "  It will not be activated.\n",
        io == 0 ? "I/O address" : "IRQ");
        return NULL;
    }
    // get pci device information by loading the pci_dev structure
    if (pci_device = pci_find_slot(bus, devfn), pci_device == NULL)
    {
        // error reading the pci device structure
        DEBUG(SHOW_ERROR_MESSAGES, "ERROR: isl38xx could not get PCI device "
            "information \n");
        return NULL;
    }

    // determine what the supported DMA memory region is
    while( pci_set_dma_mask( pci_device, dma_mask ) != 0 )
    {
        // range not supported, shift the mask and check again
        if( dma_mask >>=  1, dma_mask == 0 )
        {
            // mask is zero, DMA memory not supported by PCI
            DEBUG(SHOW_ERROR_MESSAGES, "DMA Memory not supported\n" );
            return NULL;
        }
    }

    DEBUG(SHOW_TRACING, "DMA Memory support mask is 0x%x \n", dma_mask );

    // setup the network device interface and its structure
    if (nw_device = isl38xx_probe(NULL, (long) io, (int) irq),
        nw_device == NULL)
    {
        // error configuring the driver as a network device
        DEBUG(SHOW_ERROR_MESSAGES, "ERROR: isl38xx could not configure "
        "network device \n");
        return NULL;
    }

    if (isl38xx_config(pci_device, nw_device) == -1)
    {
        // error configuring the PCI bus
        DEBUG(SHOW_ERROR_MESSAGES, "ERROR: isl38xx could not configure "
        "PCI device \n");
        return NULL;
    }
    // save the interrupt request line and use the remapped device base address
    // as the device identification both for uniqueness and parameter passing
    // to the interrupt handler
    private_config = nw_device->priv;
    private_config->pci_irq = irq;
    private_config->pci_dev_id = dev_id;
    private_config->device_id = private_config->remapped_device_base;
    private_config->pci_device = pci_device;

    // request for the interrupt before uploading the firmware
    if (rvalue = request_irq(irq, &isl38xx_interrupt, SA_INTERRUPT | SA_SHIRQ,
        DRIVER_NAME, private_config), rvalue != 0)
    {
        // error, could not hook the handler to the irq
        DEBUG(SHOW_ERROR_MESSAGES, "ERROR: isl38xx could not install "
            "IRQ-handler \n");
        return NULL;
    }

    // select the firmware file depending on the device id, take for default
    // the 3877 firmware file
    if( dev_id == PCIDEVICE_ISL3890 )
    	strcpy( firmware, ISL3890_IMAGE_FILE );
    else
    	strcpy( firmware, ISL3877_IMAGE_FILE );

    DEBUG(SHOW_TRACING, "Firmware file: %s \n", firmware );

    if (isl38xx_upload_firmware( pci_device, firmware, private_config ) == -1)
    {
        // error uploading the firmware
        DEBUG(SHOW_ERROR_MESSAGES, "ERROR: isl38xx could not upload the "
        "firmware \n");
        return NULL;
    }

    // finally setup the node structure with the device information
    node = kmalloc(sizeof(dev_node_t), GFP_KERNEL);
    strcpy(node->dev_name, nw_device->name);
    node->major = 0;
    node->minor = 0;
    node->next = NULL;
    MOD_INC_USE_COUNT;
    return node;
}

void isl38xx_suspend(dev_node_t * node)
{
    struct net_device **devp, **next;

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_suspend(%s)\n", node->dev_name);

    for (devp = &root_isl38xx_device; *devp; devp = next)
    {
        next = &((isl38xx_private *) (*devp)->priv)->next_module;
        if (strcmp((*devp)->name, node->dev_name) == 0)
        break;
    }

    if (*devp)
    {
//        long ioaddr = (*devp)->base_addr;
//        isl38xx_pause(*devp);
    }
}

void isl38xx_resume(dev_node_t * node)
{
    struct net_device **devp, **next;

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_resume(%s)\n", node->dev_name);

    for (devp = &root_isl38xx_device; *devp; devp = next)
    {
        next = &((isl38xx_private *) (*devp)->priv)->next_module;
        if (strcmp((*devp)->name, node->dev_name) == 0)
        break;
    }

    if (*devp)
    {
//        isl38xx_restart(*devp);
    }
}

void isl38xx_detach(dev_node_t * node)
{
    struct net_device **devp, **next;
    isl38xx_private *private_config;

    DEBUG(SHOW_FUNCTION_CALLS, "isl38xx_detach(%s)\n", node->dev_name);

    for (devp = &root_isl38xx_device; *devp; devp = next)
    {
        next = &((isl38xx_private *) (*devp)->priv)->next_module;
        if (strcmp((*devp)->name, node->dev_name) == 0)
        break;
    }

    if (*devp)
    {
        // free the interrupt request
        private_config = (*devp)->priv;
        writel(0, private_config->remapped_device_base + ISL38XX_INT_EN_REG);
        free_irq(private_config->pci_irq, private_config);

        // free the node
        kfree(node);
        MOD_DEC_USE_COUNT;
    }
}

int init_module(void)
{
    DEBUG(SHOW_ANYTHING, "Init module: %s", version);

    register_driver(&isl38xx_ops);
    mgt_init();
    mgt_indication_handler ( DEV_NETWORK, 0, 0, mgtc_indication );

    return 0;
}

void cleanup_module(void)
{
    struct net_device *next_dev;
    isl38xx_private *private_config;

    DEBUG(SHOW_FUNCTION_CALLS, "Cleanup module \n");

    mgt_cleanup();
    unregister_driver(&isl38xx_ops);

    /* No need to check MOD_IN_USE, as sys_delete_module() checks. */
    while (root_isl38xx_device)
    {
        private_config = (isl38xx_private *) root_isl38xx_device->priv;
        next_dev = private_config->next_module;

        DEBUG(SHOW_TRACING, "Cleanup netdevice \n");

        // unregister the network device
        unregister_netdev(root_isl38xx_device);

        // free the PCI memory and unmap the remapped page
        pci_free_consistent(private_config->pci_device, ISL38XX_HOST_MEM_BLOCK,
            private_config->driver_mem_address,
            private_config->device_host_address);
        iounmap(private_config->remapped_device_base);

        // free the separately allocated areas
        kfree(private_config);
        kfree(root_isl38xx_device);

        // change the root device pointer to the next device for clearing
        root_isl38xx_device = next_dev;
    }
}

#else

int isl38xx_probe_pci(struct pci_dev *pci_device,
    const struct pci_device_id *id)
{
    void *phymem;
    u16 irq;
    u8 bus, devfn;
    struct net_device *nw_device;
    isl38xx_private *private_config;
    int rvalue, dev_id;
    int dma_mask = 0xffffffff;
    char firmware[256];

    // Enable the pci device
    if (pci_enable_device(pci_device))
    {
        DEBUG(SHOW_ERROR_MESSAGES, "%s: pci_enable_device() failed.\n",
            DRIVER_NAME );
        return -EIO;
    }
    // Figure out our resources
    irq = pci_device->irq;
    bus = pci_device->bus->number;
    devfn = pci_device->devfn;
    dev_id = pci_device->device;

    // determine what the supported DMA memory region is
    while( pci_set_dma_mask( pci_device, dma_mask ) != 0 )
    {
        // range not supported, shift the mask and check again
        if( dma_mask >>=  1, dma_mask == 0 )
        {
            // mask is zero, DMA memory not supported by PCI
            DEBUG(SHOW_ERROR_MESSAGES, "DMA Memory not supported\n" );
            return -EIO;
        }
    }

    DEBUG(SHOW_TRACING, "DMA Memory support mask is 0x%x \n", dma_mask );

    // call for the physical address to address the PCI device
    phymem = (void *) pci_resource_start(pci_device, 0);

    // request the pci device for regions ???
    if (rvalue = pci_request_regions(pci_device, DRIVER_NAME ), rvalue)
    {
        DEBUG(SHOW_ERROR_MESSAGES, "pci_request_regions failure, rvalue %i\n",
            rvalue );
        return -EIO;
    }

    // enable PCI bus-mastering
    pci_set_master(pci_device);

    // Log the device resources information
    DEBUG(SHOW_TRACING, "isl38xx device: phymem:0x%lx, irq:%d \n",
        (long) phymem, irq);

    // setup the network device interface and its structure
    if (nw_device = isl38xx_probe(NULL, (long) phymem, (int) irq),
        nw_device == NULL)
    {
        // error configuring the driver as a network device
        DEBUG(SHOW_ERROR_MESSAGES, "ERROR: isl38xx could not configure "
            "network device \n");
        return -EIO;
    }

    // configure the device memory resources
    if (isl38xx_config(pci_device, nw_device) == -1)
    {
        // error configuring the PCI bus
        DEBUG(SHOW_ERROR_MESSAGES, "ERROR: isl38xx could not configure "
            "PCI device \n");
        return -EIO;
    }

    // save the interrupt request line and use the remapped device base address
    // as the device identification both for uniqueness and parameter passing
    // to the interrupt handler
    private_config = nw_device->priv;
    private_config->pci_irq = irq;
    private_config->pci_dev_id = dev_id;
    private_config->device_id = private_config->remapped_device_base;
    private_config->pci_device = pci_device;

    // request for the interrupt before uploading the firmware
    if (rvalue = request_irq(irq, &isl38xx_interrupt, SA_INTERRUPT | SA_SHIRQ,
        DRIVER_NAME, private_config), rvalue != 0)
    {
        // error, could not hook the handler to the irq
        DEBUG(SHOW_ERROR_MESSAGES, "ERROR: isl38xx could not install "
            "IRQ-handler \n");
        return -EIO;
    }

    // select the firmware file depending on the device id, take for default
    // the 3877 firmware file
    if( dev_id == PCIDEVICE_ISL3890 )
    	strcpy( firmware, ISL3890_IMAGE_FILE );
    else
    	strcpy( firmware, ISL3877_IMAGE_FILE );

    DEBUG(SHOW_TRACING, "Firmware file: %s \n", firmware );

    if (isl38xx_upload_firmware( pci_device, firmware, private_config ) == -1)
    {
        // error uploading the firmware
        DEBUG(SHOW_ERROR_MESSAGES, "ERROR: isl38xx could not upload the "
            "firmware \n");
        return -EIO;
    }

    MOD_INC_USE_COUNT;

    return 0;
}

int init_module(void)
{
    DEBUG(SHOW_ANYTHING, "%s", version);
    DEBUG(SHOW_TRACING, "Loaded %s\n", DRIVER_NAME );

    if (pci_register_driver(&isl38xx_pci_drv_id) <= 0)
    {
        DEBUG(SHOW_ERROR_MESSAGES,
            "isl38xx: No devices found, driver not installed.\n");
        pci_unregister_driver(&isl38xx_pci_drv_id);
        return -ENODEV;
    }
    
    mgt_init();
    mgt_indication_handler ( DEV_NETWORK, 0, 0, mgtc_indication );

    return 0;
}


void cleanup_module(void)
{
    struct net_device *next_dev;
    isl38xx_private *private_config;

    MOD_DEC_USE_COUNT;

    mgt_cleanup();
    
    // No need to check MOD_IN_USE, as sys_delete_module() checks
    while (root_isl38xx_device)
    {
        private_config = (isl38xx_private *) root_isl38xx_device->priv;
        next_dev = private_config->next_module;

        writel(0,private_config->remapped_device_base + ISL38XX_INT_EN_REG);
        free_irq(private_config->pci_irq, private_config);

        // unregister the network device
        unregister_netdev(root_isl38xx_device);

        // free the PCI memory and unmap the remapped page
        pci_free_consistent(private_config->pci_device, ISL38XX_HOST_MEM_BLOCK,
            private_config->driver_mem_address,
            private_config->device_host_address);
        iounmap(private_config->remapped_device_base);
/*      pci_release_regions(private_config->pci_device);
*/
        // free the separately allocated areas
        kfree(private_config);
        kfree(root_isl38xx_device);

        // change the root device pointer to the next device for clearing
        root_isl38xx_device = next_dev;
    }

    pci_unregister_driver(&isl38xx_pci_drv_id);

    DEBUG(SHOW_TRACING, "Unloaded %s\n", DRIVER_NAME );

    return;
}

#endif      // TARGET_SYSTEM

/******************************************************************************
    Frame (re)format functions
******************************************************************************/
static void pimfor_encode_header(int operation, unsigned long oid,
                                 int device_id, int flags, int length,
                                 pimfor_header * header)
{
    // byte oriented members
    header->version = PIMFOR_VERSION;
    header->operation = operation;
    header->device_id = device_id;
    header->flags = flags;

    // word oriented members with byte order depending on the flags
    if (flags & PIMFOR_FLAG_LITTLE_ENDIAN)
    {
        // use little endian coding
        header->oid[3] = oid >> 24;
        header->oid[2] = oid >> 16;
        header->oid[1] = oid >> 8;
        header->oid[0] = oid;
        header->length[3] = 0;
        header->length[2] = 0;
        header->length[1] = length >> 8;
        header->length[0] = length;
    }
    else
    {
        // use big endian coding
        header->oid[0] = oid >> 24;
        header->oid[1] = oid >> 16;
        header->oid[2] = oid >> 8;
        header->oid[3] = oid;
        header->length[0] = 0;
        header->length[1] = 0;
        header->length[2] = length >> 8;
        header->length[3] = length;
    }
}

static void pimfor_decode_header(pimfor_header * header, int *operation,
                                 unsigned long *oid, int *device_id,
                                 int *flags, int *length)
{
    // byte oriented members
    *operation = header->operation;
    *device_id = header->device_id;
    *flags = header->flags;

    // word oriented members with byte order depending on the flags
    if (*flags & PIMFOR_FLAG_LITTLE_ENDIAN)
    {
        // use little endian coding
        *oid = ((int) header->oid[3] << 24) |
            ((int) header->oid[2] << 16) |
            ((int) header->oid[1] << 8) | ((int) header->oid[0]);
        *length = ((int) header->length[1] << 8) |
            ((int) header->length[0]);
    }
    else
    {
        // use big endian coding
        *oid = ((int) header->oid[0] << 24) |
            ((int) header->oid[1] << 16) |
            ((int) header->oid[2] << 8) | ((int) header->oid[3]);
        *length = ((int) header->length[2] << 8) |
            ((int) header->length[3]);
    }
}

/******************************************************************************
    Memory Management functions
******************************************************************************/
void init_queue(queue_root * queue)
{
    queue->host_address = (u32) NULL;
    queue->dev_address = (u32) NULL;
    queue->size = 0;
    queue->fragments = 0;
}

int put_queue(void *device, queue_root * queue, queue_entry * entry)
{
    queue_entry *last;

    // the size and next members are both accessed from kernel code and from
    // interrupt handler, disable the update interrupt as short as possible
    disable_update_interrupt(device);

    // check whether the queue is empty or not
    if (queue->size == 0)
    {
        // queue is empty, add the entry at the start of the queue
        queue->next = (void *) entry;
        queue->last = (void *) entry;
    }
    else
    {
        // the queue is not empty, add the entry at the back
        last = (queue_entry *) queue->last;
        last->next = (void *) entry;
        queue->last = (void *) entry;
    }

    // increment the queue size
    queue->size++;

    // enable the update interrupt again
    enable_update_interrupt(device);

    return 0;
}

int get_queue(void *device, queue_root * queue,
                     queue_entry ** entry)
{
    queue_entry *first;

    // the size and next members are both accessed from kernel code and from
    // interrupt handler, disable the update interrupt as short as possible
    disable_update_interrupt(device);

    // check whether the queue is empty or not
    if (queue->size == 0)
    {
        // queue is empty, nothing to get
        // enable the update interrupt again
        enable_update_interrupt(device);

        return -1;
    }
    else
    {
        // the queue is not empty, get the entry at the start
        first = (queue_entry *) queue->next;
        queue->next = first->next;
        first->next = NULL;
        *entry = first;
    }

    // decrement the queue size
    queue->size--;

    // enable the update interrupt again
    enable_update_interrupt(device);

    return 0;
}

int queue_size(void *device, queue_root * queue)
{
    int qsize;

    // disable the update interrupt of the device
    disable_update_interrupt(device);

    // read the size of the queue from the root entry
    qsize = (int) queue->size;

    // enable the update interrupt of the device
    enable_update_interrupt(device);

    return qsize;
}

/******************************************************************************
    Miscellaneous
******************************************************************************/

void display_buffer(char *buffer, int length)
{
    if( ( pc_debug & SHOW_BUFFER_CONTENTS ) == 0 )
        return;

    while (length > 0)
    {
        printk( "[%02x]", *buffer & 255 );
        length--;
        buffer++;
    }

    printk( "\n" );
}

void print_frame(struct sk_buff *skb)
{
    int offset;

    printk("pr_frame: size %d, skb->data %p, skb %p\n",
        skb->len, skb->data, skb);
    for (offset = 0; offset < skb->len; offset++)
        printk("%02X ", skb->data[offset]);
    printk("\n");
}


void string_to_macaddress( unsigned char *string, unsigned char *mac )
{
    unsigned char *wp = mac, *rp;
    unsigned int value;
    unsigned int count = 0;

    for( rp = string; *rp != '\0'; rp++, wp++, count++ )
    {
        if( count == 6 )
            break;

        if( *rp >= '0' && *rp <= '9' )
            value = *rp - '0';
        else if( *rp >= 'a' && *rp <= 'f' )
            value = *rp - 'a' + 10;
        else if( *rp >= 'A' && *rp <= 'F' )
            value = *rp - 'A' + 10;
        else
            break;

        value <<= 4;
            rp++;
        if( *rp >= '0' && *rp <= '9' )
            value |= *rp - '0';
        else if( *rp >= 'a' && *rp <= 'f')
            value |= *rp - 'a' + 10;
        else if( *rp >= 'A' && *rp <= 'F')
            value |= *rp - 'A' + 10;
        else
            break;

        *wp = (unsigned char) value;
    }
}

int address_compare( unsigned char *address1, unsigned char *address2 )
{
    u32 *address1_hi, *address2_hi;
    u16 *address1_lo, *address2_lo;

    address2_hi = (u32 *) address2;
    address1_hi = (u32 *) address1;
    address2_lo = (u16 *) (address2 + 4);
    address1_lo = (u16 *) (address1 + 4);

    DEBUG(SHOW_TRACING, "address_compare: a1 %08x:%04x a2 %08x:%04x\n",
        *address1_hi, *address1_lo, *address2_hi, *address2_lo );

    return ((*address2_hi == *address1_hi) && (*address2_lo == *address1_lo));
}


