/*
 * linux/include/asm-armnommu/proc-armv/mpu.h
 *
 * ARM Memory Protection Unit support.
 */
/*  $Header$
 *
 *  Copyright (C) 2002 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef __ASM_PROC_MPU_H
#define __ASM_PROC_MPU_H

/*
 * ARM 9 Memory Protection Unit
 *
 * MPU regions are defined as follows:
 *
 * bit 31:12 	base address
 * bit 11:6     unused
 * bit 5:1      area size
 * bit 0	region enable
 *
 * The area size is coded as:
 *
 * 00000 to 01010 	reserved
 * 01011		 4 KB
 * 01100		 8 KB
 * 01101		16 KB
 * ....			 ....
 * 11110 		 2 GB
 * 11111		 4 GB
 *
 * We assume that all requests are properly aligned for the ARM9 MPU,
 * which is the case when the pages are allocated using the standard
 * buddy page allocator (an N-KB area must be aligned on an N-KB boundary).
 *
 */

/* Architecture specific MPU functions */
#include <asm/arch/mpu.h>

#define ARM_MPU_REGION( addr, order ) 	(addr + (order << 1) + 0x17)
#define ARM_MPU_DISABLED_REGION		(0x1f << 1)
#define ARM_MPU_REGION_ENABLED(mpu_reg)	(mpu_reg & 0x01)

#define ARM_MPU_AREA_START(mpu_reg)	(mpu_reg & 0xfffff000)
#define ARM_MPU_AREA_SIZE(mpu_reg)	(2 << ((mpu_reg & 0x3E) >> 1))
#define ARM_MPU_AREA_END(mpu_reg)	((mpu_reg & 0xfffff000) + \
					(2 << ((mpu_reg & 0x3E) >> 1)) - 1)

#define ARM_MAX_MPU_AREA_ORDER		20

#define MPU_CONTEXT_NEXT_IDX(idx)	((idx + 1 >= MAX_ACTIVE_MPU_AREAS) ? \
						MAX_STATIC_USER_MPU_AREAS : (idx + 1))

/* Enternal Interfaces to MPU functions */
extern void fixup_mpu(void);
extern void init_mpu_areas(struct mpu_context_t *);
extern int mpu_area_alloc(unsigned long , unsigned long , int , int);
extern void mpu_area_free(struct mpu_context_t *, unsigned long , unsigned long );
extern void mpu_area_exit(struct mpu_context_t *mpu_context);
extern void program_mpu_areas_kernel(void);
extern void program_mpu_areas_user(void);
extern int reprogram_mpu_areas(unsigned long , struct mpu_context_t *, int);

static inline int in_mpu_area(unsigned long addr, unsigned long mpu_reg)
{
	if(ARM_MPU_REGION_ENABLED(mpu_reg) &&
	   (addr >= ARM_MPU_AREA_START(mpu_reg)) &&
	   (addr <= ARM_MPU_AREA_END(mpu_reg)) )
		return 1;
	else
		return 0;
}

#endif /* __ASM_PROC_MPU_H */
