/*
 * include/asm-v850/bitops.h -- Bit operations
 *
 *  Copyright (C) 2001,2002  NEC Corporation
 *  Copyright (C) 2001,2002  Miles Bader <miles@gnu.org>
 *  Copyright (C) 1992  Linus Torvalds.
 *
 * This file is subject to the terms and conditions of the GNU General
 * Public License.  See the file COPYING in the main directory of this
 * archive for more details.
 */

#ifndef __V850_BITOPS_H__
#define __V850_BITOPS_H__


#include <linux/config.h>
#include <asm/byteorder.h>	/* swab32 */
#include <asm/system.h>		/* save_flags */


#ifdef __KERNEL__

/*
 * The __ functions are not atomic
 */

/*
 * ffz = Find First Zero in word. Undefined if no zero exists,
 * so code should check against ~0UL first..
 */
extern __inline__ unsigned long ffz (unsigned long word)
{
	unsigned long result = 0;

	while (word & 1) {
		result++;
		word >>= 1;
	}
	return result;
}


/*
 * Note -- the following definitions generate warnings of the following
 * form:  `asm operand 0 probably doesn't match constraints', which are
 * certainly bogus, because they are all register constrainst, so any
 * operand should just be forced into a register.  However I don't know
 * how to disable the warning.
 */

#define __const_bit_op(op, nr, addr)					      \
   ({ __asm__ (op " %0, %1"						      \
	       :: "i" ((nr) & 0x7),					      \
	          "m" (*((char *)(addr) + ((nr) >> 3)))			      \
	       : "memory"); })
#define __var_bit_op(op, nr, addr)					      \
   ({ int __nr = (nr);							      \
      __asm__ (op " %0, [%1]"						      \
	       :: "r" (__nr & 0x7),					      \
	          "r" ((char *)(addr) + (__nr >> 3))			      \
	       : "memory"); })
#define __bit_op(op, nr, addr)						      \
   (__builtin_constant_p (nr)						      \
    ? __const_bit_op (op, nr, addr)					      \
    : __var_bit_op (op, nr, addr))

#define __set_bit(nr, addr)		__bit_op ("set1", nr, addr)
#define __clear_bit(nr, addr)		__bit_op ("clr1", nr, addr)
#define __change_bit(nr, addr)		__bit_op ("not1", nr, addr)

/* The bit instructions used by `non-atomic' variants are actually atomic.  */
#define set_bit __set_bit
#define clear_bit __clear_bit
#define change_bit __change_bit

/* clear_bit doesn't provide any barrier for the compiler.  */
#define smp_mb__before_clear_bit()	barrier ()
#define smp_mb__after_clear_bit()	barrier ()


#define __const_tns_bit_op(op, nr, addr)				      \
   ({ int __tns_res;							      \
      __asm__ ("tst1 %1, %2; setf nz, %0; " op " %1, %2"		      \
	       : "=&r" (__tns_res)					      \
	       : "i" ((nr) & 0x7),					      \
	         "m" (*((char *)(addr) + ((nr) >> 3)))			      \
	       : "memory");						      \
      __tns_res;							      \
   })
#define __var_tns_bit_op(op, nr, addr)					      \
   ({ int __nr = (nr);							      \
      int __tns_res;							      \
      __asm__ ("tst1 %1, [%2]; setf nz, %0; " op " %1, [%2]"		      \
	       : "=&r" (__tns_res)					      \
	       : "r" (__nr & 0x7),					      \
	         "r" ((char *)(addr) + (__nr >> 3))			      \
	       : "memory");						      \
      __tns_res;							      \
   })
#define __tns_bit_op(op, nr, addr)					      \
   (__builtin_constant_p (nr)						      \
    ? __const_tns_bit_op (op, nr, addr)					      \
    : __var_tns_bit_op (op, nr, addr))
#define __tns_atomic_bit_op(op, nr, addr)				      \
   ({									      \
	   int __tns_atomic_res, __tns_atomic_flags;			      \
	   save_flags_cli (__tns_atomic_flags);				      \
	   __tns_atomic_res = __tns_bit_op (op, nr, addr);		      \
	   restore_flags (__tns_atomic_flags);				      \
	   __tns_atomic_res;						      \
   })

#define __test_and_set_bit(nr, addr)	__tns_bit_op ("set1", nr, addr)
#define test_and_set_bit(nr, addr)	__tns_atomic_bit_op ("set1", nr, addr)

#define __test_and_clear_bit(nr, addr)	__tns_bit_op ("clr1", nr, addr)
#define test_and_clear_bit(nr, addr)	__tns_atomic_bit_op ("clr1", nr, addr)

#define __test_and_change_bit(nr, addr)	__tns_bit_op ("not1", nr, addr)
#define test_and_change_bit(nr, addr)	__tns_atomic_bit_op ("not1", nr, addr)


/*
 * This routine doesn't need to be atomic.
 */
#define __const_test_bit(nr, addr)					      \
  ({									      \
    int __test_bit_res;							      \
    __asm__ __volatile__ ("tst1 %1, %2; setf nz, %0"			      \
			  : "=r" (__test_bit_res)			      \
			  : "i" (nr & 0x7),				      \
			    "m" (*((const char *)(addr) + ((nr) >> 3))));     \
    __test_bit_res;							      \
  })
extern __inline__ int __test_bit (int nr, void *addr)
{
	int res;
	__asm__ __volatile__ ("tst1 %1, [%2]; setf nz, %0"
			      : "=r" (res)
			      : "r" (nr & 0x7), "r" (addr + (nr >> 3)));
	return res;
}
#define test_bit(nr,addr)						      \
  (__builtin_constant_p (nr)						      \
   ? __const_test_bit ((nr), (addr))					      \
   : __test_bit ((nr), (addr)))


#define find_first_zero_bit(addr, size) \
  find_next_zero_bit ((addr), (size), 0)

extern __inline__ int find_next_zero_bit (void *addr, int size, int offset)
{
	unsigned long *p = ((unsigned long *) addr) + (offset >> 5);
	unsigned long result = offset & ~31UL;
	unsigned long tmp;

	if (offset >= size)
		return size;
	size -= result;
	offset &= 31UL;
	if (offset) {
		tmp = * (p++);
		tmp |= ~0UL >> (32-offset);
		if (size < 32)
			goto found_first;
		if (~tmp)
			goto found_middle;
		size -= 32;
		result += 32;
	}
	while (size & ~31UL) {
		if (~ (tmp = * (p++)))
			goto found_middle;
		result += 32;
		size -= 32;
	}
	if (!size)
		return result;
	tmp = *p;

 found_first:
	tmp |= ~0UL >> size;
 found_middle:
	return result + ffz (tmp);
}

#define ffs(x) generic_ffs (x)

/*
 * hweightN: returns the hamming weight (i.e. the number
 * of bits set) of a N-bit word
 */
#define hweight32(x) 			generic_hweight32 (x)
#define hweight16(x) 			generic_hweight16 (x)
#define hweight8(x) 			generic_hweight8 (x)

#define ext2_set_bit			test_and_set_bit
#define ext2_clear_bit			test_and_clear_bit
#define ext2_test_bit			test_bit
#define ext2_find_first_zero_bit	find_first_zero_bit
#define ext2_find_next_zero_bit		find_next_zero_bit

/* Bitmap functions for the minix filesystem.  */
#define minix_test_and_set_bit		test_and_set_bit
#define minix_set_bit			set_bit
#define minix_test_and_clear_bit	test_and_clear_bit
#define minix_test_bit 			test_bit
#define minix_find_first_zero_bit 	find_first_zero_bit

#endif /* __KERNEL__ */

#endif /* __V850_BITOPS_H__ */
