/*  $Header$
 *
 *  Copyright (C) 2001 Intersil Americas Inc
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <ctype.h>
#include <stdio.h>

/*****************************************************************************
 * NAME:
 *	read_text_line
 *
 * DESCRIPTION:
 * 	read a text line from a file, comments (starting with a #)
 *	are removed as well as whitespace.
 * 	the resulting tokens are seperated using #
 * 	and stored in a buffer and the size is returned
 *
 * PARAMETERS:
 * 	char * buffer		: pointer to buffer to store parsed line
 * 	unsigned int * bufsize	: IN - size of buffer, OUT - size of line
 * 	FILE * fp		: IN - file descriptor of file to parse
 *
 * RETURN:
 *	None
 *****************************************************************************/
void
read_text_line(
    char *buffer,
    unsigned int *bufsize,
    FILE *fp
)
{
    int c = 0, length = 0;

    /*
     * Eat whitespace, blank lines, and comment lines.
     */
top:
    c = fgetc(fp);
    if (c < 0) {
	goto done;			/* Exit if end-of-file */
    }
    if (isspace(c)) {
	goto top;			/* Skip over whitespace */
    }
    if (c == '#') {
	while( 1 ) {			/* Eat comments after # */
	    c = fgetc(fp);
	    if (c < 0) {
		goto done;		/* Exit if end-of-file */
	    }
	    if (c == '\n' ) {
		goto top;		/* Try to read the next line */
	    }
	}
    }
    ungetc(c, fp);	/* Other character, push it back to reprocess it */


    /*
     * Now we're actually reading a data entry.  Get each character and
     * assemble it into the data buffer, processing special characters like
     * double quotes (") and backslashes (\).
     */

mainloop:
    c = fgetc(fp);
    switch (c) {
    case EOF:
    case '\n':
	goto done;			/* Exit on EOF or newline */
    case '#':
	while( 1 ) {			/* Eat out comments after # */
	    c = fgetc( fp );
	    if( c < 0 || c == '\n' ) {
		goto done;		/* Exit on EOF or newline */
	    }
	}
    case ' ':				/* Eat out whitespace */
    case '\r':
    case '\f':
    case '\t':
    case '\v':
	while( 1 ) {
	    c = fgetc( fp );
	    if( c < 0 || c == '\n' ) {
		goto done;
	    } else if( c == '#' ) {	/* Eat out comments after # */
		while( 1 ) {
		    c = fgetc( fp );
		    if( c < 0 || c == '\n' ) {
			goto done;
		    }
		}
	    } else if( !isspace( c ) ) {
		    /* use delimitter = #  to seperate tokens */
		*buffer++ = '#'; length++;
		if( length >= *bufsize - 1 ) {
		    goto done;
		}
		*buffer++ = c; length++;
		if( length >= *bufsize - 1 ) {
		    goto done;
		}
		goto mainloop;
	    }
	}
	/* fall through if "other" character */
    default:
	*buffer++ = c;		/* Store other characters */
	length++;
	if (length >= *bufsize - 1) {
	    goto done;
	}
    }
    goto mainloop;			/* Keep going */

done:
    *buffer = '\0';			/* Terminate string */
    *bufsize = length;			/* Tell the caller its length */
}

