/*
 * gAVLTree.c: Source code for gAVLTrees.
 * Copyright (C) 1998,2001  Michael H. Buselli
 * This is version 0.1.3 (alpha).
 * Generated from $Id: xAVLTree.c.sh,v 1.5 2001/06/07 06:58:28 cosine Exp $
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The author of this library can be reached at the following address:
 * Michael H. Buselli
 * 30051 N. Waukegan Rd. Apt. 103
 * Lake Bluff, IL  60044-5412
 *
 * Or you can send email to <cosine@cosine.org>.
 * The official web page for this product is:
 * http://www.cosine.org/project/AVLTree/
 */

#include <stdlib.h>
#include "gAVLTree.h"

static gAVLNode *gAVLCloseSearchNode (gAVLTree const *avltree, gAVLKey key);
static void gAVLRebalanceNode (gAVLTree *avltree, gAVLNode *avlnode);
static void gAVLFreeBranch (gAVLNode *avlnode, void (freeitem)(void *item));
static void gAVLFillVacancy (gAVLTree *avltree,
        gAVLNode *origparent, gAVLNode **superparent,
        gAVLNode *left, gAVLNode *right);

#define MAX(x, y)      ((x) > (y) ? (x) : (y))
#define MIN(x, y)      ((x) < (y) ? (x) : (y))
#define L_DEPTH(n)     ((n)->left ? (n)->left->depth : 0)
#define R_DEPTH(n)     ((n)->right ? (n)->right->depth : 0)
#define CALC_DEPTH(n)  (MAX(L_DEPTH(n), R_DEPTH(n)) + 1)


/*
 * AVLAllocTree:
 * Allocate memory for a new AVL tree and set the getkey function for
 * that tree.  The getkey function should take an item and return an
 * AVLKey that is to be used for indexing this object in the AVL tree.
 * On success, a pointer to the malloced AVLTree is returned.  If there
 * was a malloc failure, then NULL is returned.
 */
gAVLTree *gAVLAllocTree (gAVLKey (*getkey)(void const *item),
        gAVLCompare cmp)
{
  gAVLTree *rc;

  rc = malloc(sizeof(gAVLTree));
  if (rc == NULL)
    return NULL;

  rc->top = NULL;
  rc->count = 0;
  rc->getkey = getkey;
  rc->cmp = cmp;
  return rc;
}


/*
 * AVLFreeTree:
 * Free all memory used by this AVL tree.  If freeitem is not NULL, then
 * it is assumed to be a destructor for the items reference in the AVL
 * tree, and they are deleted as well.
 */
void gAVLFreeTree (gAVLTree *avltree, void (freeitem)(void *item))
{
  if (avltree->top)
    gAVLFreeBranch(avltree->top, freeitem);
  free(avltree);
}


/*
 * AVLInsert:
 * Create a new node and insert an item there.
 *
 * Returns  0 on success,
 *         -1 on malloc failure,
 *          3 if duplicate key.
 */
int gAVLInsert (gAVLTree *avltree, void *item)
{
  gAVLNode *newnode;
  gAVLNode *node;
  gAVLNode *balnode;
  gAVLNode *nextbalnode;
  int cmp_ret;

  newnode = malloc(sizeof(gAVLNode));
  if (newnode == NULL)
    return -1;

  newnode->key = avltree->getkey(item);
  newnode->item = item;
  newnode->depth = 1;
  newnode->left = NULL;
  newnode->right = NULL;
  newnode->parent = NULL;

  if (avltree->top != NULL) {
    node = gAVLCloseSearchNode(avltree, newnode->key);

    cmp_ret = gAVLKey_cmp(avltree, newnode->key, node->key);
    
    if (!cmp_ret) {
      free(newnode);
      return 3;
    }

    newnode->parent = node;

    if (cmp_ret < 0) {
      node->left = newnode;
      node->depth = CALC_DEPTH(node);
    }

    else {
      node->right = newnode;
      node->depth = CALC_DEPTH(node);
    }

    for (balnode = node->parent; balnode; balnode = nextbalnode) {
      nextbalnode = balnode->parent;
      gAVLRebalanceNode(avltree, balnode);
    }
  }

  else {
    avltree->top = newnode;
  }

  avltree->count++;
  return 0;
}


/*
 * gAVLSearch:
 * Return a pointer to the item with the given key in the AVL tree.  If
 * no such item is in the tree, then NULL is returned.
 */
void *gAVLSearch (gAVLTree const *avltree, gAVLKey key)
{
  gAVLNode *node;

  node = gAVLCloseSearchNode(avltree, key);

  if (node && !gAVLKey_cmp(avltree, node->key, key))
    return node->item;

  return NULL;
}


/*
 * ID Extended
 *
 * gAVLSearchCursor:
 * Return a pointer to the item with the given key in the AVL tree.  If
 * no such item is in the tree, then NULL is returned.
 */
 
void *gAVLSearchCursor (gAVLTree const *avltree, gAVLKey key, gAVLCursor *cursor)
{
  gAVLNode *node;

  node = gAVLCloseSearchNode(avltree, key);

  if (node && !gAVLKey_cmp(avltree, node->key, key)) {
    cursor->avltree = avltree;
    cursor->curnode = node;
    return node->item;
  }

  return NULL;
}


/*
 * gAVLDelete:
 * Deletes the node with the given key.  Does not delete the item at
 * that key.  Returns 0 on success and -1 if a node with the given key
 * does not exist.
 */
int gAVLDelete (gAVLTree *avltree, gAVLKey key)
{
  gAVLNode *avlnode;
  gAVLNode *origparent;
  gAVLNode **superparent;

  avlnode = gAVLCloseSearchNode(avltree, key);
  if (avlnode == NULL || gAVLKey_cmp(avltree, avlnode->key, key))
    return -1;

  origparent = avlnode->parent;

  if (origparent) {
    if (gAVLKey_cmp(avltree, avlnode->key, avlnode->parent->key) < 0)
      superparent = &(avlnode->parent->left);
    else
      superparent = &(avlnode->parent->right);
  }
  else
    superparent = &(avltree->top);

  gAVLFillVacancy(avltree, origparent, superparent,
                  avlnode->left, avlnode->right);
  free(avlnode);
  avltree->count--;
  return 0;
}


/*
 * gAVLFirst:
 * Initializes an gAVLCursor object and returns the item with the lowest
 * key in the gAVLTree.
 */
void *gAVLFirst (gAVLCursor *avlcursor, gAVLTree const *avltree)
{
  const gAVLNode *avlnode;

  avlcursor->avltree = avltree;

  if (avltree->top == NULL) {
    avlcursor->curnode = NULL;
    return NULL;
  }

  for (avlnode = avltree->top;
       avlnode->left != NULL;
       avlnode = avlnode->left);
  avlcursor->curnode = avlnode;
  return avlnode->item;
}


/*
 * gAVLNext:
 * Called after an gAVLFirst() call, this returns the item with the least
 * key that is greater than the last item returned either by gAVLFirst()
 * or a previous invokation of this function.
 */
void *gAVLNext (gAVLCursor *avlcursor)
{
  const gAVLNode *avlnode;

  avlnode = avlcursor->curnode;

  if (avlnode->right != NULL) {
    for (avlnode = avlnode->right;
         avlnode->left != NULL;
         avlnode = avlnode->left);
    avlcursor->curnode = avlnode;
    return avlnode->item;
  }

  while (avlnode->parent && avlnode->parent->left != avlnode) {
    avlnode = avlnode->parent;
  }

  if (avlnode->parent == NULL) {
    avlcursor->curnode = NULL;
    return NULL;
  }

  avlcursor->curnode = avlnode->parent;
  return avlnode->parent->item;
}

/*
 * gAVLPrev:
 * This returns the item with the largest key that is smaller
 * than the key of the node pointed to by the cursor.
 */
void *gAVLPrev (gAVLCursor *avlcursor)
{
  const gAVLNode *avlnode;

  avlnode = avlcursor->curnode;

  if (avlnode->left != NULL) {
    for (avlnode = avlnode->left;
         avlnode->right != NULL;
         avlnode = avlnode->right);
    avlcursor->curnode = avlnode;
    return avlnode->item;
  }

  while (avlnode->parent && avlnode->parent->right != avlnode) {
    avlnode = avlnode->parent;
  }

  if (avlnode->parent == NULL) {
    avlcursor->curnode = NULL;
    return NULL;
  }

  avlcursor->curnode = avlnode->parent;
  return avlnode->parent->item;
}


/*
 * gAVLSearch:
 * Return a pointer to the item in the AVL tree with the smallest key
 * that is larger than the passed key. Returns NULL if no such item exists.
 */
void *gAVLSearchNext(gAVLTree const *avltree, gAVLKey key, gAVLCursor *avlcursor)
{
  gAVLNode *node;

  avlcursor->avltree = avltree;
  avlcursor->curnode = NULL;
  node = avltree->top;

  if (!node)
    return NULL;

  node = gAVLCloseSearchNode(avltree, key);
  avlcursor->curnode = node;

  if (gAVLKey_cmp(avltree, node->key, key) <= 0)
    return gAVLNext(avlcursor);
  else
    return node->item;
}



/*
 * gAVLCloseSearchNode:
 * Return a pointer to the node closest to the given key.
 * Returns NULL if the AVL tree is empty.
 */
gAVLNode *gAVLCloseSearchNode (gAVLTree const *avltree, gAVLKey key)
{
  gAVLNode *node;
  int cmp_ret;

  node = avltree->top;

  if (!node)
    return NULL;

  for (;;) {

    cmp_ret = gAVLKey_cmp(avltree, node->key, key);
    
    if (!cmp_ret)
      return node;

    if (cmp_ret < 0) {
      if (node->right)
        node = node->right;
      else
        return node;
    }

    else {
      if (node->left)
        node = node->left;
      else
        return node;
    }
  }
}


/*
 * gAVLRebalanceNode:
 * Rebalances the AVL tree if one side becomes too heavy.  This function
 * assumes that both subtrees are AVL trees with consistant data.  This
 * function has the additional side effect of recalculating the depth of
 * the tree at this node.  It should be noted that at the return of this
 * function, if a rebalance takes place, the top of this subtree is no
 * longer going to be the same node.
 */
void gAVLRebalanceNode (gAVLTree *avltree, gAVLNode *avlnode)
{
  long depthdiff;
  gAVLNode *child;
  gAVLNode *gchild;
  gAVLNode *origparent;
  gAVLNode **superparent;

  origparent = avlnode->parent;

  if (origparent) {
    if (gAVLKey_cmp(avltree, avlnode->key, avlnode->parent->key) < 0)
      superparent = &(avlnode->parent->left);
    else
      superparent = &(avlnode->parent->right);
  }
  else
    superparent = &(avltree->top);

  depthdiff = R_DEPTH(avlnode) - L_DEPTH(avlnode);

  if (depthdiff <= -2) {
    child = avlnode->left;

    if (L_DEPTH(child) >= R_DEPTH(child)) {
      avlnode->left = child->right;
      if (avlnode->left != NULL)
        avlnode->left->parent = avlnode;
      avlnode->depth = CALC_DEPTH(avlnode);
      child->right = avlnode;
      if (child->right != NULL)
        child->right->parent = child;
      child->depth = CALC_DEPTH(child);
      *superparent = child;
      child->parent = origparent;
    }

    else {
      gchild = child->right;
      avlnode->left = gchild->right;
      if (avlnode->left != NULL)
        avlnode->left->parent = avlnode;
      avlnode->depth = CALC_DEPTH(avlnode);
      child->right = gchild->left;
      if (child->right != NULL)
        child->right->parent = child;
      child->depth = CALC_DEPTH(child);
      gchild->right = avlnode;
      if (gchild->right != NULL)
        gchild->right->parent = gchild;
      gchild->left = child;
      if (gchild->left != NULL)
        gchild->left->parent = gchild;
      gchild->depth = CALC_DEPTH(gchild);
      *superparent = gchild;
      gchild->parent = origparent;
    }
  }

  else if (depthdiff >= 2) {
    child = avlnode->right;

    if (R_DEPTH(child) >= L_DEPTH(child)) {
      avlnode->right = child->left;
      if (avlnode->right != NULL)
        avlnode->right->parent = avlnode;
      avlnode->depth = CALC_DEPTH(avlnode);
      child->left = avlnode;
      if (child->left != NULL)
        child->left->parent = child;
      child->depth = CALC_DEPTH(child);
      *superparent = child;
      child->parent = origparent;
    }

    else {
      gchild = child->left;
      avlnode->right = gchild->left;
      if (avlnode->right != NULL)
        avlnode->right->parent = avlnode;
      avlnode->depth = CALC_DEPTH(avlnode);
      child->left = gchild->right;
      if (child->left != NULL)
        child->left->parent = child;
      child->depth = CALC_DEPTH(child);
      gchild->left = avlnode;
      if (gchild->left != NULL)
        gchild->left->parent = gchild;
      gchild->right = child;
      if (gchild->right != NULL)
        gchild->right->parent = gchild;
      gchild->depth = CALC_DEPTH(gchild);
      *superparent = gchild;
      gchild->parent = origparent;
    }
  }

  else {
    avlnode->depth = CALC_DEPTH(avlnode);
  }
}


/*
 * gAVLFreeBranch:
 * Free memory used by this node and its item.  If the freeitem argument
 * is not NULL, then that function is called on the items to free their
 * memory as well.  In other words, the freeitem function is a
 * destructor for the items in the tree.
 */
void gAVLFreeBranch (gAVLNode *avlnode, void (freeitem)(void *item))
{
  if (avlnode->left)
    gAVLFreeBranch(avlnode->left, freeitem);
  if (avlnode->right)
    gAVLFreeBranch(avlnode->right, freeitem);
  if (freeitem)
    freeitem(avlnode->item);
  free(avlnode);
}


/*
 * gAVLFillVacancy:
 * Given a vacancy in the AVL tree by it's parent, children, and parent
 * component pointer, fill that vacancy.
 */
void gAVLFillVacancy (gAVLTree *avltree,
                      gAVLNode *origparent, gAVLNode **superparent,
                      gAVLNode *left, gAVLNode *right)
{
  gAVLNode *avlnode;
  gAVLNode *balnode;
  gAVLNode *nextbalnode;

  if (left == NULL) {
    if (right)
      right->parent = origparent;

    *superparent = right;
    balnode = origparent;
  }

  else {
    for (avlnode = left; avlnode->right != NULL; avlnode = avlnode->right);

    if (avlnode == left) {
      balnode = avlnode;
    }
    else {
      balnode = avlnode->parent;
      balnode->right = avlnode->left;
      if (balnode->right != NULL)
        balnode->right->parent = balnode;
      avlnode->left = left;
      left->parent = avlnode;
    }

    avlnode->right = right;
    if (right != NULL)
      right->parent = avlnode;
    *superparent = avlnode;
    avlnode->parent = origparent;
  }

  for (; balnode; balnode = nextbalnode) {
    nextbalnode = balnode->parent;
    gAVLRebalanceNode(avltree, balnode);
  }
}
