/*  $Header$
 *
 *  Copyright (C) 2002 Intersil Americas Inc
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <string.h>
#include <rockhopper.h>
#include <flash_layout.h>

#include "inl_debug.h"
#include "sockutil.h"
#include "mtd.h"
#include "prismoids.h"

#include "br.h"
#include "pda.h"
#include "image.h"
#include "identification/image_header.h"
#include "identification/identification.h"
#include "identification/bootrec.h"
#include "image_check.h"
#include "linux_mvc_oids.h"

/* Define for debugging output */
#define D(x) 

#ifndef offsetof
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

/* interface name for the extension card. */
#define ext_card_ifname "eth3"

typedef enum {
        BR_TYPE_INVALID = -1,
        BR_TYPE_DAP1 = 0,
        BR_TYPE_LTM1,
        BR_TYPE_ITM1,
        BR_TYPE_DMT1,
        BR_TYPE_DBO1,
        BR_TYPE_DFS1,
        BR_TYPE_HW,
	BR_TYPE_DBCO,
        NOF_BOOTREC_TYPES
} bootrec_type_t;

/* local function prototypes */
static int find_bootrec(const struct s_bootrecord *bootrec_start, unsigned long bootrec_code,
                        struct s_bootrecord **matched_bootrec);
static upg_image_check_t failed_reason(struct s_bootrecord *failed_record);
static int match_interfaces(const struct s_bootrecord *dep_record,
                        const struct s_bootrecord *exp_record);
static int find_matching_if(const struct s_bootrecord *match_record, long *bootrecs[]);
static upg_image_check_t image_interfaces_check(long *current_bootrecs[], long *new_bootrecs[]);
static void ext_card_default_bootrecord(void *bootrecord_ext_card);
static long *get_ext_card_bootrecord(void);

/*****************************************************************************
 * NAME:
 *  bootrec_type
 *
 * DESCRIPTION:
 *  Convert "magic2" from the image header to a bootrecord type or (in case of DUALBAND AP):
 *  Convert a component_id to a bootrecord type
 *
 * PARAMETERS:
 *  unsigned long magic2        - IN -        "magic2" from image header
 *  for DUALBAND AP: s_bootrec_comp_id *comp_id
 *
 * RETURN:
 *   returns BR_TYPE_xxx or BR_TYPE_INVALID for an invalid magic2
 *****************************************************************************/

static bootrec_type_t bootrec_type(unsigned long component_id)
{
   switch (component_id) {
      case BR_COMP_ID_DAP1:
            return BR_TYPE_DAP1;
      case BR_COMP_ID_LTM1:
            return BR_TYPE_LTM1;
      case BR_COMP_ID_ITM1:
            return BR_TYPE_ITM1;
      case BR_COMP_ID_DMT1:
            return BR_TYPE_DMT1;
      case BR_COMP_ID_DBO1:
            return BR_TYPE_DBO1;
      case BR_COMP_ID_DFS1:
            return BR_TYPE_DFS1;
      case BR_COMP_ID_ISL39300A:
         return BR_TYPE_HW;
      case BR_COMP_ID_DBCO:
	  return BR_TYPE_DBCO;
      default:
            warn("Unknown image type : %08X", component_id);
   }

   return BR_TYPE_INVALID;
}

/*****************************************************************************
 * NAME:
 *  find_bootrec
 *
 * DESCRIPTION:
 *  Search the bootrecord area starting at "bootrec_start" for a bootrecord
 *  with code "bootrec_code". The first occurrence that is found is returned.
 *
 * PARAMETERS:
 *  struct s_bootrecord *bootrec_start
 *                         - IN - start of bootrecord area to search for
 *  unsigned long bootrec_code
 *                         - IN - bootrecord code to search for
 *  struct s_bootrecord **matched_bootrec
 *                         - OUT - if a bootrecord is found, this is a pointer to it.
 *
 * RETURN:
 *   0 if bootrecord is found
 *   1 if no bootrecord is found
 *  <0 for failure
 *****************************************************************************/
static int find_bootrec(const struct s_bootrecord *bootrec_start, unsigned long bootrec_code,
                        struct s_bootrecord **matched_bootrec)
{
        struct s_bootrecord *record = (struct s_bootrecord *) bootrec_start;
        int last_record = 0;
        
        do {
                if((!record) || (valid_bootrecord(record) < 0)) {
                        error("Find bootrec: No valid bootrecord");
                        return -1;
                }

                if(record->hdr.code == bootrec_code) {
                        *matched_bootrec = record;
                        return 0;
                } else if(record->hdr.code == BR_CODE_END_OF_BRA) {
                        last_record = 1;
                }
                record = next_bootrecord(record);
        } while(!last_record);

        /* no match */
        return 1;
}

static upg_image_check_t failed_reason(struct s_bootrecord *failed_record)
{
        if(failed_record->hdr.code == BR_CODE_DEPENDENT_IF) {
                struct s_bootrec_dep_if *dependent_if =
                        (struct s_bootrec_dep_if *)&(failed_record->start_of_data);
                switch(dependent_if->if_id) {
                        case BR_IF_ID_ISL39300A:
                        case BR_IF_ID_ISL39000:	/* MAC Controller HW */
                        case BR_IF_ID_ISL37700M:
                                return upg_image_unsupported_hardware_revs;
                        case BR_IF_ID_PRODUCT_IF:
                                return upg_image_unsupported_product;
                        case BR_IF_ID_OEM_IF:
                                return upg_image_unsupported_oem_id;
                        case BR_IF_ID_TFTP:	/* TFTP Interface Application <-> Firmware */
                                return upg_image_unsupported_tftp_if;
                        case BR_IF_ID_DBO1:	/* Interface between BR_COMP_ID_DBO1 <-> Firmware */
                                return upg_image_unsupported_boot;
                        case BR_IF_ID_DAP1_MAP:	/* Interface between BR_COMP_ID_DAP1 <-> Micro AP Firmware */
                                return upg_image_unsupported_uapfw;
                        case BR_IF_ID_DAP1_FS:	/* Interface between BR_COMP_ID_DAP1 <-> File Set */
			case BR_IF_ID_SNMP_IF:  /* Interface between BR_COMP_ID_DAP1 <-> File Set */
                                return upg_image_unsupported_webfileset;
                        default:
                                warn("Unknown Dependent Interface (%d) Failed reason", dependent_if->if_id);
                                return upg_image_invalid;
                }
        }
        warn("Unknown Failed reason");
        return upg_image_invalid;
}

/*****************************************************************************
 * NAME:
 *        match_interfaces
 *
 * DESCRIPTION:
 *        Find a matching exposed interface exp_record for the dependent
 *        interfaces passed in dep_record. A match is found if the interface
 *        code of the dependent and exposed interface match or if the variant
 *        of the dependent interface is within the compatibility range of the
 *        exposed interface (with matching interface role and ID).
 *
 * PARAMETERS:
 *        struct s_bootrecord *dep_record
 *                                 - IN -        bootrecord with dependent interfaces
 *        struct s_bootrecord *exp_record
 *                                - IN -        bootrecord with exposed interfaces
 * RETURN:
 *        0  found a matching exposed interface
 *        1  found no matching exposed interface
 *        <0 for failure
 *****************************************************************************/

static int match_interfaces(const struct s_bootrecord *dep_record,
                            const struct s_bootrecord *exp_record)
{
        int i, j, nof_dep_if, nof_exp_if;
        struct s_bootrec_exp_if *exposed_if;
        struct s_bootrec_dep_if *dependent_if;

        if(((nof_dep_if = nof_dep_interfaces(dep_record)) < 0) ||
           ((nof_exp_if = nof_exp_interfaces(exp_record)) < 0))
                return -1;

        dependent_if = (struct s_bootrec_dep_if *)&(dep_record->start_of_data);
        for(i = 0; i < nof_dep_if; i++, dependent_if++) {
                exposed_if = (struct s_bootrec_exp_if *)&(exp_record->start_of_data);
                for(j = 0; j < nof_exp_if; j++, exposed_if++) {
                        D(note("Matching dependent IF [%2.2d, %2.2d, %2.2d] vs exposed IF [%2.2d, %2.2d, %2.2d, %2.2d, %2.2d]",
                                dependent_if->role, dependent_if->if_id, dependent_if->variant,
                                exposed_if->role, exposed_if->if_id, exposed_if->variant,
                                exposed_if->btm_compat, exposed_if->top_compat));

                        if((dependent_if->role == exposed_if->role) &&
                           (dependent_if->if_id == exposed_if->if_id)) {
                                /* Exact Match */
                                if(dependent_if->variant == exposed_if->variant) {
                                        D(note("Exact match\n"));
                                        return 0;
                                }
                                /* Match within compatibility range */
                                if((dependent_if->variant >= exposed_if->btm_compat) &&
                                   (dependent_if->variant <= exposed_if->top_compat)) {
                                        D(note("Match within compatibility range.\n"));
                                        return 0;
                                }
                        }
                        D(note("NO Match\n"));
                }
        }
        /* No match found */
        return 1;
}

/*****************************************************************************
 * NAME:
 *        find_matching_if
 *
 * DESCRIPTION:
 *        Find a matching exposed interface in bootrecs[] for the dependent
 *        interfaces passed in match_record. A match is found if at least one
 *        of the dependent interfaces maches with an exposed interface
 *
 * PARAMETERS:
 *        struct s_bootrecord *match_record
 *                                 - IN -        bootrecord with dependent interface
 *        char *bootrecs[]        - IN -        array of bootrecords to        search for
 *                                        exposed interface that matches the
 *                                        dependent interface.
 * RETURN:
 *        0  found a matching exposed interface
 *        1  found no matching exposed interface
 *        <0 for failure
 *****************************************************************************/
static int find_matching_if(const struct s_bootrecord *match_record, long *bootrecs[])
{
        int i, ret;
        struct s_bootrecord *start_record;
        struct s_bootrecord *record = NULL;

        if(match_record->hdr.code != BR_CODE_DEPENDENT_IF)
                return -1;

        for(i = 0; (start_record = (struct s_bootrecord *)bootrecs[i]); i++) {
                while((ret = find_bootrec(start_record, BR_CODE_EXPOSED_IF, &record)) <= 0) {
                        if((ret < 0) || (ret = match_interfaces(match_record, record)) <= 0) {
                                /* match or failure, return in either case */
                                return ret;
                        }
                        /* Continue searching after the found record */
                        start_record = next_bootrecord(record);
                }
        }
        return 1;
}
/*****************************************************************************
 * NAME:
 *        image_interfaces_check
 *
 * DESCRIPTION:
 *        Check if the images which bootrecords are passed in "new_bootrecs"
 *        are supported by this device as configured in the hardconfig and if
 *        all required interfaces are found in the bootrecords of the images.
 *        A required bootrecord of an image in "new_bootrecs" needs to be
 *        exported by another image in "new_bootrecs" (e.g. an interface that is
 *        needed by the new system may be available in the new MVC).
 *        For images that are not available in "new_bootrecs", the bootrecords in
 *        "current_bootrecs" are searched for a matching interface.
 *
 * PARAMETERS:
 *        long *current_bootrecs[] - IN -        array of bootrecords of components
 *                                        that are currentlt in Flash, terminated
 *                                        with a NULL pointer. NOTE: Components
 *                                        that will be replaced by components in
 *                                        "new_bootrecs" should not occur in this
 *                                        array.
 *        long *new_bootrecs[]     - IN - array of bootrecords of components
 *                                        that are new (i.e. uploaded), terminated
 *                                        with a NULL pointer.
 *
 * PRECONDITION:
 *        The Hardconfig must be cached prior to calling this function.
 *
 * RETURN:
 *        upg_image_ok (=0) for a valid image
 *        (upg_image_check_t) >0 for invalid image or failure
 *****************************************************************************/
static upg_image_check_t
image_interfaces_check(long *current_bootrecs[], long *new_bootrecs[])
{
        int i, ret;
        struct s_bootrecord *start_record;
        struct s_bootrecord *record = NULL;

        /* Check dependent interfaces in new_bootrecs against the
         * exposed interfaces in hardconfig, new and current bootrecords
         */
         for(i = 0; (start_record = (struct s_bootrecord *)new_bootrecs[i]); i++) {
                while((ret = find_bootrec(start_record, BR_CODE_DEPENDENT_IF, &record)) <= 0) {
                        if(ret < 0)
                                return upg_image_invalid;

                        if(find_matching_if(record, current_bootrecs) &&
                           find_matching_if(record, new_bootrecs)) {
                                   return failed_reason(record);
                        }
                        /* Continue searching after the found record */
                        start_record = next_bootrecord(record);
                }
        }

        /* Check dependent interfaces in current_bootrecs against the 
         * exposed interfaces in hardconfig, current and new bootrecords.
         */
         for(i = 0; (start_record = (struct s_bootrecord *)current_bootrecs[i]); i++) {
                while((ret = find_bootrec(start_record, BR_CODE_DEPENDENT_IF, &record)) <= 0) {
                        if(ret < 0)
                                return upg_image_invalid;

                        if(find_matching_if(record, current_bootrecs) &&
                           find_matching_if(record, new_bootrecs)) {
                                   return failed_reason(record);
                        }
                        /* Continue searching after the found record */
                        start_record = next_bootrecord(record);
                }
         }

        return upg_image_ok;
}


/*
 * Read 'len' bytes from file descriptor 'fd' starting at 'offset'
 * Returns 0 on success and -1 on failure.
 */
int image_read(int fd, int offset, char *buf, int len)
{
	int cnt;

	if(len == 0)
		return -1;

	if(lseek(fd, offset, SEEK_SET) != offset)
		return -1;

	if((cnt = readn(fd, buf, len)) != len)
		return -1;

	return 0;
}


/*****************************************************************************
 * NAME:
 *  get_current_bootrecords
 *
 * DESCRIPTION:
 *  Retrieve the bootrecords of all components that are currently in the Flash
 *  and that are not passed in "new_bootrecs".
 *
 * PARAMETERS:
 *  long *current_bootrecs[] - OUT - array of bootrecords of components
 *                        that are currently in Flash.
 *  long *new_bootrecs[]     - IN - array of bootrecords of components
 *                        that are new (i.e. uploaded), terminated
 *                        with a NULL pointer. For retrieving all bootrecords,
 *                        this should be a NULL pointer.
 * PRECONDITION:
 *  The Hardconfig must be cached prior to calling this function.
 *
 * RETURN:
 *  the number fo bootrecords that were found.
 *  <0 for failure
 *****************************************************************************/
int get_current_bootrecords(const char *search_devices[], long *current_bootrecs[],
                            long *new_bootrecs[])
{
        int i, fd;
        int type, count;
        image_header_t image_header;
        
        struct s_bootrec_comp_id *comp_id = NULL; /* Keep the compiler happy */
        struct s_bootrec_comp_id *new_comp_id =  NULL;
        long *tmp_bootrecs[NOF_BOOTREC_TYPES];

        memset(tmp_bootrecs, 0, sizeof(tmp_bootrecs));
        /* Loop over all devices that can contain components */
        for(i = 0; search_devices[i]; i++) {
                unsigned long offset;

                D(note("Searching device %s for valid components", search_devices[i]));

                if(strstr(search_devices[i], MTD_BOOTLOADER) ||
                   strstr(search_devices[i], "mtd0")) {
                        if(pda_Cache(MTD_BOOTLOADER)) {
                                error("Caching of PDA failed");
                        } else {
                                D(note("Getting hardconfig bootrecord for bootloader"));
				tmp_bootrecs[BR_TYPE_HW] = (long *)pdaGetBootRecords();
                        }
                }

                if ((fd = open(search_devices[i], O_RDONLY )) < 0) {
                        warn("Failed opening device %s: %m", search_devices[i]);
                        continue;
                }

                /* Search the entire device for valid components */
                for(offset = 0;
                    image_read(fd, offset, (char *)&image_header, sizeof(image_header)) == 0;
                    offset += 0x200) {
                        if (image_header.magic1 != MAGIC_OKAY) {
                           continue;
                        }
                        type = bootrec_type(image_header.bootrec[2]);

                        /* Check if invalid or we already found an image of this type */
                        if (type == BR_TYPE_INVALID || tmp_bootrecs[type]) {
                           continue;
                        }

                        D(note("Found component at offset 0x%.8x", offset));
                        /* Now read the bootrecords for this image */
                        tmp_bootrecs[type] = (long *)malloc(BR_MAX_BRA_SIZE);
                        if(image_read(fd, offset + offsetof(struct image_header_st, bootrec),
                                      (char *)tmp_bootrecs[type], BR_MAX_BRA_SIZE) < 0) {
                                warn("Failed to read bootrecords from device %s at offset %d",
                                     search_devices[i],
                                     offset + offsetof(struct image_header_st, bootrec));
                                continue;
                        }

                        /* Check if the component in Flash is a component that is uploaded */
                        if(find_bootrec((struct s_bootrecord *)tmp_bootrecs[type],
                                        BR_CODE_COMPONENT_ID, (struct s_bootrecord **)&comp_id)) {
                                warn("Could not find Component ID at offset %d, device %s",
                                     offset + offsetof(struct image_header_st, bootrec),
                                     search_devices[i]);
                                free(tmp_bootrecs[type]);
                                tmp_bootrecs[type] = NULL;
                        } else {
                                int j, found = 0;

                                D(note(" Component ID = 0x%.8lx", comp_id->id));
                                for(j = 0; new_bootrecs && new_bootrecs[j] && !found; j++) {
                                        if(find_bootrec((struct s_bootrecord *)new_bootrecs[j],
                                                        BR_CODE_COMPONENT_ID, (struct s_bootrecord **)&new_comp_id)) {
                                                /* This probably means an invalid pointer, so we fail */
                                                warn("Could not find Component ID in new bootrecs %d", j);
                                                close(fd);
                                                return -1;
                                        }
                                        if(new_comp_id->id == comp_id->id) {
                                                free(tmp_bootrecs[type]);
                                                tmp_bootrecs[type] = NULL;
                                                found = 1;
                                        }
                                }
                        }
                }
                close(fd);
        }
        
        if(tmp_bootrecs[BR_TYPE_ITM1] == 0) {
		/* Add the bootrecord of the extension card. */
        	tmp_bootrecs[BR_TYPE_ITM1] = get_ext_card_bootrecord();
	}
        
        for(i = 0, count = 0; i < NOF_BOOTREC_TYPES; i++) {
                if(tmp_bootrecs[i]) {
                        current_bootrecs[count] = tmp_bootrecs[i];
                        count++;
                }
        }
        
        current_bootrecs[count] = NULL;
        return count;
}

/*****************************************************************************
 * NAME:
 *        image_file_check
 *
 * DESCRIPTION:
 *        Check if the given images are supported by this device as configured in
 *        the PDA. The bootrecord fields after each header are searched.
 *        For correct functioning the needed PDA parameters should have
 *        been cached.
 *
 * PARAMETERS:
 *        image_header_t *        - IN  - array of pointers to file headers,
 *                                        terminated with a NULL pointer
 *
 * RETURN:
 *        upg_image_ok (=0) for a valid image
 *        >0 for invalid image
 *****************************************************************************/
upg_image_check_t
image_file_check(image_header_t *hdr[])
{
   int i = 0;

   long  *new_bootrecs[NOF_BOOTREC_TYPES];
   long  *current_bootrecs[NOF_BOOTREC_TYPES];

   upg_image_check_t   ret;

   struct s_bootrec_comp_id  *component_id_ptr;

   const char *search_devs[] = {
	   MTD_BOOTLOADER,
	   MTD_APFIRMWARE,
	   FILE_WEBFILESET,
	   FILE_UAP_FIRMWARE,
	   NULL};

   if (hdr[i] == NULL) {
      return upg_image_invalid;
   }

   while(hdr[i]) {
      if (hdr[i]->magic1 != MAGIC_OKAY) {
         return upg_image_invalid;
      }

      /*--------------------*
       | Check Component ID |
       *--------------------*/
      component_id_ptr = (struct s_bootrec_comp_id *)(&(hdr[i]->bootrec[0]));

      if ((component_id_ptr->id != BR_COMP_ID_DAP1) &&
          (component_id_ptr->id != BR_COMP_ID_LTM1) &&
          (component_id_ptr->id != BR_COMP_ID_ITM1) &&
          (component_id_ptr->id != BR_COMP_ID_DFS1) &&
          (component_id_ptr->id != BR_COMP_ID_DBCO)) {

         return upg_image_invalid;
      }

      new_bootrecs[i] = (long *)&(hdr[i]->bootrec[0]);
      i++;
   }
   new_bootrecs[i] = NULL;

   if (get_current_bootrecords(search_devs, current_bootrecs, new_bootrecs) < 0) {
      return upg_image_invalid;
   }
   
   ret = image_interfaces_check(current_bootrecs, new_bootrecs);

   for (i = 0; current_bootrecs[i]; i++) {
       free(current_bootrecs[i]);
   }

   return ret;
}

/* 
 *  Copy an extension card default bootrecord into bootrecord_ext_card.
 */
static void ext_card_default_bootrecord(void *bootrecord_ext_card)
{
    memcpy(bootrecord_ext_card, &ext_card_def_boot_record, sizeof(ext_card_def_boot_record));
}


/**************************************************************************
 *
 *   Get the bootrecord of the extension card from the driver.
 *   If no extension card is available or the firmware for the 
 *   extension card is not suitable for this card,
 *   a default set of exposed interfaces will be presented.
 *   
 **************************************************************************/
static long *get_ext_card_bootrecord(void)
{
    long *bootrecord_ext_card;
    
    /* Malloc a buffer instead of pointing to the struct, as all bootrecords are freed after use. */
    bootrecord_ext_card = (long *)malloc((size_t)BR_MAX_BRA_SIZE);

    if (!bootrecord_ext_card) {
        error("Couldn't malloc extension bootrecord buffer.");
        return(NULL);
    }

    /* TEMP HACK TO FIX UPGRADE PROBLEMS */
    ext_card_default_bootrecord(bootrecord_ext_card);
    return(bootrecord_ext_card);
    
    if (get_prismoid(ext_card_ifname, BLOB_OID_EXT_CARD_BOOTREC, (long*)bootrecord_ext_card, BR_MAX_BRA_SIZE) < 0) {
        
        D(note("Couldn't get bootrecord from extension card. Use default bootrecord."));
        ext_card_default_bootrecord(bootrecord_ext_card);
    }
    else {
        D(note("Got bootrecord from extension card.")); 
    }
    
    return(bootrecord_ext_card);

}

