/*  $Header$
 *
 *  Copyright (C) 2001 Intersil Americas Inc
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*===================================================================*
 | PDA_TEST must be set to 0 if a real PDA is present in the flash   |
 | If set to 1 it provides a test PDA structure (see pdaTestData.h)  |
 | and some debugging functionality. Handy when switching from       |
 | HardConfig to Pda.                                                |
 *-------------------------------------------------------------------*/
#define PDA_TEST     0

/*----------------------------------------------------------------------------*
 |                                    INCLUDES                                |
 *----------------------------------------------------------------------------*/
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>

#include <flash_layout.h>
#include <identification/interfaces.h>
#include <identification/bootrec.h>
#include <identification/components.h>

#include "mtd.h"
#include "inl_debug.h"
#include "dispatch.h"
#include "pda.h"

#ifdef PDA_TEST
#include "pdaTestData.h"
#endif


/*----------------------------------------------------------------------------*
 |                    GLOBAL DECLARATIONS and DEFINITIONS                     |
 *----------------------------------------------------------------------------*/

#ifndef offsetof
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

#ifndef TRUE
#define TRUE    1
#endif
#ifndef FALSE
#define FALSE   0
#endif


static void  *Pda       = 0;

/*----------------------------------------------------------------------------*
 |                             FUNCTION DEFINITIONS                           |
 *----------------------------------------------------------------------------*/

void *
pda_Get_PdaPtr (void)
{
	return (Pda);
}


/*=======================================================================================*
 | NAME:
 |    pda_Cache
 |
 | DESCRIPTION:
 |    Cache (copy to ram) the PDA
 |
 | PARAMETERS:
 |    char *mtdevice    - IN - Pointer to name of MDT device to use e.g. /dev/mtdr0
 |
 | RETURN:
 |    int   0 if success, 1 if failure
 |
 *=======================================================================================*/
int
pda_Cache (char *mtddevice)
{

   int fd         = -1;
   int rc         =  0;
   int offset;

   u_long   pda_addr;

   static char pda_cached = 0;

   /*---------------------------*
    | Caching once is enough... |
    *---------------------------*/
   if (pda_cached) {
      return 0;
   }

#if (PDA_TEST != 0)
   /*----------------------------------------------------------------------------*
    | For Testing, use the static PDA test structure instead of the one in Flash |
    *----------------------------------------------------------------------------*/
   Pda = (void *)PdaTestData;
   pda_cached = 1;
#else

   fd = open (mtddevice, O_RDONLY);
   if (fd < 0) {
      warn ("open %s failed: %m", mtddevice);
      return 1;
   }

   offset = DISPATCH_TABLE + (DISPATCH_HC_START * sizeof(unsigned char *));
   rc = mtd_read_ulong (fd, offset, &pda_addr);
   if (rc == 1) {
      close (fd);
      return 1;
   }
   Pda = (void *)malloc((size_t)PDA_SIZE);
   if (Pda) {
      unsigned long mask, flashsize;
      /* convert the PDA address to an offset from the start of flash */
      flashsize = lseek(fd, 0, SEEK_END);
      for(mask = (unsigned long)-1;; mask >>= 1) {
         offset = pda_addr & mask;
         if(offset < flashsize)
            break;
      }
      rc = mtd_read_uchar_array (fd, offset, (u_char *)Pda, PDA_SIZE);
      if (rc == 1) {
         close (fd);
         warn ("PDA could not be cached, reading from device %sfailed.\n", mtddevice);
         free(Pda);
         return 1;
      }
      pda_cached = 1;
      close (fd);
      return 0;
   } else {
      warn ("PDA could not be cached, malloc failed.\n");
      close (fd);
      return 1;
   }

#endif
}


/*=======================================================================================*
 | NAME:
 |    pda_Find_Pdr
 |
 | DESCRIPTION:
 |    In the PDA, search for a PDR matching a given PDR Code
 |
 | PARAMETERS:
 |    pdrCode    - IN -  find the PDR matching this PDR-code
 |                       Note: If multiple entries exist in the PDA, use this function
 |                             to find the first entry by passing a NULL for second argument,
 |                             find the next entries by passing the PDR ptr as a second argument.
 | RETURN:
 |    void *    A pointer to the PDR start memory location if PDR was found
 |              Note: a NULL pointer is returned if PDR was not found.
 | EXAMPLE:
 |    pdrPtr = NULL;
 |    while ((pdrPtr = (struct Pdr_Struct *)pda_Find_Pdr(PDR_INTERFACE_LIST, pdrPtr)) != NULL) {
 |       printf("Found an Entry\n");
 |    }
 |
 *=======================================================================================*/
struct Pdr_Struct *
pda_Find_Pdr (u_short pdrCode, struct Pdr_Struct *Prev)
{
   unsigned short   pdaOffset = 0;
   unsigned short   *pdaPtr;
   struct Pdr_Struct *pdrPtr;

   if (!Pda)
   {
      // Pda is not set, call Pda_Cache to initialise the ptr
      if ( pda_Cache (MTD_BOOTLOADER) )
         return (struct Pdr_Struct *)0;
   }
   if (Prev == NULL) {
      /*------------------*
       | Find first Entry |
       *------------------*/
      pdaPtr = (unsigned short *)Pda;
      pdrPtr = (struct Pdr_Struct *)pdaPtr;
   } else {
      /*-----------------*
       | Find next Entry |
       *-----------------*/
      if (Prev->PdrCode != PDR_END) {
         pdaOffset = (Prev->PdrLength + 1);
         // Next statement results in a warning but should work ok.
         pdaPtr = (unsigned short *)Prev + pdaOffset;
         pdrPtr = (struct Pdr_Struct *)pdaPtr;
      } else {
         return (struct Pdr_Struct *)0;
      }
   }

  /*----------------------------------------------*
   | Walk through the cached PDA end check each   |
   | PDR for it's PDRcode to match the one we are |
   | looking for.                                 |
   *----------------------------------------------*/
   while (pdrPtr->PdrCode != PDR_END) {
      if (pdrPtr->PdrCode == pdrCode) {
         return pdrPtr;
      } else {
         pdaOffset =pdrPtr->PdrLength + 1;
         /*----------------------------------------------------------------*
          | If we get past 8k or we have an invalid PdrLength, return NULL |
          | since this indicates an invalid PDA.                           |
          *----------------------------------------------------------------*/
         if ((pdaOffset > 8192) || (pdrPtr->PdrLength == 65535) || (pdrPtr->PdrLength == 0)) {
            return (struct Pdr_Struct *)0;
         }
        /*----------------------*
         | point to next record |
         *----------------------*/
         pdaPtr = (unsigned short *)pdaPtr + pdaOffset;
         pdrPtr = (struct Pdr_Struct *)pdaPtr;
      }
   };

   if (pdrCode == PDR_END) {
      /*---------------------------------------------*
       | If we were looking for PDR_END, we found it |
       *---------------------------------------------*/
      return pdrPtr;
   } else {
      /*-----------------------------------------------------*
       | We were looking for another PDR and did not find it |
       *-----------------------------------------------------*/
      return (struct Pdr_Struct *)0;
   }
}

/*----------------------------------------------------*
 |        Retrieve Bootloader Records from PDA        |
 *----------------------------------------------------*/
struct s_bootloader_boot_record * pdaGetBootRecords(void)
{
   struct Pdr_Struct *PdrIfList;
   struct Pdr_Struct *PdrHWCompId;
   unsigned char  *Ptr;
   unsigned long  *lPtr;
   unsigned short *sPtr;
   unsigned short  nrRecords;
   unsigned short  cnt;
   unsigned short  itemsInStruct;

   struct s_bootrec_exp_if *ExpIfPtr;

   PdrIfList = pda_Find_Pdr (PDR_INTERFACE_LIST, 0);
   PdrHWCompId = pda_Find_Pdr(PDR_HW_PLATFORM_COMPONENT_ID, 0);
   if (PdrIfList != 0) {
      Ptr = (unsigned char *)malloc(BR_MAX_BRA_SIZE);
      if (Ptr) {
         /*------------------*
          | Compose a Header |
          *------------------*/
         lPtr = (unsigned long *)Ptr;
         *lPtr++ = BR_CODE_COMPONENT_ID;
         *lPtr++ = 0x00000001;
         *lPtr++ = BR_COMP_ID_ISL39300A;

         *lPtr++ = BR_CODE_COMPONENT_VERSION;
         *lPtr++ = VER_STRING_SIZE;
         /* A revision string (if any) is attached after the 4-byte component ID */
         if(PdrHWCompId && (PdrHWCompId->PdrLength > 3)) {
            memcpy(lPtr, (char *)PdrHWCompId->PdrData + 4,
            		sizeof(short) * (PdrHWCompId->PdrLength - 3));
         } else {
            memset(lPtr, 0, VER_STRING_SIZE);
         }
         lPtr += VER_STRING_SIZE;

         /*------------------------------------------------------*
          | Now parse the PDR data to find Interface definitions |
          *------------------------------------------------------*/
         ExpIfPtr = (struct s_bootrec_exp_if *)(PdrIfList->PdrData);
         itemsInStruct = sizeof(struct s_bootrec_exp_if) / sizeof(short);
         nrRecords = (PdrIfList->PdrLength-1) / itemsInStruct;
         if (nrRecords > 0) {
            *lPtr++ = BR_CODE_EXPOSED_IF;
            *lPtr++ = PdrIfList->PdrLength / 2;
            sPtr = (unsigned short *)lPtr;
            for (cnt=0; cnt<nrRecords; cnt++) {
               *sPtr++ = ExpIfPtr->role;
               *sPtr++ = ExpIfPtr->if_id;
               *sPtr++ = ExpIfPtr->variant;
               *sPtr++ = ExpIfPtr->btm_compat;
               *sPtr++ = ExpIfPtr->top_compat;
               ExpIfPtr++;
            }
            if (nrRecords & 1) {
               sPtr++;                   // align to nearest long
            }
         }

         /*-----------------------------*
          | Write the END OF BRA Record |
          *-----------------------------*/
         lPtr = (unsigned long *)sPtr;
         *lPtr++ = BR_CODE_END_OF_BRA;
         *lPtr++ = 0x00000005;                   // length
         *lPtr++ = 0x00000000;                   // crc16
         *lPtr++ = 0x00000000;                   // MD5 Digest
         *lPtr++ = 0x00000000;
         *lPtr++ = 0x00000000;
         *lPtr++ = 0x00000000;

         return (struct s_bootloader_boot_record *)Ptr;

      } else {
         return (struct s_bootloader_boot_record *)NULL;
      }
   } else {
      return (struct s_bootloader_boot_record *)NULL;
   }
}


/*----------------------------------------------------*
 |        Retrieve MacAddress Record from PDA         |
 *----------------------------------------------------*/
unsigned char * pdaGetMacAddress(void)
{
   struct Pdr_Struct *pdrStructPtr;

   pdrStructPtr = pda_Find_Pdr (PDR_MAC_ADDRESS, (struct Pdr_Struct *)0);
   if (pdrStructPtr) {
      return (unsigned char *)(pdrStructPtr->PdrData);
   } else {
      return (unsigned char *)0;
   }
}



