/*  $Header$
 *
 *  Copyright (C) 2001 Intersil Americas Inc
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation
 *  version 2.1 of the License
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <string.h>
#include <unistd.h>
#include "../../../../../kernel/include/linux/kernel.h"

/*
  block in use: 	[flags+size][data ..........][size]
  free block: 		[flags+size][next][prev][...][size]
*/

/*
 * Retrieve the page size from the kernel
 */
#define PAGE_SIZE getpagesize()

enum
{
   FIRST_FLAG = 0x10000000,
   LAST_FLAG  = 0x20000000,
   FREE_FLAG  = 0x40000000,
   SIZE_MASK  = 0x0fffffff
};

typedef struct block Block;

struct block
{
   size_t size;
   Block *next, *prev;
};

#ifdef L_malloc
Block free_list = { 0, &free_list, &free_list };
static size_t	min_mmap_size = 4096;		/* Minimum size of mmaps performed by malloc */
#else
extern Block free_list;
#endif /* L_malloc */

//#define VERBOSE_MALLOC 1
#ifdef VERBOSE_MALLOC
extern void show_free( void );
#endif /* VERBOSE_MALLOC */

#ifndef L_malloc
extern void set_size( Block *head, size_t size );
extern void add_free( Block *b );
extern void unlink_block( Block *b );
#else /*L_malloc*/

long *get_tail( Block *head )
{
   return( (long *) head + (head->size & SIZE_MASK) - 1 );
}

void set_size( Block *head, size_t size )
{
   long *tail = (long *) head + size - 1;

   head->size = (head->size & ~SIZE_MASK) | (*tail = size);
}

void insert( Block *prev, Block *b )
{
   Block *next = prev->next;

   b->prev = prev;
   b->next = next;
   prev->next = b;
   next->prev = b;
}

void add_free( Block *b )
{
   b->size |= FREE_FLAG;
   insert( &free_list, b );
}

void unlink_block( Block *b )
{
   Block *prev = b->prev;
   Block *next = b->next;

   prev->next = next;
   next->prev = prev;
}


Block *bestfit( Block *list, size_t size )
{
   Block *block, *best = 0;
   size_t best_size = 0x7fffffff;
   size_t block_size;

   for( block = list->next ; block != list; block = block->next )
   {
      block_size = block->size & SIZE_MASK;
      if( block_size >= size && block_size < best_size )
      {
         best = block;
	 best_size = block_size;
         if ( block_size == size)
           break;
      }
   }
   if( best )
   {
      unlink_block( best );
      return( best );
   }
   return( 0 );
}

#ifdef VERBOSE_MALLOC
void show_free( void )
{
   Block *p;
   int i;

   fprintf(stderr, "free list %p:\n", &free_list );
   for( p = free_list.next, i = 0; p != &free_list; p = p->next, i++ )
   {
      fprintf(stderr, "%p: size %d %p (%ld)\n", p, p->size & SIZE_MASK, get_tail(p), *get_tail(p) );
   }
}
#endif /* VERBOSE_MALLOC */

void set_malloc_mmap_size(size_t size)
{
      for (min_mmap_size = PAGE_SIZE;; min_mmap_size <<= 1)
      {
        if (min_mmap_size >= size)
          break;
      }
#ifdef VERBOSE_MALLOC
      fprintf(stderr, "MALLOC: New minimum malloc mmap size set to %d\n", min_mmap_size);
#endif
}

void *malloc( size_t size )
{
   Block *block;

//   if (size == 0)
//     return (NULL);

   /* Need 2 longs for size at beginning and end of block
    * and align size to 4 byte boundary. Note that the size
    * stored in the block is in longs.
    */
   size = (size + 8 + 3) >> 2;
   if( size < 4 )
      size = 4;

   block = bestfit( &free_list, size );

   if( !block )
   {
      size_t blocksize;

      /* Calculate the blocksize (multiple of PAGE_SIZE)
       * that corresponds with the requested size
       */
      for (blocksize = PAGE_SIZE;; blocksize <<= 1)
      {
        if (blocksize >= size << 2)
          break;
      }

      /* If we don't have a suitable block, first try if we can use the
       * memory just after the application data segment, instead of
       * mmapping a new area.
       */
#ifdef VERBOSE_MALLOC
      fprintf(stderr, "BRK of size %d = ", blocksize);
#endif /* VERBOSE_MALLOC  */

      block = sbrk( blocksize );
#ifdef VERBOSE_MALLOC
      fprintf(stderr, "%p\n", block);
#endif /* VERBOSE_MALLOC  */
      if ( block == (void *) -1 )
	  {
         /* Couldn't brk a block, so we have to mmap a block of memory.
          * First adjust the blocksize to take the minimum mmap size into account.
          */
         if (blocksize < min_mmap_size)
            blocksize = min_mmap_size;
#ifdef VERBOSE_MALLOC
         fprintf(stderr, "Doing mmap of size %d = ", blocksize);
#endif /* VERBOSE_MALLOC  */

         block  = mmap((void *) 0, blocksize,
                       PROT_READ | PROT_WRITE,
#ifdef __UCLIBC_HAS_MMU__
                       MAP_PRIVATE | MAP_ANONYMOUS,
#else /* No MMU */
                       MAP_SHARED | MAP_ANONYMOUS,
#endif
                       0, 0);

#ifdef VERBOSE_MALLOC
         fprintf(stderr, "%p\n", block);
#endif /* VERBOSE_MALLOC  */
         if (block == (void *) -1)
         {
           return(NULL);
         }
	  }
      /* We have a block. Set the correct size */
      block->size = FIRST_FLAG | LAST_FLAG;
      set_size( block, blocksize >> 2 );
   }
   if( block )
   {
      size_t leftover = (block->size & SIZE_MASK) - size;

      if( leftover >= 4 )
      {
	 Block *split;

	 split = (Block *) ((long *) block + size);
	 split->size = block->size & LAST_FLAG;
	 set_size( split, leftover );
	 add_free( split );
	 block->size &= FIRST_FLAG;
	 set_size( block, size );
      }
      block->size &= ~FREE_FLAG;

#ifdef VERBOSE_MALLOC
      //show_free();
#endif /* VERBOSE_MALLOC */

      return( (long *) block + 1 );
   }
   return( 0 );
}
#endif /* L_malloc */


#ifdef L_free

void free( void *p )
{
   long *l;
   size_t flags, size;

   if (p == NULL)
       return;

   l = (long *) p - 1;
   flags = l[0];
   size = flags & SIZE_MASK;

   if( !(flags & LAST_FLAG) )
   {
      if( l[size] & FREE_FLAG )
      {
         flags |= (l[size] & LAST_FLAG);
         unlink_block( (Block *) (l + size) );
         size += l[size] & SIZE_MASK;
      }
   }
   if( !(flags & FIRST_FLAG) )
   {
      long *prev = l - l[-1];

      if( prev[0] & FREE_FLAG )
      {
         flags |= (prev[0] & FIRST_FLAG);
	 unlink_block( (Block *) prev );
	 size += prev[0] & SIZE_MASK;
         l = prev;
      }
   }
   l[0] = flags;
   set_size( (Block *) l, size );
   add_free( (Block *) l );

#ifdef VERBOSE_MALLOC
      //show_free();
#endif /* VERBOSE_MALLOC */
}

#endif /* L_free */

#ifdef L_calloc

void *calloc(size_t unit, size_t quantity)
{
        void *p;

        unit *= quantity;

        if ((p = malloc(unit)) == NULL)
                return NULL;
        memset(p, 0, unit);
        return p;
}

#endif /* L_calloc */


#ifdef L_realloc

void *realloc(void *ptr, size_t size)
{
    void *newptr = ptr;
    size_t copy_size;
    long *l;

    /* Get the size of the old block for doing a copy to the new block */
    if (ptr != NULL)
    {
        l = (long *) ptr - 1;
        copy_size = l[0] & SIZE_MASK;
        copy_size = copy_size - 2;	// do not copy the 2 size administrations longs
    }

    if ((size > 0) || ((size ==0) && (ptr == NULL)))
    {
        newptr = malloc(size);
        if (newptr && ptr)
        {
            memcpy(newptr, ptr, copy_size * 4);		// copy_size is in longs
            free(ptr);
        }
    }
    else
	free(ptr);

    return newptr;
}

#endif /* L_realloc  */

#ifdef L_calloc_dbg

void *calloc_dbg(size_t num, size_t size, char *function, char *file,
				 int line)
{
	void *ptr;

	fprintf(stderr, "calloc of %ld bytes at %s @%s:%d = ",
			(long) (num * size), function, file, line);
	ptr = calloc(num, size);
	fprintf(stderr, "%p\n", ptr);
	return ptr;
}

#endif /* L_calloc_dbg */

#ifdef L_malloc_dbg

void *malloc_dbg(size_t len, char *function, char *file, int line)
{
	void *result;

	fprintf(stderr, "malloc of %ld bytes at %s @%s:%d = ", (long) len,
			function, file, line);
	result = malloc(len);
	fprintf(stderr, "%p\n", result);
	return result;
}

#endif /* L_malloc_dbg */

#ifdef L_realloc_dbg

void *realloc_dbg(void *ptr, size_t len, char *function, char *file, int line)
{
	void *result;

	fprintf(stderr, "realloc at %p of %ld bytes at %s @%s:%d = ", ptr, (long) len,
			function, file, line);
	result = realloc(ptr, len);
	fprintf(stderr, "%p\n", result);
	return result;
}

#endif /* L_realloc_dbg */

#ifdef L_free_dbg

void free_dbg(void *ptr, char *function, char *file, int line)
{
	fprintf(stderr, "free of %p at %s @%s:%d\n", ptr, function, file,
			line);
	free(ptr);
}

#endif /* L_free_dbg */
